# -*- coding: iso-8859-1 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#                   Courbes                    #
##--------------------------------------#######
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2007  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# version unicode

from pylib import *
from mathlib import *
from objet import *
from labels import *
from objets import *



################################################################################
    
# Courbes
    
class Courbe_generique(Objet):
    u"""Un courbe gnrique.
    
    Usage interne : la classe mre pour tous les types de courbes :
    - courbes reprsentatives de fonctions,
    - courbes paramtres,
    - courbes polaires."""
    

    
    def __init__(self, **kw):
        Objet.__init__(self)
        self.style(**param.points)
        #self.etiquette = Label_point(self)
        self.attributs = []
        self.vassaux = WeakList()   # lors de sa creation, l'objet n'a aucun vassal (aucun objet ne depend de lui)
        self.recalculer = True
        
    def _interpreter_ensemble(self, chaine):
        return formatage_ensemble(chaine)
        
    def _interpreter_formule(self, formule, variables_libres = ""):
        # FORMATAGE DE LA CHAINE (ex: 2x -> 2*x)
        return formatage_formule(formule, variables_libres)
    



    
class Courbe_fonction(Courbe_generique):
    u"""La courbe reprsentative d'une fonction numrique  variable relle, dfinie par une formule, et un ensemble de dfinition."""

    def __init__(self, formule, ensemble = "R", **kw):
        Courbe_generique.__init__(self)
        self._set_val(formule, ensemble)
            
        
        

    def _set_val(self, formule, ensemble = "R"):
        u"""val() propose une interface standard pour afficher une valeur
        (ou la modifier, si c'est une valeur libre), indpendamment de son origine.
        
        Pour comprendre ce qui se passe, il faut savoir que le contenu de la variable est stock dans self.valeur.
        S'il s'agit d'un nombre, il est stock tel quel, mais s'il s'agit d'une expression,
        le code qui suit va remplacer toutes les noms d'objets par les objets eux-mmes.
        Ainsi, si un objet est renomm, le contenu reste pertinent.
        Exemple :
        "u.norme() + v.norme()" devient ["", u, ".norme() + ", v, ".norme()"]
        Le premier terme de la liste est toujours une chane (ventuellement vide), et le dernier aussi.
        Le nombre d'lments de la liste est donc impair, avec une alternance de chanes et d'objets.
        
        La connaissance de ce fonctionnement interne est rarement requise, mais peut-tre utile
        pour court-circuiter le fonctionnement normal de Variable par exemple.
        """
                

        # on supprime la courbe de la liste des vassaux pour les objets dont elle ne dpendra plus desormais:
        for objet in self.ancetres(): 
            objet.vassaux.remove(self)
            
        self.attributs = ["formule", "ensemble"]


        # FORMULE :
        #if isinstance(formule, basestring):
        formule = self._interpreter_formule(formule)
        
        liste = regexp("[^.A-Za-z0-9_][A-Za-z_][A-Za-z0-9_]*", " " + formule.replace(" ",""), "x[1:] <> 'x' and (x[0]+'$'+x[1:]+'$') or x").split("$")
        for i in range(len(liste)):
            if i%2:
                if liste[i]:
                    liste[i] = self.__feuille__.objets[liste[i]]
        formule = liste

        # La premire tape, c'est d'liminer les dpendances circulaires.
        # Par exemple, u = Variable(3); A = Point("u", "2*u"); u("A.x")
        # A deviendrait alors  la fois hritier de u (ie. on a besoin de u pour calculer A), et anctre de u (ie. on a besoin de A pour calculer u !)
        heritiers = self.heritiers() + [self]
        for ancetre in formule:
            if isinstance(ancetre, Objet):
                for heritier in heritiers:
                    if ancetre is heritier:
                        liste = valeur[:]
                        for i in range(len(liste)):
                            if i%2:
                                liste[i] = [key for key, value in self.__feuille__.objets.__dict__.items() if value is liste[i]][0]
                        raise RuntimeError, "Dfinition circulaire dans %s : l'objet %s se retrouve dpendre de lui-mme." %("".join(liste).strip(), ancetre)
                
        self.formule = list(formule)


        # ENSEMBLE :
        
        #if ensemble is None:
        #    ensemble = "R"
        #if isinstance(ensemble, basestring):
        ensemble = self._interpreter_ensemble(ensemble)
        
        liste = regexp("[^.A-Za-z0-9_][A-Za-z_][A-Za-z0-9_]*", " " + ensemble.replace(" ",""), "x[0]+'$'+x[1:]+'$'").split("$")
        for i in range(len(liste)):
            if i%2:
                if liste[i]:
                    liste[i] = self.__feuille__.objets[liste[i]]
        ensemble = liste

        # La premire tape, c'est d'liminer les dpendances circulaires.
        # Par exemple, u = Variable(3); A = Point("u", "2*u"); u("A.x")
        # A deviendrait alors  la fois hritier de u (ie. on a besoin de u pour calculer A), et anctre de u (ie. on a besoin de A pour calculer u !)
        heritiers = self.heritiers() + [self]
        for ancetre in ensemble:
            if isinstance(ancetre, Objet):
                for heritier in heritiers:
                    if ancetre is heritier:
                        liste = valeur[:]
                        for i in range(len(liste)):
                            if i%2:
                                liste[i] = [key for key, value in self.__feuille__.objets.__dict__.items() if value is liste[i]][0]
                        raise RuntimeError, "Dfinition circulaire dans %s : l'objet %s se retrouve dpendre de lui-mme." %("".join(liste).strip(), ancetre)
                
        self.ensemble = list(ensemble)




        dependances = self.ancetres()
        # plus self._hierarchie est faible, plus l'objet est haut place dans la _hierarchie
        if dependances: self.hierarchie(max([objet._hierarchie for objet in dependances]) + 1)
        else:           self._hierarchie = 0 # au sommet, les objets libres (ou suzerains)
        
        for objet in dependances:   # l'objet est vassal de chacun des objets dont il depend
            objet.vassaux.append(self)



            
            
    def _get_val(self):
        liste = self.formule[:]
        for i in range(len(liste)):
            if i%2: # annuaire inverse ;) (on retrouve le nom de l'objet)
                try:
                    liste[i] = [key for key, value in self.__feuille__.objets.__dict__.items() if value is liste[i]][0]
                except:
                    print u"Objet introuvable sur la feuille (id %s) : %s" %(id(self.__feuille__), liste[i])
                    raise
        instruction = "def fonction(x):return" + "".join(liste)
        print instruction, self.__feuille__.objets.__dict__
        local = {}
        exec(instruction, self.__feuille__.objets.__dict__, local)

        liste = self.ensemble[:]
        for i in range(len(liste)):
            if i%2: # annuaire inverse ;) (on retrouve le nom de l'objet)
                try:
                    liste[i] = [key for key, value in self.__feuille__.objets.__dict__.items() if value is liste[i]][0]
                except:
                    print u"Objet introuvable sur la feuille (id %s) : %s" %(id(self.__feuille__), liste[i])
                    raise
        ensemble = eval("".join(liste), self.__feuille__.objets.__dict__)
        return (local["fonction"], ensemble)



    def __call__(self, x):
        pass


