# -*- coding: iso-8859-1 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#                   Labels                    #
##--------------------------------------#######
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2007  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# version unicode

from pylib import *
from mathlib import *
from texte import *



class Label_generique(Texte):
    u"""Un label (tiquette accole  l'objet)
    
    Le label est cre automatiquement lors de la cration de l'objet.
    Le label n'est pas un 'vrai' objet, il n'est pas enregistr sur la feuille."""

    def __init__(self, parent):
        Objet.__init__(self)
        self.parent = weakref.ref(parent) # self.parent est une fonction qui renvoit parent si il existe encore. Cela permet de ne pas le maintenir en vie artificiellement ( pas de rfrence circulaire).
        self.style(**param.labels)
        self.attributs = []
        self.suffixe = self.__class__.__name__.split("_", 1)[1]
        self.initialiser_coordonnees()
        # pas d'enregistrement  la fin : un label n'est pas un 'vrai' objet.
        
    def initialiser_coordonnees(self):
        pass
    
    def epix2coo(self, x, y):
        u"Conversion des carts de pixels en carts de coordonnes."
        return self.__canvas__().coeff(0)*x, self.__canvas__().coeff(1)*y
        
    def ecoo2pix(self, x, y):
        u"Conversion des carts de coordonnes en carts de pixels."
        return x/self.__canvas__().coeff(0), y/self.__canvas__().coeff(1)
    
    def _get_coordonnees(self):
        return NotImplemented    # sera implemente differemment pour chaque type de label
        
    def label(self):
        return self.parent().label()
        

    def affiche(self, actualiser = False): # pas besoin d'effectuer autant de tests que pour les objets ordinaires
        parent = self.parent()
        if parent.existe() and parent.style("visible"):
            self._affiche()
            #extra = self.style("extra")
            #if extra:
            #    for representation in self.__representation__:
            #        representation.set(**extra)
            if parent.en_gras():
                self.__representation__[0]._fontproperties.set_size(1.15*self.style("taille"))
        else:
            self.__canvas__().graph.supprimer(self.__representation__)
            self.__representation__ = []
        if actualiser:
            self.__canvas__().actualiser()

#    def _affiche(self):
#        if not self.__representation__:
#            self.__representation__ = [self.__canvas__().text(va = "bottom", ha = "left", zorder = 4), self.__canvas__().fill()]
#            self.__representation__[0]._fontproperties = copy.copy(self.__empty_font__)
#        Texte._affiche(self)
    
        
#    def coordonnees(self, *args, **kw):
#        if self._modifiable and (args or kw):
#            self.recalculer = True
#            self._set_coordonnees(*args, **kw)
#        if self.recalculer:
#            if self.parent().existe():
#                self._xy = self._get_coordonnees(*args, **kw)
#            else:
#                self._xy = None # si l'objet n'existe pas, renvoit None
#            self.recalculer = False
#        return self._xy


    def coordonnees(self, *args, **kw):
        if self._modifiable and (args or kw):
            self.recalculer = True
            try:
                self.etiquette_a_rafraichir()
                self._set_coordonnees(*args, **kw)
                if self.__feuille__ is not None and self.__feuille__.action_principale is self:
                    rafraichir = no_twin(self.__feuille__.a_rafraichir) # on limine les doublons
                    for obj in rafraichir:
                        obj.affiche() # actualise la reprsentation graphique de chaque objet (sans rafraichir la fentre)
                    self.affiche(True) # 'True' pour redessiner la fentre effectivement
            finally: # l'objet qui a plac un verrou doit imprativement l'enlever (sinon, plus aucun rafraichissement n'a lieu !!)
                if self.__feuille__ is not None and self.__feuille__.action_principale is self:
                    self.__feuille__.action_principale = None
        if self.recalculer:
            if self.parent().existe():  # ATTENTION : c'est l'existence du l'objet qui dtermine celle de son tiquette
                self._xy = self._get_coordonnees()
            else:
                self._xy = None # si l'objet n'existe pas, renvoie None
            self.recalculer = False
        return self._xy


    def etiquette_a_rafraichir(self):
        u"""Indique que l'etiquette devra tre rafraichi."""
        if self.__feuille__ is not None:
            if self.__feuille__.action_principale is None:
                # on place un verrou : tant que cette action n'est pas termine, aucun rafraichissement n'a lieu
                # l'ide est d'viter qu'un mme objet soit rafraichi plusieurs fois par diffrentes sous-actions.
                # le rafraichissement n'a lieu qu'une seule fois, et les sous-actions ne font qu'indiquer les objets qu'elles veulent voir rafraichis.
                self.__feuille__.action_principale = self
                self.__feuille__.a_rafraichir = [self]
            else:
                self.__feuille__.a_rafraichir.append(self)





    def distance_inf(self, x, y, d):
        return self._distance_inf(x, y, d)

    def refresh(self):
        self.parent().refresh()

    def existe(self):
        return self.parent().existe()

    _call = coordonnees















class Label_point(Label_generique):
    u"L'tiquette d'un point."

    def initialiser_coordonnees(self):
        self.style(_angle_ = pi/4) # les noms de style sont entre "_" pour viter des conflits avec les styles de Texte.
        self.style(_rayon_ = 7) # distance au point en pixels
        
    def _get_coordonnees(self):
        parent = self.parent()
        r = self.style("_rayon_"); a = self.style("_angle_")
        rx, ry = self.epix2coo(r*cos(a), r*sin(a))
        return  parent.abscisse() + rx, parent.ordonnee() + ry
        
    def _set_coordonnees(self, x = None, y = None):
        if x is not None:
            parent = self.parent()
            rx, ry = self.ecoo2pix(x - parent.abscisse(), y - parent.ordonnee())
            rayon = math.hypot(rx, ry)
            if rayon:
                self.style(_angle_ = math.acos(rx/rayon)*(cmp(ry, 0) or 1))
            self.style(_rayon_ = min(rayon, 50)) # distance maximale entre le point et son etiquette : 25 pixels




###########################
# NOTE IMPORTANTE:
# Les classes suivantes sont bases sur les classes Glisseur de geolib.
# Si la classe utilise l'objet Glisseur_segment par exemple, elle DOIT s'appeler Label_segment.
# Si la classe utilise l'objet Glisseur_machin, elle DOIT s'appeler Label_machin. Etc.
# (La dtection de l'objet  utiliser se fait automatiquement selon le nom de la classe).

class Label_segment(Label_generique):
    u"""L'tiquette d'un segment.
    
    Attention, la classe ne doit pas tre renomme."""
    defaut = 0.5
    
    def initialiser_coordonnees(self):
        self.style(_angle_ = pi/4) # les noms de style sont entre "_" pour viter des conflits avec les styles de Texte.
        self.style(_rayon_ = 7) # distance au point en pixels
        self.style(_k_ = self.defaut) # Sur le segment [AB], l'tiquette sera relative au point fictif M = k*A + (1-k)*B
        self._M = None # il serait assez dlicat d'initialiser _M d'office (risque de rcursion infinie)...
        
        
    def _get_coordonnees(self):
        if self._M is None:
            import objets
            self._M = getattr(objets, "Glisseur_" + self.suffixe)(self.parent(), k = self.style("_k_"))
        x0, y0 =  self._M.coordonnees()
        r = self.style("_rayon_"); a = self.style("_angle_")
        rx, ry = self.epix2coo(r*cos(a), r*sin(a)); 
        return  x0 + rx, y0 + ry

    def _set_coordonnees(self, x = None, y = None):
        if self._M is None:
            import objets
            self._M = getattr(objets, "Glisseur_" + self.suffixe)(self.parent(), k = self.style("_k_"))
        if x is not None:
            self.recalculer = True
            x0, y0 = self._M.coordonnees(x, y)
            self.style(_k_ = self._M.parametre())
            rx, ry = self.ecoo2pix(x - x0, y - y0)
            rayon = math.hypot(rx, ry)
            if rayon:
                self.style(_angle_ = math.acos(rx/rayon)*(cmp(ry, 0) or 1))
            self.style(_rayon_ = min(rayon, 50)) # distance maximale entre le point et son etiquette : 25 pixels



class Label_droite(Label_segment):
    u"""L'tiquette d'une droite.
    
    Attention, la classe ne doit pas tre renomme."""


class Label_demidroite(Label_segment):
    """L'tiquette d'une demi-droite.
    
    Attention, la classe ne doit pas tre renomme."""



class Label_cercle(Label_segment):
    """L'tiquette d'un cercle.
    
    Attention, la classe ne doit pas tre renomme."""
    defaut = 0
    

class Label_arc_cercle(Label_segment):
    """L'tiquette d'un arc de cercle.
    
    Attention, la classe ne doit pas tre renomme."""


#   Autres classes
#
###########################




class Label_polygone(Label_generique):
    u"L'tiquette d'un polygone."
    
    def initialiser_coordonnees(self):
        self.style(_angle_ = pi/4) # les noms de style sont entre "_" pour viter des conflits avec les styles de Texte.
        self.style(_rayon_ = 7) # distance au point en pixels
        
    def _set_coordonnees(self, x = None, y = None):
        if x is not None:
            self.recalculer = True
            parent = self.parent()
            x1, x2, y1, y2 = parent._espace_vital()
            x = max(min(x, x2), x1)
            y = max(min(y, y2), y1)
            rx, ry = self.ecoo2pix(x - parent.centre.abscisse(), y - parent.centre.ordonnee())
            rayon = math.hypot(rx, ry)
            if rayon:
                self.style(_angle_ = math.acos(rx/rayon)*(cmp(ry, 0) or 1))
            self.style(_rayon_ = rayon) # distance maximale entre le point et son tiquette : 25 pixels


    def _get_coordonnees(self):
        parent = self.parent()
        r = self.style("_rayon_"); a = self.style("_angle_")
        rx, ry = self.epix2coo(r*cos(a), r*sin(a))
        return  parent.centre.abscisse() + rx, parent.centre.ordonnee() + ry




class Label_angle(Label_generique):
    u"L'tiquette d'un angle."
    
    def initialiser_coordonnees(self):
        self.style(_k_ = 0.5) # les noms de style sont entre "_" pour viter des conflits avec les styles de Texte.
        self.style(_rayon_ = param.codage["rayon"] + 10) # distance au point en pixels
        
    def _get_coordonnees(self):
        parent = self.parent()
        r = self.style("_rayon_"); k = self.style("_k_")
        #xx, yy = self.parent.point.position()
        u = self.epix2coo(*parent.vecteur1)
        v = self.epix2coo(*parent.vecteur2)
        i = (1, 0)
        a = angle_vectoriel(i, u)
        b = angle_vectoriel(i, v)
        if parent.sens() == u"non dfini" and parent._sens() < 0:
            a, b = b, a
        if b < a:
            b += 2*pi
        c = k*b + (1 - k)*a
        rx, ry = self.epix2coo(r*cos(c), r*sin(c))
        return  parent.point.abscisse() + rx, parent.point.ordonnee() + ry


    def _set_coordonnees(self, x = None, y = None):
        if x is not None:
            self.recalculer = True
            parent = self.parent()
            rx, ry = self.ecoo2pix(x - parent.point.abscisse(), y - parent.point.ordonnee())
            rayon = math.hypot(rx, ry)
            if rayon:
                u = self.epix2coo(*parent.vecteur1)
                v = self.epix2coo(*parent.vecteur2)
                i = (1, 0)
                a = angle_vectoriel(i, u)
                b = angle_vectoriel(i, v)
                if parent.sens() == u"non dfini" and parent._sens() < 0:
                    a, b = b, a
                c = angle_vectoriel(i, (rx, ry))
                if a <> b:
                    if b < a:
                        b += 2*pi
                    if c < a:
                        c += 2*pi
                    self.style(_k_ = (c-a)/(b-a))
            self.style(_rayon_ = min(rayon, param.codage["rayon"] + 25))
