# -*- coding: iso-8859-1 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#                   Objet                     #
##--------------------------------------#######
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2007  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# version unicode

from pylib import *
from mathlib import *
import objets as __objets__
#################################################################################################################


# NOTE : les vassaux ne sont pas parfaitement bien geres :
#    Un objet ne doit pas etre a la fois vassal et suzerain d'un autre.
#    Une telle situation provoquerait un plantage lors de la suppression d'un des deux objets (boucle recursive infinie).
#    Ceci ne pose a priori pas de problemes, sauf un peu de gaspillage de memoire :
#    en effet, les objets geometriques intermediaires introduits pour construire un objet ne sont pas systematiquement detruits quand l'objet est supprime.


class Nom(object):
    """Nom d'un objet.
    
    Affiche le nom de l'objet quand on le met sous forme de chane.
    Exemple :
    >>> u = [Nom(A), Nom(B), Nom(C)]
    >>> print u
     ['A', 'B', 'C']
    """
    def __init__(self, objet):
        self.objet = objet
        
    def __repr__(self):
        return repr(self.objet.nom())


class Liste_de_formules(list):
    u"""Liste de textes et formules.
    
    Quand on le met sous forme de chane, on obtient la formule sous forme "interprtable".
    """
    def __repr__(self):
        s = ""
        for elt in self:
            if isinstance(elt, basestring):
                s += elt
            else:
                try:
                    s += "{" + str(elt) + "}"
                except:
                    if param.debug:
                        print_error()
                        print elt.valeur
                        for obj in elt.valeur:
                            if isinstance(obj, Objet):
                                print obj.nom()
                    s += "{<?>}"
        return repr(unicode2(s)) # retourne le texte de la formule

        
    def __unicode__(self):
        s = ""
        for elt in self:
            if isinstance(elt, basestring):
                s += elt
            else:
                try:
                    s += str(nchiffres(float(elt), param.chiffres_significatifs))
                except:
                    if param.debug:
                        print_error()
                        print elt.valeur
                        for obj in elt.valeur:
                            if isinstance(obj, Objet):
                                print obj.nom()
                    s += "<?>"
        return unicode2(s) # Retourne le texte avec les expressions values
        
        
    def __str__(self):
        return unicode(self).encode(param.encodage)



class Objet(object):
    u"""Un objet gomtrique."""
    
    __feuille__ = None
    # NOTES:
    # Apres l'importation du module, on peut ainsi changer la __feuille__ utilisee par defaut par les objets crees.
    # Pour changer la feuille par defaut, il suffit donc de modifier l'attribut de classe.
    # Si la __feuille__ est nulle (None), les objets ne sont pas enregistres.
    # Cela sert essentiellement a creer des objets temporaires, ou a faire de la geometrie sans figures (sic!)
    
    def __init__(self, **kw):
        self.synchroniser_feuille()

        # Parametres par defauts
        # (Note: verifier si ils sont definis, lors d'heritages multiples, il ne faut pas tout reinitialiser)
        # (exemple, lors de la creation d'un vecteur, on appelle Segment.__init__(), et Vecteur_generique.__init__())

        if not hasattr(self, "_style"):
            self._style = param.defaut_objets.copy()
                    
        if not hasattr(self, "_pointable"):
            self._pointable = hasattr(self, "_get_coordonnees")

        if not hasattr(self, "etiquette"):
            self.etiquette = None
            
        if not hasattr(self, "_enregistrer"):
            self._enregistrer = True   # indique si un objet a t cr automatiquement (et ne doit donc pas tre sauvegard)

        if not hasattr(self, "_modifiable"):
            self._modifiable = hasattr(self, "_set_coordonnees") or hasattr(self, "_set_val") or hasattr(self, "_set_equation")
        
        # Parametres identiques pour tous les objets

        self.__representation__ = []
        self._en_gras = False
        self.recalculer = True
        self.__trace__ = []
        

        
# Fonctions d'initialisation de l'objet :
#########################################

    def synchroniser_feuille(self):
        if self.__feuille__:
            if self.__feuille__.parent:
                self.__class__.__feuille__ = self.__feuille__.parent.parent.actuel().actuelle
        self.__feuille__ = self.__class__.__feuille__ 
        # quand on change d'onglet dans WxGeometrie, ou de feuille courante, la feuille par defaut se synchronise (ouf!)

# La fonction suivante rpertorie certaines informations utiles ( dpendances...)
# ATTENTION : elle doit tre appele A LA FIN de la fonction __init__(), pour CHAQUE objet !
# (sauf ventuellement pour les objets gnriques, qui ne sont jamais crs tels quels...)

    def enregistre(self, *types):
        """Finalise la cration d'un objet.
        
        Cette mthode effectue, dans l'ordre, les actions suivantes :
        - vrifie la prsence de l'argument **kw
        - met  jour les styles  partir de kw
        - transforme toutes les variables locales en attributs de l'objet (et vrifie qu'il n'y a pas de conflits de noms)
        - met  jour le systme des dpendances entre objets,  partir des arguments entrs. (*)
        - vrifie que les arguments entrs sont biens du type donn. (*) (**)
        
        Notes:
        (*) si un objet hrite d'un autre objet,ces actions ne sont effectues que pour la mthode __init__ de la classe finale.
        
        (**) Le format de *types est le suivant :
         chaque argument de __init__, doit correspondre :
        - soit un type d'objet (on vrifie alors que isinstance(arg, type) == True). Ex: Point_generique, (Segment, Vecteur), etc.
        - soit None (l'argument peut-tre de n'importe quel type). Ex: None.
        - soit le type d'objet entre crochets (l'argument doit tre une liste d'objets du type indiqu. Ex: [Point], [(int, long)], etc.
        """
        niveau = 1 # on va remonter dans les appels de mthodes jusqu' trouver la dernire mthode __init__ appele.
        # le niveau 0 correspond  la mthode actuelle (ie. Objet.enregistre), le niveau 1  la mthode depuis laquelle Objet.enregistre est appele, etc...
        contexte = sys._getframe(niveau) # contexte de la mthode de niveau i
        while contexte.f_code.co_name <> "__init__":
            niveau += 1
            contexte = sys._getframe(niveau)
        precedent = sys._getframe(niveau) # contexte de la dernire mthode __init__ appele.
        avant_precedent = sys._getframe(niveau + 1) # contexte depuis lequel cette mthode __init__ avait t appele.
        test = avant_precedent.f_code.co_name == "__init__" and inspect.getargvalues(avant_precedent)[3].get('self') is inspect.getargvalues(precedent)[3].get('self')
        # si le test est positif, enregistre() est lance via une commande style:
        # Objet.__init__(self)
        # dans ce cas, il ne faut pas donner pour l'instant de valeur a self.attributs
        # en effet, un autre enregistre suivra.


        # Effectue "self.parametre = parametre" pour tous les parametres, sauf "self":
        vars = inspect.getargvalues(precedent)[-1]
        vars.pop("self")
        #kw = vars.pop("kw", {})
        try:
            kw = vars.pop("kw")
        except KeyError:
            print u"Attention, classe %s : initialisation incorrecte !" %self.classe()
            print u"La methode %s.__init__() doit avoir **kw comme dernier argument." %self.classe()
            kw = {}
        
        self._style.update(kw)
        if param.debug:
            for key, val in vars.items(): # on vrifie les ventuels conflits de nom
                if hasattr(self, key) and getattr(self, key) is not val and getattr(self, key) is not NotImplemented:
                    print u"Attention, classe %s : le nom d'attribut '%s' est dj utilis, pour un autre objet, par une classe parente !" %(self.classe(), key)
                    print u"Renommer le paramtre '%s' de la methode %s.__init__()." %(key, self.classe())
        self.__dict__.update(vars)


        # on liste les paramtres dans self.attributs (sert entre autres pour la sauvegarde)
        if not test:  # initialisation de l'objet
            args = inspect.getargspec(self.__init__) # arguments de la fonction __init__          
            self.attributs = args[0][1:]
            if args[1]:                 # arguments rentrs sous la forme "*args"
                self.attributs.append(args[1])
                    
            dependances = self.ancetres()
            # plus self._hierarchie est faible, plus l'objet est haut plac dans la hirarchie
            if dependances: self._hierarchie = max([objet._hierarchie for objet in dependances]) + 1
            else:           self._hierarchie = 0 # au sommet, les objets libres (ou suzerains)
            
            for objet in dependances:   # l'objet est vassal de chacun des objets dont il depend
                objet.vassaux.append(self)
                
            self.vassaux = WeakList()   # lors de sa cration, l'objet n'a, lui, aucun vassal (aucun objet ne dpend de lui)
            # La cration d'une WeakList plutt que d'une liste permet d'viter les pertes de mmoire.
        
            # on vrifie le type des paramtres:
            for nom, type in zip(self.attributs, types): # ne le faire que lors de l'initialisation de l'objet
                obj = getattr(self, nom) # sinon, self.attributs ne correspond pas aux paramtres entrs en cas d'hritage
                if type == None:
                    pass
                elif isinstance(type, list):
                    for o in obj:
                        self.verifier_type(o, type[0])
                else:
                    self.verifier_type(obj, type)


    
    #def __del__(self):
    #    print "delete"
    #    for objet in self.ancetres():   # l'objet est vassal de chacun des objets dont il depend
    #        objet.vassaux.remove(self)  # ne gnre jamais d'erreur, mme si self n'est pas dans la WeakList (contrairement au ".remove()" d'une liste)
    
    def verifier_type(self, obj, typ):
        if not isinstance(obj, typ):
            if hasattr(typ, "__name__"):
                s = str(typ.__name__)
            else:
                s = str(typ)
            if hasattr(obj, "__class__"):
                t = str(obj.__class__.__name__)
            else:
                t = str(type(obj))
            raise TypeError, str2(u"L'objet %s doit etre du type %s, et non du type %s." %(str(obj), s, t))
    

# Styles et informations sur l'objet
####################################    
        
    def label(self, label = None, formule = False):
        u"""Affiche le label (ou etiquette) de l'objet.
        
        La chaine renvoye dpendra de la valeur du style legende:
        - si legende = 0, renvoie ''
        - si legende = 1, renvoie le nom de l'objet
        - si legende = 2, renvoie le label proprement dit de l'objet
        - si legende = 4, renvoie le label interprt comme une formule
        
        Si le paramtre label est spcifi, le label est modifi, et le style de lgende est fix  2."""
        if label is not None:
            if formule:
                self.modifier(label = label, legende = wxgeo.FORMULE) # on bascule l'affichage en mode formule en mme temps
            else:
                self.modifier(label = label, legende = wxgeo.TEXTE) # si on change le label, celui-ci s'affiche par dfaut en mme temps.
        legende = self.style("legende")
        if legende == wxgeo.NOM:
            nom = self.nom()
            if re.match("[A-Za-z]_[0-9]+$", nom):
                return "$" + nom[0] +"_{" + nom[2:] +"}$"
            elif re.match("[A-Za-z][0-9]+$", nom):
                return "$" + nom[0] +"_{" + nom[1:] +"}$"
            elif nom.endswith("_prime"):
                return nom[:-6] + "'"
            elif nom.endswith("_seconde"):
                return nom[:-8] + '"'
            elif nom.endswith("_tierce"):
                return nom[:-7] + "'''"
            else:
                return nom
        elif legende == wxgeo.TEXTE:
            txt = self.style("label")
            if param.unicode:
                txt = unicode2(txt)
            return txt
        elif legende == wxgeo.FORMULE:
            return unicode(self._style["label"]) # Retourne le texte avec les expressions values
            
        else:
            return ""
    
    def style(self, nom = None, refresh = False, **kwargs):
        u"""- Si nom == None:
        Change le style de l'objet (mais ne rafraichit pas l'affichage par dfaut).
        - Si nom est chaine de caractre:
        Recherche la proprit correspondante dans self._style."""
        if kwargs:
            self._style.update(kwargs)
            label = self._style["label"]
            if self._style["legende"] == wxgeo.FORMULE:
                if isinstance(label, basestring):
                    label = Liste_de_formules(re.sub("[{][^}]+[}]", lambda x: "#__objets__.Formule(self, '"+x.group()[1:-1]+"')#", label).split("#"))
                    if len(label) == 1: # En mode formule, "A.x" seul est interprt comme "{A.x}".
                        label = Liste_de_formules(("", __objets__.Formule(self, label[0]), ""))
                    else:
                        for i in xrange(len(label)):
                            if i%2:
                                label[i] = eval(label[i])
                    self._style["label"] = label
            else:
                if isinstance(label, list):
                    self._style["label"] = repr(self._style["label"]) # label contient maintenant le texte de la formule
            if refresh:
                self.refresh()
        if nom is not None:
            if nom == "label" and self._style["legende"] == wxgeo.FORMULE:
                return eval(repr(self._style["label"])) # retourne le texte de la formule
            return self._style.get(nom)
        return self._style
        
    def modifier(self, **kwargs):
        u"Change le style et rafraichit l'affichage."
        self.style(refresh = True, **kwargs)    
    
    
    def cacher(self):
        self.modifier(visible = False)
        self.info("cache")
        
    def voir(self):
        self.modifier(visible = True)
        self.info("visible")
    
    def fixe(self):
        self.style(fixe = True)
    
    def libre(self):
        self.style(fixe = False)
    
    def nom(self):
        if self.__feuille__: return self.__feuille__.nom_objet(self)
        return ""

    def nom_corrige(self):
        u"""Renvoie le nom en remplaant, le cas chant, _prime par ', _seconde par ", et _tierce par '''."""
        return self.nom().replace("_prime", "'").replace("_seconde", "\"").replace("_tierce", "'''")
        
    def nom_complet(self):
        return self.__class__.__name__.split("_")[0] + " " + self.nom_corrige()
        
    def definition(self):
        u"""Dfinition en franais correct de l'objet.
        
        Exemple: 'Perpendiculaire  d passant par M'"""
        return self.nom_complet() #  surclasser en gnral

    def type(cls):
        return cls.__name__.lower()
    
    type = classmethod(type)


    def classe(cls):
        return cls.__name__

    classe = classmethod(classe)
    
    
    def titre(cls, article = "un", point_final = True):
        u"""Affichage format du type de l'objet... (un peu de grammaire!).
        Article peut tre 'un', 'le', ou 'du', l'accord se faisant automatiquement.
        Le formatage est respect (essayez 'un', 'UN', 'Un').
        
        >>> u = Vecteur_libre();u.titre()
        "un vecteur libre."
        """
        
        titre = cls.__doc__.split("\n")[0].rstrip(".").lower()
        if point_final: titre += "."
        
        if not (article.lower() == u"un"):
            titre = " ".join(titre.split()[1:])
            
        if article.lower() == u"le":
            if titre[0] in "aeiouy":
                titre = u"l'" + titre
            elif cls.genre() == "m":
                titre = u"le " + titre
            else:
                titre = u"la " + titre
                
        if article.lower() == u"du":
            if titre[0] in "aeiouy":
                titre = u"de l'" + titre
            elif cls.genre() == "m":
                titre = u"du " + titre
            else:
                titre = u"de la " + titre            
        
        if article.istitle():   titre = titre.capitalize()
        if article.isupper():   titre = titre.upper()
        return titre

    titre = classmethod(titre)

    
    def titre_complet(self, article = "un", point_final = False):
        return self.titre(article, point_final) + " " + self.nom_corrige()
    
    
    def genre(cls):
        det = cls.__doc__.split()[0]
        if det.lower() == u"un":
            return "m"
        else:
            return "f"

    genre = classmethod(genre)

        
        
    
    def ancetres(self): # retourne l'ensemble des objets dont depend l'objet.
        attributs = [getattr(self, nom) for nom in self.attributs if not isinstance(getattr(self, nom), (list, tuple))]
        listes = [getattr(self, nom) for nom in self.attributs if isinstance(getattr(self, nom), (list, tuple))]
        for liste in listes: attributs += list(liste)
        return no_twin([objet for objet in attributs if isinstance(objet, Objet)])
        
    def heritiers(self, classer = False):
        u"Retourne l'ensemble des objets qui dpendent de cet objet, ventuellement classs selon l'attribut '_hierarchie'."
        heritiers = self.vassaux.values()
        vassaux = tuple(heritiers)
        for vassal in vassaux:
            heritiers += vassal.heritiers()
        heritiers = no_twin(heritiers)
        if classer:
            heritiers.sort(key=lambda x:getattr(x,"_hierarchie"))
        return heritiers
        

    def hierarchie(self, valeur = None):
        if valeur is None:
            return self._hierarchie
        else:
            self._hierarchie = valeur
            for obj in self.vassaux.values():
                if obj._hierarchie <= self._hierarchie:
                    obj.hierarchie(self._hierarchie + 1)
        

# Acces a l'environnement exterieur
###################################

    def __canvas__(self):
        return self.__feuille__ and self.__feuille__.canvas
        # __canvas__() est une methode, afin de pouvoir s'adapter a un changement de feuille
        # je sais pas si c'est encore bien utile avec les changements d'implementation...
        
        
    def coo2pix(self, couple):
        return self.__canvas__().XYcoo2pix(couple, -1)
        
    def pix2coo(self, couple):
        return self.__canvas__().XYpix2coo(couple, -1)

    def message(self, message):
        if self.__feuille__:
            self.__feuille__.message(message)
        else:
            print message

    def erreur(self, message):
        self.__feuille__.erreur(message)

    def info(self, info):
        self.message(self.nom_complet() + " " + info + ".")
    
# API graphique (affichage et gestion des coordonnees)
######################################################
    
    def _trace(self):
        if self.style("trace"):
            x, y = self.coordonnees()
            if self.__trace__:
                self.__trace__[0]._x.append(x)
                self.__trace__[0]._y.append(y)
            else:
                self.__trace__ = [self.__canvas__().plot([x], [y], color = self.style("couleur"), marker = ".", linestyle = "None", markersize = 10)]
        else:
            self.__trace__ = []
        return self.__trace__
    
    # Meme pour les objets invisibles (variables), cette fonction est definie.
    # L'idee est de creer un rudiment d'API.
    def _affiche(self):
        pass

    def affiche(self, actualiser = False):
        u"Construit la reprsentation graphique de l'objet (mais ne rafraichit pas l'affichage par dfaut)."
        # si l'objet n'a pas de nom (''), c'est que c'est une construction intermediaire
        # il ne faut donc pas l'afficher.
        if self.__feuille__ is not None:
            nom = self.nom()
            canvas = self.__canvas__()
            if (nom or self in self.__feuille__.__objets_temporaires__) and canvas:
                if self.etiquette:
                    self.etiquette.affiche()
                if self.existe() and self.style("visible"):
                    self._affiche()
                    extra = self.style("extra")
                    if extra:
                        for representation in self.__representation__:
                            representation.set(**extra)
                    if self.en_gras():
                        e = self.style("epaisseur")
                        t = self.style("taille")
                        for plot in self.__representation__:
                            if hasattr(plot, "_linewidth") and e is not None:
                                plot._linewidth = 2*e
                                if hasattr(plot, "_markeredgewidth") and e is not None:
                                        plot._markeredgewidth = 2*e
                                        if t is not None:
                                            plot._markersize = 1.15*t
                            elif hasattr(plot, "_fontproperties") and t is not None:
                                plot._fontproperties.set_size(1.15*t)
                    if self.style("trace"):
                        x, y = self.coordonnees()
                        if self.__trace__:
                            self.__trace__[0]._x.append(x)
                            self.__trace__[0]._y.append(y)
                        else:
                            self.__trace__ = [self.__canvas__().plot([x], [y], color = self.style("couleur"), marker = ".", linestyle = "None", markersize = 10)]
                    else:
                        self.__trace__ = []
                        
                    #    x, y = self.coordonnees()
                    #    canvas.plot([x], [y], color = self.style("couleur"), marker = ".", markersize = 10)
    
                else:
                    canvas.graph.supprimer(self.__representation__)
                    self.__representation__ = []
            if actualiser and canvas:
                canvas.actualiser()
                    #print self.nom()
        

    # Certains objets n'existent pas toujours.
    # Par exemple, l'intersection de 2 droites est parfois vide.
    # Un objet construit a partir du point d'intersection de 2 droites peut donc ne plus exister temporairement.
    # En particulier, il est important de verifier que l'objet existe avant de l'afficher.
    def _existe(self):
        return False not in [obj.existe() for obj in self.ancetres()] and False not in self.conditions_existence()
        # /!\ self.conditions_existence() ne doit PAS etre evalue si un ancetre n'existe pas (resultat imprevisible)
    
    def existe(self):
        if self.recalculer:
            self._existence = self._existe()
        return self._existence
    

    def conditions_existence(self): # conditions specifiques pour que l'objet existe
        if self.recalculer:
            self._conditions = self._conditions_existence()
        return self._conditions

    def _conditions_existence(self): # conditions specifiques pour que l'objet existe, a definir pour chaque objet
        return [] # (exemple: la mediatrice de [AB] existe ssi A <> B)


    def _espace_vital(self):
        return None
    
    def espace_vital(self):
        if self.existe():   return self._espace_vital()


#    La gestion de l'affichage est complique... en particulier, dans geolib, les "auto_refresh" sont peut-tre  revoir.
#    L'ide est que quand on modifie un objet (style except), on passe par 3 mthodes :
#    _set_val, _set_equation, et _set_coordonnees.
#    Une fois ces mthodes appeles, on doit rafraichir l'affichage de toutes les dpendances, aprs les avoir recalcules.
#    La difficult, c'est que quand on modifie par exemple les coordonnes d'un point, on modifie aussi les variables
#    x et y de ce point, et donc, on rafraichit inutilement plusieurs fois l'affichage.
#    Une ide, c'est de simplement marquer les objets  rafficher, et ensuite de rafrachir l'affichage " la fin".
#    Mais comment dtecter "la fin" ? Une solution est de crer un verrou que pose la premire mthode  s'xcuter.


    def equation(self, *args, **kw):
        heritiers = None
        if self._modifiable and (args or kw):
            self.recalculer = True
            heritiers = self.heritiers() # pour viter d'appeler systmatiquement la mthode self.heritiers()
            self.heritiers_a_recalculer(heritiers)
            try:
                self.heritiers_a_rafraichir(heritiers) # place un verrou pour bloquer le rafraichissement le temps de l'action (sauf si un verrou est dj plac)
                self._set_equation(*args, **kw)
                if self.__feuille__ is not None and self.__feuille__.action_principale is self:
                    rafraichir = no_twin(self.__feuille__.a_rafraichir) # on limine les doublons
                    for obj in rafraichir:
                        obj.affiche() # actualise la reprsentation graphique de chaque objet (sans rafraichir la fentre)
                    self.affiche(True) # 'True' pour redessiner la fentre effectivement
            finally: # l'objet qui a plac un verrou doit imprativement l'enlever (sinon, plus aucun rafraichissement n'a lieu !!)
                if self.__feuille__ is not None and self.__feuille__.action_principale is self:
                    self.__feuille__.action_principale = None
        if self.recalculer:
            #if heritiers is None:
            #    heritiers = self.heritiers()
            #self.heritiers_a_recalculer(heritiers) # utile ??
            if self.existe():
                self._eq = self._get_equation()
            else:
                self._eq = None # si l'objet n'existe pas, renvoie None
            self.recalculer = False
        return self._eq
        
        
    def coordonnees(self, *args, **kw):
        heritiers = None
        if self._modifiable and (args or kw):
            self.recalculer = True
            heritiers = self.heritiers() # pour viter d'appeler systmatiquement la mthode self.heritiers()
            self.heritiers_a_recalculer(heritiers)
            try:
                self.heritiers_a_rafraichir(heritiers) # place un verrou pour bloquer le rafraichissement le temps de l'action (sauf si un verrou est dj plac)
                self._set_coordonnees(*args, **kw)
                if self.__feuille__ is not None and self.__feuille__.action_principale is self:
                    rafraichir = no_twin(self.__feuille__.a_rafraichir) # on limine les doublons
                    for obj in rafraichir:
                        obj.affiche() # actualise la reprsentation graphique de chaque objet (sans rafraichir la fentre)
                    self.affiche(True) # 'True' pour redessiner la fentre effectivement
            finally: # l'objet qui a plac un verrou doit imprativement l'enlever (sinon, plus aucun rafraichissement n'a lieu !!)
                if self.__feuille__ is not None and self.__feuille__.action_principale is self:
                    self.__feuille__.action_principale = None
        if self.recalculer:
            #if heritiers is None:
            #    heritiers = self.heritiers()
            #self.heritiers_a_recalculer(heritiers)
            if self.existe():
                self._xy = self._get_coordonnees()
            else:
                self._xy = None # si l'objet n'existe pas, renvoie None
            self.recalculer = False
        return self._xy
    
    
    def intersections(self):    # NON OPTIMISE (calcul systmatique)
        if self.existe():
            return self._get_intersections()
        else:
            return None # si l'objet n'existe pas, renvoie None
        
    
    def val(self, *args, **kw):
        #d_.deb()
        heritiers = None
        if self._modifiable and (args or kw):
            self.recalculer = True
            heritiers = self.heritiers() # pour viter d'appeler systmatiquement la mthode self.heritiers()
            self.heritiers_a_recalculer(heritiers)
            try:
                self.heritiers_a_rafraichir(heritiers) # place un verrou pour bloquer le rafraichissement le temps de l'action (sauf si un verrou est dj plac)
                self._set_val(*args, **kw)
                if self.__feuille__ is not None and self.__feuille__.action_principale is self:
                    rafraichir = no_twin(self.__feuille__.a_rafraichir) # on limine les doublons
                    print "rafraichir:"
                    print rafraichir
                    for obj in rafraichir:
                        obj.affiche() # actualise la reprsentation graphique de chaque objet (sans rafraichir la fentre)
                    self.affiche(True) # 'True' pour redessiner la fentre effectivement
            finally: # l'objet qui a plac un verrou doit imprativement l'enlever (sinon, plus aucun rafraichissement n'a lieu !!)
                if self.__feuille__ is not None and self.__feuille__.action_principale is self:
                    self.__feuille__.action_principale = None
        #d_.fin("val : bloc 1")
        if self.recalculer:
            #if heritiers is None:
            #    heritiers = self.heritiers()
            #self.heritiers_a_recalculer(heritiers)
            if self.existe():
                self._valeur = self._get_val()
            else:
                self._valeur = None # si l'objet n'existe pas, renvoie None
            self.recalculer = False
        #d_.fin("val : bloc 2")
        return self._valeur
    
    
    def heritiers_a_recalculer(self, heritiers):
        u"""Indique que l'objet a t modifi, et que les coordonnes de tous les objets
        qui en dpendent doivent tre recalcules."""
        #self.recalculer = True  # pour les conditions d'existence
        if self.etiquette:
            self.etiquette.recalculer = True
        for objet in heritiers:#classer = True):
            objet.recalculer = True
            if objet.etiquette:
                objet.etiquette.recalculer = True
    
    def heritiers_a_rafraichir(self, heritiers):
        u"""Indique que l'objet devra tre rafraichi, ainsi que tous les objets qui en dpendent."""
        if self.__feuille__ is not None:
            if self.__feuille__.action_principale is None:
                # on place un verrou : tant que cette action n'est pas termine, aucun rafraichissement n'a lieu
                # l'ide est d'viter qu'un mme objet soit rafraichi plusieurs fois par diffrentes sous-actions.
                # le rafraichissement n'a lieu qu'une seule fois, et les sous-actions ne font qu'indiquer les objets qu'elles veulent voir rafraichis.
                self.__feuille__.action_principale = self
                self.__feuille__.a_rafraichir = heritiers
                self.__feuille__.a_rafraichir.append(self)
            else:
                self.__feuille__.a_rafraichir.append(self)
                self.__feuille__.a_rafraichir += heritiers

    
    # Notes concernant le code prcdent :
    # - Un objet qui "n'existe pas" (au sens gomtrique) a des coordonnes fixes  None.
    # - Un objet libre (c'est--dire qui accepte des arguments pour ses mthodes coordonnees() ou equation())
    #   ne doit pas avoir de conditions d'existence.
    
    def __iter__(self): # definit tuple(objet) si l'objet renvoit des coordonnees
        if hasattr(self, "_get_coordonnees"):
            coo = self.coordonnees()
            if coo is not None:
                return iter(coo)
        raise TypeError, str2(u"Conversion impossible, l'objet n'a pas de coordonnes.")
    
    def coo(self):
        coo = array(self.coordonnees())
        if coo == None:
            self.message(u"L'objet %s n'est pas dfini" % self.nom())
        return coo

    def position(self):
        return array(self.__canvas__().XYcoo2pix(self.coo(), -1))

    
    def refresh(self, souffler = False):
        u"""self.recalculer = True"""
        heritiers = self.heritiers()#classer = True)
        """def cmp(x, y): # relation d'ordre utilise pour le tri
            if x._hierarchie < y._hierarchie: return -1
            if x._hierarchie > y._hierarchie: return 1
            return 0
            
        heritiers.sort(cmp) # on doit recalculer les objets dans le bon ordre (suivant la _hierarchie)
        for objet in heritiers:
            objet.recalculer = True  # on recalcule (au besoin) la position des objets avant d'actualiser l'affichage"""
        for objet in heritiers:
            objet.affiche()
        self.affiche(True)
        #if self.__canvas__():
        #    self.__canvas__().actualiser()
        # dans une boucle, permet de rendre la main si necessaire au programme principal, qui reste reactif:
        if souffler:
            wx.YieldIfNeeded()
    
    def distance_inf(self, x, y, d):
        u"Teste si la distance ( l'ecran, en pixels) entre l'objet et (x, y) est infrieure  d."
        if hasattr(self, "_distance_inf") and self.existe():
            return self._distance_inf(x, y, d)
        else:
            return False
        
    def affiche_en_gras(self, booleen = True):
        u"Met l'objet 'en gras', et actualise l'affichage."
        self.en_gras(booleen)
        self.affiche(True)
        

    def en_gras(self, booleen = None):
        u"Renvoie, aprs l'avoir ventuellement modifi, l'tat de l'objet (en gras ou non)."
        if booleen is not None:
            self._en_gras = booleen
#            if self.etiquette is not None:
#                self.etiquette._en_gras = booleen
        return self._en_gras

    
# API non graphique
###################
                              
    
    def _call(self, *args, **kw):
        raise "_call n'est pas defini pour ce type d'objet", NotImplementedError
        
    def __call__(self, *args, **kw):
        if not self.style("fixe"):
            return self._call(*args, **kw)
        else:
            return self._call()


    def renomme(self, nom, **kw):
        u"Permet de renommer l'objet, et ventuellement de changer en mme temps son style."
        nom_actuel = self.nom()
        if nom_actuel <> nom:
            nom = self.__feuille__.objets.__verifier_nom__(nom)
            objet = self.__feuille__.objets.__dict__.pop(nom_actuel)
            self.__feuille__.objets.__dict__[nom] = objet
        self.style(**kw)
        if self.__canvas__():
            self.affiche(True)
            #self.__canvas__().actualiser()

    def supprimer(self):
        nom = self.nom_complet() # le nom doit tre rcupr AVANT la suppression
        self._supprime()
        if self.__canvas__():
            self.__canvas__().actualiser()
        self.message(u"%s supprim." %nom)

    
    def _supprime(self):
        for objet in self.ancetres():   # l'objet est supprim de la liste des vassaux, pour chaque objet dont il depend
            objet.vassaux.remove(self)  # ne gnre jamais d'erreur, mme si self n'est pas dans la WeakList (contrairement au ".remove()" d'une liste)
        for obj in self.vassaux.values():
            try: obj._supprime()
            except KeyError: pass # il se peut que l'objet n'existe deja plus.
        if self.__feuille__:
            if self.__canvas__():
                self.__canvas__().graph.supprimer(self.__representation__)
                if self.etiquette:
                    self.__canvas__().graph.supprimer(self.etiquette.__representation__)
            for key in self.__feuille__.nom_objet(self, liste = True):
                self.__feuille__.objets.__dict__.pop(key)
                
    def redefinir(self, valeur):
        self.__feuille__.redefinir(self, valeur)


    def __repr__(self, styles = True):  # ceci sert pour sauvegarder les objets
        def formater(objet):
            if isinstance(objet, Objet):
                if objet.nom():
                    return objet.nom()
                else:
                    return objet.__repr__(styles)
                #if isinstance(objet, Variable): return repr(objet.val())
            if isinstance(objet, (list, tuple)):
                return "[" + ",".join([formater(item) for item in objet]) + "]"
            return repr(objet)
                
        if styles:
            return self.classe() + "(" + ",".join([formater(getattr(self, attribut)) for attribut in self.attributs]) + ", **" + repr(self.style()) + ")"
        else:
            return self.classe() + "(" + ",".join([formater(getattr(self, attribut)) for attribut in self.attributs]) + ")"

        
    def __str__(self): # Pour l'affichage (ne retourne pas les styles)
        def formater(objet):
            if isinstance(objet, Objet):
                if objet.nom():    return objet.nom()
                #if isinstance(objet, Variable): return repr(objet.val())
            if isinstance(objet, (list, tuple)):
                return "[" + ",".join([formater(item) for item in objet]) + "]"
            return str(objet)
                
        return self.classe() + "(" + ",".join([formater(getattr(self, attribut)) for attribut in self.attributs]) + ")"
    
    
    def copy(self):
        styles = self.style().copy()
        styles.pop("noms", None)
        print "copie de %s :" % self
        print self.style()
        return self.__class__(*(getattr(self, attribut) for attribut in self.attributs), **self.style())
        
    
    def copier_style(self, objet, refresh = False):
        """Applique le style de 'objet'  l'objet (self).
        
        Si les objets sont de type diffrent, seuls certains styles communs entrent en compte."""
        style = self.style()
        for key, value in objet.style().items():
            if style.has_key(key):
                if key in param.styles_a_signification_variable:
                    #print "clef:", key, style["categorie"], objet.style("categorie")
                    if style["categorie"] == objet.style("categorie"):
                        style[key] = value
                elif key not in param.styles_a_ne_pas_copier:
                    style[key] = value
                    
        if objet.etiquette and self.etiquette:
            style_etiquette = objet.etiquette.style().copy()
            for key in param.styles_a_ne_pas_copier:
                style_etiquette.pop(key, None)
            self.etiquette.style(**style_etiquette)
            
        if refresh:
            self.refresh()


    
# Utilitaires de conversion (permettant d'accepter plusieurs types d'entrees pour plus de souplesse).
###########################

    def nb2var(self, val):    # permet a l'utilisateur de rentrer au choix une Variable, ou un entier, ou... sans difference de traitement  (un nombre est "converti" en variable temporaire).
        if isinstance(val, __objets__.Variable):
            return val.copy()
        return __objets__.Variable(val)
    
    def lign2bip(self, ligne): # convertit en bipoint (couple de points) n'importe quel type de ligne.
        if isinstance(ligne, __objets__.Ligne): ligne = (ligne.point1, ligne.point2)
        return ligne

# Routines rapides
#
# Les routines suivantes sont optimises pour la vitesse, afin d'viter de passer par des classes (Vecteur, Segment, ...) lors des calculs internes.
# Elles essaient de concilier ceci avec une grande souplesse d'utilisation, en acceptant des arguments de types trs divers.
# Par exemple, un point peut tre donn sous la forme d'une instance de la classe Point, d'un tuple, d'une liste, d'un array.
############################

    @staticmethod
    def _distance(A, B):
        u"Distance entre les points A et B."
        return math.hypot(*(array(tuple(B))-array(tuple(A))))

    @staticmethod
    def _vec(A, B):
        u"Coordonnes (array) du vecteur A>B."
        return array(tuple(B))-array(tuple(A))

    @staticmethod
    def _angle_oriente(vec1, vec2): # DESUET ! Utiliser la fonction angle_vectoriel() de pylib maintenant
        u"Valeur en radian de l'angle dfinit par les vecteurs vec1, et vec2."
        x1, y1 = vec1; x2, y2 = vec2
        signe = cmp(x1*y2-x2*y1, 0) # signe du dterminant
        if signe is 0:
            signe = 1 # Peut importe le signe quand les vecteurs sont colineaires (0 = -0 et pi = -pi [2pi])
        return signe*arccos(produit_scalaire(vec1, vec2)/(math.hypot(*vec1)*math.hypot(*vec2)))






class Objet_numerique(Reel, Objet):
    u"Ensemble de mthodes propres au produit scalaire, aux angles, et autres objets numriques."
    
    def __init__(self, *args, **kw):
        Objet.__init__(self, *args, **kw)
    
    def __float__(self):
        return float(self.val())

    def __int__(self):
        return int(self.val())

    _call = Objet.val
    
    for operation, symbole in (("add", "+"), ("sub", "-"), ("mul", "*"), ("div", "/"), ("truediv", "/"), ("pow", "**")):

        def __op__(self, y, operation = "__" + operation + "__"):
            return getattr(float(self), operation)(y)
    
        def __rop__(self, y, operation = "__r" + operation + "__"):
            return getattr(float(self), operation)(y)

        def __iop__(self, y, op = operation, symbole = symbole):
            self.val(getattr(float(self), "__%s__" %op)(float(y)))
            return self
        
        exec("__%s__=__op__; __r%s__=__rop__; __i%s__=__iop__" %(operation, operation, operation))
        
    del __op__, __rop__

    def __neg__(self):
        return 0 - self
        
    def __pos__(self):
        return 0 + self
        
    def __eq__(self, y):
        return hasattr(y, "__float__") and abs(float(self) - float(y)) < param.tolerance

    def __ne__(self, y):
        return not self == y

    def __nonzero__(self):
        return self <> 0
        
    def __gt__(self, y):
        return hasattr(y, "__float__") and float(self) > float(y)
        
    def __ge__(self, y):
        return self > y or self == y
