# -*- coding: iso-8859-1 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#                   Objets                    #
##--------------------------------------#######
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2007  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# version unicode

from pylib import *
from mathlib import *
from objet import *
from labels import *
from widgets import *




################################################################################

# Points

class Point_generique(Objet):
    u"""Un point gnrique.

    Usage interne : la classe mre pour tous les types de points (libres, barycentres, intersection...)."""


    def __init__(self, **kw):
        Objet.__init__(self)
        self.style(**param.points)
        self.etiquette = Label_point(self)


    def _affiche(self):
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().plot()]
        x, y = self.coordonnees()
        plot = self.__representation__[0]
        plot._x = array([x]) ; plot._y = array([y])
        plot._markeredgecolor = plot._markerfacecolor= self.style("couleur")
        plot._marker = self.style("style")
        plot._markersize = self.style("taille")
        plot._markeredgewidth = self.style("epaisseur")
        plot.zorder = self.style("niveau")


    def _espace_vital(self):
        x, y = self.coordonnees()
        return (x, x, y, y)

    def _get_coordonnees(self):
        return NotImplemented    # sera implemente differemment pour chaque type de point

    def abscisse(self):
        return self.coordonnees()[0]

    def ordonnee(self):
        return self.coordonnees()[1]

    x = abscisse
    y = ordonnee

    def __sub__(self, y):
        return self + (-y)


    def __add__(self, y):
        if isinstance(y, Vecteur_generique): return Point_final(self, [y])
        raise TypeError, "vecteur attendu"

    # A>B est un alias de Vecteur(A,B) - attention aux parentheses pour 2*(A>B) !
    # A ne pas utiliser en interne (code peu lisible)
    def __gt__(self, point2):
        return Vecteur(self, point2)

    _call = Objet.coordonnees


    def _distance_inf(self, x, y, d):
        x0, y0 = self.position()
        return (x - x0)**2 + (y - y0)**2 < d**2





class Point(Point_generique):
    u"""Un point libre."""

    def __init__(self, x = None, y = None, **kw):
        if x is None:
            self.synchroniser_feuille()
            if self.__canvas__():
                xmin, xmax, ymin, ymax = self.__canvas__().fen()
                x = random.uniform(xmin,xmax)
                y = random.uniform(ymin,ymax)
                del xmin, xmax, ymin, ymax # evite d'enregistrer les variables xmin, xmax...
            else:
                x = y = 0

        x, y = self.nb2var(x), self.nb2var(y)
        # on convertit l'entree en Variable (voir la definition de la Classe Variable en fin de page)
        # l'entree peut-etre un nombre, un(e longueur de) segment, un (rayon de) cercle
        Point_generique.__init__(self)
        self.style(**param.points_deplacables)
        self.enregistre(Variable, Variable)

    x = NotImplemented  # supprime l'avertissement qui est affich lorsqu'on "crase" une mthode ou un attribut d'une classe parente.
    y = NotImplemented # (ici, des mthodes 'x' et 'y' sont dj dfinies par la classe 'Point_generique')

    def _get_coordonnees(self):
        return self.x.val(), self.y.val()


    def _set_coordonnees(self, x = None, y = None):
        if x is not None:
            self.recalculer = True
            if y is None: # _set_coordonnees((-1,2)) est accept
                x, y = x
            self.x(x)
            self.y(y)




class Barycentre(Point_generique):
    u"""Un barycentre.

    Un barycentre de n points."""

    def __init__(self, points, coeffs = None, **kw):
        if coeffs == None:
            coeffs = [1 for point in points]
        coeffs = map(self.nb2var,coeffs)
        Point_generique.__init__(self)
        self.enregistre([Point_generique], [Variable])
        self.verifier_coeffs(coeffs)


    def _get_coordonnees(self):  # renvoie les coordonnes du barycentre en fonction de celles des points
        coeffs = self.coefficients() # valeur des coefficients
        return builtin.sum(k*M.coo() for k, M in zip(coeffs, self.points))/self.__total

    def coefficients(self, *coeffs, **kw):
        if coeffs:
            self.verifier_coeffs(coeffs)
            for coeff, valeur in zip(self.coeffs(), coeffs):
                coeff.val(valeur) # l'objet coeff (de type "Variable") reoit la valeur "valeur".
        return [coeff() for coeff in self.coeffs]

    #def somme(self): return sum(self.coeffs_get_val())

    def verifier_coeffs(self, coeffs = []):
        if len(self.points) <> len(coeffs):
            self.erreur("nombre de coefficients incorrect.")
        if not sum([coeff() for coeff in coeffs]):
            self.erreur("la somme des coefficients est nulle.")

    def _conditions_existence(self):
        self.__total = sum(self.coefficients()) # on le stocke pour viter de le calculer 2 fois.
        return [self.__total <> 0]


    #def changer_coeffs(self, coeffs = []):  # A ADAPTER
    #    if len(self.points) <> len(coeffs):
    #        self.erreur("nombre de coefficients incorrect.")
    #    if not sum(coeffs):
    #        self.erreur("la somme des coefficients est nulle.")
    #    self.coeffs = map(self.nb2var,coeffs)



class Milieu(Barycentre):
    u"""Un milieu.

    Le milieu de 2 points"""

    def __init__(self, point1, point2, **kw):
        Barycentre.__init__(self, points = [point1, point2])
        self.enregistre(Point_generique, Point_generique)




class Point_final(Point_generique):
    u"""Un point dfini par une relation vectorielle.

    Point dfini par une relation vectorielle.
                              ->      ->     ->
    Exemple : point N tel que AN = -3 AB + 2 BC
    N = Point_final(A, [A>B, B>C], [-3,2])"""

    # A ADAPTER POUR ACCEPTER DES VARIABLES COMME COEFFICIENTS

    def __init__(self, depart, vecteurs = [], coeffs = None, **kw):
        if coeffs == None:
            coeffs = [1 for vecteur in vecteurs]
        coeffs = map(self.nb2var,coeffs)
        Point_generique.__init__(self)
        self.enregistre(Point_generique, [Vecteur_generique], [Variable])


    def _get_coordonnees(self):
        return self.depart.coo() + builtin.sum(coeff()*vecteur.coo() for coeff, vecteur in zip(self.coeffs, self.vecteurs))




class Point_translation(Point_final):
    u"""Une image d'un point par translation."""

    def __init__(self, point, translation, **kw):
        Point_final.__init__(self, point, [translation.vecteur])
        self.enregistre(Point_generique, Translation)




class Point_rotation(Point_generique):
    u"""Une image d'un point par rotation.

    Point construit  partir d'un autre via une rotation d'angle et de centre donn."""

    def __init__(self, point, rotation, **kw):
        Point_generique.__init__(self)
        self.enregistre(Point_generique, Rotation)

    def _get_coordonnees(self):
        x0, y0 = self.rotation.centre.coordonnees()
        xA, yA = self.point.coordonnees()
        a = self.rotation.radian()
        sina = math.sin(a) ; cosa = math.cos(a)
        return (-sina*(yA - y0) + x0 + cosa*(xA - x0), y0 + cosa*(yA - y0) + sina*(xA - x0))

    def _get_coordonnees2(self): # un poil plus lent
        x0, y0 = self.rotation.centre.coordonnees()
        xA, yA = self.point.coordonnees()
        z = ((xA - x0) + (yA - y0)*1j)*cmath.exp(1j*self.rotation.radian()) + x0 + y0*1j
        return z.real, z.imag



class Point_homothetie(Point_generique):
    u"""Une image d'un point par homothtie.

    Point construit  partir d'un autre via une homothtie de rapport et de centre donn."""

    def __init__(self, point, homothetie, **kw):
        Point_generique.__init__(self)
        self.enregistre(Point_generique, Homothetie)

    def _get_coordonnees(self):
        x0, y0 = self.homothetie.centre.coordonnees()
        xA, yA = self.point.coordonnees()
        k = self.homothetie.rapport()
        return x0 + k*(xA-x0), y0 + k*(yA-y0)




class Point_reflexion(Point_generique):
    u"""Une image d'un point par rflexion.

    Point construit  partir d'un autre via une symtrie d'axe donn."""

    def __init__(self, point, reflexion, **kw):
        Point_generique.__init__(self)
        self.enregistre(Point_generique, Reflexion)

    def _get_coordonnees(self):
        x0, y0 = self.reflexion.droite.point1
        x1, y1 = self.reflexion.droite.point2
        x, y = self.point
        z = x1 - x0 + (y1 - y0)*1j
        M = (x - x0 + (y0 - y)*1j)*z/z.conjugate() + x0 + y0*1j
        return M.real, M.imag







class Projete_generique(Point_generique):
    u"""Un projet gnrique.

    Classe mre des diffrents types de projets orthogonaux (sur droite, sur cercle ou sur segment)."""

    def __init__(self, point, objet, **kw):
        Point_generique.__init__(self)
        self.enregistre(Point_generique, Objet)


    def distance(self):
        return self._distance(self, self.point)




class Projete_droite(Projete_generique):
    u"""Un projet orthogonal sur une droite.

    Projet orthogonal d'un point sur une droite."""

    def __init__(self, point, droite, **kw):
        Projete_generique.__init__(self, point = point, objet = droite)
        self.enregistre(Point_generique, Droite)

    def _conditions_existence(self):
        self.__perpendiculaire = Perpendiculaire(self.droite, self.point)
        return [self.__perpendiculaire.existe()]

    def _get_coordonnees(self):
        # il faudra prevoir le cas ou la perpendiculaire n'existe pas (si le point appartient a la droite)
        return Intersection_droites(self.droite, self.__perpendiculaire).coordonnees()



class Projete_cercle(Projete_generique):
    u"""Un projet orthogonal sur un cercle.

    Projet orthogonal d'un point sur un cercle."""

    def __init__(self, point, cercle, **kw):
        Projete_generique.__init__(self, point = point, objet = cercle)
        self.enregistre(Point_generique, Cercle_generique)

    def _conditions_existence(self):
    # cas de non existence : si le point est au centre du cercle
        self.__droite = Droite(self.point, self.cercle.centre)
        return [self.__droite.existe()]


    def _get_coordonnees(self):
        return Intersection_droite_cercle(self.__droite, self.cercle, self.point).coordonnees()



class Projete_arc_cercle(Projete_generique):
    u"""Un projet orthogonal sur un arc de cercle.

    Projet orthogonal d'un point sur un arc de cercle."""

    # il faudrait prevoir le cas de non existence (si le point est au centre du cercle)
    def __init__(self, point, arc, **kw):
        Projete_generique.__init__(self, point = point, objet = arc)
        self.enregistre(Point_generique, Arc_generique)

    def _conditions_existence(self):
        self.__inter = Intersection_droite_cercle(Droite(self.point, self.arc.centre), self.arc, self.point)
        return [self.__inter.existe()]


    def _get_coordonnees(self):
        return __inter.coordonnees()



class Projete_segment(Projete_generique):
    u"""Un projet orthogonal sur un segment.

    Projet orthogonal d'un point sur un segment."""

    # il faudrait prevoir le cas de non existence (si le point est au centre du cercle)
    def __init__(self, point, segment, **kw):
        Projete_generique.__init__(self, point = point, objet = segment)
        self.enregistre(Point_generique, Segment)



    def _conditions_existence(self):
        self.__M = Intersection_droites(Droite(self.segment.point1, self.segment.point2), Perpendiculaire(self.segment, self.point))
        return [self.__M.existe()]


    def _get_coordonnees(self):
        A = self.segment.point1; B = self.segment.point2
        M = self.__M
        p = produit_scalaire(self._vec(A, B), self._vec(A, M))
        if p < 0:
            return A.coordonnees()
        if p > self._distance(A, B)**2:
            return B.coordonnees()
        return M.coordonnees()


class Projete_demidroite(Projete_generique):
    u"""Un projet orthogonal sur une demi-droite.

    Projet orthogonal d'un point sur une demi-droite."""

    # il faudrait prevoir le cas de non existence (si le point est au centre du cercle)
    def __init__(self, point, demidroite, **kw):
        Projete_generique.__init__(self, point = point, objet = demidroite)
        self.enregistre(Point_generique, Demidroite)


    def _conditions_existence(self):
        self.__M = Intersection_droites(Droite(self.demidroite.point1, self.demidroite.point2), Perpendiculaire(self.demidroite, self.point))
        return [self.__M.existe()]


    def _get_coordonnees(self):
        A = self.demidroite.point1; B = self.demidroite.point2
        M = self.__M
        p = produit_scalaire(self._vec(A, B), self._vec(A, M))
        if p < 0:
            return A.coordonnees()
        return M.coordonnees()


class Centre_polygone_generique(Point_generique):
    u"""Un centre d'un triangle.

    Classe mre des diffrents centres d'un triangle."""


    def __init__(self, polygone, **kw):
        Point_generique.__init__(self)
        self.enregistre(Polygone)



class Centre_gravite(Centre_polygone_generique):
    u"""un centre de gravit.

    Centre de gravit d'un polygone (l'intersection des mdianes dans le cas d'un triangle)."""

    def __init__(self, polygone, **kw):
        Centre_polygone_generique.__init__(self, polygone)
        self.enregistre(Polygone)

    def _conditions_existence(self):
        return [self.polygone.centre_gravite.existe()]

    def _get_coordonnees(self):
        return self.polygone.centre_gravite()



class Orthocentre(Centre_polygone_generique):
    u"""Un orthocentre.

    Orthocentre d'un triangle (intersection des hauteurs)."""

    def __init__(self, triangle, **kw):
        Centre_polygone_generique.__init__(self, triangle)
        self.enregistre(Triangle)

    def _conditions_existence(self):
        return [self.triangle.orthocentre.existe()]

    def _get_coordonnees(self):
        return self.triangle.orthocentre()


class Centre_cercle_circonscrit(Centre_polygone_generique):
    u"""Un centre du cercle circonscrit.

    Centre du cercle circonscrit d'un triangle (intersection des mdiatrices)."""

    def __init__(self, triangle, **kw):
        Centre_polygone_generique.__init__(self, triangle)
        self.enregistre(Triangle)

    def _conditions_existence(self):
        return [self.triangle.centre_cercle_circonscrit.existe()]

    def _get_coordonnees(self):
        return self.triangle.centre_cercle_circonscrit()


class Centre_cercle_inscrit(Centre_polygone_generique):
    u"""Un centre du cercle inscrit.

    Centre du cercle inscrit d'un triangle (intersection des bissectrices)."""

    def __init__(self, triangle, **kw):
        Centre_polygone_generique.__init__(self, triangle)

    def _conditions_existence(self):
        return [self.triangle.centre_cercle_inscrit.existe()]

    def _get_coordonnees(self):
        return self.triangle.centre_cercle_inscrit()


class Centre(Point_generique):
    u"""Un centre d'un cercle."""

    def __init__(self, cercle, **kw):
        Point_generique.__init__(self)
        self.enregistre(Cercle_generique)

    def _conditions_existence(self):
        return [self.cercle.centre.existe()]

    def _get_coordonnees(self):
        return self.cercle.centre()


class Point_equidistant(Point_generique):
    u"""Point quidistant de 3 points.

    Utilis surtout pour un usage interne, le calcul des coordonnes est plus rapide qu'en passant par les mdiatrices."""

    def __init__(self, point1, point2, point3, **kw):
        Point_generique.__init__(self)
        self.enregistre(Point_generique, Point_generique, Point_generique)

    def _conditions_existence(self):
        xA, yA = self.__A = self.point1.coordonnees()
        xB, yB = self.__B = self.point2.coordonnees()
        xC, yC = self.__C = self.point3.coordonnees()
        a = xB - xA
        b = yB - yA
        c = xC - xB
        d = yC - yB
        self.__det2 = 2*(a*d - b*c)
        return [abs(self.__det2) > param.tolerance or abs(xB - xC) + abs(yB - yC) < param.tolerance or abs(xA - xC) + abs(yA - yC) < param.tolerance or abs(xB - xA) + abs(yB - yA) < param.tolerance]

    def _get_coordonnees(self):
        xA, yA = self.__A
        xB, yB = self.__B
        xC, yC = self.__C
        if abs(self.__det2) > param.tolerance:
            a = xB - xA
            b = yB - yA
            c = xC - xB
            d = yC - yB
            p1 = a*d; p2 = b*c
            return (p1*(xA+xB) - p2*(xB+xC) - b*d*(yC-yA))/self.__det2, (a*c*(xC-xA) - p2*(yA+yB) + p1*(yB+yC))/self.__det2
        elif abs(xB - xC) + abs(yB - yC) < param.tolerance:
            return (.5*(xA+xB), .5*(yA+yB))
        elif abs(xA - xC) + abs(yA - yC) < param.tolerance:
            return (.5*(xA+xB), .5*(yA+yB))
        elif abs(xB - xA) + abs(yB - yA) < param.tolerance:
            return (.5*(xA+xC), .5*(yA+yC))

##########################################################################################

# Lignes



class Ligne(Objet):
    u"""Une ligne gnrique.

    Usage interne : la classe mere pour les droites, segments, vecteurs."""

    def __init__(self, point1, point2, **kw):
        Objet.__init__(self)
        self.enregistre(Point_generique, Point_generique)


    def _longueur(self):
        return norm2((self.point2.coo() - self.point1.coo()))


    def _espace_vital(self):
        x1, y1 = self.point1.coordonnees()
        x2, y2 = self.point2.coordonnees()
        return (min(x1, x2), max(x1, x2), min(y1, y2), max(y1, y2))


    def _get_equation(self):
        # retourne un triplet (a, b, c), tel que ax + by + c = 0 soit une equation de droite de la ligne.
        xA, yA = self.point1.coo()
        xB, yB = self.point2.coo()
        return (yA - yB, xB - xA, xA*yB - yA*xB)



class Segment(Ligne):
    u"""Un segment.

    Un segment dfini par deux points"""

    def __init__(self, point1, point2, **kw):
        Ligne.__init__(self, point1 = point1, point2 = point2)
        self.etiquette = Label_segment(self)
        self.style(**param.segments)
        self.enregistre(Point_generique, Point_generique)


    def _affiche(self):
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().plot(), self.__canvas__().plot(), self.__canvas__().plot(), self.__canvas__().plot()]
        x1, y1 = self.point1.coordonnees()
        x2, y2 = self.point2.coordonnees()

        for elt_graphique in self.__representation__:
            elt_graphique._visible = True

        plot = self.__representation__[0]
        plot._x = array([x1, x2]) ; plot._y = array([y1, y2])
        plot._color = self.style("couleur")
        plot._linestyle = self.style("style")
        plot._linewidth = self.style("epaisseur")
        niveau = self.style("niveau")
        plot.zorder = niveau

        # Toute la suite concerne les codages utiliss pour indiquer les segments de mme longueur
        if not self.style("codage"):
            for plot in self.__representation__[1:]:
                plot._visible = False
                plot.zorder = niveau + 0.01

        elif self.style("codage") == "o":
            plot = self.__representation__[1]
            plot._marker = "o"
            plot._markersize = param.codage["taille"]
            plot._markerfacecolor = None
            plot._markeredgecolor = self.style("couleur")
            plot._markeredgewidth = self.style("epaisseur")
            plot._x = [.5*(x1 + x2)]
            plot._y = [.5*(y1 + y2)]
            for plot in self.__representation__[2:]:
                plot._visible = False

        else:
            A = array(self.coo2pix((x1, y1)))
            B = array(self.coo2pix((x2, y2)))
            G = .5*(B + A)
            r, a = param.codage["taille"], param.codage["angle"]
            vec = (A - B)/norm2(A - B)
            C = G + r*vec
            if self.style("codage") == "x":
                M = (dot([[cos(a), -sin(a)], [sin(a), cos(a)]], transpose([C - G])) + transpose([G]))[:,0]
                N = (dot([[cos(a), sin(a)], [-sin(a), cos(a)]], transpose([C - G])) + transpose([G]))[:,0]
                plot1 = self.__representation__[1]
                plot2 = self.__representation__[2]
                plot1._x, plot1._y = self.pix2coo(zip(M, 2*G-M))
                plot2._x, plot2._y = self.pix2coo(zip(N, 2*G-N))
                plot1._marker = plot2._marker = 'None'
                plot1._color = plot2._color = self.style("couleur")
                plot1._linewidth = plot2._linewidth = self.style("epaisseur")
                self.__representation__[3]._visible = False

            elif self.style("codage") == "/":
                M = (dot([[cos(a), sin(a)], [-sin(a), cos(a)]], transpose([C - G])) + transpose([G]))[:,0]
                plot1 = self.__representation__[1]
                plot1._x, plot1._y = self.pix2coo(zip(M, 2*G-M))
                plot1._marker = 'None'
                plot1._color = self.style("couleur")
                plot1._linewidth = self.style("epaisseur")
                for plot in self.__representation__[2:]:
                    plot._visible = False

            elif self.style("codage") == "//":
                M = (dot([[cos(a), sin(a)], [-sin(a), cos(a)]], transpose([C - G])) + transpose([G]))[:,0]
                plot1 = self.__representation__[1]
                plot2 = self.__representation__[2]
                vec = 2*vec
                plot1._x, plot1._y = self.pix2coo(zip(M + vec, 2*G - M + vec))
                plot2._x, plot2._y = self.pix2coo(zip(M - vec, 2*G - M - vec))
                plot1._marker = plot2._marker = 'None'
                plot1._color = plot2._color = self.style("couleur")
                plot1._linewidth = plot2._linewidth = self.style("epaisseur")
                self.__representation__[3]._visible = False

            elif self.style("codage") == "///":
                M = (dot([[cos(a), sin(a)], [-sin(a), cos(a)]], transpose([C - G])) + transpose([G]))[:,0]
                plot1 = self.__representation__[1]
                plot2 = self.__representation__[2]
                plot3 = self.__representation__[3]
                vec = 3*vec
                plot1._x, plot1._y = self.pix2coo(zip(M, 2*G - M))
                plot2._x, plot2._y = self.pix2coo(zip(M + vec, 2*G - M + vec))
                plot3._x, plot3._y = self.pix2coo(zip(M - vec, 2*G - M - vec))
                plot1._marker = plot2._marker = plot3._marker = 'None'
                plot1._color = plot2._color = plot3._color = self.style("couleur")
                plot1._linewidth = plot2._linewidth = plot3._linewidth = self.style("epaisseur")

    def vect_affiche(self):
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().plot(), self.__canvas__().plot()]
        plot1, plot2 = self.__representation__
        x1, y1 = self.point1.coordonnees()
        x2, y2 = self.point2.coordonnees()
        A = array(self.coo2pix((x1, y1)))
        B = array(self.coo2pix((x2, y2)))
        k = self.style("position")
        G = k*B + (1-k)*A   # par dfaut, k = 1, et la flche est positionne  l'extrmit du vecteur (point B)

        #                                                 M
        #                                               C +--\
        # l'ideal est ici un dessin:   A +---------------+----+ G=B
        #                                                 +--/
        #                                                 N
        # On calcule les coordonnees de C, puis par rotation autour de G, celles de M et N.

        r, a = self.style("taille"), self.style("angle")*pi/360
        C = G + r*(A - B)/norm2(A - B) # prendre B et non G, au cas o G = A !
        M = (dot([[cos(a), -sin(a)], [sin(a), cos(a)]], transpose([C - G])) + transpose([G]))[:,0]
        N = (dot([[cos(a), sin(a)], [-sin(a), cos(a)]], transpose([C - G])) + transpose([G]))[:,0]
        x, y = self.pix2coo(zip(M, G, N))
        plot1._x = array([x1, x2]) ; plot1._y = array([y1, y2])
        plot2._x = array(x) ; plot2._y = array(y)
        plot1._color = plot2._color = self.style("couleur")
        plot1._linestyle = plot2._linestyle = self.style("style")
        plot1._linewidth = plot2._linewidth = self.style("epaisseur")




    def _distance_inf(self, x, y, d):
        # cf. "distance_point_segment.odt" dans "doc/developpeurs/maths/"
        xA, yA = self.point1.position()
        xB, yB = self.point2.position()
        x1 = min(xA, xB) - d; x2 = max(xA, xB) + d
        y1 = min(yA, yB) - d; y2 = max(yA, yB) + d
        if x1<x<x2 and y1<y<y2:
            norme2 = ((xB-xA)**2+(yB-yA)**2)
            if norme2 > param.tolerance:
                return ((yA-yB)*(x-xA)+(xB-xA)*(y-yA))**2/norme2 < d**2
            else:   # les extrmits du segment sont confondues
                return (x - xA)**2 + (y - yA)**2 < d**2
        else:
            return False



    def __contains__(self, M):
        #if not isinstance(M, Point_generique):
        #    return False
        A = self.point1
        B = self.point2
        xu, yu = self._vec(A, B)
        xv, yv = self._vec(A, M)
        if abs(xu*yv-xv*yu) > param.tolerance:
            return False
        if xu:
            k = xv/xu
        elif yu:
            k = yv/yu
        else:  # A == B
            return M == A
        return 0 <= k <= 1

    def longueur(self):
        u"""Alias de _longueur, disponible pour indiquer que l'objet a vraiment
        une longueur au sens mathematique du terme (pas comme une droite !)"""
        return self._longueur()




class Demidroite(Ligne):
    u"""Une demi-droite.

    Une demi-droite dfinie par son origine et un deuxime point"""


    def __init__(self, origine, point, **kw):
        Ligne.__init__(self, point1 = origine, point2 = point)
        self.etiquette = Label_demidroite(self)
        self.style(**param.droites)
        self.enregistre(Point_generique, Point_generique)


    def _conditions_existence(self):
        return [sum(abs(self.point1.coo() - self.point2.coo())) > param.tolerance]
        # EDIT: les conditions d'existence en amont sont dsormais bien dfinies (ou devraient l'tre !!)
        # return [self.point1.coo() is not None and self.point2.coo() is not None and sum(abs(self.point1.coo() - self.point2.coo())) > param.tolerance]
        # si les conditions d'existence en amont sont mal definies, il se peut que les coordonnees d'un point valent None


    def x_y(self, nbr = 0, flag = 1):
        # si flag = 1, donne le point (x,y) sur la droite en fonction de x.
        # si flag = 0, donne le point (x,y) sur la droite en fonction de y.
        [x1, y1], [x2, y2] = self.point1.coordonnees(), self.point2.coordonnees()

        # si la pente est presque verticale et que flag = 1, ou presque horizontale et que flag = 0, le calcul peut exploser... l'option flag = 2 adapte flag a la pente.
        if flag == 2:  flag = self.pente(1)

        # dans le cas d'une droite horizontale ou verticale, flag est ignore et prend la seule valeur correcte :
        flag = not (y1 - y2) and 1 or ((x1 - x2) and flag or 0)

        if flag:  return (nbr,((y1-y2)*(nbr-x1)+(x1-x2)*y1)/(x1-x2))
        else:  return (((x1-x2)*(nbr-y1)+(y1-y2)*x1)/(y1-y2),nbr)


    def pente(self, flag = 0):
         # si flag = 1, retourne la "tendance" de la pente (verticale : 0, horizontale : 1)
        x1, y1 = self.point1.coordonnees()
        x2, y2 = self.point2.coordonnees()
        if flag: return (abs(x2 - x1) > abs(y2 - y1)) and 1 or 0
        try: return (y2 - y1)/(x2 - x1)
        except ZeroDivisionError: return None


    def _affiche(self):
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().plot()]

        p = self.pente(1) # "tendance" de la pente (verticale : 0, horizontale : 1)
        fen = self.__feuille__.fenetre() # il faut tracer depuis le bord de fenetre
        x, y = self.origine.coordonnees()
        x0, y0 = self.point.coordonnees()
        x1, y1 = self.x_y((fen[2], fen[0])[p], p)
        x2, y2 = self.x_y((fen[3], fen[1])[p], p)
        plot = self.__representation__[0]

        if produit_scalaire((x1 - x, y1 - y), (x0 - x, y0 - y)) > produit_scalaire((x2 - x, y2 - y), (x0 - x, y0 - y)):
            plot._x = array([x1, x]) ; plot._y = array([y1, y])
        else:
            plot._x = array([x, x2]) ; plot._y = array([y, y2])
        plot._color = self.style("couleur")
        plot._linestyle = self.style("style")
        plot._linewidth = self.style("epaisseur")
        plot.zorder = self.style("niveau")

#    def _affiche(self):
#        fen = self.__feuille__.fenetre() # il faut tracer depuis le bord de fenetre
#        p = self.pente(1)
#        A, B = self.x_y((fen[2], fen[0])[p],p), self.x_y((fen[3], fen[1])[p],p)
#        self.__canvas__().Ligne(A, B, self.style("couleur"))

    def _distance_inf(self, x, y, d):
        # cf. "distance_point_segment.odt" dans "doc/developpeurs/maths/"
        xA, yA = self.point1.position()
        xB, yB = self.point2.position()
        x1 = min(xA, xB); x2 = max(xA, xB)
        y1 = min(yA, yB); y2 = max(yA, yB)
        if (xA == x1 and x>x1-d or xA == x2 and x<x2+d) and (yA == y1 and y>y1-d or yA == y2 and y<y2+d):
            return ((yA-yB)*(x-xA)+(xB-xA)*(y-yA))**2/((xB-xA)**2+(yB-yA)**2) < d**2
        else:
            return False

    """def _distance_inf(self, x, y, d): # A REECRIRE !
        x0, y0 = self.point1.position()
        x1, y1 = self.point2.position()
        A = Point(x0, y0) ; B = Point(x1, y1) ; M = Point(x, y)
        AB = Demidroite(A, B)
        return Projete_demidroite(M, AB).distance() < d"""

    def __contains__(self, M):
        #if not isinstance(M, Point_generique):
        #    return False
        A = self.origine
        B = self.point
        xu, yu = self._vec(A, B)
        xv, yv = self._vec(A, M)
        if abs(xu*yv-xv*yu) > param.tolerance:
            return False
        if xu:
            k = xv/xu
        elif yu:
            k = yv/yu
        else:  # A == B
            return M == A
        return 0 <= k

    def mise_en_equation(self):
        u"Equation sous forme lisible par l'utilisateur."
        eq = self.equation()
        if eq is None:
            return u"L'objet n'est pas dfini."
        a, b, c = (str(nchiffres(coeff, param.chiffres_significatifs)) for coeff in eq)
        pente = self.pente(1) # "tendance" de la pente (verticale : 0, horizontale : 1)
        if pente:
            if (self.point.ordonnee() - self.origine.ordonnee()) > param.tolerance:
                ajout = "y > " + str(nchiffres(self.origine.ordonnee(), param.chiffres_significatifs))
            elif (self.point.ordonnee() - self.origine.ordonnee()) < param.tolerance:
                ajout = "y < " + str(nchiffres(self.origine.ordonnee(), param.chiffres_significatifs))
            else:
                return u"Prcision insuffisante."
        else:
            if (self.point.abscisse() - self.origine.abscisse()) > param.tolerance:
                ajout = "x > " + str(nchiffres(self.origine.abscisse(), param.chiffres_significatifs))
            elif (self.point.abscisse() - self.origine.abscisse()) < param.tolerance:
                ajout = "x < " + str(nchiffres(self.origine.abscisse(), param.chiffres_significatifs))
            else:
                return u"Prcision insuffisante."

        return "%s*x + %s*y + %s = 0 et %s" %(a, b, c, ajout) # on ne garde "que" 7 chiffres significatifs




class Droite(Ligne):
    u"""Une droite.

    Une droite dfinie par deux points"""

    def __init__(self, point1, point2, **kw):
        Ligne.__init__(self, point1 = point1, point2 = point2)
        self.etiquette = Label_droite(self)
        self.style(**param.droites)
        self.enregistre(Point_generique, Point_generique)


    def _conditions_existence(self):
        return [sum(abs(self.point1.coo() - self.point2.coo())) > param.tolerance]
        # EDIT: les conditions d'existence en amont sont dsormais bien dfinies (ou devraient l'tre !!)
        # return [self.point1.coo() is not None and self.point2.coo() is not None and sum(abs(self.point1.coo() - self.point2.coo())) > param.tolerance]
        # si les conditions d'existence en amont sont mal definies, il se peut que les coordonnees d'un point valent None


    def x_y(self, nbr = 0, flag = 1):
        # si flag = 1, donne le point (x,y) sur la droite en fonction de x.
        # si flag = 0, donne le point (x,y) sur la droite en fonction de y.
        [x1, y1], [x2, y2] = self.point1.coordonnees(), self.point2.coordonnees()

        # si la pente est presque verticale et que flag = 1, ou presque horizontale et que flag = 0, le calcul peut exploser... l'option flag = 2 adapte flag a la pente.
        if flag == 2:  flag = self.pente(1)

        # dans le cas d'une droite horizontale ou verticale, flag est ignore et prend la seule valeur correcte :
        flag = not (y1 - y2) and 1 or ((x1 - x2) and flag or 0)

        if flag:  return (nbr,((y1-y2)*(nbr-x1)+(x1-x2)*y1)/(x1-x2))
        else:  return (((x1-x2)*(nbr-y1)+(y1-y2)*x1)/(y1-y2),nbr)


    def pente(self, flag = 0):
         # si flag = 1, retourne la "tendance" de la pente (verticale : 0, horizontale : 1)
        x1, y1 = self.point1.coordonnees()
        x2, y2 = self.point2.coordonnees()
        if flag: return (abs(x2 - x1) > abs(y2 - y1)) and 1 or 0
        try: return (y2 - y1)/(x2 - x1)
        except ZeroDivisionError: return None


    def _affiche(self):
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().plot()]
        x1, y1 = self.point1.coordonnees()
        x2, y2 = self.point2.coordonnees()
        p = self.pente(1) # "tendance" de la pente (verticale : 0, horizontale : 1)
        fen = self.__feuille__.fenetre() # il faut tracer depuis le bord de fenetre
        x1, y1 = self.x_y((fen[2], fen[0])[p], p)
        x2, y2 = self.x_y((fen[3], fen[1])[p], p)
        plot = self.__representation__[0]
        plot._x = array([x1, x2]) ; plot._y = array([y1, y2])
        plot._color = self.style("couleur")
        plot._linestyle = self.style("style")
        plot._linewidth = self.style("epaisseur")
        plot.zorder = self.style("niveau")

#    def _affiche(self):
#        fen = self.__feuille__.fenetre() # il faut tracer depuis le bord de fenetre
#        p = self.pente(1)
#        A, B = self.x_y((fen[2], fen[0])[p],p), self.x_y((fen[3], fen[1])[p],p)
#        self.__canvas__().Ligne(A, B, self.style("couleur"))

    def _distance_inf(self, x, y, d):
        # cf. "distance_point_segment.odt" dans "doc/developpeurs/maths/"
        xA, yA = self.point1.position()
        xB, yB = self.point2.position()
        return ((yA-yB)*(x-xA)+(xB-xA)*(y-yA))**2/((xB-xA)**2+(yB-yA)**2) < d**2


    """def _distance_inf(self, x, y, d):
        x0, y0 = self.point1.position()
        x1, y1 = self.point2.position()
        A = Point(x0, y0) ; B = Point(x1, y1) ; M = Point(x, y)
        AB = Droite(A, B)
        return Projete_droite(M, AB).distance() < d"""

    def __contains__(self, M):
        #if not isinstance(M, Point_generique):
        #    return False
        A = self.point1
        B = self.point2
        xu, yu = self._vec(A, B)
        xv, yv = self._vec(A, M)
        return abs(xu*yv-xv*yu) < param.tolerance

    def mise_en_equation(self):
        u"Equation sous forme lisible par l'utilisateur."
        eq = self.equation()
        if eq is None:
            return u"L'objet n'est pas dfini."
        a, b, c = (str(nchiffres(coeff, param.chiffres_significatifs)) for coeff in eq)
        return "%s*x + %s*y + %s = 0" %(a, b, c) # on ne garde "que" 7 chiffres significatifs




class Droite_vectorielle(Droite):
    u"""Une droite vectorielle.

    Une droite dfinie par un vecteur directeur (sous forme de bipoint, ou de vecteur, ou ...), et un point.
    NOTE : convertir en bipoint, puis en vecteur, permet d'accepter des droites comme argument."""

    def __init__(self, vecteur, point, **kw):
        vecteur = self.lign2bip(vecteur) # conversion en bipoint
        vecteur = Vecteur(vecteur[0], vecteur[1])  # conversion en vecteur
        Droite.__init__(self, point1 = point, point2 = Point_final(point, [vecteur]))
        self.enregistre(Vecteur_generique, Point_generique)



class Parallele(Droite_vectorielle):
    u"""Une parallle.

    Une droite parallle  ... (droite, bipoint, vecteur...) passant par le point ...
    C'est presque un alias de Droite_vectorielle."""

    def __init__(self, droite, point, **kw):
        droite = self.lign2bip(droite) # conversion en bipoint
        droite = Droite(droite[0], droite[1])  # conversion en droite
        Droite.__init__(self, point1 = point, point2 = Point_final(point, [Vecteur(droite.point1, droite.point2)]))
        self.enregistre(Droite, Point_generique)



class Droite_rotation(Droite):
    u"""Une image d'une droite par rotation.

    Une droite obtenue par rotation d'une autre droite, ou d'un bipoint, ou..."""

    def __init__(self, droite, rotation, **kw):
        droite = self.lign2bip(droite) # conversion en bipoint
        droite = Droite(droite[0], droite[1])  # conversion en droite
        Droite.__init__(self, point1 = Point_rotation(droite.point1, rotation), point2 = Point_rotation(droite.point2, rotation))
        self.enregistre(Droite, Rotation)



class Mediatrice(Droite_rotation):
    u"""Une mdiatrice.

    La mdiatrice d'un segment (ou d'un bipoint, ...)"""

    def __init__(self, objet1, objet2 = None, **kw):
        if objet2 == None:
            objet1, objet2 = self.lign2bip(objet1) # a la place de deux objets, on peut entrer un segment
        Droite_rotation.__init__(self, droite = (objet1, objet2), rotation = Rotation(Milieu(objet1, objet2), 90., "d"))
        self.enregistre(Point_generique, Point_generique)



class Perpendiculaire(Droite_vectorielle):
    u"""Une perpendiculaire.

    Une droite perpendiculaire  une autre passant par un point."""

    def __init__(self, droite, point, **kw):
        Droite_vectorielle.__init__(self, vecteur = Droite_rotation(droite = droite, rotation = Rotation(Point(0., 0.), 90., "d")), point= point)
        self.enregistre((Droite, Segment, Vecteur_generique, Demidroite), Point_generique)



class Droite_equation(Droite):
    u"""Une droite dfinie par une quation.

    Une droite d'quation donne sous forme d'un triplet (a, b, c). (ax + by + c = 0)"""

    def __init__(self, coeffs = None, **kw):
        Droite.__init__(self, point1 = Point(0, 0), point2 = Point(1, 0))
        self.enregistre(tuple)
        self.equation(coeffs, False)


    def verifier_coeffs(self, coeffs = []):
        if len(coeffs) <> 3:
            self.erreur(u"nombre de coefficients incorrect.")
        if coeffs[0] == coeffs[1] == 0:
            self.erreur(u"les deux premiers coefficients sont nuls.")


    def _get_equation(self):
        # Retourne un triplet (a, b, c), tel que ax + by + c = 0 soit une equation de droite de la ligne.
        # On peut aussi modifier l'equation de la droite.
        xA, yA = self.point1.coordonnees()
        xB, yB = self.point2.coordonnees()
        return (yA - yB, xB - xA, xA*yB - yA*xB)

    def _set_equation(self, coeffs = None):
        if coeffs is not None:
            self.recalculer = True
            self.verifier_coeffs(coeffs)
            a, b, c = coeffs
            self.coeffs = coeffs
            if b:
                xA = 0; yA = -c/b
                xB = 1; yB = -a/b - c/b
            else:
                xA = xB = -c/a
                yA = 0 ; yB = 1
            self.point1.coordonnees(xA, yA)
            self.point2.coordonnees(xB, yB)



class Bissectrice(Mediatrice):
    u"""Une bissectrice.

    La bissectrice d'un angle dfini par 3 points.
    On peut, au choix, entrer un angle ou 3 points comme argument."""

    def __init__(self, pt1, pt2 = None, pt3 = None, **kw):
        if pt2 == None: # point1 est un angle
            pt2 = pt1.point2
            pt3 = pt1.point3
            pt1 = pt1.point1


        A = pt2 + Vecteur_unitaire(Vecteur(pt2, pt1))
        B = pt2 + Vecteur_unitaire(Vecteur(pt2, pt3))
        Mediatrice.__init__(self, A, B)
        #del A, B
        self.enregistre(Point_generique, Point_generique, Point_generique)

    def _conditions_existence(self):
        # EDIT: les conditions d'existence en amont sont dsormais bien dfinies (ou devraient l'tre !!)
        # if None in (self.pt1.coo(), self.pt2.coo(), self.pt3.coo()):
        #     return [False]
        return [   sum(abs(self.pt1.coo() - self.pt2.coo())) > param.tolerance,
                    sum(abs(self.pt2.coo() - self.pt3.coo())) > param.tolerance,
                    sum(abs(self.pt1.coo() - self.pt3.coo())) > param.tolerance]






class Tangente(Perpendiculaire):
    u"""Une tangente.

    Une des deux tangente  un cercle passant par un point extrieur au cercle.
    Le dernier paramtre (True/False) sert  distinguer les deux tangentes.
    (Voir la classe Intersection_cercles pour plus d'infos)."""

    def __init__(self, cercle, pt, angle_positif = None, **kw):
        self.synchroniser_feuille()
        if self.__feuille__:            # si l'on cre 2 fois de suite un tangente de mmes cercle et point,
            if angle_positif == None:   # alors on doit obtenir les deux tangentes diffrentes possibles.
                tangentes = [objet for objet in self.__feuille__.objets() if isinstance(objet, Tangente) and objet.cercle is cercle and objet.pt is pt]
                if tangentes:
                    angle_positif = not tangentes[0].angle_positif # on cree l'autre tangente
                else:
                    angle_positif = True
                del tangentes
        else:
            angle_positif = True

        # on a deux cas :
        # si le point est sur le cercle, on prend le point (Attention  bien prvoir tous les cas)
        # si le point n'est pas sur le cercle, on construit une intersection de cercles

        # la dichotomie n'est pas sensas, car il se peut que M soit tres prs du cercle...
        if isinstance(pt, Glisseur_cercle) and pt.cercle is cercle or isinstance(cercle, Cercle_diametre) and (cercle.point1 is pt or cercle.point2 is pt) or isinstance(cercle, Cercle_rayon) and cercle.point is pt:
            point_tangence = pt
        else:
            point_tangence = Intersection_cercles(cercle, Cercle_diametre(cercle.centre, pt), angle_positif)
        Perpendiculaire.__init__(self, Droite(cercle.centre, point_tangence), point_tangence)
        self.enregistre(Cercle_generique, Point_generique)

    def _conditions_existence(self):
        return [self.point_tangence.existe() and Perpendiculaire._conditions_existence(self)]
        # EDIT: les conditions d'existence en amont sont dsormais bien dfinies (ou devraient l'tre !!)
        # return [self.point1.coo() is not None and self.point2.coo() is not None and sum(abs(self.point1.coo() - self.point2.coo())) > param.tolerance]
        # si les conditions d'existence en amont sont mal definies, il se peut que les coordonnees d'un point valent None


##########################################################################################

# Vecteurs


class Vecteur_generique(Objet):
    u"""Un vecteur gnrique.

    Usage interne : la classe mre pour les diffrents vecteurs"""

    def __init__(self, **kw):
        Objet.__init__(self)

    def _get_coordonnees(self):
        return self.point2.coo()- self.point1.coo()

    def abscisse(self):
        return self.coordonnees()[0]

    def ordonnee(self):
        return self.coordonnees()[1]


    def norme(self):
        return norm2(self.coo())

    def __add__(self, y):
        if isinstance(y, Vecteur_generique):
            return Somme_vecteurs([self, y])
        raise TypeError, "vecteur attendu"

    def __sub__(self, y):
        return self + (-y)

    def __mul__(self, y):
        if isinstance(y, Vecteur_generique):
            return Produit_scalaire(self, y)
        else:
            return NotImplemented

    def __rmul__(self, y):
        return Somme_vecteurs([self], [y])

    _call = Objet.coordonnees

    def __neg__(self, *args):
        return (-1)*self

    def __div__(self, y):
        return (1./y)*self

    def __truediv__(self, y):
        return self.__div__(y)

    def __eq__(self, y):
        return isinstance(y, Vecteur_generique) and tuple(self.coordonnees()) == tuple(y.coordonnees())

    def __nonzero__(self):
        return tuple(self.coordonnees()) <> (0, 0)

    def __ne__(self, y):
        return not self == y



class Vecteur_rotation(Point_rotation, Vecteur_generique): # en pratique, c'est presque un alias de Point_rotation...
    u"""Une image d'un vecteur par une rotation.

    Vecteur construit  partir d'un autre via une rotation d'angle et de centre donn."""

    def __init__(self, vecteur, rotation, **kw):
        Point_rotation.__init__(self, point = vecteur, rotation = rotation)
        self.enregistre(Vecteur_generique, Rotation)

    def _espace_vital(self):
        pass

    def _affiche(self):  # pour que le vecteur ne soit pas affich comme point
        pass


class Vecteur(Segment, Vecteur_generique):
    u"""Un vecteur.

    Un vecteur dfini par deux points."""

    def _affiche(self):
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().plot(), self.__canvas__().plot(), self.__canvas__().plot()]
        plot1, plot2, plot3 = self.__representation__
        x1, y1 = self.point1.coordonnees()
        x2, y2 = self.point2.coordonnees()
        A = array(self.coo2pix((x1, y1)))
        B = array(self.coo2pix((x2, y2)))
        k = self.style("position")
        G = k*B + (1-k)*A   # par dfaut, k = 1, et la flche est positionne  l'extrmit du vecteur (point B)

        #                                                 M
        #                                               C +--\
        # l'ideal est ici un dessin:   A +---------------+----+ G=B
        #                                                 +--/
        #                                                 N
        # On calcule les coordonnees de C, puis par rotation autour de G, celles de M et N.

        r, a = self.style("taille"), self.style("angle")*pi/360
        C = G + r*(A - B)/norm2(A - B) # prendre B et non G, au cas o G = A !
        M = (dot([[cos(a), -sin(a)], [sin(a), cos(a)]], transpose([C - G])) + transpose([G]))[:,0]
        N = (dot([[cos(a), sin(a)], [-sin(a), cos(a)]], transpose([C - G])) + transpose([G]))[:,0]
        x, y = self.pix2coo(zip(M, G, N))
        plot1._x = array([x1, x2]) ; plot1._y = array([y1, y2])
        plot2._x = array(x) ; plot2._y = array(y)

        if self.style("double_fleche"):
            plot3._visible = True
            G = k*A + (1-k)*B   # par dfaut, k = 1, et la flche est positionne  l'extrmit du vecteur (point B)

            #                                                 M
            #                                               C +--\
            # l'ideal est ici un dessin:   B +---------------+----+ G=A
            #                                                 +--/
            #                                                 N
            # On calcule les coordonnees de C, puis par rotation autour de G, celles de M et N.

            r, a = self.style("taille"), self.style("angle")*pi/360
            C = G + r*(B - A)/norm2(B - A) # prendre B et non G, au cas o G = A !
            M = (dot([[cos(a), -sin(a)], [sin(a), cos(a)]], transpose([C - G])) + transpose([G]))[:,0]
            N = (dot([[cos(a), sin(a)], [-sin(a), cos(a)]], transpose([C - G])) + transpose([G]))[:,0]
            x, y = self.pix2coo(zip(M, G, N))
            plot3._x = array(x) ; plot3._y = array(y)
        else:
            plot3._visible = False

        plot1._color = plot2._color = plot3._color = self.style("couleur")
        plot1._linestyle = plot2._linestyle = plot3._linestyle = self.style("style")
        plot1._linewidth = plot2._linewidth = plot3._linewidth = self.style("epaisseur")
        plot1.zorder = plot2.zorder = plot3.zorder = self.style("niveau")


    def __init__(self, point1, point2, **kw):
        Segment.__init__(self, point1 = point1, point2 = point2)
        self.style(**param.vecteurs)
        Vecteur_generique.__init__(self)
        self.enregistre(Point_generique, Point_generique)





class Vecteur_unitaire(Vecteur_generique):
    u"""Un vecteur unitaire.

    Un vecteur dfini en normalisant un autre vecteur.
    Il aura donc mme sens et mme direction, mais sera de norme 1."""

    def __init__(self, vecteur, **kw):
        Vecteur_generique.__init__(self)
        self.enregistre(Vecteur_generique)

    def _get_coordonnees(self):
        return (self.vecteur.abscisse()/self.__norme, self.vecteur.ordonnee()/self.__norme)

    def _conditions_existence(self):
        self.__norme = self.vecteur.norme()
        return [self.__norme > param.tolerance]



class Vecteur_libre(Vecteur):
    u"""Un vecteur libre.

    Un vecteur dfini par ses coordonnes."""

    def __init__(self, x = None, y = None, **kw):
        if x == None:
            self.synchroniser_feuille()
            if self.__feuille__:
                xmin, xmax, ymin, ymax = self.__canvas__().fen()
                x = random.uniform(xmin,xmax)
                y = random.uniform(ymin,ymax)
                del xmin, xmax, ymin, ymax # evite d'enregistrer les variables xmin, xmax...
            else:
                x = y = 0

        x, y = self.nb2var(x), self.nb2var(y)
        # on convertit l'entree en Variable (voir la definition de la Classe Variable en fin de page)
        # l'entree peut-etre un nombre, un(e longueur de) segment, un (rayon de) cercle
        Vecteur.__init__(self, point1 = Point(0., 0.), point2 = Point(x, y))
        self.enregistre(Variable, Variable)

    def _affiche(self):
        pass  # les vecteurs libres ne sont pas affiches.

    def _get_coordonnees(self):
        return self.point2._get_coordonnees()

    def _set_coordonnees(self, x = None, y = None):
        self.recalculer = True
        return self.point2._set_coordonnees(x, y)



class Representant(Vecteur):
    u"""Un reprsentant d'un vecteur.

    Un reprsentant d'un vecteur, ayant pour origine un point donn."""
    def __init__(self, vecteur, point = None, **kw):
        if point is None:
            point = Point(0, 0)
        Vecteur.__init__(self, point, Point_final(point, [vecteur]))
        self.enregistre(Vecteur_generique, Point_generique)
        if self.__feuille__ and not self.style("previsualisation"):
            if self.point2 not in self.__feuille__.objets():
                noms = self.style("noms")
                if noms is None:
                    noms = {"point_final": ''}
                self.point2._enregistrer = False # objet cr automatiquement
                self.point2.style(legende = wxgeo.RIEN)
                setattr(self.__feuille__.objets, noms["point_final"] or self.__feuille__.nom_alea("M"), self.point2)
                self.style(noms = {"point_final": Nom(self.point2)})
                # ce style spcial sert  bien regnrer les mmes noms aprs une sauvegarde.






class Somme_vecteurs(Vecteur_generique):
    u"""Une somme de vecteurs.

    Les arguments sont des couples coefficient, vecteur
    Exemple: Somme_vecteurs([u,v,A>B], [3,2,-5])."""

    def __init__(self, vecteurs = [], coeffs = None, **kw):

        if coeffs == None:
            coeffs = [1 for vecteur in vecteurs]
        coeffs = map(self.nb2var, coeffs)

        _coeffs_, _vecteurs_ = [], [] # le code suivant evite de multiplier les sommes de vecteurs imbriquees
        for i in range(len(vecteurs)):
            if isinstance(vecteurs[i], Somme_vecteurs):
                _coeffs_.extend(list(coeffs[i].mult(k) for k in vecteurs[i].coeffs))
                _vecteurs_.extend(vecteurs[i].vecteurs)
            else:
                _coeffs_.append(coeffs[i])
                _vecteurs_.append(vecteurs[i])
        coeffs, vecteurs = _coeffs_, _vecteurs_

        Vecteur_generique.__init__(self)
        self.enregistre([Vecteur_generique], [Variable])


    def _get_coordonnees(self):
        return sum([self.coeffs[i].val()*self.vecteurs[i].coo() for i in range(len(self.coeffs))])





##########################################################################################

# Cercles

class Arc_generique(Objet):
    u"""La classe mre de tous les arcs de cercles."""

    def __init__(self, **kw):
        Objet.__init__(self)
        self.vecteur_unite = Vecteur_libre(1, 0)
        self.style(**param.arcs)
        self.etiquette = Label_arc_cercle(self)

    def rayon(self):
        try:
            return self._distance(self.centre, self.point1)
        except TypeError:
            return 0

    def diametre(self):
        return 2*self.rayon()

    def _get_equation(self):
        # retourne a, b, c tels que x**2 + y**2 + ax + by + c = 0
        a, b = self.centre(); c = self.rayon()
        return -2*a, -2*b, a**2 + b**2 - c**2


    def mise_en_equation(self):
        u"Equation sous forme lisible par l'utilisateur."
        eq = self.equation()
        if eq is None:
            return u"L'objet n'est pas dfini."
        a, b, c = (str(nchiffres(coeff, param.chiffres_significatifs)) for coeff in eq) # on ne garde "que" 7 chiffres significatifs
        xmin, xmax, ymin, ymax = (str(nchiffres(extremum, param.chiffres_significatifs)) for extremum in self._espace_vital())
        return "x^2 + y^2 + %s*x + %s*y + %s = 0 avec %s<x<%s et %s<y<%s" %(a, b, c, xmin, xmax, ymin, ymax)

    def _distance_inf(self, x, y, d):
        x0, y0 = self.centre.position()
        a, b = self._intervalle()
        t = arange(a, b, .003)
        rx, ry = [self.rayon()/self.__canvas__().coeff(i) for i in (0, 1)]
        u = x - x0 - rx*cos(t)
        v = y - y0 + ry*sin(t)
        m = min(u*u + v*v) # u*u est beaucoup plus rapide que u**2 (sic!)
        return m < d**2


    def _intervalle(self):
        u"Renvoie deux nombres a < b. L'arc est l'ensemble des points (r*cos(t), r*sin(t)) pour t apartenant  [a, b]."
        pass

    def _sens(self):
        return 1

    def _longueur(self):
        a, b = self._intervalle()
        return (b-a)*self.rayon()

    def longueur(self):
        try:
            return self._longueur()
        except TypeError:
            return 0

    def _affiche(self):
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().plot(), self.__canvas__().plot(), self.__canvas__().plot(), self.__canvas__().plot()]

        plot = self.__representation__[0]
        x, y = self.centre()
        r = self.rayon()
        a, b = self._intervalle()
        t = concatenate((arange(a, b, self.__canvas__().pas()), [b]))
        plot._x = x + r*cos(t)
        plot._y = y + r*sin(t)
        plot._color = self.style("couleur")
        plot._linestyle = self.style("style")
        plot._linewidth = self.style("epaisseur")
        niveau = self.style("niveau")
        plot.zorder = niveau

        for elt_graphique in self.__representation__:
            elt_graphique._visible = True

        # Gestion du codage des arcs de cercle (utilis pour indiquer les arcs de cercles de mme longeur)
        if not self.style("codage"):
            for plot in self.__representation__[1:]:
                plot._visible = False
                plot.zorder = niveau + 0.01

        else:
            c = .5*(a + b)
            x0 = x + r*cos(c); y0 = y + r*sin(c)

            if self.style("codage") == "o":
                plot = self.__representation__[1]
                plot._marker = "o"
                plot._markersize = param.codage["taille"]
                plot._markerfacecolor = None
                plot._markeredgecolor = self.style("couleur")
                plot._markeredgewidth = self.style("epaisseur")
                plot._x = [x0]
                plot._y = [y0]
                for plot in self.__representation__[2:]:
                    plot._visible = False

            else:
                r, a = param.codage["taille"], param.codage["angle"]
                G = array(self.coo2pix((x0, y0)))
                coeff0, coeff1 = self.__canvas__().coeffs()
                vecteur = array(((y0 - y)/coeff0, (x0 - x)/coeff1)) # vecteur orthogonal au rayon
                vec = vecteur/norm2(vecteur)
                C = G + r*vec
                if self.style("codage") == "x":
                    M = (dot([[cos(a), -sin(a)], [sin(a), cos(a)]], transpose([C - G])) + transpose([G]))[:,0]
                    N = (dot([[cos(a), sin(a)], [-sin(a), cos(a)]], transpose([C - G])) + transpose([G]))[:,0]
                    plot1 = self.__representation__[1]
                    plot2 = self.__representation__[2]
                    plot1._x, plot1._y = self.pix2coo(zip(M, 2*G-M))
                    plot2._x, plot2._y = self.pix2coo(zip(N, 2*G-N))
                    plot1._marker = plot2._marker = 'None'
                    plot1._color = plot2._color = self.style("couleur")
                    plot1._linewidth = plot2._linewidth = self.style("epaisseur")
                    self.__representation__[3]._visible = False

                elif self.style("codage") == "/":
                    M = (dot([[cos(a), sin(a)], [-sin(a), cos(a)]], transpose([C - G])) + transpose([G]))[:,0]
                    plot1 = self.__representation__[1]
                    plot1._x, plot1._y = self.pix2coo(zip(M, 2*G-M))
                    plot1._marker = 'None'
                    plot1._color = self.style("couleur")
                    plot1._linewidth = self.style("epaisseur")
                    for plot in self.__representation__[2:]:
                        plot._visible = False

                elif self.style("codage") == "//":
                    M = (dot([[cos(a), sin(a)], [-sin(a), cos(a)]], transpose([C - G])) + transpose([G]))[:,0]
                    plot1 = self.__representation__[1]
                    plot2 = self.__representation__[2]
                    vec = 2*vec
                    plot1._x, plot1._y = self.pix2coo(zip(M + vec, 2*G - M + vec))
                    plot2._x, plot2._y = self.pix2coo(zip(M - vec, 2*G - M - vec))
                    plot1._marker = plot2._marker = 'None'
                    plot1._color = plot2._color = self.style("couleur")
                    plot1._linewidth = plot2._linewidth = self.style("epaisseur")
                    self.__representation__[3]._visible = False

                elif self.style("codage") == "///":
                    M = (dot([[cos(a), sin(a)], [-sin(a), cos(a)]], transpose([C - G])) + transpose([G]))[:,0]
                    plot1 = self.__representation__[1]
                    plot2 = self.__representation__[2]
                    plot3 = self.__representation__[3]
                    vec = 3*vec
                    plot1._x, plot1._y = self.pix2coo(zip(M, 2*G - M))
                    plot2._x, plot2._y = self.pix2coo(zip(M + vec, 2*G - M + vec))
                    plot3._x, plot3._y = self.pix2coo(zip(M - vec, 2*G - M - vec))
                    plot1._marker = plot2._marker = plot3._marker = 'None'
                    plot1._color = plot2._color = plot3._color = self.style("couleur")
                    plot1._linewidth = plot2._linewidth = plot3._linewidth = self.style("epaisseur")


    def _espace_vital(self):
        x0, y0 = self.centre.coordonnees()
        a, b = self._intervalle()
        t = arange(a, b, .003)
        r = self.rayon()
        u = x0 + r*cos(t)
        v = y0 + r*sin(t)
        return min(u), max(u), min(v), max(v)


    def __contains__(self, M):
        O = self.centre
        a, b = self._intervalle()
        c = angle_vectoriel(self.vecteur_unite, self._vec(O, M))
        if c < a:
            c += 2*pi
        return abs(self._distance(O, M) - self.rayon()) < 5*param.tolerance and a < c < b



class Arc_cercle(Arc_generique):
    u"""Un arc de cercle.

    Un arc de cercle orient, dfini par son centre et ses extremits, dans le sens direct."""


    def __init__(self, centre, point1, point2, **kw):
        Arc_generique.__init__(self)
        angle1 = Angle_vectoriel(self.vecteur_unite, Vecteur(centre, point1))
        angle2 = Angle_vectoriel(self.vecteur_unite, Vecteur(centre, point2))
        #angle_au_centre = Angle_vectoriel(u, v)
        self.enregistre(Point_generique, Point_generique, Point_generique)



    def _intervalle(self):
        a = self.angle1.val()
        b = self.angle2.val()
        if b < a:
            b += 2*pi
        return a, b







class Arc_points(Arc_generique):
    u"""Un arc dfini par 3 points.

    Un arc de cercle, dfini par ses extrmits, et un autre point."""


    def __init__(self, point1, point2, point3, **kw):
        Arc_generique.__init__(self)
        centre = Point_equidistant(point1, point2, point3)
        angle1 = Angle_vectoriel(self.vecteur_unite, Vecteur(centre, point1))
        angle2 = Angle_vectoriel(self.vecteur_unite, Vecteur(centre, point2))
        angle3 = Angle_vectoriel(self.vecteur_unite, Vecteur(centre, point3))
        self.enregistre(Point_generique, Point_generique, Point_generique)


    def _intervalle(self):
        a = self.angle1.val() # mesure des angles dans ]-pi;pi]
        b = self.angle2.val()
        c = self.angle3.val()
        if b < a:
            b += 2*pi
        if c < a:
            c += 2*pi
        if b < c:
            return a, c
        return c, a + 2*pi

    def _sens(self):
        u"Sens de parcours de l'arc : direct (1) ou indirect (-1)"
        a = self.angle1.val() # mesure des angles dans ]-pi;pi]
        b = self.angle2.val()
        c = self.angle3.val()
        if b < a:
            b += 2*pi
        if c < a:
            c += 2*pi
        if b < c:
            return 1
        return -1


    def _conditions_existence(self):
        return [self.centre.existe()]




class Arc_oriente(Arc_points):
    u"""Un arc de cercle orient.

    Un arc de cercle orient, dfini par ses extrmits, et un autre point."""


    def __init__(self, point1, point2, point3, **kw):
        Arc_points.__init__(self, point1, point2, point3)
        self.style(**param.arcs_orientes)
        self.enregistre(Point_generique, Point_generique, Point_generique)

    def sens(self):
        if self._sens() is 1:
            return "direct"
        return "indirect"

    def _affiche(self):
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().plot(), self.__canvas__().plot(), self.__canvas__().plot()]
        plot1, plot2, plot3 = self.__representation__

        x, y = self.centre()
        r = self.rayon()
        a, b = self._intervalle()
        t = concatenate((arange(a, b, self.__canvas__().pas()), [b]))
        plot1._x = x + r*cos(t)
        plot1._y = y + r*sin(t)

        # Dessin de la flche (cf. le code pour la classe Vecteur, et les explications qui y sont donnes)
        k = self.style("position")
        coeff0, coeff1 = self.__canvas__().coeffs()
        if self._sens() is 1: # vecteur donnant le sens de la flche (orthogonal au rayon)
            c = k*b + (1-k)*a
            x0 = x + r*cos(c); y0 = y + r*sin(c)
            vec = array(((y0 - y)/coeff0, (x0 - x)/coeff1))
        else:
            c = (1-k)*b + k*a
            x0 = x + r*cos(c); y0 = y + r*sin(c)
            vec = array(((y - y0)/coeff0, (x - x0)/coeff1))

        taille, ang = self.style("taille"), self.style("angle")*pi/360
        G = array(self.coo2pix((x0, y0)))
        C = G + taille*vec/norm2(vec)
        M = (dot([[cos(ang), -sin(ang)], [sin(ang), cos(ang)]], transpose([C - G])) + transpose([G]))[:,0]
        N = (dot([[cos(ang), sin(ang)], [-sin(ang), cos(ang)]], transpose([C - G])) + transpose([G]))[:,0]
        xx, yy = self.pix2coo(zip(M, G, N))
        plot2._x = array(xx) ; plot2._y = array(yy)

        if self.style("double_fleche"):
            plot3._visible = True
            if self._sens() is 1: # vecteur donnant le sens de la flche (orthogonal au rayon)
                c = (1-k)*b + k*a
                x0 = x + r*cos(c); y0 = y + r*sin(c)
                vec = array(((y - y0)/coeff0, (x - x0)/coeff1))
            else:
                c = k*b + (1-k)*a
                x0 = x + r*cos(c); y0 = y + r*sin(c)
                vec = array(((y0 - y)/coeff0, (x0 - x)/coeff1))

            G = array(self.coo2pix((x0, y0)))
            C = G + taille*vec/norm2(vec)
            M = (dot([[cos(ang), -sin(ang)], [sin(ang), cos(ang)]], transpose([C - G])) + transpose([G]))[:,0]
            N = (dot([[cos(ang), sin(ang)], [-sin(ang), cos(ang)]], transpose([C - G])) + transpose([G]))[:,0]
            xx, yy = self.pix2coo(zip(M, G, N))
            plot3._x = array(xx) ; plot3._y = array(yy)
        else:
            plot3._visible = False

        plot1._color = plot2._color = plot3._color = self.style("couleur")
        plot1._linestyle = plot2._linestyle = plot3._linestyle = self.style("style")
        plot1._linewidth = plot2._linewidth = plot3._linewidth = self.style("epaisseur")
        plot1.zorder = plot2.zorder = plot3.zorder = self.style("niveau")








class Demicercle(Arc_cercle):
    u"""Un demi-cercle.

    Un demi-cercle orient, dfini par ses extrmits, dans le sens direct."""


    def __init__(self, point1, point2, **kw):
        Arc_cercle.__init__(self, Milieu(point1, point2), point1, point2)
        self.enregistre(Point_generique, Point_generique, Point_generique)





class Cercle_generique(Objet):
    u"""Un cercle gnrique.

    Usage interne : la classe mre pour tous les types de cercles."""


    def __init__(self, **kw):
        Objet.__init__(self)
        self.style(**param.cercles)
        self.etiquette = Label_cercle(self)


    def _affiche(self):
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().plot()]

        plot = self.__representation__[0]
        x, y = self.centre()
        r = self.rayon()
        t = concatenate((arange(0, 2*pi , self.__canvas__().pas()), [0]))
        plot._x = x + r*cos(t)
        plot._y = y + r*sin(t)
        plot._color = self.style("couleur")
        plot._linestyle = self.style("style")
        plot._linewidth = self.style("epaisseur")
        plot.zorder = self.style("niveau")


    def perimetre(self):
        return 2*pi*self.rayon()

    _longueur = perimetre

    def diametre(self):
        return 2*self.rayon()

    def _get_equation(self):
        # retourne a, b, c tels que x**2 + y**2 + ax + by + c = 0
        a, b = self.centre(); c = self.rayon()
        return -2*a, -2*b, a**2 + b**2 - c**2

    def _distance_inf(self, x, y, d):
        x0, y0 = self.centre.position()
        t = arange(0, 2*pi, .003)
        rx, ry = [self.rayon()/self.__canvas__().coeff(i) for i in (0, 1)]
        u = x - x0 - rx*sin(t)
        v = y - y0 + ry*cos(t)
        m = min(u*u + v*v) # u*u est beaucoup plus rapide que u**2 (sic!)
        return m < d**2

    def __contains__(self, M):
        O = self.centre
        return abs(self._distance(O, M) - self.rayon()) < param.tolerance


    def _espace_vital(self):
        x, y = self.centre.coordonnees()
        r = self.rayon()
        return (x-r, x+r, y-r, y+r)


    def mise_en_equation(self):
        u"Equation sous forme lisible par l'utilisateur."
        eq = self.equation()
        if eq is None:
            return u"L'objet n'est pas dfini."
        a, b, c = (str(nchiffres(coeff, param.chiffres_significatifs)) for coeff in eq)
        return "x^2 + y^2 + %s*x + %s*y + %s = 0" %(a, b, c)  # on ne garde "que" 7 chiffres significatifs





class Cercle(Cercle_generique):
    u"""Un cercle.

    Un cercle dfini par son centre et son rayon."""

    #def rayon(self, ray = None):  # retourne le rayon (eventuellement apres l'avoir modifie)
    #    return self._rayon.val(ray)

    def __init__(self, centre = None, rayon = None, **kw):

        if rayon is None:
            self.synchroniser_feuille()
            if self.__feuille__:
                xmin, xmax, ymin, ymax = self.__canvas__().fen()
                rayon = .5*random.uniform(0, min(abs(xmin - xmax), abs(ymin - ymax)))
            else:
                rayon = 1
        rayon = self.nb2var(rayon)
        if centre is None:
            self.synchroniser_feuille()
            if self.__feuille__:
                xmin, xmax, ymin, ymax = self.__canvas__().fen()
                r = rayon()
                centre = Point(random.uniform(xmin + r, xmax - r), random.uniform(ymin + r, ymax - r))
            else:
                centre = Point(0, 0)
        Cercle_generique.__init__(self)
        self.enregistre(Point_generique, Variable)



class Cercle_rayon(Cercle_generique):
    u"""Un cercle dfini par un rayon.

    Un cercle dfini par un rayon [OM]."""

    def __init__(self, centre, point, **kw):
        Cercle_generique.__init__(self)
        self.enregistre(Point_generique, Point_generique)
        #self.centre = self.point1 = centre
        #self.point = self.point2 = point
        #self.rayon = Variable()

    def rayon(self):
        return self._distance(self.centre, self.point)



class Cercle_diametre(Cercle_generique):
    u"""Un cercle dfini par un diamtre.

    Un cercle dfini par un diamtre [AB]."""


    def __init__(self, point1, point2, **kw):
        Cercle_generique.__init__(self)
        centre = Milieu(point1, point2)
        self.enregistre(Point_generique, Point_generique)
        #self.point1 = point1
        #self.point2 = point2
        #self.centre = Milieu(point1, point2)
        #self.rayon = Variable(Segment(self.centre, point2))

    def rayon(self):
        return self._distance(self.centre, self.point2)



class Cercle_points(Cercle_generique):
    u"""Un cercle dfini par 3 points.

    Un cercle dfini par la donne de 3 points du cercle."""

    def __init__(self, point1, point2, point3, **kw):
        centre = Point_equidistant(point1, point2, point3)
        Cercle_generique.__init__(self)
        self.style(**param.cercles)
        self.enregistre(Point_generique, Point_generique, Point_generique)

    def rayon(self):
        try:
            return self._distance(self.centre, self.point1)
        except TypeError:
            return 0

    def _conditions_existence(self):
        return [self.centre.existe()]




class Cercle_equation(Cercle_generique):
    u"""Un cercle dfini par une quation.

    Un cercle d'quation donne sous forme d'un triplet (a, b, c). (x**2 + y**2 + ax + by + c = 0)"""

    def __init__(self, *coeffs, **kw):
        Cercle_generique.__init__(self)
        centre = Point(0, 0, fixe = True)
        self.enregistre(tuple)
        self.equation(coeffs, False)


    def verifier_coeffs(self, coeffs = []):
        # (x + a/4)**2 + (y + b/4)**2 = -c + a/4 + b/4
        if len(coeffs) <> 3:
            self.erreur(u"nombre de coefficients incorrect.")
        a, b, c = coeffs
        if a/4 + b/4 - c < 0:
            self.erreur(u"le rayon du cercle doit tre positif.")

    def rayon(self):
        a, b, c = self.coeffs
        return sqrt(a/4 + b/4 - c)


    def _get_equation(self, coeffs = None):
        # Retourne un triplet (a, b, c), tel que x**2 + y**2 + ax + by + c = 0 soit une equation du cercle.
        # On peut aussi modifier l'equation du cercle.
        return self.coeffs

    def _set_equation(self, coeffs = None):
        if coeffs is not None:
            self.recalculer = True
            self.verifier_coeffs(coeffs)
            self.coeffs = coeffs
            self.centre.coordonnees(-coeffs[0]/4, -coeffs[1]/4)




class Disque(Cercle_generique):
    u"""Un disque.

    Un disque dfini par le cercle le dlimitant."""

    def __init__(self, cercle, **kw):
        Cercle_generique.__init__(self)
        centre = cercle.centre
        self.style(**param.polygones)
        self.enregistre(Cercle_generique)

    def _affiche(self):
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().fill()]

        fill = self.__representation__[0]
        x, y = self.centre()
        r = self.rayon()
        t = concatenate((arange(0, 2*pi , self.__canvas__().pas()), [0]))
        fill.xy = zip(x + r*cos(t), y + r*sin(t))
        fill._alpha = self.style("alpha")
        fill._color = self.style("couleur")
        fill._linestyle = self.style("style")
        fill._linewidth = self.style("epaisseur")
        fill.zorder = self.style("niveau")


    def rayon(self):
        return self.cercle.rayon()

    def _distance_inf(self, x, y, d):
        x, y = self.pix2coo((x, y))
        return self._distance(self.centre, (x, y)) <= self.rayon()

    def __contains__(self, M):
        return self._distance(self.centre, M) <= self.rayon()

    def aire(self):
        return pi*self.rayon()**2


    def mise_en_equation(self):
        u"Equation sous forme lisible par l'utilisateur."
        eq = self.equation()
        if eq is None:
            return u"L'objet n'est pas dfini."
        a, b, c = (str(nchiffres(coeff, param.chiffres_significatifs)) for coeff in eq)  # on ne garde "que" 7 chiffres significatifs
        eq = "x + y + %s x + %s y + %s <= 0" %(a, b, c)
        return eq.replace("+ -", "-")



##########################################################################################

# Coniques (exprimental)

class Conique_generique(Objet):
    u"""Une conique gnrique.

    Usage interne : la classe mre pour tous les types de coniques."""

    def __init__(self, **kw):
        Objet.__init__(self)
        self.style(**param.cercles) # param.coniques
        # self.etiquette = Label_conique(self)


    def _affiche(self):
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().plot(), self.__canvas__().plot()]

        #plot = self.__representation__[0]
        #x, y = self.centre()
        #r = self.rayon()
        #t = concatenate((arange(0, 2*pi , self.__canvas__().pas()), [0]))
        #plot._x = x + r*cos(t)
        #plot._y = y + r*sin(t)
        #plot._color = self.style("couleur")
        #plot._linestyle = self.style("style")
        #plot._linewidth = self.style("epaisseur")
        #plot.zorder = self.style("niveau")



    #def perimetre(self):
    #    return 2*pi*self.rayon()

    #_longueur = perimetre

    #def diametre(self):
    #    return 2*self.rayon()


    #def _get_equation(self):
        # retourne a, b, c tels que x**2 + y**2 + ax + by + c = 0
    #    a, b = self.centre(); c = self.rayon()
    #    return -2*a, -2*b, a**2 + b**2 - c**2

    #def _distance_inf(self, x, y, d):
    #    x0, y0 = self.centre.position()
    #    t = arange(0, 2*pi, .003)
    #    rx, ry = [self.rayon()/self.__canvas__().coeff(i) for i in (0, 1)]
    #    u = x - x0 - rx*sin(t)
    #    v = y - y0 + ry*cos(t)
    #    m = min(u*u + v*v) # u*u est beaucoup plus rapide que u**2 (sic!)
    #    return m < d**2

    #def __contains__(self, M):
    #    O = self.centre
    #    return abs(self._distance(O, M) - self.rayon()) < param.tolerance


    #def _espace_vital(self):
    #    x, y = self.centre.coordonnees()
    #    r = self.rayon()
    #    return (x-r, x+r, y-r, y+r)


    def mise_en_equation(self):
        u"Equation sous forme lisible par l'utilisateur."
        eq = self.equation()
        if eq is None:
            return u"L'objet n'est pas dfini."
        a, b, c, d, e, f = (str(nchiffres(coeff, param.chiffres_significatifs)) for coeff in eq)
        return "%s*x^2 + %s*xy + %s*y^2 + %s*x + %s*y + %s = 0" %(a, b, c, d, e, f)  # on ne garde "que" 7 chiffres significatifs





class Conique(Conique_generique):
    u"""Une conique.

    Une conique dfinie par 5 points."""

    #def rayon(self, ray = None):  # retourne le rayon (eventuellement apres l'avoir modifie)
    #    return self._rayon.val(ray)

    def __init__(self, centre = None, rayon = None, **kw):
        Conique_generique.__init__(self)
        self.enregistre(Point_generique, Point_generique, Point_generique, Point_generique, Point_generique)

    def _get_equation(self):
        u"Retourne a, b, c, d, e, f tels que a*x**2 + b*xy + c*y**2 + d*x + e*y + f = 0, o f=0 ou 1."





##########################################################################################

# Divers

class Transformation(Objet):
    u"""Une transformation.

    La classe mre de toutes les transformations (usage interne)."""
    def __init__(self, **kw):
        Objet.__init__(self, **kw)

    def __call__(self, objet):
        if isinstance(objet, Point_generique):
            return self.Point_image(objet, self)
        elif isinstance(objet, Objet) and not isinstance(objet, (Objet_numerique, Transformation)):
            return objet.__class__(*(self(getattr(objet, attribut)) for attribut in objet.attributs))
        else:
            return objet


class Rotation(Transformation):
    u"""Une rotation.

    Une rotation dfinie par son centre, et un angle en radian (r), grad (g) ou degr (d)."""

    Point_image = Point_rotation

    def __init__(self, centre = None, angle = pi, unite = "r", **kw):
        # l'unite par defaut est le radian.
        if centre == None:
            centre = Point(0, 0)
        if not isinstance(angle, Angle_generique):
            angle = Angle_libre(angle, unite)
        #valeur = self.nb2var(valeur)
        Transformation.__init__(self, **kw)
        self.enregistre(Point_generique, Angle_generique, basestring)

    #def centre(self):
    #    return self._centre.coordonnees()

    #def angle(self, valeur = None, unite = None, auto_refresh = True):
    #    if unite is not None:
    #        self.unite = self.choix_unite(unite)
    #    return self.valeur(valeur, auto_refresh)

    def radian(self):
        return self.angle.radian()

    conversion = {"r":1., "g":pi/200., "d": pi/180.}

    @staticmethod
    def choix_unite(unite):
        u"Gestion des alias."
        if unite is None:
            return "r"
        if unite == u"" or unite.startswith("d"):
            return "d"
        if unite.startswith("g"):
            return "g"
        return "r"


# A DEVELOPPER (image d'un objet avec __call__) !!!





class Translation(Transformation):
    u"""Une translation.

    Une translation dfinie par un vecteur."""

    Point_image = Point_translation

    def __init__(self, vecteur, **kw):
        Transformation.__init__(self, **kw)
        if isinstance(vecteur, (tuple, list, types.ArrayType)):
            vecteur = Vecteur_libre(*vecteur)
        self.enregistre(Vecteur_generique)



class Reflexion(Transformation):
    u"""Une symtrie axiale.

    Une symtrie axiale (rflexion) dfinie par une droite."""

    Point_image = Point_reflexion

    def __init__(self, droite, **kw):
        Transformation.__init__(self, **kw)
        self.enregistre(Ligne)


class Homothetie(Transformation):
    u"""Une homothtie.

    Une homothtie dfinie par son centre et son rapport."""

    Point_image = Point_homothetie

    def __init__(self, centre = None, rapport = 1, **kw):
        Transformation.__init__(self, **kw)
        if centre == None:
            centre = Point(0, 0)
        rapport = self.nb2var(rapport)
        self.enregistre(Point_generique, Variable)



class Symetrie_centrale(Homothetie):
    u"""Une symtrie centrale.

    Une symtrie centrale dfinie par un point."""

    def __init__(self, centre = None, **kw):
        Homothetie.__init__(self, centre, **kw)
        self.enregistre(Point_generique)


class Produit_scalaire(Objet_numerique):
    u"""Un produit scalaire.

    Le produit scalaire de deux vecteurs"""

    def __init__(self, vecteur1, vecteur2, **kw):
        Objet.__init__(self)
        self.enregistre(Vecteur_generique, Vecteur_generique)

    def _get_val(self):
        return sum(self.vecteur1.coo()*self.vecteur2.coo())

    _call = Objet.val


# NOTE :
# A l'avenir, une autre methode pourrait etre d'implementer le produit scalaire comme multiplication directement dans l'objet vecteur (def __mul__(self, vecteur): etc...).
# On pourrait aussi imaginer que 2*un vecteur soit defini, ou encore vecteur + vecteur mais c'est un peu complique.
# Il y a differents types de vecteurs avec des dependances differentes, quel serait le type final ?



# A FAIRE :
# - GERER LES DIFFERENTES UNITES POUR LES ANGLES.
# - POUVOIR ENTRER UN ANGLE COMME ARGUMENT POUR LES ROTATIONS.

class Angle_generique(Objet_numerique):
    u"""Un angle gnrique.

    Classe mre des angles orients ou non. L'unit par dfaut est le radian."""

    def __init__(self, unite = "r", **kw):
        # l'unite par defaut est le radian.
        unite = self.choix_unite(unite)
        Objet.__init__(self)


    def _get_val(self):
        return angle_vectoriel(self.vecteur1, self.vecteur2)*self.conversion["r"][self.unite]


    _call = Objet.val


    def radian(self):
        return self.val()*self.conversion[self.unite]["r"]


    def degre(self):
        return self.val()*self.conversion[self.unite]["d"]


    def grad(self):
        return self.val()*self.conversion[self.unite]["g"]

    def sin(self):
        return math.sin(self.radian())

    def cos(self):
        return math.cos(self.radian())

    def tan(self):
        return math.tan(self.radian())

    def _conditions_existence(self):
        return [self.vecteur1.norme() > param.tolerance, self.vecteur2.norme() > param.tolerance]

    def changer_unite(self, unite = "r"):
        u"Change l'unit de l'angle en effectuant les adaptations ncessaires."
        self.recalculer = True
        for objet in self.heritiers():#classer = True):
            objet.recalculer = True
            if objet.etiquette:
                objet.etiquette.recalculer = True
        self.unite = self.choix_unite(unite)



    conversion = {  "r": {"r":1., "d":180./pi, "g":200./pi},
                    "g": {"r":pi/200., "d":180./200., "g":1.},
                    "d": {"r":pi/180., "d":1., "g":200./180.}}
    # utilisation: t*conversation["r"]["d"] pour convertir t de radian en degr.

    @staticmethod
    def choix_unite(unite):
        "Gestion des alias."
        if unite is None:
            return "r"
        if unite == u"" or unite.startswith("d"):
            return "d"
        if unite.startswith("g"):
            return "g"
        return "r"




class Secteur_angulaire(Angle_generique):
    u"Classe mre de tous les angles 'affichables'."
    def __init__(self, point, vecteur1, vecteur2, unite = "r", **kw):
        # l'unite par defaut est le radian.
        unite = self.choix_unite(unite)
        Angle_generique.__init__(self, unite)
        self.style(**param.angles)
        self.enregistre(Point_generique, Vecteur_generique, Vecteur_generique, basestring)
        self.etiquette = Label_angle(self)

    def _sens(self):
        x1, y1 = self.vecteur1
        x2, y2 = self.vecteur2
        return cmp(x1*y2-x2*y1, 0)

    def sens(self):
        if self._sens() < 0:
            return "indirect"
        return "direct"

    def _affiche(self):
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().fill(), self.__canvas__().plot(), self.__canvas__().plot(), self.__canvas__().plot()]

        for elt_graphique in self.__representation__:
            elt_graphique._visible = True

        x1, y1 = self.vecteur1
        xx, yy = self.point
        x2, y2 = self.vecteur2
        coeff0, coeff1 = self.__canvas__().coeffs()
        u = (x1/coeff0, y1/coeff1)
        v = (x2/coeff0, y2/coeff1)
        i = (1, 0)
        a = angle_vectoriel(i, u)
        b = angle_vectoriel(i, v)
        if isinstance(self, Angle) and self._sens() < 0:
            a, b = b, a
        if b<a:
            b += 2*pi
        r = param.codage["rayon"]
        x0, y0 = self.coo2pix((xx, yy))
        t = concatenate((arange(a, b, self.__canvas__().pas()), [b]))
        niveau = self.style("niveau")


        codage = self.style("codage")
        if not codage:
            for plot in self.__representation__[1:]:
                plot._visible = False
                plot.zorder = niveau
            x = self.__canvas__().Xpix2coo(x0 + r*cos(t))
            y = self.__canvas__().Ypix2coo(y0 - r*sin(t), -1)
        else:
            if codage == "^":
                u = .5*r*array(u)/norm2(u)
                v = .5*r*array(v)/norm2(v)
                plot = self.__representation__[1]
                x = self.__canvas__().Xpix2coo((x0 + u[0], x0 + u[0] + v[0], x0 + v[0]))
                y = self.__canvas__().Ypix2coo((y0 - u[1], y0 - u[1] - v[1], y0 - v[1]), -1)
                plot._x = x
                plot._y = y
                plot._color = self.style("couleur")
                plot._linestyle = self.style("style")
                plot._linewidth = self.style("epaisseur")
                plot._marker = 'None'
                for plot in self.__representation__[2:]:
                    plot._visible = False
                    plot._marker = 'None'
                    plot._color = self.style("couleur")
                    plot._linestyle = self.style("style")
                    plot._linewidth = self.style("epaisseur")

            else:
                n = len(codage)
                for i in xrange(n):
                    plot = self.__representation__[1+i]
                    if i == 0:
                        plot._x = x = self.__canvas__().Xpix2coo(x0 + (r-3*i)*cos(t))
                        plot._y = y = self.__canvas__().Ypix2coo(y0 - (r-3*i)*sin(t), -1)
                    else:
                        plot._x = self.__canvas__().Xpix2coo(x0 + (r-3*i)*cos(t))
                        plot._y = self.__canvas__().Ypix2coo(y0 - (r-3*i)*sin(t), -1)
                    plot._color = self.style("couleur")
                    plot._linestyle = self.style("style")
                    plot._linewidth = self.style("epaisseur")
                    plot._marker = 'None'
                for plot in self.__representation__[1+n:]:
                    plot._visible = False
                    plot._marker = 'None'
                    plot._color = self.style("couleur")
                    plot._linestyle = self.style("style")
                    plot._linewidth = self.style("epaisseur")

                if codage in ("/", "x", "o"):
                    c = .5*(a+b)
                    xc, yc = x0 + r*cos(c), y0 - r*sin(c)
                    taille = param.codage["taille"]
                    if codage == "o":
                        plot = self.__representation__[2]
                        plot._marker = "o"
                        plot._markersize = taille
                        plot._markerfacecolor = None
                        plot._markeredgecolor = self.style("couleur")
                        plot._markeredgewidth = self.style("epaisseur")
                        plot._x = [self.__canvas__().Xpix2coo(xc)]
                        plot._y = [self.__canvas__().Ypix2coo(yc, -1)]
                    else:
                        u = taille*(xc - x0)/r, taille*(yc- y0)/r
                        if codage == "/":
                            plot = self.__representation__[2]
                            xa, ya = self.pix2coo((xc + u[0], yc + u[1]))
                            xb, yb = self.pix2coo((xc - u[0], yc - u[1]))
                            plot._x = [xa, xb]
                            plot._y = [ya, yb]
                        elif codage == "x":
                            a = param.codage["angle"]
                            plot1 = self.__representation__[2]
                            plot2 = self.__representation__[3]
                            u = array(u)
                            G = array((xc, yc))
                            C = G + u
                            M = (dot([[sin(a), -cos(a)], [cos(a), sin(a)]], transpose([C - G])) + transpose([G]))[:,0]
                            N = (dot([[sin(a), cos(a)], [-cos(a), sin(a)]], transpose([C - G])) + transpose([G]))[:,0]
                            plot1._x, plot1._y = self.pix2coo(zip(M, 2*G-M))
                            plot2._x, plot2._y = self.pix2coo(zip(N, 2*G-N))


        fill = self.__representation__[0]
        fill.xy = zip(x, y) + [(xx, yy)]
        fill._alpha = self.style("alpha")
        fill._facecolor = self.style("couleur")
        fill._linestyle = self.style("style")
        fill._linewidth = self.style("epaisseur")
        fill.zorder = niveau - 0.01



    def _espace_vital(self):
        x, y = self.point.coordonnees()
        x1, y1, = self.vecteur1.coordonnees()
        x2, y2 = self.vecteur2.coordonnees()
        xmin = min(x, x + x1, x + x2)
        xmax = max(x, x + x1, x + x2)
        ymin = min(y, y + y1, y + y2)
        ymax = max(y, y + y1, y + y2)
        return xmin, xmax, ymin, ymax




    def _distance_inf(self, x, y, d):
        x0, y0 = self.point.position()
        #print (x - x0)**2 + (y - y0)**2
        if (x - x0)**2 + (y - y0)**2 > (param.codage["rayon"] + d)**2:
            return False
        x, y = self.pix2coo((x, y))
        x0, y0 = self.point.coordonnees()
        u = (x - x0, y - y0)
        if isinstance(self, Angle) and self._sens() < 0:
            a = angle_vectoriel(self.vecteur2, self.vecteur1)
            b = angle_vectoriel(self.vecteur2, u)
        else:
            a = angle_vectoriel(self.vecteur1, self.vecteur2)
            b = angle_vectoriel(self.vecteur1, u)
        #print a, b
        if a < 0:
            return not a < b < 0
        else:
            return 0 <= b <= a





class Angle_oriente(Secteur_angulaire):
    u"""Un angle orient.

    Un angle orient dfini par 3 points A, B, C -> angle (BA>, BC>)."""

    def __init__(self, point1, point2, point3, unite = "r", **kw):
        # l'unite par defaut est le radian.
        unite = self.choix_unite(unite)
        Secteur_angulaire.__init__(self, point2, Vecteur(point2, point1), Vecteur(point2, point3))
        self.enregistre(Point_generique, Point_generique, Point_generique)




class Angle(Secteur_angulaire):
    u"""Un angle.

    Un angle non orient, dfini par 3 points A, B, C -> angle /ABC\."""

    def __init__(self, point1, point2, point3, unite = "r", **kw):
        # l'unite par defaut est le radian.
        unite = self.choix_unite(unite)
        Secteur_angulaire.__init__(self, point2, Vecteur(point2, point1), Vecteur(point2, point3))
        self.enregistre(Point_generique, Point_generique, Point_generique)

    def _get_val(self):
        return abs(Angle_generique._get_val(self))


    def sens(self):
        return u"non dfini" # attention en cas de modification : cette valeur est utilise dans la classe Label_angle


class Angle_libre(Angle_generique):
    u"""Un angle libre.

    Un angle dfini par sa valeur."""

    def __init__(self, valeur = None, unite = "r", **kw):
        unite = self.choix_unite(unite)
        if valeur is None:
            valeur = pi/2

        valeur = self.nb2var(valeur)
        # on convertit l'entree en Variable (voir la definition de la Classe Variable en fin de page)
        # l'entree peut-etre un nombre, un(e longueur de) segment, un (rayon de) cercle
        Angle_generique.__init__(self, unite)
        self.enregistre(Variable, basestring)

    def _affiche(self):
        pass  # les angles libres ne sont pas affichs.

    def _set_val(self, val = None):
        self.recalculer = True
        self.valeur._set_val(val)

    def _get_val(self):
        return self.valeur._get_val()

    def _conditions_existence(self):
        return [True]

    def changer_unite(self, unite = "r"):
        u"Change l'unit de l'angle en effectuant les adaptations ncessaires."
        unite = self.choix_unite(unite)
        self.valeur(self.valeur.mult(conversion[self.unite][unite]))
        # le code ci-dessus n'est pas parfait. Il marche bien pour les variable contenant simplement une valeur numrique.
        # mais pour une variable style "A.x", des conversions successives de degr en radian vont faire qqch du style "((((A.x)*0.017453292519943295)*57.295779513082323)*0.017453292519943295)"
        # En effet, une variable de "type compos" ne sait pas simplifier son code interne pour l'instant.
        Angle_generique.changer_unite(self, unite)





class Angle_vectoriel(Angle_generique):
    u"""Un angle dfini par des vecteurs.

     Un angle orient dfini par deux vecteurs (non affich)."""
    def __init__(self, vecteur1, vecteur2, unite = "r", **kw):
        # l'unite par defaut est le radian.
        unite = self.choix_unite(unite)
        Angle_generique.__init__(self, unite)
        self.enregistre(Vecteur_generique, Vecteur_generique, basestring)






##########################################################################################

# Intersections

class Intersection_generique(Point_generique):
    u"""Une intersection gnrique.

    La classe mre des diffrents types d'intersections"""

    def __init__(self, objet1, objet2, **kw):
        Point_generique.__init__(self)
        self.enregistre(Objet, Objet)







class Intersection_droites(Intersection_generique):
    u"""une intersection de droites.

    L'intersection de 2 droites, segments, demi-droites..."""

    def __init__(self, droite1, droite2, **kw):
        Intersection_generique.__init__(self, objet1 = droite1, objet2 = droite2)
        self.enregistre((Droite, Segment, Demidroite), (Droite, Segment, Demidroite))


    def _conditions_existence(self):
        a, b, c = self.droite1.equation()
        d, e, f = self.droite2.equation()
        #self.__coeffs = a, b, c, d, e, f
        self.__determinant = a*e - b*d
        if abs(self.__determinant) <= param.tolerance:
            return [False]
        x = self.__x = (f*b - c*e)/self.__determinant
        y = self.__y = (d*c - a*f)/self.__determinant
        M = (x, y)
        return [isinstance(self.droite1, Droite) or M in self.droite1, isinstance(self.droite2, Droite) or M in self.droite2]


    def _get_coordonnees(self):
        return (self.__x, self.__y)



class Intersection_droite_cercle(Intersection_generique):
    u"""Une intersection d'une droite et d'un cercle.

    Un des deux points M et N d'intersection d'une droite et d'un cercle.
    Supposons qu'on le note M, et que la droite soit (AB).
    L'argument point_proche (optionnel) doit tre un des deux points de la droite (A par exemple).
    Il sert  indiquer "de quel cot" de la droite (AB) se trouve le point M.
    (Plus exactement, dans l'exemple choisi, MA + NB est minimal).
    """

    def __init__(self, droite, cercle, point_proche = None, **kw):
        # l'intersection de la droite et du cercle (si elle existe) est deux points (ventuellement confondus).
        # lorsque l'utilisateur cr deux fois deux suite un point d'intersection, ce ne doit pas tre le mme.
        self.synchroniser_feuille()
        if point_proche is None and self.__feuille__:
            intersections = [objet for objet in self.__feuille__.objets() if isinstance(objet, Intersection_droite_cercle) and objet.droite is droite and objet.cercle is cercle]
            if intersections: # une intersection existe dj, on va construire l'autre
                point_proche = intersections[0].point_lointain
            del intersections

        if point_proche is None:  # aucune intersection n'existait, ou bien on ne pouvait pas le savoir (pas de feuille de travail dfinie).
            # On construit une intersection qui existe actuellement, si possible.
            if Intersection_droite_cercle(droite, cercle, droite.point1).existe():
                point_proche = droite.point1
            else:
                point_proche = droite.point2

        self.point_lointain = (droite.point1 == point_proche) and droite.point2 or droite.point1
        Intersection_generique.__init__(self, objet1 = droite, objet2 = cercle)
        self.enregistre((Droite, Segment, Demidroite), (Cercle_generique, Arc_generique), Point_generique)



    def _get_coordonnees(self):
        return self.__sol1
        # on a dj calcul les coordonnes en cherchant les conditions d'existence

    def _get_intersections(self):
        u"""Retourne les coordonnes des deux points d'intersection.
        (Dans l'ordre: celles du point, puis celles du 2me point d'intersection).
        """
        return self.__sol1, self.__sol2


    def _conditions_existence(self):
        # Trouver les conditions d'existence revient pratiquement  chercher les coordonnes.
        # Tout le calcul de coordonnes se fait donc ici, pour simplifier.

        # Cas de l'intersection d'une tangente  un cercle avec "son" cercle :
        if isinstance(self.droite, Tangente) and self.droite.cercle is self.cercle:
            self.__sol1 = self.__sol2 = self.droite.point_tangence.coordonnees()
            return [True]



        a, b, c = self.droite.equation()
        d, e, f = self.cercle.equation()

        if b: # La droite n'est pas verticale
            m = -a/b; n = -c/b # equation de la droite : y = mx + n

            #p = Polynome([m**2 + 1, 2*n*m + d + e*m, n**2 + e*n + f])
            sols = racines(m**2 + 1, 2*n*m + d + e*m, n**2 + e*n + f)

            if len(sols) == 2: # 2 points d'intersection.
                x1, x2 = sols
                M = (x1, m*x1+n); N = (x2, m*x2+n)
            elif len(sols) == 1: # 1 seul point
                x = sols[0]
                M = self.__sol1 = self.__sol2 = (x, m*x+n)
                return [isinstance(self.droite, Droite) or M in self.droite, isinstance(self.cercle, Cercle_generique) or M in self.cercle]
            else:
                return [False] # intersection vide.

        else: # La droite est verticale
            x = -c/a
            #p = Polynome([1, e, x**2 + d*x + f])
            sols = racines(1, e, x**2 + d*x + f)

            if len(sols) == 2: # 2 points d'intersection.
                y1, y2 = sols
                M = (x, y1); N = (x, y2)
            elif len(sols) == 1:
                y = sols[0]
                M = self.__sol1 = self.__sol2 = (x, y)
                return [isinstance(self.droite, Droite) or M in self.droite, isinstance(self.cercle, Cercle_generique) or M in self.cercle]
            else:
                return [False] # intersection vide.

        # La partie la plus complexe consiste a distinguer les deux points d'intersection a l'aide du "point proche".
        # Pour comprendre ce qui s'y passe, il est conseille de faire un dessin !

        A, B = self.point_proche, self.point_lointain
        u = self._vec(A, B)
        if produit_scalaire(u, self._vec(A, M)) < produit_scalaire(u, self._vec(A, N)):
            self.__sol1, self.__sol2 = M, N
            return [isinstance(self.droite, Droite) or M in self.droite, isinstance(self.cercle, Cercle_generique) or M in self.cercle]
        self.__sol1, self.__sol2 = N, M
        return [isinstance(self.droite, Droite) or N in self.droite, isinstance(self.cercle, Cercle_generique) or N in self.cercle]






class Intersection_cercles(Intersection_generique):
    u"""Une intersection de cercles.

    Un des deux points M et N d'intersection des deux cercles.
    L'argument angle_positif permet de distinguer les deux points.
    Si on note A et B les centres respectifs des 2 cercles, l'angle orient (AB>, AM>) doit tre de signe constant.
    """

    def __init__(self, cercle1, cercle2, angle_positif = None, **kw):
        # l'intersection de deux cercles (si elle existe) est deux points (eventuellement confondus).
        # lorsque l'utilisateur cree deux fois de suite un point d'intersection, ce ne doit pas etre le meme.
        self.synchroniser_feuille()
        if angle_positif is None and self.__feuille__:
            intersections = [objet for objet in self.__feuille__.objets() if isinstance(objet, Intersection_cercles) and objet.cercle1 is cercle1 and objet.cercle2 is cercle2]
            if intersections:
                angle_positif = intersections[0].angle == -1 # pour ne pas crer le mme point.
            del intersections


        if angle_positif is None:  # aucune intersection n'existait, ou bien on ne pouvait pas le savoir (pas de feuille de travail dfinie).
            # On construit une intersection qui existe actuellement, si possible.
            angle_positif = Intersection_cercles(cercle1, cercle2, angle_positif = True).existe()

        angle = angle_positif and 1 or -1
        Intersection_generique.__init__(self, objet1 = cercle1, objet2 = cercle2)
        self.enregistre((Cercle_generique, Arc_generique), (Cercle_generique, Arc_generique), bool)







    def _conditions_existence(self):
        a, b, c = self.cercle1.equation()
        d, e, f = self.cercle2.equation()
        print a, b, c, d, e, f
        if (a, b, c) == (d, e, f):
            return None

        # (a-d)x + (b-e)y + (c-f) = 0 : on soustrait les 2 equations de cercle -> equation d'une droite.
        d = Droite_equation((a - d, b - e, c - f))
        print d.equation()
        if isinstance(self.cercle1, Cercle_generique):
            c = self.cercle1
        else: # cas d'un arc de cercle (on doit calculer les intersections avec le cercle tout entier.)

            c = Cercle_rayon(self.cercle1.centre, self.cercle1.point1)
        M = Intersection_droite_cercle(d, c, d.point1)
        if M.existe():
            self.__sol1, self.__sol2 = M.intersections()
            print self.__sol1, self.__sol2
        else:
            return [False]

        # La encore, la partie la plus complexe consiste a distinguer les deux points d'intersection, a l'aide cette fois du signe de l'angle.
        # Pour comprendre ce qui s'y passe, il est conseille de faire un dessin !
        A, B = self.cercle1.centre, self.cercle2.centre
        a = angle_vectoriel(self._vec(A, B), self._vec(A, M))

        signe_a = cmp(a, 0)


        if signe_a <> self.angle:
            M = self.__sol1, self.__sol2 = self.__sol2, self.__sol1

        return [isinstance(self.cercle1, Cercle_generique) or M in self.cercle1, isinstance(self.cercle2, Cercle_generique) or M in self.cercle2]




    def _get_coordonnees(self):
        return self.__sol1
        # on a dj calcul les coordonnes en cherchant les conditions d'existence


    def _get_intersections(self):
        u"""Retourne les coordonnes des deux points d'intersection.
        (Dans l'ordre: celles du point, puis celles du 2me point d'intersection).
        """
        return self.__sol1, self.__sol2








class Intersection_cercles(Intersection_generique):
    u"""Une intersection de cercles.

    Un des deux points M et N d'intersection des deux cercles.
    L'argument angle_positif permet de distinguer les deux points.
    Si on note A et B les centres respectifs des 2 cercles, l'angle orient (AB>, AM>) doit tre de signe constant.
    """

    def __init__(self, cercle1, cercle2, angle_positif = None, **kw):
        # l'intersection de deux cercles (si elle existe) est deux points (eventuellement confondus).
        # lorsque l'utilisateur cree deux fois de suite un point d'intersection, ce ne doit pas etre le meme.
        self.synchroniser_feuille()
        if angle_positif is None and self.__feuille__:
            intersections = [objet for objet in self.__feuille__.objets() if isinstance(objet, Intersection_cercles) and objet.cercle1 is cercle1 and objet.cercle2 is cercle2]
            if intersections:
                angle_positif = intersections[0].angle == -1 # pour ne pas crer le mme point.
            del intersections


        if angle_positif is None:  # aucune intersection n'existait, ou bien on ne pouvait pas le savoir (pas de feuille de travail dfinie).
            # On construit une intersection qui existe actuellement, si possible.
            angle_positif = Intersection_cercles(cercle1, cercle2, angle_positif = True).existe()

        angle = angle_positif and 1 or -1
        Intersection_generique.__init__(self, objet1 = cercle1, objet2 = cercle2)
        self.enregistre((Cercle_generique, Arc_generique), (Cercle_generique, Arc_generique), bool)







    """def _conditions_existence(self):
        a, b, c = self.cercle1.equation()
        d, e, f = self.cercle2.equation()
        print a, b, c, d, e, f
        if (a, b, c) == (d, e, f):
            return None

        # (a-d)x + (b-e)y + (c-f) = 0 : on soustrait les 2 equations de cercle -> equation d'une droite.
        d = Droite_equation((a - d, b - e, c - f))
        print d.equation()
        if isinstance(self.cercle1, Cercle_generique):
            c = self.cercle1
        else: # cas d'un arc de cercle (on doit calculer les intersections avec le cercle tout entier.)

            c = Cercle_rayon(self.cercle1.centre, self.cercle1.point1)
        M = Intersection_droite_cercle(d, c, d.point1)
        if M.existe():
            self.__sol1, self.__sol2 = M.intersections()
            print self.__sol1, self.__sol2
        else:
            return [False]

        # La encore, la partie la plus complexe consiste a distinguer les deux points d'intersection, a l'aide cette fois du signe de l'angle.
        # Pour comprendre ce qui s'y passe, il est conseille de faire un dessin !
        A, B = self.cercle1.centre, self.cercle2.centre
        a = angle_vectoriel(self._vec(A, B), self._vec(A, M))

        signe_a = cmp(a, 0)


        if signe_a <> self.angle:
            self.__sol1, self.__sol2 = self.__sol2, self.__sol1
        M = Point(*self.__sol1)

        return [isinstance(self.cercle1, Cercle_generique) or M in self.cercle1, isinstance(self.cercle2, Cercle_generique) or M in self.cercle2]"""


    def _conditions_existence(self):
        # Trouver les conditions d'existence revient pratiquement  chercher les coordonnes.
        # Tout le calcul de coordonnes se fait donc ici, pour simplifier.

        a, b, c = self.cercle1.equation()
        d, e, f = self.cercle2.equation()
        #print a, b, c, d, e, f
        if (a, b, c) == (d, e, f):
            return None

        # (a-d)x + (b-e)y + (c-f) = 0 : on soustrait les 2 equations de cercle -> equation d'une droite.
        # On se ramne donc au cas de l'intersection d'un cercle et d'une droite
        a, b, c = a - d, b - e, c - f

        if b: # La droite n'est pas verticale
            m = -a/b; n = -c/b # equation de la droite : y = mx + n

            #p = Polynome([m**2 + 1, 2*n*m + d + e*m, n**2 + e*n + f])
            sols = racines(m**2 + 1, 2*n*m + d + e*m, n**2 + e*n + f)

            if len(sols) == 2: # 2 points d'intersection.
                x1, x2 = sols
                self.__sol1 = (x1, m*x1+n); self.__sol2 = (x2, m*x2+n)
            elif len(sols) == 1: # 1 seul point
                x = sols[0]
                self.__sol1 = self.__sol2 = (x, m*x+n)
            else:
                return [False] # intersection vide.

        else: # La droite est verticale
            x = -c/a
            #p = Polynome([1, e, x**2 + d*x + f])
            sols = racines(1, e, x**2 + d*x + f)

            if len(sols) == 2: # 2 points d'intersection.
                y1, y2 = sols
                self.__sol1 = (x, y1); self.__sol2 = (x, y2)
            elif len(sols) == 1:
                y = sols[0]
                self.__sol1 = self.__sol2 = (x, y)
            else:
                return [False] # intersection vide.

        # L encore, la partie la plus complexe consiste  distinguer les deux points d'intersection,  l'aide cette fois du signe de l'angle.
        # Pour comprendre ce qui s'y passe, il est conseill de faire un dessin !
        A = self.cercle1.centre.coordonnees()
        B = self.cercle2.centre.coordonnees()
        M = self.__sol1
        a = angle_vectoriel(self._vec(A, B), self._vec(A, M))

        signe_a = cmp(a, 0)


        if signe_a <> self.angle:
            self.__sol1, self.__sol2 = self.__sol2, self.__sol1
        M = self.__sol1

        return [isinstance(self.cercle1, Cercle_generique) or M in self.cercle1, isinstance(self.cercle2, Cercle_generique) or M in self.cercle2]


    def _get_coordonnees(self):
        return self.__sol1
        # on a dj calcul les coordonnes en cherchant les conditions d'existence


    def _get_intersections(self):
        u"""Retourne les coordonnes des deux points d'intersection.
        (Dans l'ordre: celles du point, puis celles du 2me point d'intersection).
        """
        return self.__sol1, self.__sol2








##########################################################################################

# Glisseurs


class Glisseur_generique(Point_generique):
    u"""Un glisseur gnrique.

    Classe mre des diffrents types de glisseurs"""

    def __init__(self, objet, k = 0, **kw):
        if isinstance(k, (list, tuple)):
            pt = k
            k = self.nb2var(0)
        else:
            pt = None
            k = self.nb2var(k)
        Point_generique.__init__(self)
        self.style(**param.points_deplacables)
        self.enregistre(Objet, Variable)
        if pt is not None:
            self._set_coordonnees(pt[0], pt[1])

    def parametre(self, coeff = None):
        u"""Cette fonction devrait surtout tre utilise en lecture.
        Alias plus 'expressif' de self.k(coeff)."""
        if coeff is not None: self.k(coeff)
        return self.k.val()


    def coo(self):
        return (1 - self.k.val())*self.objet.point1.coo() + self.k.val()*self.objet.point2.coo()

    def _set_coordonnees(self):
        return NotImplemented



class Glisseur_droite(Glisseur_generique):
    u"""Un point sur une droite.

    Point pouvant 'glisser' sur une droite. k est un coefficient barycentrique.
    Si A et B sont les deux points de rfrence de la droite, k le coefficient,
    les coordonnes du glisseur M seront determines par la formule M = kA + (1-k)B."""

    def __init__(self, droite, k = None, **kw):
        if k is None:
            k = normalvariate(0.5, 0.5)
        # k = self.nb2var(k)
        Glisseur_generique.__init__(self, objet = droite, k = k)
        del k # k a ete enregistre apres modification dans Glisseur_generique (evitons de perdre les modifications !).
        # self.k = k # sinon, k a ete remplace par une copie (et l'enregistrement ne marche pas)
        self.enregistre(Droite, Variable)


    def _set_coordonnees(self, x = None, y = None):
        if x is not None:
            self.recalculer = True
            # permet d'utiliser le meme code pour Glisseur_droite, Glisseur_segment, Glisseur_demidroite
            if isinstance(self, Glisseur_segment):
                Projete = Projete_segment
            elif isinstance(self, Glisseur_demidroite):
                Projete = Projete_demidroite
            else:
                Projete = Projete_droite
            M = Projete(Point(x, y), self.objet)
            A = self.objet.point1; B = self.objet.point2
            self.k.val(produit_scalaire(self._vec(A, M), self._vec(A, B))/self._distance(A, B)**2)



    def _get_coordonnees(self):
        return self.coo()





class Glisseur_segment(Glisseur_droite):
    u"""Un point sur un segment.

    Point pouvant 'glisser' sur un segment. k est un coefficient barycentrique.
    Si A et B sont les deux points de rfrence de la droite, k le coefficient,
    les coordonnes du glisseur M seront determines par la formule M = kA + (1-k)B."""

    def __init__(self, segment, k = None, **kw):
        if k is None:
            k = random.uniform(0, 1)
        Glisseur_droite.__init__(self, droite = segment, k = k)
        del k # k a ete enregistre apres modification dans Glisseur_droite (evitons de perdre les modifications !).
        self.enregistre(Segment, Variable)




class Glisseur_demidroite(Glisseur_droite):
    u"""Un point sur une demi-droite.

    Point pouvant 'glisser' sur un segment. k est un coefficient barycentrique.
    Si A et B sont les deux points de rfrence de la droite, k le coefficient,
    les coordonnes du glisseur M seront determines par la formule M = kA + (1-k)B."""

    def __init__(self, demidroite, k = None, **kw):
        if k is None:
            k = abs(normalvariate(0.5, 0.5))
        Glisseur_droite.__init__(self, droite = demidroite, k = k)
        del k # k a ete enregistre apres modification dans Glisseur_droite (evitons de perdre les modifications !).
        self.enregistre(Demidroite, Variable)



class Glisseur_cercle(Glisseur_generique):
    u"""Un point sur un cercle.

    Point pouvant 'glisser' sur un cercle. k est un angle.
    Si O et r sont les centres et rayons du cercle, k l'angle,
    les coordonnes du glisseur M seront dtermines par la formule M = O + r*(cos(k), sin(k))."""

    def __init__(self, cercle, k = None, **kw):
        if k is None:
            k = 2*pi*rand()
        # k = self.nb2var(k)
        Glisseur_generique.__init__(self, objet = cercle, k = k)
        del k # k a ete enregistre apres modification dans Glisseur_generique (evitons de perdre les modifications !).
        # self.k = k # sinon, k a ete remplace par une copie (et l'enregistrement ne marche pas)
        self.enregistre(Cercle_generique, Variable)


    def coo(self):
        return self.objet.centre.coo() + self.objet.rayon()*array((cos(self.parametre()), sin(self.parametre())))


    def _set_coordonnees(self, x = None, y = None):
        if x is not None:
            self.recalculer = True
            O = self.objet.centre; M = (x, y)
            self.k(angle_vectoriel((1, 0), self._vec(O, M)))


    def _get_coordonnees(self):
        return self.coo()




class Glisseur_arc_cercle(Glisseur_cercle):
    u"""Un point sur un arc de cercle.

    Point pouvant 'glisser' sur un cercle. k est un angle.
    Si O et r sont les centres et rayons de l'arc, k un coefficient compris entre 0 et 1,
    et a et b les mesures d'angle marquant le dbut et la fin de l'arc,
    les coordonnes du glisseur M seront dtermines par la formule M = O + r*(cos(ka+(1-k)b), sin(ka+(1-k)b)),
    si l'arc est de sens direct, et M = O + r*(cos(kb+(1-k)a), sin(kb+(1-k)a)) sinon."""

    def __init__(self, arc, k = None, **kw):
        if k is None:
            k = random.uniform(0, 1)
        # k = self.nb2var(k)
        Glisseur_cercle.__init__(self, cercle = arc, k = k)
        del k # k a ete enregistre apres modification dans Glisseur_generique (evitons de perdre les modifications !).
        # self.k = k # sinon, k a ete remplace par une copie (et l'enregistrement ne marche pas)
        self.enregistre(Arc_generique, Variable)


    def coo(self):
        a, b = self.objet._intervalle()
        k = self.parametre()
        if self.objet._sens() == 1:
            t = a*k + b*(1 - k)
        else:
            t = b*k + a*(1 - k)
        return self.objet.centre.coo() + self.objet.rayon()*array((cos(t), sin(t)))


    def _get_coordonnees(self):
        return self.coo()


    def _set_coordonnees(self, x = None, y = None):
        self.recalculer = True
        M = (x, y)
        O = self.objet.centre;
        a, b = self.objet._intervalle()
        c = angle_vectoriel((1, 0), self._vec(O, M))
        while c < a:
            c += 2*pi
        # La mesure d'angle c est donc dans l'intervalle [a; a+2*pi[
        if c > b: # c n'appartient pas  [a;b] (donc M est en dehors de l'arc de cercle)
            if c - b > 2*pi + a - c: # c est plus proche de a+2*pi
                c = a
            else:   # c est plus proche de b
                c = b
        if b - a:
            if self.objet._sens() == 1:
                k = (b - c)/(b - a)
            else:
                k = (c - a)/(b - a)
        else:
            k = 0.5
        self.k.val(k)



##########################################################################################

# Polygones





class Cote(Segment):
    u"""Un cot.

    Un cot d'un polygone, reliant le point numero 'n' au point numero 'n + 1'.
    Note: n commence  0.
    L'objet est cr automatiquement lors de la cration du polygone.
    De plus, si l'objet est supprim, le polygone est automatiquement supprim."""

    def __init__(self, polygone, n, **kw):
        Segment.__init__(self, polygone.points[n], polygone.points[(n + 1)%len(polygone.points)])
        self.style(**polygone.style())
        self.enregistre(Polygone, int)
        self._enregistrer = False # l'objet ne doit pas tre sauvegard

    def supprimer(self):
        self.polygone.supprimer()





class Sommet(Point_generique):
    u"""Un sommet.

    Le nime sommet d'un polygone.
    Note: n commence  0.
    L'objet est cr automatiquement lors de la cration du polygone.
    De plus, si l'objet est supprim, le polygone est automatiquement supprim."""

    def __init__(self, polygone, n, **kw):
        Point_generique.__init__(self)
        self.style(legende = wxgeo.RIEN)
        point = polygone.points[n]
        self.enregistre(Polygone, int)
        self._enregistrer = False

    def supprimer(self):
        self.polygone.supprimer()

    def _get_coordonnees(self):
        return self.point.coordonnees()

    def _conditions_existence(self):
        return [self.point.existe()]




class Polygone(Objet):
    u"""Un polygone.

    Classe mre de tous les polygones."""

    def __new__(cls, *points, **kw):
        if cls is Polygone:
            if len(points) == 3 :
                return object.__new__(Triangle, *points, **kw)
            elif len(points) == 4:
                return object.__new__(Quadrilatere, *points, **kw)
        return object.__new__(cls, *points, **kw)


    def __init__(self, *points, **kw):
        if kw.get("previsualisation"):
            Objet.__init__(self)
            self.style(**param.polygones)
        else:
            if isinstance(points[0], (list, tuple)):
                points = tuple(points[0])
            sommets = points # alias
            centre = centre_gravite = Barycentre(points)
            Objet.__init__(self)
            self.etiquette = Label_polygone(self)
            self.style(**param.polygones)
        self.enregistre([Point_generique])


    def enregistre(self, *args, **kw):
        Objet.enregistre(self, *args, **kw)
        if not self.style("previsualisation"): # inutile de crer des objets cts pour une simple prvisualisation
            self.creer_cotes()


    def creer_cotes(self):
        u"Doit tre appel aprs l'enregistrement, pour crer les cts, et les sommets."
        if hasattr(self, "_hierarchie") and not hasattr(self, "cotes"):
        # un triangle, un paralllogramme, etc. peuvent subir plusieurs initialisations
            n = len(self.points)
            self.cotes = [Cote(self, i) for i in xrange(n)]
            self.sommets = [Sommet(self, i) for i in xrange(n)]
            if self.__feuille__:
                noms = self.style("noms")
                if noms is None:
                    noms = {"cotes": ['' for i in xrange(n)], "sommets": ['' for i in xrange(n)]}
                for i in xrange(n):
                    setattr(self.__feuille__.objets, noms["cotes"][i] or self.__feuille__.nom_alea("c"), self.cotes[i])
                    if self.sommets[i].point not in self.__feuille__.objets(type = Point_generique):
                        setattr(self.__feuille__.objets, noms["sommets"][i] or self.__feuille__.nom_alea("S"), self.sommets[i])

                self.style(noms = {"cotes": [Nom(cote) for cote in self.cotes], "sommets": [Nom(sommet) for sommet in self.sommets]})
                # ce style spcial sert  bien regnrer les mmes noms aprs une sauvegarde.




    def _affiche(self):
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().fill(), self.__canvas__().plot(zorder = 1)]

        fill = self.__representation__[0]
        plot = self.__representation__[1]
        niveau = self.style("niveau")
        points = self.points + (self.points[0],)
        xy = [pt.coordonnees() for pt in points]
        x, y = zip(*xy)
        fill.xy = xy
        if self.style("previsualisation"):
            x = x[:-1]
            y = y[:-1]
        plot._x = x; plot._y = y
        fill._alpha = self.style("alpha")
        hachures = self.style("hachures")
        if hachures is not None:
            fill._hatch = hachures
        fill._edgecolor = fill._facecolor = plot._color = self.style("couleur")
        fill.zorder = niveau - 0.01
        plot._linestyle = fill._linestyle = self.style("style")
        plot._linewidth = fill._linewidth = self.style("epaisseur")
        plot.zorder = niveau




    def _distance_inf(self, x, y, d):
        xy = [pt.position() for pt in self.points]
        return point_dans_polygone((x,y), xy)

    def __contains__(self, M):
        xy = [pt.coordonnees() for pt in self.points]
        return True in [(M in cote) for cote in self.cotes] or point_dans_polygone(tuple(M), xy)

    def aire(self):
        u"""D'aprs David Chandler, Area of a general polygon.
        http://www.davidchandler.com/AreaOfAGeneralPolygon.pdf"""

        if self.existe():
            xy = [pt.coordonnees() for pt in (self.points + (self.points[0],))]
            s1 = s2 = 0
            for i in range(len(self.points)):
                s1 += xy[i][0]*xy[i+1][1]
                s2 += xy[i][1]*xy[i+1][0]
            return .5*abs(s1-s2)
        return None

    def perimetre(self):
        return builtin.sum(cote.longueur() for cote in self.cotes)

    def _espace_vital(self):
        x1 = min(pt.abscisse() for pt in self.points)
        x2 = max(pt.abscisse() for pt in self.points)
        y1 = min(pt.ordonnee() for pt in self.points)
        y2 = max(pt.ordonnee() for pt in self.points)
        return (x1, x2, y1, y2)







class Triangle(Polygone):
    u"""Un triangle."""

    def __init__(self, point1, point2, point3, **kw):
        Polygone.__init__(self, point1, point2, point3, **kw)
        if not kw.get("previsualisation"):
            d1 = Perpendiculaire(Droite(point1, point2), point3)
            d2 = Perpendiculaire(Droite(point2, point3), point1)
            orthocentre = Intersection_droites(d1, d2)
            centre_cercle_circonscrit = Point_equidistant(point1, point2, point3)
            d1 = Bissectrice(point1, point2, point3)
            d2 = Bissectrice(point3, point1, point2)
            centre_cercle_inscrit = Intersection_droites(d1, d2)
            del d1, d2
        self.enregistre(Point_generique, Point_generique, Point_generique)


    def enregistre(self, *args, **kw):
        Polygone.enregistre(self, *args, **kw)
        self.point1 = self.points[0] # utile pour certaines classes drives  hritage multiple
        self.point2 = self.points[1] # qui ne font pas appel  Triangle.__init__ lors de leur initialisation
        self.point3 = self.points[2] # (par exemple, Triangle_equilateral_centre)




class Quadrilatere(Polygone):
    u"""Un quadrilatre."""

    def __init__(self, point1, point2, point3, point4, **kw):
        Polygone.__init__(self, point1, point2, point3, point4, **kw)
        self.enregistre(Point_generique, Point_generique, Point_generique, Point_generique)

    def enregistre(self, *args, **kw):
        Polygone.enregistre(self, *args, **kw)
        self.point1 = self.points[0]
        self.point2 = self.points[1]
        self.point3 = self.points[2]
        self.point4 = self.points[3]



class Parallelogramme(Quadrilatere):
    u"""Un paralllogramme."""

    def __init__(self, point1, point2, point3, **kw):
        Quadrilatere.__init__(self, point1, point2, point3, Barycentre((point1, point2, point3), (1, -1, 1)), **kw)
        self.enregistre(Point_generique, Point_generique, Point_generique)


class Rectangle(Parallelogramme):
    u"""Un rectangle."""

    #  dfinir


class Losange(Parallelogramme):
    u"""Un losange."""

    #  dfinir



class Polygone_regulier_centre(Polygone):
    u"""Un polygone rgulier.

    Un polygone rgulier dfini par son centre, un sommet, et le nombre de cts."""

    def __new__(cls, *args, **kw):
        if cls is Polygone_regulier_centre:
            l = len(args)
            if l:
                kw["centre"] = args[0]
                if l > 1:
                    kw["sommet"] = args[1]
                    if l >2:
                        kw["n"] = args[2]
            centre = kw["centre"]
            sommet = kw["sommet"]
            n =  kw["n"]
            kw.pop("centre", None)
            kw.pop("sommet", None)
            kw.pop("n", None)
            if n == 3:
                return object.__new__(Triangle_equilateral_centre, centre, sommet, **kw)
            elif n == 4:
                return object.__new__(Carre_centre, centre, sommet, **kw)
        return object.__new__(cls, *args, **kw)

    def __init__(self, centre, sommet, n, **kw):
        n = int(n)
        Polygone.__init__(self, sommet, *(Rotation(centre, 2*pi*i/n)(sommet) for i in xrange(1, n)), **kw)
        self.centre = NotImplemented # supprime l'avertissement qui est affich lorsqu'on "crase" une mthode ou un attribut d'une classe parente.
        self.enregistre(Point_generique, Point_generique, int)



class Triangle_equilateral_centre(Polygone_regulier_centre, Triangle):
    u"""Un triangle quilatral.

    Un triangle quilatral dfini par son centre et un sommet."""

    def __init__(self, centre, sommet, **kw):
        Polygone_regulier_centre.__init__(self, centre, sommet, 3, **kw)
        self.enregistre(Point_generique, Point_generique)



class Carre_centre(Polygone_regulier_centre, Rectangle, Losange):
    u"""Un carr.

    Un carr dfini par son centre et un sommet."""

    def __init__(self, centre, sommet, **kw):
        Polygone_regulier_centre.__init__(self, centre, sommet, 4, **kw)
        self.enregistre(Point_generique, Point_generique)






class Polygone_regulier(Polygone):
    u"""Un polygone rgulier.

    Un polygone rgulier dfini par deux points conscutif (sens direct)."""

    def __new__(cls, *args, **kw):
        if cls is Polygone_regulier:
            l = len(args)
            if l:
                kw["point1"] = args[0]
                if l > 1:
                    kw["point2"] = args[1]
                    if l >2:
                        kw["n"] = args[2]
            point1 = kw["point1"]
            point2 = kw["point2"]
            n =  kw["n"]
            kw.pop("point1", None)
            kw.pop("point2", None)
            kw.pop("n", None)
            if n == 3:
                return object.__new__(Triangle_equilateral, point1, point2, **kw)
            elif n == 4:
                return object.__new__(Carre, point1, point2, **kw)
        return object.__new__(cls, *args, **kw)

    def __init__(self, point1, point2, n, **kw):
        n = int(n)
        points = [point1, point2]
        for i in xrange(n - 2):
            points.append(Rotation(points[-1], -(n - 2)*pi/n)(points[-2])) # image d'un sommet par une rotation de centre le sommet suivant
        Polygone.__init__(self, *points, **kw)
        del points
        self.enregistre(Point_generique, Point_generique, int)



class Triangle_equilateral(Polygone_regulier, Triangle):
    u"""Un triangle quilatral.

    Un triangle quilatral dfini par deux points conscutif (sens direct)."""

    def __init__(self, point1, point2, **kw):
        Polygone_regulier.__init__(self, point1, point2, 3, **kw)
        self.enregistre(Point_generique, Point_generique)



class Carre(Polygone_regulier, Rectangle, Losange):
    u"""Un carr.

    Un carr dfini par deux points conscutif (sens direct)."""

    def __init__(self, point1, point2, **kw):
        Polygone_regulier.__init__(self, point1, point2, 4, **kw)
        self.enregistre(Point_generique, Point_generique)








##########################################################################################

# Divers




class Repere(Objet): # A TERMINER
    u"""Un repre.

    Un repre dfini par un point et deux vecteurs."""

    def __init__(self, origine, vecteur1, vecteur2, **kw):
        self.enregistre(Point_generique, Vecteur_generique, Vecteur_generique)


    def _affiche(self):
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().plot()]
        x1, y1 = self.point1.coordonnees()
        x2, y2 = self.point2.coordonnees()
        p = self.pente(1) # "tendance" de la pente (verticale : 0, horizontale : 1)
        fen = self.__feuille__.fenetre() # il faut tracer depuis le bord de fenetre
        x1, y1 = self.x_y((fen[2], fen[0])[p], p)
        x2, y2 = self.x_y((fen[3], fen[1])[p], p)
        plot = self.__representation__[0]
        plot._x = array((x1, x2)) ; plot._y = array((y1, y2))
        plot._color = self.style("couleur")
        plot._linestyle = self.style("style")
        plot._linewidth = self.style("epaisseur")
        plot.zorder = self.style("niveau")




class Interpolation_generique(Objet):
    u"""Classe mre de toute les interpolations."""
    
    def __init__(self, *points, **kw):
        if len(points) == 1 and isinstance(points[0], (list, tuple)):
            points = tuple(points[0])
        debut = kw.pop("debut", None)
        if debut is None:
            debut = kw.pop("deb", None)
        fin = kw.pop("fin", None)
        Objet.__init__(self)
        self.style(**param.segments)
        self.enregistre([Point_generique])

    def _extremite_arc(self, x, y, vecteur):
        coeff0, coeff1 = self.__canvas__().coeffs()
        if coeff0 == 0 or coeff1 == 0:
            warning("Rsolution d'affichage incorrecte (division par zro).")
            return
        vecteur = vecteur[0]/-coeff0, vecteur[1]/coeff1
        if vecteur[0] == 0:
            if vecteur[1] > 0:
                angle = pi/2
            else:
                angle = -pi/2
        else:
            angle = arctan(vecteur[1]/vecteur[0])
            if vecteur[0] < 0:
                angle += pi
        # donne l'angle d'incidence a l'extremite
        t = arange(angle - pi/2, angle + pi/2, 0.05)
        R = self.__canvas__().taille["("]
        
        x, y = self.coo2pix((x, y))
        return self.pix2coo((x + R*(cos(t) - cos(angle)), y + R*(sin(t) - sin(angle))))
        
    
    def _affiche_extremites(self, vec_deb = None, vec_fin = None):    
        couleur = self.style("couleur")
        niveau = self.style("niveau")
        # Dbut de la courbe
        plot = self.__representation__[-2]
        x, y = self.points[0].coordonnees()
        a, b = self.points[1].coordonnees()
        if vec_deb is None:
            vec_deb = (x-a, y-b)
        if self.debut == True:
            plot._visible = True
            plot._x, plot._y = array((x,)), array((y,))
            plot._color = couleur
            plot._marker = "o"
            plot._markeredgecolor = couleur
            plot._markersize = 2*self.__canvas__().taille["o"]
            plot.zorder = niveau
        elif self.debut == False:
            plot._visible = True
            plot._x, plot._y = self._extremite_arc(x, y, vec_deb)
            plot._color = couleur
            plot.zorder = niveau
        else:
            plot._visible = False
            
        # Fin de la courbe
        plot = self.__representation__[-1]
        x, y = self.points[-1].coordonnees()
        a, b = self.points[-2].coordonnees()
        if vec_fin is None:
            vec_fin = (x-a, y-b)
        if self.fin == True:
            plot._visible = True
            plot._x, plot._y = array((x,)), array((y,))
            plot._color = couleur
            plot._marker = "o"
            plot._markeredgecolor = couleur
            plot._markersize = 2*self.__canvas__().taille["o"]
            plot.zorder = niveau
        elif self.fin == False:
            plot._visible = True
            plot._x, plot._y = self._extremite_arc(x, y, vec_fin)
            plot._color = couleur
            plot.zorder = niveau
        else:
            plot._visible = False
        
        
        
        

class Interpolation_lineaire(Interpolation_generique):
    u"""Une interpolation linaire.
    
    Interpolation entre les points donns par des segments joints (courbe de classe C0).
    """
    
    
    def __init__(self, *points, **kw):
        Interpolation_generique.__init__(self, *points, **kw)
        #~ self.points = NotImplemented
        del points
        self.enregistre([Point_generique])

    def _affiche(self):
        n = len(self.points)
        couleur = self.style("couleur")
        niveau = self.style("niveau")
        style = self.style("style")
        epaisseur = self.style("epaisseur")
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().plot() for i in xrange(n + 1)]
            
        if n < 2:
            return

        for i in xrange(n - 1):
            plot = self.__representation__[i]
            x1, y1 = self.points[i].coordonnees()
            x2, y2 = self.points[i+1].coordonnees()
            plot._x = array((x1, x2))
            plot._y = array((y1, y2))
            plot._color = couleur
            plot._linestyle = style
            plot._linewidth = epaisseur
            plot.zorder = niveau        
        
        self._affiche_extremites()



class Interpolation_quadratique(Interpolation_generique):
    u"""Une interpolation quadratique.
    
    Interpolation des points donns par une courbe polynomiale par morceaux, et de classe C1.
    Pour chaque morceau, x(t)=at+bt+c et y(t)=dt+et+f, o t appartient  [0;1].
    """
    
    def __init__(self, *points, **kw):
        Interpolation_generique.__init__(self, *points, **kw)
        #~ self.points = NotImplemented
        del points
        self.enregistre([Point_generique])


    def _affiche(self):
        n = len(self.points)
        couleur = self.style("couleur")
        niveau = self.style("niveau")
        style = self.style("style")
        epaisseur = self.style("epaisseur")
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().plot() for i in xrange(n + 1)]
            
        if n < 2:
            return

        pas = self.__canvas__().pas()
        t = concatenate((arange(0, 1, pas), (1,)))
        for i in xrange(n - 1):
            plot = self.__representation__[i]
            x0, y0 = self.points[i].coordonnees()
            x1, y1 = self.points[i+1].coordonnees()
            if i == 0:
                dx0, dy0 = x1 - x0, y1 - y0
            a, b, c = x1 - dx0 - x0, dx0, x0
            d, e, f = y1 - dy0 - y0, dy0, y0
            plot._x = (a*t + b)*t + c
            plot._y = (d*t + e)*t + f
            plot._color = couleur
            plot._linestyle = style
            plot._linewidth = epaisseur
            plot.zorder = niveau        
            
            dx0 = 2*a + b
            dy0 = 2*d + e
            
        self._affiche_extremites(vec_fin = (dx0, dy0))




class Interpolation_cubique_old(Interpolation_generique):
    u"""Une interpolation cubique.
    
    Interpolation des points donns par une courbe polynomiale par morceaux, de vecteur tangent horizontal aux sommets, et de classe C1 en gnral (ie. si x_n<>x_{n-1}).
    Pour chaque morceau, x(t)=at^3+bt^2+ct+d et y(t)=et^3+ft^2+gt+h, o t appartient  [0;1], et y'(0)=y'(1)=0.
    """
    
    def __init__(self, *points, **kw):
        courbure = kw.pop("courbure", 1)
        Interpolation_generique.__init__(self, *points, **kw)
        #~ self.points = NotImplemented
        del points
        self.enregistre([Point_generique])


    def _affiche(self):
        n = len(self.points)
        couleur = self.style("couleur")
        niveau = self.style("niveau")
        style = self.style("style")
        epaisseur = self.style("epaisseur")
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().plot() for i in xrange(n + 1)]
            
        if n < 2:
            return

        pas = self.__canvas__().pas()
        t = concatenate((arange(0, 1, pas), (1,)))
        for i in xrange(n - 1):
            plot = self.__representation__[i]
            x0, y0 = self.points[i].coordonnees()
            x1, y1 = self.points[i+1].coordonnees()
            dx0 = dx1 = self.courbure*abs(x1 - x0)
            dy0 = dy1 = 0
            a = -2*(x1 - dx0 - x0)
            b = 3*(x1 - dx0 - x0)
            c = dx0
            d = x0
            e = -2*(y1 - dy0 - y0)
            f = 3*(y1 - dy0 - y0)
            g = dy0
            h = y0
            plot._x = ((a*t + b)*t + c)*t + d
            plot._y = ((e*t + f)*t + g)*t + h
            plot._color = couleur
            plot._linestyle = style
            plot._linewidth = epaisseur
            plot.zorder = niveau        
            
        self._affiche_extremites(vec_deb = (dx0, dy0), vec_fin = (dx1, dy1))




class Interpolation_cubique(Interpolation_generique):
    u"""Une interpolation cubique.
    
    Interpolation des points donns par une courbe polynomiale par morceaux, de vecteur tangent horizontal aux sommets, et de classe C1 en gnral (ie. si x_n<>x_{n-1}).
    Pour chaque morceau, x(t)=at^3+bt^2+ct+d et y(t)=et^3+ft^2+gt+h, o t appartient  [0;1], et y'(0)=y'(1)=0.
    """
    
    def __init__(self, *points, **kw):
        courbure = kw.pop("courbure", 1)
        Interpolation_generique.__init__(self, *points, **kw)
        #~ self.points = NotImplemented
        del points
        self.enregistre([Point_generique])


    def _affiche(self):
        n = len(self.points)
        couleur = self.style("couleur")
        niveau = self.style("niveau")
        style = self.style("style")
        epaisseur = self.style("epaisseur")
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().plot() for i in xrange(n + 1)]
            
        if n < 2:
            return

        pas = self.__canvas__().pas()
        t = concatenate((arange(0, 1, pas), (1,)))
        for i in xrange(n - 1):
            plot = self.__representation__[i]
            x0, y0 = self.points[i].coordonnees()
            x1, y1 = self.points[i+1].coordonnees()
            if i == 0:
                dx0 = x1 - x0
                dy0 = y1 - y0
            else:
                dx0 = self.courbure*abs(x1 - x0)
                if dx1 <> 0:
                    dy0 = dy1/dx1*dx0
                else:
                    dx0 = dx1
                    dy0 = dy1
            if i < n - 2:
                x2, y2 = self.points[i+2].coordonnees()
                if cmp(y0, y1) == cmp(y1, y2):
                    dy1 = y1 - y0
                    dx1 = x1 - x0
                else:
                    dy1 = 0
                    dx1 = self.courbure*abs(x1 - x0)
            else:
                dy1 = y1 - y0
                dx1 = x1 - x0
            a = 2*(x0 - x1) + dx0 + dx1
            b = 3*(x1 - x0) -2*dx0 - dx1
            c = dx0
            d = x0
            e = 2*(y0 - y1) + dy0 + dy1
            f = 3*(y1 - y0) -2*dy0 - dy1
            g = dy0
            h = y0
            plot._x = ((a*t + b)*t + c)*t + d
            plot._y = ((e*t + f)*t + g)*t + h
            plot._color = couleur
            plot._linestyle = style
            plot._linewidth = epaisseur
            plot.zorder = niveau        
            
        self._affiche_extremites(vec_deb = (dx0, dy0), vec_fin = (dx1, dy1))





class Interpolation_cubique(Interpolation_generique):
    u"""Une interpolation cubique.
    
    Interpolation des points donns par une courbe polynomiale par morceaux, de vecteur tangent horizontal aux sommets, et de classe C1 en gnral (ie. si x_n<>x_{n-1}).
    Pour chaque morceau, x(t)=at^3+bt^2+ct+d et y(t)=et^3+ft^2+gt+h, o t appartient  [0;1], et y'(0)=y'(1)=0.
    """
    
    def __init__(self, *points, **kw):
        courbure = kw.pop("courbure", 1)
        Interpolation_generique.__init__(self, *points, **kw)
        #~ self.points = NotImplemented
        del points
        self.enregistre([Point_generique])


    def _affiche(self):
        n = len(self.points)
        couleur = self.style("couleur")
        niveau = self.style("niveau")
        style = self.style("style")
        epaisseur = self.style("epaisseur")
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().plot() for i in xrange(n + 1)]
            
        if n < 2:
            return

        pas = self.__canvas__().pas()
        t = concatenate((arange(0, 1, pas), (1,)))
        for i in xrange(n - 1):
            plot = self.__representation__[i]
            x0, y0 = self.points[i].coordonnees()
            x1, y1 = self.points[i+1].coordonnees()
            if i == 0:
                dx0 = x1 - x0
                dy0 = y1 - y0
            else:
                dx0 = self.courbure*abs(x1 - x0)
                if dx1 <> 0:
                    dy0 = dy1/dx1*dx0
                else:
                    dx0 = dx1
                    dy0 = dy1
            if i < n - 2:
                x2, y2 = self.points[i+2].coordonnees()
                if cmp(y0, y1) == cmp(y1, y2):
                    dy1 = y2 - y0
                    dx1 = x2 - x0
                else:
                    dy1 = 0
                    dx1 = self.courbure*abs(x1 - x0)
            else:
                dy1 = y1 - y0
                dx1 = x1 - x0
            a = 2*(x0 - x1) + dx0 + dx1
            b = 3*(x1 - x0) -2*dx0 - dx1
            c = dx0
            d = x0
            e = 2*(y0 - y1) + dy0 + dy1
            f = 3*(y1 - y0) -2*dy0 - dy1
            g = dy0
            h = y0
            plot._x = ((a*t + b)*t + c)*t + d
            plot._y = ((e*t + f)*t + g)*t + h
            plot._color = couleur
            plot._linestyle = style
            plot._linewidth = epaisseur
            plot.zorder = niveau        
            
        self._affiche_extremites(vec_deb = (dx0, dy0), vec_fin = (dx1, dy1))



#~ class FastPolygone(Polygone):  # DESUET (utiliser le style 'previsualisation' dsormais)
    #~ u"""Un polygone sans ct (usage interne).

    #~ Polygone sans cts destin  un affichage rapide (prvisualisation des constructions)."""

    #~ def __init__(self, *points, **kw):
        #~ if isinstance(points[0], (list, tuple)):
            #~ points = tuple(points[0])
        #~ sommets = points # alias
        #~ centre = centre_gravite = Barycentre(points)
        #~ Objet.__init__(self)
        #~ self.etiquette = Label_polygone(self)
        #~ self.style(**param.polygones)
        #~ self.enregistre([Point_generique])



