# -*- coding: iso-8859-1 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#                    Texte                    #
##--------------------------------------#######
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2007  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# version unicode

from variable import *

class Formule(Variable):
    u"Une formule contenue dans un texte."
    def __init__(self, parent, valeur, **kw):
        self.parent = weakref.ref(parent) # self.parent est une fonction qui renvoit parent si il existe encore. Cela permet de ne pas le maintenir en vie artificiellement ( pas de rfrence circulaire).
        try:
            Variable.__init__(self, valeur, **kw)
        except:
            if param.debug:
                print_error()
            self.valeur = valeur
            
        
    def affiche(self):
        self.parent().affiche()
        
    def _affiche(self):
        self.parent()._affiche()
        
#    def val(self, *args, **kw):
#        return Variable.val(self, *args, **kw)
            
#    def __repr__(self):
#        return Variable.__repr__(self)
    
#    def __str__(self):
#        return Variable.__str__(self)




class Texte(Objet):
    u"""Un texte.
    
    Un texte  afficher"""
    
    try:
        __empty_font__ = matplotlib.font_manager.FontProperties()
    except NameError:
        pass
    
    def __init__(self, texte = "", x = None, y = None, **kw):
        if x == None:
            self.synchroniser_feuille()
            if self.__canvas__():
                xmin, xmax, ymin, ymax = self.__canvas__().fen()
                x = random.uniform(xmin,xmax)
                y = random.uniform(ymin,ymax)
                del xmin, xmax, ymin, ymax
            else:
                x = y = 0

        # on convertit l'entree en Variable
        # l'entree peut-etre un nombre, un(e longueur de) segment, un (rayon de) cercle
        x, y = self.nb2var(x), self.nb2var(y)
        if param.unicode:
            texte = unicode2(texte, param.encodage)
        Objet.__init__(self)
        self.style(**param.textes)
        if texte:   self.style(label = texte)
        self.enregistre((str, unicode), Variable, Variable)


    def _get_coordonnees(self):
        return self.x.val(), self.y.val()


    def _set_coordonnees(self, x = None, y = None):
        if x is not None:
            self.recalculer = True
            if y is None: # _set_coordonnees((-1,2)) est accept
                x, y = x
            self.x(x)
            self.y(y) # False : ne rafraichit l'affichage qu'a la fin (deja que c'est lent !!).


    def _affiche(self):
        #self.__canvas__().Texte(self.texte, self.coo())
        if not self.__representation__:
            self.__representation__ = [self.__canvas__().text(), self.__canvas__().fill()]
            self.__representation__[0]._fontproperties = copy.copy(self.__empty_font__)
        x, y = self.coordonnees()
        text = self.__representation__[0]
        text._x = x ; text._y = y
        text._color = self.style("couleur")
        font = text._fontproperties
        font.set_size(self.style("taille"))
        #font.set_stretch(self.style("largeur"))  # mal gere par matploltib (version 0.87)
        font.set_weight(100*min(9, max(1, int(self.style("epaisseur")))))
        font.set_style(self.style("style"))
        font.set_family(self.style("famille"))
        text._rotation = self.style("angle")
        text._text = self.label()
        text._verticalalignment = self.style("alignement_vertical")
        text._horizontalalignment = self.style("alignement_horizontal")
        text.zorder = self.style("niveau") + .001
        fill = self.__representation__[1]
        if self.style("fond") is None:
            fill._visible = False
        else:
            fill._visible = True
            can = self.__canvas__()
            box = text.get_window_extent(can.get_renderer())
            print box.width(), box.height()
            w = .5*box.width()*can.coeff(0)
            h = .5*box.height()*can.coeff(1)
            if text._verticalalignment == "left":
                x += w 
            elif text._verticalalignment == "right":
                x -= w
            if text._horizontalalignment == "top":
                y -= h
            elif text._horizontalalignment == "bottom":
                y += h
            fill.xy = [(x - w, y - h), (x - w, y + h), (x + w, y + h), (x + w, y - h)]
            fill._facecolor = fill._edgecolor = self.style("fond")
            fill.zorder = self.style("niveau")
        #debug(font.__dict__)



    def _espace_vital(self):
        #renderer = self.__representation__[0]._renderer
        #print renderer
        #if renderer is None:
        #    x, y = self.coordonnees()
        #    return (x, x, y, y)
        #else:
        box = self.__representation__[0].get_window_extent(self.__canvas__().get_renderer())
        can = self.__canvas__()
        return (can.Xpix2coo(box.xmin()), can.Xpix2coo(box.xmax()), can.Ypix2coo(box.ymin()), can.Ypix2coo(box.ymax()))
    


    def _distance_inf(self, x, y, d):
        x1, x2, y1, y2 = self._espace_vital()
        can = self.__canvas__()
        #print can.Xcoo2pix(x1), can.Xcoo2pix(x2), can.Ycoo2pix(y1, -1), can.Ycoo2pix(y2, -1)
        return can.Xcoo2pix(x1) - d < x < can.Xcoo2pix(x2) + d and can.Ycoo2pix(y2, -1) - d < y < can.Ycoo2pix(y1, -1) + d
        
        """renderer = self.__representation__[0]._renderer
        if renderer is None:
            x0, y0 = self.position()
            return (x - x0)**2 + (y - y0)**2 < d**2
        else:
            box = self.__representation__[0].get_window_extent()
            return box.xmin() - d < x < box.xmax() + d and box.ymin() - d < y < box.ymax() + d"""

    _call = Objet.coordonnees
