# -*- coding: iso-8859-1 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#                  Variable                   #
##--------------------------------------#######
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2007  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# version unicode

from objet import *


class Variable(Objet_numerique):
    u"""Une variable.
    
    Une variable numrique ; l'argument peut tre un nombre, ou une expression sous forme de chaine de caractres.
    Exemple: Variable(17.5), Variable('AB.longeur()+1').
    Dans ce dernier cas, il est ncessaire qu'une feuille de travail soit dfinie."""
    
    def __init__(self, valeur = 0, **kw):
            
        #debug(valeur)
        #if type(valeur) ==  str: raise TypeError
        self.attributs = []
        self.vassaux = WeakList()   # lors de sa creation, l'objet n'a aucun vassal (aucun objet ne depend de lui)
        self.recalculer = True
        self.etiquette = False
        Objet.__init__(self) #  effectuer avant self.val (synchronise la feuille !!!)
        self._set_val(valeur)
        # l'enregistrement des variables est tres particulier, et se fait avec self.val...
    
    
        #self.enregistre((int, float, long, Cercle_generique, Segment, Glisseur_generique))
    def soustype(self):
        return isinstance(self.valeur, list) and "compose" or "simple"

    def _set_val(self, valeur = None):
        u"""val() propose une interface standard pour afficher une valeur
        (ou la modifier, si c'est une valeur libre), indpendamment de son origine.
        
        Pour comprendre ce qui se passe, il faut savoir que le contenu de la variable est stock dans self.valeur.
        S'il s'agit d'un nombre, il est stock tel quel, mais s'il s'agit d'une expression,
        le code qui suit va remplacer toutes les noms d'objets par les objets eux-mmes.
        Ainsi, si un objet est renomm, le contenu reste pertinent.
        Exemple :
        "u.norme() + v.norme()" devient ["", u, ".norme() + ", v, ".norme()"]
        Le premier terme de la liste est toujours une chane (ventuellement vide), et le dernier aussi.
        Le nombre d'lments de la liste est donc impair, avec une alternance de chanes et d'objets.
        
        La connaissance de ce fonctionnement interne est rarement requise, mais peut-tre utile
        pour court-circuiter le fonctionnement normal de Variable par exemple.
        """
                
        if valeur is not None:
            #self.recalculer = True
            if isinstance(valeur, Variable): # si u est une variable, Variable(u) renvoie une copie de u
                #val = valeur.valeur
                #if isinstance(val, list):
                #    val = str(valeur)
                #valeur = val
                valeur = valeur.valeur

            # on supprime la variable de la liste des vassaux pour les objets dont elle ne dpendra plus desormais:
            for objet in self.ancetres(): 
                objet.vassaux.remove(self)
            
            self.attributs = ["valeur"]
            
            if isinstance(valeur, basestring):
                #debug(valeur)
                liste = regexp("[^.A-Za-z0-9_][A-Za-z_][A-Za-z0-9_]*", " " + valeur.replace(" ",""), "x[0]+'$'+x[1:]+'$'").split("$")
                #debug(liste)
                for i in range(len(liste)):
                    if i%2:
                        if liste[i]:
                            liste[i] = self.__feuille__.objets[liste[i]]
                valeur = liste

            if isinstance(valeur, (list, tuple)):
            
                # La premire tape, c'est d'liminer les dpendances circulaires.
                # Par exemple, u = Variable(3); A = Point("u", "2*u"); u("A.x")
                # A deviendrait alors  la fois hritier de u (ie. on a besoin de u pour calculer A), et anctre de u (ie. on a besoin de A pour calculer u !)
                heritiers = self.heritiers() + [self]
                for ancetre in valeur:
                    if isinstance(ancetre, Objet):
                        for heritier in heritiers:
                            if ancetre is heritier:
                                liste = valeur[:]
                                for i in range(len(liste)):
                                    if i%2:
                                        liste[i] = [key for key, value in self.__feuille__.objets.__dict__.items() if value is liste[i]][0]
                                self.erreur(u"Dfinition circulaire dans %s : l'objet %s se retrouve dpendre de lui-mme." %("".join(liste).strip(), ancetre))
                                raise RuntimeError, "Definition circulaire dans %s : l'objet %s se retrouve dependre de lui-meme." %("".join(liste).strip(), ancetre)
                    
                self.valeur = list(valeur)
                dependances = self.ancetres()
                # plus self._hierarchie est faible, plus l'objet est haut place dans la _hierarchie
                if dependances: self.hierarchie(max([objet._hierarchie for objet in dependances]) + 1)
                else:           self._hierarchie = 0 # au sommet, les objets libres (ou suzerains)
                
                for objet in dependances:   # l'objet est vassal de chacun des objets dont il depend
                    objet.vassaux.append(self)
                    
            elif isinstance(valeur, (int, float, long)):
                self.valeur = valeur
                self._hierarchie = 0
            
            else:
                self.valeur = float(valeur)
                self._hierarchie = 0
                #raise TypeError, (u"Les arguments d'une variable doivent tre de type str, int, float ou list, et non %s" %type(valeur)).encode(param.encodage)

                
            
    def _get_val(self):
        if self.soustype() == "compose":
            liste = self.valeur[:]
            for i in range(len(liste)):
                if i%2: # annuaire inverse ;) (on retrouve le nom de l'objet)
                    try:
                        liste[i] = [key for key, value in self.__feuille__.objets.__dict__.items() if value is liste[i]][0]
                    except:
                        print u"Objet introuvable sur la feuille (id %s) : %s" %(id(self.__feuille__), liste[i])
                        raise
            val = eval("".join(liste), self.__feuille__.objets.__dict__)
            if isinstance(val, Variable): # Variable("u+1") devient un alias pratique de Variable("u()+1")
                val = val.val()           # (mais d'un autre cote, ca entretient la confusion entre u et u()...)
            return val
        else:
            return 1.*self.valeur

    _call = Objet.val
    
    # donne une representation affichable de l'objet qui sert a generer la variable
    def __repr__(self, styles = True):
        u"""Attention, pour les variables, l'argument styles n'a pas la mme signification.
        Il indique si l'affichage doit tre complet ou non.
        
        >>> Variable(7.2).__repr__(True)
        "Variable(7.2)"
        >>> Variable(7.2).__repr__(False)
        "7.2"
        """
        if isinstance(self.valeur, (int, float, long)):
            if styles:
                return "Variable(%s)" %self.val()
            else:
                return repr(self.val())
        else:
            liste = self.valeur[:]
            for i in range(len(liste)):
                if i%2:
                    liste[i] = [key for key, value in self.__feuille__.objets.__dict__.items() if value is liste[i]][0]
            return "Variable('%s')" % "".join(liste).strip()
            

    
    def __str__(self):
        if isinstance(self.valeur, (int, float, long)):
            return str(self.val())
        else:
            liste = self.valeur[:]
            for i in range(len(liste)):
                if i%2:
                    liste[i] = [key for key, value in self.__feuille__.objets.__dict__.items() if value is liste[i]][0]
            return "".join(liste).strip() # renvoit la chaine qui sert a initialiser la variable
    
    
    
#    def varier(self, debut = 0, fin = 1, pas = 0.02, periode = 0.03):
#        if self.__canvas__():
#            self.variation = [wx.Timer(self.__canvas__()), debut, fin, pas]
#            self.variation[0].Start(1000*periode)
#            
#            self.__canvas__().Bind(wx.EVT_TIMER, self._varier)
        
#    def _varier(self, event = None):
#        self(self.variation[1])
#        self.variation[1] += self.variation[3]
#        if self.variation[3]*(self.variation[1] - self.variation[2]) > 0:
#        self.variation[0].Stop()


# Cette portion de code marche, mais son comportement des que je modifie qqchose est surrealiste !!!        
        

    def varier(self, debut = 0, fin = 1, pas = 0.02, periode = 0.03):
        if self.__feuille__ is not None:
            self.__feuille__.start()
            for i in concatenate((arange(debut, fin, pas), (fin,))):
                t = time.clock()
                self.val(i)
                while time.clock() < t + periode:
                    wx.YieldIfNeeded()
                    if self.__feuille__.__stop__:
                        break
                wx.YieldIfNeeded()
                if self.__feuille__.__stop__:
                    break


    #def varier(self):
    #    self.timer = wx.FutureCall(1, self.nvarier)
        
    #def nvarier(self):
    #    self.bidon = 3

#    def copy(self):
#        if self.type() == "compose":
#            return Variable(repr(
    
    
        
#    def __add__(self, y):
#        if isinstance(y, (float, long, int)):
#            return Variable("%s()+%s" % (self.nom(), y))
#        if isinstance(y, self.__class__):
#            return Variable("%s()+%s()" % (self.nom(), y.nom()))
#        raise TypeError
    
#    def __radd__(self, y):
#        if isinstance(y, (float, long, int)):
#            return Variable("%s+%s()" % (y, self.nom()))
#        if isinstance(y, self.__class__):
#            return Variable("%s()+%s()" % (self.nom(), y.nom()))
#        raise TypeError
        
    
    # On definit __add__, __radd__, __sub__, etc...
#    for operation, symbole in (("add", "+"), ("sub", "-"), ("mul", "*"), ("div", "*1./"), ("pow", "**")):

#        def __op__(self, y, operation = "__" + operation + "__", symbole = symbole):
#            msg = "unsupported operand types for %s: 'Variable' and '%s'" %(symbole, type(y))
#            if self.type() == "simple":
#                if isinstance(y, (float, long, int)):
#                    return Variable(getattr(self(), operation)(y))
#                elif isinstance(y, Variable):
#                    if y.type() == "simple":
#                        return Variable(getattr(self(), operation)(y()))
#                    else:
#                        return Variable("%s%s%s" % (self, symbole, y))
#                else:
#                    raise TypeError, msg
#            elif isinstance(y, (float, long, int, Variable)):
#                return Variable("%s%s%s" % (self, symbole, y))
#            else:
#                raise TypeError, msg
#    
#        def __rop__(self, y, operation = "__r" + operation + "__", symbole = symbole):
#            msg = "unsupported operand types for %s: '%s' and 'Variable'" %(symbole, type(y))
#            if self.type() == "simple":
#                if isinstance(y, (float, long, int)):
#                    return Variable(getattr(self(), operation)(y))
#                elif isinstance(y, Variable):
#                    if y.type() == "simple":
#                        return Variable(getattr(self(), operation)(y()))
#                    else:
#                        return Variable("%s%s%s" % (y, symbole, self))
#                else:
#                    raise TypeError, msg
#            elif isinstance(y, (float, long, int, Variable)):
#                return Variable("%s%s%s" % (y, symbole, self))
#            else:
#                raise TypeError, msg
            
        
    
#    for nom in fonctions_matplotlib:
#        exec("""
#def %s(self, *args):
#    if self.type() == "simple":
#    arguments = []
#    for arg in args:
#        arguments.append(float(arg))
#    return Variable(%s(float(self), *arguments))""" %(nom, nom))

#    def __float__(self):
#        return self.val()
        
#    def __int__(self):
#        return int(float(self))
        
#    def __neg__(self):
#        return 0 - self
        
#    def __pos__(self):
#        return 0 + self

#    def __eq__(self, y):
        #print type(self), type(y)
#        return hasattr(y, "__float__") and float(self) == float(y)
        
#    def __ne__(self, y):
#        return not self == y
        
#    def __nonzero__(self):
#        return self <> 0
        
#    def __gt__(self, y):
#        return hasattr(y, "__float__") and float(self) > float(y)
        
#    def __ge__(self, y):
#        return self > y or self == y
        
    def add(self, y):
        u"Addition lie (le rsultat est une variable qui reste toujours gale  la somme des 2 valeurs)."
        if self.soustype() == "simple":
            if isinstance(y, Variable) and y.soustype() == "simple":
                return Variable(self()+y())
            elif isinstance(y, (float, int, long, complex)):
                return Variable(self()+y)
        return Variable("(%s)+(%s)" %(self, y))
    
    def mult(self, y):
        u"Multiplication lie (le rsultat est une variable qui reste toujours gale au produit des 2 valeurs)."
        if self.soustype() == "simple":
            if isinstance(y, Variable) and y.soustype() == "simple":
                return Variable(self()*y())
            elif isinstance(y, (float, int, long, complex)):
                return Variable(self()*y)
        return Variable("(%s)*(%s)" %(self, y))
        
        
        

    def _conditions_existence(self): # conditions specifiques pour que l'objet existe, a definir pour chaque objet
        try:
            self._get_val()
            return [True]
        except:
            print_error()
            return [False]
    
    # L'ide est bonne, mais le code actuel est difficiel  faire voluer.
    # Pour l'instant, la mthode 'val()' est dfinie ainsi :
    
    #def val(self, *args, **kw):
    #    if self.recalculer or args or kw:
    #        self.recalculer = True  # pour les conditions d'existence
    #        if self.etiquette:
    #            self.etiquette.recalculer = True
    #        for objet in self.heritiers():#classer = True):
    #            objet.recalculer = True
    #            if objet.etiquette:
    #                objet.etiquette.recalculer = True
    #        if self.existe():
    #            self._valeur = self._get_val(*args, **kw)
    #        else:
    #            self._valeur = None # si l'objet n'existe pas, renvoit None
    #        self.recalculer = False
    #    return self._valeur
    
    # Notes concernant le code prcdent :
    # - Un objet qui "n'existe pas" (au sens gomtrique) a des coordonnes fixes  None.
    # - Un objet libre (c'est--dire qui accepte des arguments pour ses mthodes coordonnees() ou equation())
    #   ne doit pas avoir de conditions d'existence.

    # Ainsi, comme indiqu dans la note, un objet qui n'existe pas ne peut pas tre redfini actuellement.