# -*- coding: iso-8859-1 -*-
from __future__ import division
import re, math
#from pylib import advanced_split
from redefinir import *
import redefinir, pylib
from objets import *

def somme(*args):
    return reduce(lambda x, y:x+y, args, 0)

def produit(*args):
    return reduce(lambda x, y:x*y, args, 1)

types_numeriques = (int, float, long, complex, Reel)

def isnumtype(obj):
    return isinstance(obj, types_numeriques) or isinstance(obj, Constante) and obj.isnum()

def isknowntype(obj):
    return isinstance(obj, types_numeriques) or isinstance(obj, Objet_formel)

def isinttype(obj):
    return isnumtype(obj) and int(obj) == obj

def derivee_abs(x):
    if x == 0:
        raise ValueError, "La valeur absolue n'est pas derivable en 0"
    return cmp(x, 0)




class Objet_formel(object):
    def __neg__(self):
        return Expression_formelle("-", self)

    def __pos__(self):
        return self

    def __add__(self, y):
    # VERIFIER LES TYPES
    # IMPLEMENTER LES FONCTIONS (f+2, 2f, etc.)
        if not isknowntype(y):
            return NotImplemented
        if y == 0:
            return self
        if isnumtype(y) and y < 0:
            return Expression_formelle("-", self, y)
        return Expression_formelle("+", self, y)

    def __radd__(self, y):
        if not isknowntype(y):
            return NotImplemented
        if y == 0:
            return self
        if isnumtype(y) and y < 0:
            return Expression_formelle("-", y, self)
        return Expression_formelle("+", y, self)

    def __sub__(self, y):
        if not isknowntype(y):
            return NotImplemented
        if y == 0:
            return self
        return Expression_formelle("-", self, y)

    def __rsub__(self, y):
        if not isknowntype(y):
            return NotImplemented
        if y == 0:
            return - self
        return Expression_formelle("-", y, self)

    def __mul__(self, y):
        if not isknowntype(y):
            return NotImplemented
        if y == 0:
            return Constante(0)
        if y == 1:
            return self
        if y == -1:
            return -self
        return Expression_formelle("*", self, y)

    def __rmul__(self, y):
        if not isknowntype(y):
            return NotImplemented
        if y == 0:
            return Constante(0)
        if y == 1:
            return self
        if y == -1:
            return -self
        return Expression_formelle("*", y, self)

    def __div__(self, y):
        if not isknowntype(y):
            return NotImplemented
        if y == 0:
            raise ZeroDivisionError
        if y == 1:
            return self
        if y == -1:
            return -self
        return Expression_formelle("/", self, y)

    def __rdiv__(self, y):
        if not isknowntype(y):
            return NotImplemented
        if y == 0:
            return Constante(0)
        return Expression_formelle("/", y, self)

    __truediv__ = __div__
    __rtruediv__ = __rdiv__

    def __pow__(self, y):
        if not isknowntype(y):
            return NotImplemented
        if y == 1:
            return self
        if y == 0:
            return Constante(1)
        return Expression_formelle("^", self, y)

    def __rpow__(self, y):
        if not isknowntype(y):
            return NotImplemented
        if y == 1:
            return Constante(1)
        return Expression_formelle("^", y, self)

    def __xor__(self, y):
        if not isknowntype(y):
            return NotImplemented
        return self.__pow__(y)

    def __rxor__(self, y):
        if not isknowntype(y):
            return NotImplemented
        return self.__rpow__(y)

    def log(self):
        return Expression_formelle("ln", self)

    def exp(self):
        if not isknowntype(self):
            return NotImplemented
        return Expression_formelle("exp", self)

    def cos(self):
        return Expression_formelle("cos", self)

    def sin(self):
        return Expression_formelle("sin", self)

    def tan(self):
        return Expression_formelle("tan", self)

    def abs(self):
        return Expression_formelle("abs", self)

    __abs__ = abs

    def sqrt(self):
        return Expression_formelle("sqrt", self)


class Constante(Objet_formel):
    u"""Un conteneur pour une constante.

    L'argument valeur peut soit tre un nombre, soit le nom d'une constante connue (pi, e).

    >>> Constante("pi", 3.14)
    Constante('pi')
    >>> Constante(3.14)
    Constante(3.14)
    """

    def __init__(self, nom_ou_valeur, valeur_approchee = None):
        self.valeur = nom_ou_valeur
        if valeur_approchee is None:
            valeur_approchee = nom_ou_valeur
        if isinstance(valeur_approchee, basestring):
            raise ValueError, "Vous devez fournir une valeur approchee de la constante %s." %nom_ou_valeur
        self.valeur_approchee = valeur_approchee

    def isnum(self):
        "L'objet 'Constante' peut contenir ou bien un nombre, ou bien un nom de constante connue."
        return not isinstance(self.valeur, basestring)

    def isfloat(self):
        "L'objet 'Constante' est dfini  partir d'un nombre de type float (et est considr comme une valeur approche)."
        return isinstance(self.valeur, float)

    def __float__(self):
        if self.isnum():
            return float(self.valeur)
        return float(self.valeur_approchee)

    def __int__(self):
        if self.isnum():
            return int(self.valeur)
        return int(self.valeur_approchee)

    def __index__(self):
        if self.isnum():
            return int(self.valeur)
        return NotImplemented

    def __str__(self):
        if self.isnum():
            return str(self.valeur)
        return self.valeur

    def __repr__(self):
        if self.isnum():
            return "Constante(%s)" %self.valeur
        return "Constante('%s')" %self.valeur


    def __eq__(self, y):
        return self.isnum() and self.valeur == y or isinstance(y, Constante) and self.valeur == y.valeur

    def __gt__(self, y):
        if isinstance(y, Constante):
            return self.valeur_approchee > y.valeur_approchee
        return self.valeur_approchee > y

    def __ge__(self, y):
        return not self < y

    def __lt__(self, y):
        if isinstance(y, Constante):
            return self.valeur_approchee < y.valeur_approchee
        return self.valeur_approchee < y

    def __le__(self, y):
        return not self > y

    def __cmp__(self, y):
        if self.isnum() and isnumtype(y):
            return cmp(float(self), float(y))
        return NotImplemented

    def __neg__(self):
        if self.isnum():
            return Constante(- self.valeur)
        return Objet_formel.__neg__(self)


    def __add__(self, y):
        if self.isnum():
            if self == 0:
                return y
            elif isinstance(y, Constante) and y.isnum():
                return Constante(self.valeur + y.valeur)
            elif isinstance(y, types_numeriques):
                return Constante(self.valeur + y)
        return Objet_formel.__add__(self, y)

    def __radd__(self, y):
        if self.isnum():
            if self == 0:
                return y
            elif isinstance(y, Constante) and y.isnum():
                return Constante(self.valeur + y.valeur)
            elif isinstance(y, types_numeriques):
                return Constante(self.valeur + y)
        return Objet_formel.__radd__(self, y)

    def __sub__(self, y):
        if self.isnum():
            if self == 0:
                return -y
            elif isinstance(y, Constante) and y.isnum():
                return Constante(self.valeur - y.valeur)
            elif isinstance(y, types_numeriques):
                return Constante(self.valeur - y)
        return Objet_formel.__sub__(self, y)

    def __rsub__(self, y):
        if self.isnum():
            if self == 0:
                return y
            elif isinstance(y, Constante) and y.isnum():
                return Constante(y.valeur - self.valeur)
            elif isinstance(y, types_numeriques):
                return Constante(y - self.valeur)
        return Objet_formel.__rsub__(self, y)

    def __mul__(self, y):
        if self.isnum():
            if self == 0:
                return 0
            elif self == 1:
                return y
            elif self == -1:
                return -y
            elif isinstance(y, Constante) and y.isnum():
                return Constante(self.valeur * y.valeur)
            elif isinstance(y, types_numeriques):
                return Constante(self.valeur * y)
        return Objet_formel.__mul__(self, y)

    def __rmul__(self, y):
        if self.isnum():
            if self == 0:
                return 0
            elif self == 1:
                return y
            elif self == -1:
                return -y
            elif isinstance(y, Constante) and y.isnum():
                return Constante(self.valeur*y.valeur)
            elif isinstance(y, types_numeriques):
                return Constante(self.valeur*y)
        return Objet_formel.__rmul__(self, y)

    def __div__(self, y):
        if self.isnum():
            if self == 0:
                return 0
            elif isinstance(y, Constante) and y.isnum():
                if isinstance(self.valeur, (float, complex, Reel)):
                    return Constante(self.valeur/y.valeur)
                else:
                    return Constante(Frac(self.valeur)/y.valeur)
            elif isinstance(y, types_numeriques):
                if isinstance(self.valeur, (float, complex, Reel)):
                    return Constante(self.valeur/y)
                else:
                    return Constante(Frac(self.valeur)/y)
        return Objet_formel.__div__(self, y)

    def __rdiv__(self, y):
        if self.isnum():
            if self == 0:
                raise ZeroDivisionError
            elif self == 1:
                return y
            elif self == -1:
                return -y
            elif isinstance(y, Constante) and y.isnum():
                if isinstance(y, (float, complex, Reel)):
                    return Constante(y.valeur/self.valeur)
                else:
                    return Constante(Frac(y.valeur)/self.valeur)
            elif isinstance(y, types_numeriques):
                if isinstance(y, (complex, float, Reel)):
                    return Constante(y/self.valeur)
                else:
                    return Constante(Frac(y)/self.valeur)
        return Objet_formel.__rdiv__(self, y)

    __truediv__ = __div__
    __rtruediv__ = __rdiv__

    def __floordiv__(self, y):
        return Constante(int(float(self)/y))

    def __rfloordiv__(self, y):
        return Constante(int(y/float(self)))

    def __mod__(self, y):
        return self - y*(self//y)

    def __rmod__(self, y):
        return y - self*(y//self)

    def __pow__(self, y):
        if self.isnum():
            if self == 1:
                return 1
            elif isinstance(y, Constante) and y.isnum():
                if isinstance(self.valeur, (int, long)) and y<0:
                    return Constante(Frac(self.valeur)**y.valeur)
                return Constante(self.valeur**y.valeur)
            elif isinstance(y, types_numeriques):
                if isinstance(self.valeur, (int, long)) and y<0:
                    return Constante(Frac(self.valeur)**y)
                return Constante(self.valeur**y)
        elif self.valeur == "e":
            return Expression_formelle("exp", y)
        return Objet_formel.__pow__(self, y)

    def __rpow__(self, y):
        if self.isnum():
            if self == 0:
                return 1
            elif self == 1:
                return y
            elif isinstance(y, Constante) and y.isnum():
                return Constante(y.valeur**self.valeur)
            elif isinstance(y, types_numeriques):
                return Constante(y**self.valeur)
        return Objet_formel.__rpow__(self, y)

    def sqrt(self):
        if self.isnum():
            if self.isfloat():
                return Constante(sqrt(self.valeur))
            elif isinstance(self.valeur, (int, long, Frac)):
                return Constante(Rac(self.valeur))
        return Objet_formel.sqrt(self)

    def abs(self):
        if self < 0:
            return -self
        else:
            return self

    def sin(self):
        if self.valeur == 0:
            return Constante(0)
        elif self.valeur == "pi":
            return Constante(0)
        return Objet_formel.sin(self)

    def cos(self):
        if self.valeur == 0:
            return Constante(1)
        elif self.valeur == "pi":
            return Constante(-1)
        return Objet_formel.cos(self)

    def tan(self):
        if self.valeur == 0:
            return Constante(0)
        elif self.valeur == "pi":
            return Constante(0)
        return Objet_formel.tan(self)

    def log(self):
        if self.valeur == 1:
            return Constante(0)
        elif self.valeur == "e":
            return Constante(1)
        elif self <= 0:
            raise ValueError, "math domain error"
        return Objet_formel.log(self)

    def evaluer(self, **kw):
        return self

    def factorise(self):
        if self.isnum() and int(self.valeur) == self.valeur:
            return Expression_formelle("*", *facteurs(int(self.valeur)))
        return self




class Variable_formelle(Objet_formel):
    u"""Une variable utilisable en calcul formel.

    >>> Variable_formelle("x")
    Variable_formelle('x')
    """
    def __init__(self, nom):
        if isinstance(nom, Variable_formelle):
            nom = nom.nom
        self.nom = nom

    def __str__(self):
        return self.nom

    def __repr__(self):
        return "Variable_formelle('%s')" %self.nom

    def __eq__(self, y):
        return isinstance(y, Variable_formelle) and self.nom == y.nom

    def derive(self, variable):
        if isinstance(variable, basestring):
            variable = Variable_formelle(variable)
        if variable == self:
            return 1
        return 0

    def __cmp__(self, y):
        if isinstance(y, Variable_formelle):
            return cmp(self.nom, y.nom)
        if isnumtype(y):
            return 1
        return NotImplemented


    def evaluer(self, **kw):
        u"""Evalue l'expression en assignant une valeur donne  certaines variables.
        >>> a = Variable_formelle("x")
        >>> a.evaluer(x=5)
        5
        >>> a.evaluer(y=7)
        Variable_formelle('x')
        >>> a.evaluer(x=2, y=17)
        2
        """
        return kw.get(self.nom, self)

    def factorise(self):
        return self


class Expression_formelle(Objet_formel):
    u"""Conteneur pour une expression mathmatique quelconque.

    Exemple:
    (2+7ln(x)+x)/(x+cox(x))
    est stock sous la forme :
    ("/", ("+", 2, ("+", ("*", 7, ("ln", x)), x)), ("+", x, ("cos", x)))

    Note:
    Pour rentrer les arguments sous forme de chane de caractres, utilisez la fonction expression_formelle().

    >>> Expression_formelle("+", 7, Variable_formelle("x"), Variable_formelle("y"))
    7+Variable_formelle('x')+Variable_formelle('y')
    """

    environnement = redefinir.__dict__.copy() # Environnement de travail courant.

    def __init__(self, *arguments, **kw):
        self.contenu = list(arguments)

    def operateur(self):
        return self.contenu[0]

    def __repr__(self):
        def formater(objet):
            if isinstance(objet, Expression_formelle) and objet.operateur() in "+-/*^":
                if operateur == "+" or objet.operateur() == "^" <> operateur or objet.operateur() in "*/" and operateur in "+-":
                    return repr(objet) # la puissance est prioritaire, les parenthses sont inutiles, idem pour la multiplication ou la division face  l'addition ou la soustraction
                return "(%s)" %repr(objet) # une expression un peu complexe est mise entre parenthses, dans le doute
            return repr(objet)

        operateur = self.operateur()
        if operateur in "+-/*^":
            if operateur == "-" and len(self.contenu) == 2:
                return "-" + formater(self.contenu[1])
            else:
                return operateur.join(formater(elt) for elt in self.contenu[1:])
        return operateur + "(" + ",".join(repr(elt) for elt in self.contenu[1:]) + ")"


    def __str__(self):
        def formater(objet, enlever_parentheses = False):
            if isinstance(objet, Expression_formelle) and objet.operateur() in "+-/*^":
                if enlever_parentheses or operateur == "+" or objet.operateur() == "^" <> operateur or objet.operateur() in "*/" and operateur in "+-" or operateur == objet.operateur() =="*":
                    return str(objet) # la puissance est prioritaire, les parenthses sont inutiles, idem pour la multiplication ou la division face  l'addition ou la soustraction
                return "(%s)" %objet # une expression un peu complexe est mise entre parenthses, dans le doute
            return str(objet)

        operateur = self.operateur()
        if operateur in "+-/*^":
            if operateur == "-":
                if len(self.contenu) == 2:
                    return "-" + formater(self.contenu[1])
                else:
                    return formater(self.contenu[1], enlever_parentheses=True) + "-" + formater(self.contenu[2]) # les parenthses ne sont utiles qu'aprs le '-'
            else:
                return operateur.join(formater(elt) for elt in self.contenu[1:])
        return operateur + "(" + ",".join(str(elt) for elt in self.contenu[1:]) + ")"

    def __eq__(self, y):
        return isinstance(y, Expression_formelle) and self.contenu == y.contenu

    def __cmp__(self, y):
        if isinstance(y, Expression_formelle):
            return cmp(self.contenu, y.contenu)
        return 1

    def __pow__(self, y):
        if self.operateur() == "^" and isnumtype(self.contenu[2]) and isnumtype(y) and self.contenu[2] == int(self.contenu[2]) and y == int(y):
            return self.contenu[1]**(self.contenu[2]*y)
        elif self.operateur() == "sqrt" and isnumtype(y) and int(y) == y and y%2 == 0:
            return self.contenu[1]**(y/2) # un peu abusif (on largit l'ensemble de dfinition)
        return Objet_formel.__pow__(self, y)


    def sin(self):
        f = self.fraction_de_pi()
        if f is not None:
            pass
        return Objet_formel.sin(self)


    def cos(self):
        f = self.fraction_de_pi()
        if f is not None:
            pass
        return Objet_formel.cos(self)

    def tan(self):
        f = self.fraction_de_pi()
        if f is not None:
            pass
        return Objet_formel.tan(self)

    def sqrt(self):
        if self.operateur() == "^":
            u = self.contenu[2]
            if isnumtype(u) and int(u) == u and int(u)%2 == 0:
                return abs(self.contenu[1]**(u/2))
        return Objet_formel.sqrt(self)

    def abs(self):
        if self.operateur() == "^":
            u = self.contenu[2]
            if isnumtype(u) and int(u) == u and int(u)%2 == 0:
                return self
        return Objet_formel.abs(self)

    __abs__ = abs

    def fraction_de_pi(self):
        return # A ECRIRE
        return Frac(n, d)
        #ou
        return None

    def derive(self, variable = None):
        if variable is None:
            variables = self.variables()
            if Variable_formelle("x") in variables:
                variable = "x"
            elif Variable_formelle("y") in variables:
                variable = "y"
            elif Variable_formelle("z") in variables:
                variable = "z"
            elif Variable_formelle("t") in variables:
                variable = "t"
            elif len(variables):
                variables.sort()
                variable = variables[0]
            else:
                variable = "x"

        if isinstance(variable, basestring):
            variable = Variable_formelle(variable)
        if len(self.contenu) >= 2:
            u = self.contenu[1]
            if len(self.contenu) >= 3:
                v = self.contenu[2]
            else:
                v = None
        if self.operateur() == "+":
            return somme(*(derive(elt, variable) for elt in self.contenu[1:]))
        if self.operateur() == "-":
            if v is None:
                return - derive(u, variable)
            return derive(u, variable) - derive(v, variable)
        if self.operateur() == "*":
            return derive(u, variable)*produit(*self.contenu[2:]) + u*derive(produit(*self.contenu[2:]), variable)
        if self.operateur() == "/":
            return (derive(u, variable)*v - u*derive(v, variable))/v**2
        if self.operateur() == "^":
            if v == 0:
                return 0
            if isinstance(v, types_numeriques):
                return v*derive(u, variable)*u**(v - 1)
            return derive(v*ln(u), variable)*u**v
        if self.operateur() == "cos":
            return - derive(u, variable)*sin(u)
        if self.operateur() == "sin":
            return derive(u, variable)*cos(u)
        if self.operateur() == "ln":
            return derive(u, variable)/u
        if self.operateur() == "tan":
            return derive(u, variable)/cos(u)**2
        if self.operateur() == "exp":
            return derive(u, variable)*exp(u)
        if self.operateur() == "abs":
            return derive(u, variable)*derivee_abs(u)
        if self.operateur() == "sqrt":
            return derive(u, variable)/(2*sqrt(u))
        return NotImplemented


    def developpe(self):
        u"""Dveloppe l'expression."""
        resultat = self.copier()
        resultat._supprimer_moins()
        resultat._supprimer_puissances()
        resultat._developpe()
        return resultat


    def simplifier(self):
        u"""Simplifie (lgrement) l'criture de l'expression.

        >>> u = Expression_formelle("+", 7, 4)
        >>> u.simplifier()
        Constante(11)
        """
        resultat = self.copier()
        resultat._supprimer_moins()
        resultat._supprimer_divisions_numeriques()
        resultat._rassembler()
        resultat = resultat._compacter()
        if hasattr(resultat, "_ordonner"):
            resultat._ordonner()
        #print "avant : ", resultat.contenu
        #resultat._retablir_moins()
        #print "apres : ", resultat.contenu
        if hasattr(resultat, "_reformater"):
            return resultat._reformater()
        return resultat

    def _rassembler(self):
        u"""Regroupe les sommes (respectivement, les produits) successifs en une grande somme (respectivement un grand produit).

        Par exemple, '7*((x+3)+y)' devient '7*(x+3+y)'.
        """
        operateur = self.operateur()
        nouveau_contenu = [operateur]
        for elt in self.contenu[1:]:
            if isinstance(elt, Expression_formelle):
                elt._rassembler()
                if operateur in "+*" and elt.operateur() == operateur:
                    nouveau_contenu += elt.contenu[1:]
                else:
                    nouveau_contenu.append(elt)
            else:
                nouveau_contenu.append(elt)
        self.contenu = nouveau_contenu

    def _compacter(self):
        for i in xrange(1, len(self.contenu)):
            if isinstance(self.contenu[i], Expression_formelle):
                self.contenu[i] = self.contenu[i]._compacter()
        operateur = self.operateur()

        if operateur == "*":
            # On multiplie tous les nombres entre eux.
            total = produit(*(elt for elt in self.contenu if isnumtype(elt)))
            self.contenu = [elt for elt in self.contenu if not isnumtype(elt)]
            # On regrouppe les autres lments (par exemple, x**2 et x**5 vont tre regroups en x**7)
            autres_types = []
            quantites = []
            for elt in self.contenu[1:]:
                if isinstance(elt, Expression_formelle) and elt.operateur() == "^" and isnumtype(elt.contenu[2]):
                    # ici, il faudrait classer les termes en cas d'expression formelle, pour qu'ils apparaissent toujours dans le mme ordre, afin d'identifier
                    val = elt.contenu[2]
                    elt = elt.contenu[1]
                else:
                    val = 1
                if isinstance(elt, Expression_formelle) and elt.operateur() == "+": # oprateurs commutatifs autres que *
                    l = elt.contenu[1:]
                    l.sort()
                    elt.contenu = [elt.operateur()] + l
                if elt in autres_types:
                    quantites[autres_types.index(elt)] += val
                else:
                    autres_types.append(elt)
                    quantites.append(val)
            n = len(quantites)
            if n:
                if total == 1:
                    if n > 1:
                        self.contenu = ["*"] + [autres_types[i]**quantites[i] for i in xrange(n)]
                    else:
                        return autres_types[0]**quantites[0]
                else:
                    self.contenu = ["*", total] + [autres_types[i]**quantites[i] for i in xrange(n)]
            else:
                return total

        elif operateur == "+":
            # On additionne tous les nombres entre eux.
            total = somme(*(elt for elt in self.contenu if isnumtype(elt)))
            self.contenu = [elt for elt in self.contenu if not isnumtype(elt)]

            # On regrouppe les autres lments (par exemple, 2*x et 5*x vont tre regroups en 7*x)
            autres_types = []
            quantites = []
            for elt in self.contenu[1:]:
                if isinstance(elt, Expression_formelle) and elt.operateur() == "*" and True in [isnumtype(ss_elt) for ss_elt in elt.contenu[1:]]:
                    # ici, il faudrait classer les termes en cas d'expression formelle, pour qu'ils apparaissent toujours dans le mme ordre, afin d'identifier
                    non_num = [ss_elt for ss_elt in elt.contenu[1:] if not isnumtype(ss_elt)]
                    val = produit(*(ss_elt for ss_elt in elt.contenu[1:] if isnumtype(ss_elt)))
                    if len(non_num) > 1:
                        elt = Expression_formelle("*", *non_num)
                    else:
                        elt = non_num[0]
                else:
                    val = 1
                if isinstance(elt, Expression_formelle) and elt.operateur() == "*": # oprateurs commutatifs autres que +
                    l = elt.contenu[1:]
                    l.sort()
                    elt.contenu = [elt.operateur()] + l
                if elt in autres_types:
                    quantites[autres_types.index(elt)] += val
                else:
                    autres_types.append(elt)
                    quantites.append(val)
            n = len(quantites)
            if n:
                if total == 0:
                    if n > 1:
                        self.contenu = ["+"] + [quantites[i]*autres_types[i] for i in xrange(n)]
                    else:
                        return quantites[0]*autres_types[0]
                else:
                    self.contenu = ["+"] + [quantites[i]*autres_types[i] for i in xrange(n)] + [total]
            else:
                return total
        return self






#            def cmp(x, y):
#                if isinstance(x, Expression_formelle):
#                    return 1
#                if isinstance(y, Expression_formelle):
#                    return -1
#                if isinstance(x, Variable_formelle):
#                    return

    def _supprimer_moins(self):
        u"Supprime les soustractions : '5 - x' devient '5 + (-1)*x'"
        if self.operateur() == "-":
            # Cas -([expression])
            if len(self.contenu) == 2:
                # Sous-cas -([...]+[...])
                if isinstance(self.contenu[1], Expression_formelle) and self.contenu[1].operateur() == "+":
                    # si le membre  soustraire est une somme, on distribue le "-" :
                    self.contenu = ["+"] + [Expression_formelle("-", elt) for elt in self.contenu[1].contenu[1:]]
                elif isinstance(self.contenu[1], Expression_formelle) and self.contenu[1].operateur() == "-":
                    # Sous-cas -(-[...])
                    if len(self.contenu[1].contenu) == 2:
                        self.contenu = ["+", self.contenu[1].contenu[1]] #-- devient +
                    # Sous-cas -([...]-[...])
                    else:
                        self.contenu = ["+", Expression_formelle("-", self.contenu[1].contenu[1]), self.contenu[1].contenu[2]]
                else:
                    self.contenu = ["*", Constante(-1), self.contenu[1]]
            # Cas ([expression1]) - ([expression2])
            else:
                # Sous-cas ([expression1])-([...]+[...])
                if isinstance(self.contenu[2], Expression_formelle) and self.contenu[2].operateur() == "+":
                    # si le membre  soustraire est une somme, on distribue le "-" :
                    self.contenu = ["+", self.contenu[1]] + [Expression_formelle("-", elt) for elt in self.contenu[2].contenu[1:]]
                elif isinstance(self.contenu[2], Expression_formelle) and self.contenu[2].operateur() == "-":
                    # Sous-cas ([expression1])-(-[...])
                    if len(self.contenu[2].contenu) == 2:
                        self.contenu = ["+", self.contenu[1], self.contenu[2].contenu[1]] #-- devient +
                    # Sous-cas ([expression1])-([...]-[...])
                    else:
                        self.contenu = ["+", self.contenu[1], Expression_formelle("-", self.contenu[2].contenu[1]), self.contenu[2].contenu[2]]
                else:
                    self.contenu = ["+", self.contenu[1], Expression_formelle("*", Constante(-1), self.contenu[2])]
        
        for elt in self.contenu[1:]:
            if isinstance(elt, Expression_formelle):
                elt._supprimer_moins()



    def _supprimer_divisions_numeriques(self):
        u"Supprime les divisions par un nombre : 'x/5' devient '(1/5)*x'"
        if self.operateur() == "/" and isnumtype(self.contenu[2]):
            self.contenu = ["*", Constante(1)/self.contenu[2], self.contenu[1]]
        
        for elt in self.contenu[1:]:
            if isinstance(elt, Expression_formelle):
                elt._supprimer_divisions_numeriques()
    
    
    
    def _retablir_moins(self): # A AMELIORER (l'ordre des lments d'une somme ne devrait pas changer, pour que les monmes restent classs)
        u"Transformer +-3*x en -3*x. A faire en tout dernier."
        for elt in self.contenu[1:]:
            if isinstance(elt, Expression_formelle):
                elt._retablir__supprimer_divisions_numeriquesmoins()

        if self.operateur() == "+":
            for i in xrange(1, len(self.contenu[1:])):
                elt = self.contenu[i]
                if isinstance(elt, Expression_formelle):
                    # on traite le cas [expression1] + (-[expression2])
                    if elt.operateur() == "-" and len(elt.contenu) == 2:
                        reste = self.contenu[1:i] + self.contenu[i+1:]
                        if len(reste) > 1:
                            self.contenu = ["-", Expression_formelle("+", *reste), elt.contenu[1]]
                        else:
                            self.contenu = ["-", reste[0], elt.contenu[1]]
                        self._retablir_moins()
                        break

                    # on traite le cas [expression1] + ([expression_numerique_negative]*[expression(s)])
                    # on suppose que expression_numerique_negative est plac au dbut du produit, ce qui est en principe
                    # le cas, si les mthodes ._rasssembler() et ._compacter() ont t appeles auparavant.
                    elif elt.operateur() == "*" and isnumtype(elt.contenu[1]) and elt.contenu[1] < 0:
                        reste = self.contenu[1:i] + self.contenu[i+1:]
                        produit = Expression_formelle("*", -elt.contenu[1], *elt.contenu[2:])
                        if len(reste) > 1:
                            self.contenu = ["-", Expression_formelle("+", *reste), produit]
                        else:
                            self.contenu = ["-", reste[0], produit]
                        self.contenu = ["-", produit, ]
                        self._retablir_moins()
                        break


    def _supprimer_puissances(self):
        u"Supprime toutes les puissances entires de l'expression : '(x+3)^n' devient '(x+3)*(x+3)*(x+3)'."
        if self.operateur() == "^" and isinttype(self.contenu[2]):
            contenu = self.contenu[1]
            if isinstance(contenu, Expression_formelle):
                contenu._supprimer_puissances()
            n = int(self.contenu[2])
            self.contenu = ["*"] + n*[self.contenu[1]]
        else:
            for elt in self.contenu[1:]:
                if isinstance(elt, Expression_formelle):
                    elt._supprimer_puissances()


    def _developpe(self):
        u"Effectue le dveloppement proprement dit."
        for elt in self.contenu[1:]:
            if isinstance(elt, Expression_formelle):
                elt._developpe()

        if self.operateur() == "*":
            num = None # emplacement d'un facteur  dvelopper
            for i in xrange(1, len(self.contenu)):
                elt = self.contenu[i]
                if isinstance(elt, Expression_formelle) and elt.operateur() == "+":
                    num = i
            if num is not None:
                facteur = self.contenu[num]
                autres_facteurs = self.contenu[1:num] + self.contenu[num+1:]
                self.contenu = ["+"] + [Expression_formelle("*", elt, *autres_facteurs) for elt in facteur.contenu[1:]]

                self._developpe() # il reste peut-tre encore des choses  dvelopper.

        # on coupe les fractions (ceci permet surtout de ne pas passer  ct de dveloppements style '((x+3)/5)*x'.
        if self.operateur() == "/":
            numerateur, denominateur = self.contenu[1:]
            if isinstance(numerateur, Expression_formelle) and numerateur.operateur() == "+":
                self.contenu = ["+"] + [Expression_formelle("/", elt, denominateur) for elt in numerateur.contenu[1:]]




    def _ordonner(self):
        u"""Classe les termes des sommes en respectant l'ordre des monmes.

        Cette mthode prsuppose qu'un compactage ait dj eu lieu."""
        for elt in self.contenu[1:]:
            if isinstance(elt, Expression_formelle):
                elt._ordonner()

        if self.operateur() == "+":
            def classer(x, y):
                if x == "+": # le "+" doit rester au dbut
                    return -1
                elif y == "+":
                    return 1
                elif isnumtype(x): # les nombres en dernier
                    return 1
                elif isnumtype(y):
                    return -1
                elif isinstance(x, Constante):
                    if isinstance(y, Constante):
                        return cmp(x.valeur, y.valeur) # constantes classes par ordre alphabtique, juste devant les nombres
                    return 1
                elif isinstance(y, Constante):
                    return -1
                elif isinstance(x, (Expression_formelle, Variable_formelle)) and isinstance(y, (Expression_formelle, Variable_formelle)):
                    if isinstance(x, Variable_formelle) and isinstance(y, Variable_formelle):
                        return cmp(x.nom, y.nom)
                    elif (isinstance(x, Variable_formelle) or x.operateur() in "*^") and (isinstance(y, Variable_formelle) or y.operateur() in "*^"):
                        def degre(objet):
                            if isinstance(objet, Variable_formelle):
                                return 1
                            elif objet.operateur() == "^":
                                if isinttype(objet.contenu[2]):
                                    return int(objet.contenu[2])
                                return 0
                            elif objet.operateur() == "*":
                                return sum(degre(elt) for elt in objet.contenu[1:] if isinstance(elt, (Expression_formelle, Variable_formelle)))
                            return 0
                        return cmp(degre(y), degre(x)) # classement par degrs dcroissants

                    return 1
                return 0
            self.contenu.sort(cmp = classer)



    def _reformater(self):
        return expression_formelle(str(self))


    def factorise(self):
        u"""Factorise l'expression."""
        resultat = self.copier()
        resultat._supprimer_moins()
        resultat._supprimer_puissances()
        resultat._rassembler()
        return resultat._factorise()


    def _factorise(self):
        u"Effectue la factorisation proprement dite."
        def factor(elt):
            if hasattr(elt, "factorise"):
                return elt.factorise()
            return elt
            
        op = self.operateur()
        print "op: ", op
        if op == "*":
            return Expression_formelle("*", *(factor(elt) for elt in self.contenu[1:]))
        elif op == "/":
            return Expression_formelle("/", factor(self.contenu[1]), factor(self.contenu[2]))
        elif op == "^":
            return Expression_formelle("^", factor(self.contenu[1]), self.contenu[2])
        elif op == "sqrt":
            return Expression_formelle("sqrt", factor(self.contenu[1]))
        elif op == "+" or op == "-":
            variables = self.variables()
            for variable in variables: # on essaie de factoriser par la variable
                facteur = Expression_formelle("+")
                for elt in self.contenu[1:]:
                    if hasattr(elt, "_variables") and variable in elt._variables() and elt.operateur() == "*" and variable in elt.contenu[1:]:
                        l = elt.contenu[1:]
                        l.remove(variable)
                        if len(l) == 1:
                            facteur.contenu.append(l[0])
                        else:
                            facteur.contenu.append(Expression_formelle("*", *l))
                    elif isinstance(elt, Variable_formelle) and elt == variable:
                        facteur.contenu.append(Constante(1))
                    else:
                        facteur = None
                        break
                if facteur is not None:
                    print "factorisation par ", variable
                    return Expression_formelle("*", variable, facteur._factorise())
            constantes = self.constantes()
            for constante in constantes: # on essaie de factoriser par la constante
                facteur = Expression_formelle("+")
                for elt in self.contenu[1:]:
                    if hasattr(elt, "_constantes") and constante in elt._constantes() and elt.operateur() == "*" and constante in elt.contenu[1:]:
                        l = elt.contenu[1:]
                        l.remove(constante)
                        if len(l) == 1:
                            facteur.contenu.append(l[0])
                        else:
                            facteur.contenu.append(Expression_formelle("*", *l))
                    elif isinstance(elt, Constante) and elt == constante:
                        facteur.contenu.append(Constante(1))
                    else:
                        facteur = None
                        break
                if facteur is not None:
                    print "factorisation par ", constante
                    return Expression_formelle("*", constante, facteur._factorise())
                # il faut essayer de dtecter si tous les termes de la somme sont des multiples de la variable.
                # A ECRIRE
                # il faudrait aussi tester a pour les constantes comme pi et e.

            if len(variables) == 1: # on va essayer de reprer un polynome
                variable = variables[0]
                polynome = Polynome(0)
                X = Polynome(1, 0)
                for elt in self.contenu[1:]:
                    if isnumtype(elt):
                        polynome += elt
                    elif isinstance(elt, Variable_formelle):
                        polynome += X
                    elif isinstance(elt, Expression_formelle) and elt.operateur() == "*":
                        coeff = 1
                        n = 0
                        for facteur in elt.contenu[1:]:
                            if isinstance(facteur, Variable_formelle):
                                n += 1
                            elif isinstance(facteur, Constante):
                                if facteur.isnum():
                                    coeff *= facteur.valeur
                                else:
                                    print "coeffs inexacts"
                                    return self # les coefficients du polynme ne sont pas grs de manire exacte.
                                    # il faudrait modifier la classe polynme...
                            elif isnumtype(facteur):
                                coeff *= facteur
                            else:
                                print "ce n'est pas un polynome"
                                return self # ce n'est pas un polynme
                        polynome += coeff*X**n
                    elif isinstance(elt, Expression_formelle) and elt.operateur() == "-" and len(elt.contenu) == 2:
                        elt = elt.contenu[1]
                        if isnumtype(elt):
                            polynome += elt
                        elif isinstance(elt, Variable_formelle):
                            polynome += X
                        elif isinstance(elt, Expression_formelle) and elt.operateur() == "*":
                            coeff = -1
                            n = 0
                            for facteur in elt.contenu[1:]:
                                if isinstance(facteur, Variable_formelle):
                                    n += 1
                                elif isinstance(facteur, Constante):
                                    if facteur.isnum():
                                        coeff *= facteur.valeur
                                    else:
                                        print "coeffs inexacts"
                                        return self # les coefficients du polynme ne sont pas grs de manire exacte.
                                        # il faudrait modifier la classe polynme...
                                else:
                                    print "ce n'est pas un polynome"
                                    return self # ce n'est pas un polynme
                            polynome += coeff*X**n
                    else: # ce n'est pas un polynome
                        return self
                produit = factorise_polynome(polynome)
                expression = Expression_formelle("*")
                for facteur in produit:
                    coeffs = list(facteur.coeffs)
                    coeffs.reverse()
                    expr_facteur = Constante(0)
                    for i in xrange(len(coeffs)):
                        expr_facteur += Constante(coeffs[i])*variable**i
                    expression.contenu.append(expr_facteur)
                print "factorisation:", expression
                return expression
        
            return self




    def _variables(self):
        u"Liste des variables apparaissant dans l'expression (une variable peut apparaitre plusieurs fois)."
        liste = [elt for elt in self.contenu if isinstance(elt, Variable_formelle)]
        for elt in self.contenu:
            if isinstance(elt, Expression_formelle):
                liste += elt._variables()
        return liste
        
        
    def variables(self):
        u"Liste des variables apparaissant dans l'expression (chaque variable apparait une seule fois)."
        liste = []
        for variable in self._variables():
            if variable not in liste:
                liste.append(variable)
        return liste


    def _constantes(self):
        u"Liste des constantes (ie. pi, e, ...) apparaissant dans l'expression (une constante peut apparaitre plusieurs fois)."
        liste = [elt for elt in self.contenu if isinstance(elt, Constante) and not elt.isnum()]
        for elt in self.contenu:
            if isinstance(elt, Expression_formelle):
                liste += elt._constantes()
        return liste
        
        
    def constantes(self):
        u"Liste des constantes (ie. pi, e, ...)  apparaissant dans l'expression (chaque constante apparait une seule fois)."
        liste = []
        for constante in self._constantes():
            if constante not in liste:
                liste.append(constante)
        return liste


    def evaluer(self, simplifier = True, **kw):
        u"""Evalue l'expression en assignant une valeur donne  certaines variables.
        >>> a = expression_formelle("x+y+7")
        >>> a.evaluer(x=2, y=3)
        Constante(12)
        """
        def evaluer_si_expr(arg, **kw):
            if isinstance(arg, (Expression_formelle, Variable_formelle)):
                return arg.evaluer(**kw)
            return arg
        resultat = Expression_formelle(*(evaluer_si_expr(arg, **kw) for arg in self.contenu))
        if simplifier:
            return resultat.simplifier()
        return resultat

    def copier(self):
        return pylib.copy.deepcopy(self)

    copy = copier

    def _display(self):
        u"Parenthses systmatiques pour le dbugage."
        def formater(objet):
            if isinstance(objet, Expression_formelle):
                return objet._display()
            return str(objet)

        operateur = self.operateur()
        if operateur in "+-/*^":
            if operateur == "-" and len(self.contenu) == 2:
                return "-(" + formater(self.contenu[1]) + ")"
            else:
                return "(" + operateur.join(formater(elt) for elt in self.contenu[1:]) + ")"
        return operateur + "(" + ",".join(formater(elt) for elt in self.contenu[1:]) + ")"

    _d = _display


class Relation(Objet_formel):
    u"Relation(s) d'ordre."

    def __init__(self, *arguments):
        self.contenu = list(arguments)

    def relation(self):
        return self.contenu[0]


def expression_formelle(expression, simplifier = False, dictionnaire = None):
    u"""Convertit une chane de caractres en expression formelle.

    Exemple:
    (2+7ln(x)+x)/(x+cox(x))
    est stock sous la forme :
    ("/", ("+", 2, ("+", ("*", 7, ("ln", x)), x)), ("+", x, ("cos", x)))

    >>> expression_formelle("7+x")
    (Constante(7)+Variable_formelle('x'))

    La dclaration des variables est en principe automatique.
    On peut galement la forcer (par exemple, si une constante porte le mme nom).
    Dans ce cas, on indique  la fin de l'expression les noms des variables, spars par une virgule.

    >>> expression_formelle("x+y,x,y")
    (Variable_formelle('x')+Variable_formelle('y'))
    """
    
    if expression.strip().startswith("factorise("): # vite de simplifier 2*2*2 en 8 !
        simplifier = False
    
    def conversion(__match__):
        __chaine__ = __match__.group()
        if __chaine__[0] in "0123456789":
            return "Constante(%s)" %__chaine__[:-1] + __chaine__[-1]
        elif __chaine__[0] == ".":
            if __chaine__[1:-1].isdigit():
                return "Constante(%s)" %__chaine__[:-1] + __chaine__[-1]
            return __chaine__ # de la forme ".derive" par exemple (dans "(x**2).derive()").
        elif __chaine__[:-1] in dictionnaire.keys():
            return __chaine__
        return "Variable_formelle('%s')" %__chaine__[:-1] + __chaine__[-1]

    if dictionnaire is None:
        dictionnaire = Expression_formelle.environnement.copy()
    else:
        dictionnaire = dictionnaire.copy()
        Expression_formelle.environnement = dictionnaire


    dictionnaire["Constante"] = Constante
    dictionnaire["Variable_formelle"] = Variable_formelle
    dictionnaire["Objet_formel"] = Objet_formel
    dictionnaire["Expression_formelle"] = Expression_formelle
    dictionnaire["Fonction"] = Fonction
    dictionnaire["pi"] = Constante("pi", math.pi)
    dictionnaire["e"] = Constante("e", math.e)

    arguments = pylib.advanced_split(expression, ",")
    expression = arguments[0]
    expression = expression.replace("^", "**")
    variables = arguments[1:]
    for variable in variables:
        variable = variable.strip()
        dictionnaire[variable] = Variable_formelle(variable)
    expression = re.sub("""[A-Za-z0-9._]+[^'"]""", conversion, expression + " ")
    ancienne_expression = ""
    if expression:
        while ancienne_expression <> expression:
            ancienne_expression = expression
            expression = expression.replace("++", "+").replace("+-", "-").replace("-+", "-").replace("--", "+")
    #print expression
    resultat = eval(expression, dictionnaire)
    if simplifier and isinstance(resultat, Expression_formelle):
        resultat = resultat.simplifier()
    return resultat



def derive(expression, variable = None):
    u"Drive une expression ou une fonction suivant une variable. Syntaxe: derive(x*y**2+3, y)"
    if isinstance(expression, Fonction):
        return Fonction(expression.nom + "'", ",".join(expression.variables) + ":" + repr(derive(expression.expression, variable)))
    if isinstance(variable, Variable_formelle):
        variable = variable.nom
    if isinstance(expression, basestring):
        expression = expression_formelle(expression)
    if hasattr(expression, "derive"):
        d = expression.derive(variable)
        if isinstance(d, Expression_formelle):
            return d.simplifier()
        return d
    return 0

def developpe(expression):
    u"Dveloppe une expression. Syntaxe: developpe((x+3)*(y+5))"
    if isinstance(expression, basestring):
        expression = expression_formelle(expression)
    if hasattr(expression, "developpe"):
        exp_dvp = expression.developpe()
        if isinstance(exp_dvp, Expression_formelle):
            return exp_dvp.simplifier()
        return exp_dvp
    return expression


def factorise(expression):
    u"Tente de factoriser une expression. Syntaxe: factorise(x**2-1)"
    if isinstance(expression, basestring):
        expression = expression_formelle(expression)
    if hasattr(expression, "factorise"):
        exp_fact = expression.factorise()
        print "exp_fact", exp_fact
        if isinstance(exp_fact, Expression_formelle) and not isinstance(expression, Constante):
            return exp_fact.simplifier()
        return exp_fact
    return expression


class Fonction(object):
    u"""Fonction mathmatique.

    >>> Fonction("f", "x,y:x+y")
    Fonction('f',"x,y:Variable_formelle('x')+Variable_formelle('y')")
    """

    def __init__(self, nom, expression, dictionnaire = None):
        self.nom = nom
        variables, expression = expression.split(":")
        self.variables = variables.split(",")
        self.expression = expression_formelle(expression + "," + variables, True, dictionnaire)

    def __call__(self, *args, **kw):
        for i in xrange(len(args)):
            kw[str(self.variables[i])] = args[i]
        if isinstance(self.expression, Objet_formel):
            return self.expression.evaluer(**kw)
        return self.expression

    def __str__(self):
        return "Fonction %s:%s->%s" %(self.nom, ",".join(self.variables), self.expression)

    def __repr__(self):
        return "Fonction(" + repr(self.nom) + "," + repr(",".join(self.variables) + ":" + repr(self.expression)) + ")"
