# -*- coding: iso-8859-1 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#            fonctions mathlib                #
##--------------------------------------#######
# Fonctions plutt destines  un usage interne.
# Low-level functions, used elsewhere in wxgeometrie.
# These are not supposed to be used directly be end-user.
#
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2007  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# version unicode
  
from pylib import *
from redefinir import *


def cbrt(x):
    u"Racine cubique de x."
    return cmp(x, 0)*math.exp(math.log(abs(x))/3)

def root(x, n):
    u"""Racine nime de x.
    
    N'est dfinie pour x ngatif que si n est pair."""
    if n%2:
        return cmp(x, 0)*math.exp(math.log(abs(x))/n)
    return math.exp(math.log(x)/n)

def prod(facteurs):
    return reduce(lambda x,y:x*y, facteurs, 1)

def factoriel(n, maximum = 20000):
    if n>maximum: raise OverflowError
    return prod(range(1, n + 1))

def gcd(a, b):
    u"pgcd de a et de b"
    # algorithme d'Euclide
    a, b = max(abs(a),abs(b)), min(abs(a),abs(b))
    while b:
        a, b = b, a%b
    return a

def lcm(a, b):
    u"ppcm de a et de b"
    return a*b//gcd(a,b)

def pgcd(*termes):
    u"Le plus grand dnominateur commun  un nombre quelconque d'entiers."
    return reduce(lambda x,y:gcd(x,y), termes)

    
def ppcm(*termes):
    u"Le plus petit multiple commun  un nombre quelconque d'entiers."
    return reduce(lambda x,y:lcm(x,y), termes)


def facteurs(n):
    u"Dcompose n en produit de facteurs premiers."
    p = [] # liste des facteurs
    if n < 0:
        n = abs(n)
        p.append(-1)
    
    while n%2 == 0:
        n = n//2; p.append(2)
    
    i = 3
    m = math.sqrt(n)
    while i <= m:
        while n%i == 0:
            n = n//i; m = m//math.sqrt(i)
            p.append(i)
        i += 2
        
    if n <> 1:  p.append(n)
    
    return p    



def diviseurs(n):
    u"Donne la liste des diviseurs d'un entier."
    if n == 1: # sinon, l'algorithme retourne [1, 1] pour 1.
        return [1]
        
    p = [] # liste des diviseurs
    if n < 0:
        n = abs(n)
        p.append(-1)
    
    i = 1
    m = math.sqrt(n)
    while i <= m:
        if n%i == 0:
            p += [n//i, i]
        i += 1
        #print i
    
    return p



def n_premiers(n = 100, maximum = 50000): # securite face aux erreurs de frappe...!
    u"Donne la liste des n premiers nombres premiers."
    liste = [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47, 53, 59, 61, 67, 71, 73, 79, 83, 89, 97]
    m = liste[-1] + 2
    if n > maximum:
        raise OverflowError
    while len(liste) < n:
        r = math.sqrt(m)
        for k in liste:
            if k > r: # pas de diviseurs inferieurs a sa racine, donc le nombre est premier
                liste.append(m)
                break
            if not m%k:
                break
        m += 2
    return liste
    
    

    


def frac(x, maximum = 100000):
    u"Convertit un nombre dcimal en fraction." # (par nombre dcimal, il faut entendre float)
    for i in xrange(2, maximum):
        if (int(x*i)-i*x) == 0:
            return int(x*i), i
    for i in xrange(1, maximum):
        if (int(x*10**i)-10**i*x) == 0:
            return int(float(str(float(x*10**i)))), 10**i # le passage par str permet de supprimer les chiffres non significatifs
            
            
def bin(n):
    u"COnversion en binaire."
    s = ""
    while n:
        s = str(n%2) + s
        n //= 2
    return s
    
    
    
def old_nchiffres(x, n = 1):
    u"Arrondi x en fournissant n chiffres significatifs. Ex: nchiffres(2345, 2)."
    
    return x and round(x/10**floor(log10(abs(x))-n+1))*10**floor(log10(abs(x))-n+1) # Attention au cas x = 0 !
    
def nchiffres(x, n = 1):
    u"Arrondi x en fournissant n chiffres significatifs. Ex: nchiffres(2345, 2)."
    if x:
        k = 10**floor(log10(abs(x))-n+1)
        return round(x/k)*k
    return x # Attention au cas x = 0 !


def arrondir(x):
    u"""Arrondi le nombre : un seul chiffre significatif, compris entre 1 et 5.
    Transforme automatiquement le nombre en entier selon le cas."""
    n = nchiffres(x, sqrt(.5))
    if int(n) == n:
        n = int(n)
    return n
    
    
def point_dans_polygone(couple, polygone):
    u"""Dtermine si un point est  l'intrieur d'un polygone.
    D'aprs un algorithme de Paul Bourke."""
    x, y = couple
    dedans = False
    n = len(polygone)
    p1x, p1y = polygone[0]
    for i in range(1, n + 1):
        p2x, p2y = polygone[i % n]
        if min(p1y, p2y) < y <= max(p1y, p2y) and x <= max(p1x, p2x):
            if p1y != p2y:
                xinters = (y - p1y)*(p2x - p1x)//(p2y - p1y) + p1x
                if p1x == p2x or x <= xinters:
                    dedans = not dedans
        p1x, p1y = p2x, p2y

    return dedans
    
    
    
# Fonctions rapides destines essentiellement  golib
    
def produit_scalaire(u, v):
    u"""Calcule le produit scalaire des vecteurs u et v.
    u et v doivent tre de type list, tuple, array, ou geolib.Vecteur."""
    return builtin.sum(i*j for i, j in zip(u ,v))
    
    
    
def angle_vectoriel(u, v):
    u"""Renvoie une mesure sur ]-pi;pi] de l'angle form par les vecteurs u et v.
    u et v doivent tre de type list, tuple, array, ou geolib.Vecteur, et de dimension 2."""
    return cmath.log(complex(*v)/complex(*u)).imag

def _angle_vectoriel_bis_(u, v):
    u"Autre implmentation du produit vectoriel (un peu plus lente)."
    return (cmp(u[0]*v[1]-v[0]*u[1], 0) or 1)*math.acos(produit_scalaire(u, v)/(math.hypot(*u)*math.hypot(*v)))

def _angle_vectoriel_ter_(u, v):
    u"Autre implmentation du produit vectoriel (plus lente)."
    return angle(complex(*v)/complex(*u))


def norm2(couple):
    u"Implmentation rapide de la norme d'un vecteur."
    return math.hypot(*couple)

# Les deux fonctions qui suivent sont bien plus rapides que la classe Polynome pour fournir des solutions approchees.
# Elles servent comme routine dans geolib par exemple.

def delta(a, b, c):
    return b**2 - 4*a*c
    
def racines(a, b, c):
    d = delta(a, b, c)
    if not d:
        return [-b/(2*a)]
    if d > 0:
        rac = sqrt(d)
        return [(-rac-b)/(2*a) , (rac-b)/(2*a)]
    return []



    
def limite(f, a):
    u"""Tente de calculer la limite de f en a."""
    l = None
    h = 1e-8
    def ecart(h):
        return abs(f(a+h)-f(a-h))
    while h and ecart(h) > 1e-8:
        h/=2
        
    if h:
        return nchiffres((f(a+h)+f(a-h))/2, 5)

def lagrange(*args):
    u"""Polynome d'interpolation de Lagrange. Les arguments sont les coordonnes des points (donnes sous forme de couple)."""
    import objets
    P = objets.Polynome(0)
    X = objets.Polynome(1,0)
    n = 0
    for x, y in args:
        P += y*prod(X-args[i][0] for i in xrange(len(args)) if i != n)/prod(x-args[i][0] for i in xrange(len(args)) if i != n)
        n += 1
    return P

def derive(f, a):
    u"""Tente de calculer le nombre driv de f en a."""
    return limite(lambda h:(f(a+h)-f(a))/h, 0)
    
    
def divise_polynome(A, B):
    u"""Division euclidienne du polynme A par le polynme B."""
    from objets import Frac, Polynome
    A = array(list(A), typecode = "O") # pour optimiser le traitement
    B = array(list(B), typecode = "O")
    m = len(A)
    n = len(B)
    d = max(m-n, 0)
    Q = zeros(d+1, typecode = "O")
    R = list(A)
    while m >= n:
        k = Frac(R[0], B[0])
        Q[d - m + n] = k
        R = (R-k*concatenate((B, zeros(m - n))))[1:]
        m = len(R)
    return Polynome(Q), Polynome(R)
        
    
    
        
def factorise_polynome(polynome):
    u"""Tente de factoriser un polynome."""
    from objets import Frac, Polynome, Rac
    if not polynome:
        return [polynome]

    # D'abord, nous essayons de trouver des racines rationnelles.
    # cf. http://www-fourier.ujf-grenoble.fr/~parisse/mat249/mat249/node24.html
    # First, we try to find some rationnal root.
    racines = []
    def tester(x, polynome = polynome, racines = racines):
        if polynome(x) == 0:
            racines.append(x)
            poly = polynome.derive() # Test des racines multiples.
            while poly(x) == 0 and poly:
                racines.append(x)
                poly = poly.derive()
    i = 1
    while polynome.coeffs[-i] == 0:  # en ce cas, a0 n'a aucun diviseur ! On divise alors le polynome par X.
        racines.append(0)
        i += 1
    an = abs(polynome.coeffs[0])
    a0 = abs(polynome.coeffs[-i])
    if isinstance(a0, Frac): # on simule la multiplication du polynome par des rels, pour que a0 et an soient entiers.
        an *= a0.den        # pas besoin de multiplier les autres coeffs, a ne change rien au final
        a0 = a0.num
    if isinstance(an, Frac):
        a0 *= an.den
        an = an.num
    if isinstance(a0, (int, long)) and isinstance(an, (int, long)): # on cherche des racines rationnelles
        dn = set(diviseurs(an))
        d0 = set(diviseurs(a0))
        for p in d0:
            for q in dn:
                if gcd(p, q) is 1:
                    tester(Frac(p, q))
                    tester(Frac(-p, q))
                    #print p, q
        produit = [Polynome(1, -r) for r in racines]
        #print produit
        if produit:
            q, r = divise_polynome(polynome, prod(produit))
            if r <> 0:
                print u"Racines rationnelles : %s" %racines
                print u"%s == (%s)*(%s) + %s" %(polynome, q, prod(produit), r)
                raise ValueError, "La division euclidienne ne tombe pas juste."
            polynome = q
    else:
        produit = []
    if polynome.deg() is 2:
        racines = polynome.racines()
        if len(racines) is 1:
            produit += [Polynome(1, -racines[0]), Polynome(1, -racines[0])]
            polynome = Polynome(polynome.coeffs[0])
        elif len(racines) is 2:
            produit += [Polynome(1, -racines[0]), Polynome(1, -racines[1])]
            polynome = Polynome(polynome.coeffs[0])
    produit.append(polynome)
    return produit
                