# -*- coding: iso-8859-1 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#        fonctions utilisateurs mathlib       #
##--------------------------------------#######
# Fonctions de haut niveau, destines  l'utilisateur final.
# High-level functions, for end-user.
#
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2007  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# version unicode
  
  
from resolution import *
import calcul_formel
from calcul_formel import derive, developpe
    
# Fonctions gnralistes
########################
    
def resoudre(*args):
    u"""Rsolution d'une quation en X (dans R)."""
    return resoudre_systeme(*args)
    


def factorise(expression):
    u"""Factorisation d'un polynome."""
    if isinstance(expression, calcul_formel.Objet_formel):
        return calcul_formel.factorise(expression)
    if isinstance(expression, (int, long)):
        return Product(*facteurs(expression))
    if isinstance(expression, Polynome):
        return Product(*factorise_polynome(expression))

# Alias
#######
    
solve = resoud = resoudre
    
factor = factoriser = factorise

def racine(x, n = 3):
    u"Racine nime d'un nombre (racine cubique par dfaut)."
    return root(x, n)

def mat(*args, **kw):
    u"Matrice. Syntaxe : mat((1,2),(3,4))"
    return Mat(*args, **kw)

def somme(*args):
    u"Somme des termes d'une liste. Syntaxe : somme((1,2,3)) ou somme(1,2,3)"
    if isinstance(args[0], (tuple, list, types.ArrayType)):
        args = tuple(args[0])
    return sum(args)

def produit(*args):
    u"Produit des termes d'une liste. Syntaxe : produit((1,2,3)) ou produit(1,2,3)"
    if isinstance(args[0], (tuple, list, types.ArrayType)):
        args = tuple(args[0])
    return prod(args)
    

# Les fonctions qui suivent seront automatiquement rajoutes dans le menu de la calculatrice, dans cet ordre.
__fonctions__ = [racine, frac, mat, None, pgcd, ppcm, diviseurs, None, somme, produit, None, lagrange, None, factorise, resoudre, derive, developpe]
