# -*- coding: iso-8859-1 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#                Objets CALC                  #
##--------------------------------------#######
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2007  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# version unicode

from resolution import *

class Expression: ### DESUET ###
    u"Une chaine de caractres reprsentant une expression mathmatique."
        
    def __init__(self, chaine = "", variables = "", calcul_exact = False, conversion_crochet = True, OOo = True, changer_separateurs = False, separateurs_personnels = (",", ";")):
        self.calcul_exact = calcul_exact
        self.conversion_crochet = conversion_crochet
        self.OOo = OOo  # convertir les notations d'OpenOffice.org
        self.expression = chaine
        self.variables = variables
        self.changer_separateurs = changer_separateurs
        self.separateurs_personnels = separateurs_personnels
        deprecation("La classe Expression est desuete. Utiliser la fonction traduire_formule()  la place.")
        
    def reformuler(self):
        calcul = self.expression.strip()
        if calcul.startswith(">"):
            calcul = "ans()" + calcul
        calcul = self.expression.replace("^", "**").replace(">deg", "*180/pi").replace(">rad", "*pi/180")
        
        for s in ("frac", "hex", "oct", "bin"):
            if calcul.endswith(">%s" %s):
                calcul = "%s(%s)" %(s, calcul[:-len(s)-1])
        
        # Transforme les chaines pour pouvoir utiliser simplement factoriser(), ou un alias
        calcul = mreplace(calcul, ("factor ", "factoriser "), "factorise ")
        calcul = mreplace(calcul, ("factor(", "factoriser("), "factorise(")
        if calcul.startswith("factorise "):
            calcul = calcul.replace("factorise ", "factorise(") + ")"
        
        # Transforme les chaines pour pouvoir utiliser simplement resoudre(), ou un alias
        calcul = mreplace(calcul, ("solve ", "resoud "), "resoudre ")
        calcul = mreplace(calcul, ("solve(", "resoud("), "resoudre(")
        if calcul.startswith("resoudre "):
            calcul = calcul.replace("resoudre ", "resoudre(") + ")"

        print calcul
        # les arguments de resoudre() sont formats...
        def formatage_equations(chaine):
            def transformer(chaine):
                chaine = mreplace(chaine, (" et ", " and "), "&")
                chaine = mreplace(chaine, (" ou ", " or "), "|")
                chaine = mreplace(chaine, (" non ", " pas ", " not "), "~")
                chaine = re.sub("[-A-Za-z0-9.+/* ()]+[>=<][>=]?[-A-Za-z0-9.+/* ()]+", lambda x:"(" + x.group(0) + ")", chaine)
                # La variable doit s'appeler 'X'
                # Comme c'est un peu contraignant, s'il y a une seule variable, on la remplace par 'X'
                caracteres = re.findall("[A-Za-z]", chaine)
                if len(set(caracteres)) is 1:
                    chaine = chaine.replace(set(caracteres).pop(), "X")
                return chaine
            position = chaine.find("resoudre(")
            if position is not -1:
                # On remplace "=" par "==" si ncessaire:
                chaine = re.sub("[^=<>][=][^=]", lambda x:x.group(0).replace("=", "=="), chaine)
                liste = split_around_parenthesis(chaine, position)
                if len(liste) is 3:
                    return liste[0] + transformer(liste[1]) + formatage_equations(liste[2])
                else:
                    return transformer(liste[0])
            else:
                return chaine
        calcul = formatage_equations(calcul)
            
        # il est plus intuitif d'utiliser le symbole % pour les pourcentages, et mod pour le modulo.
        calcul = mreplace(calcul.replace("%", "/100"), (" mod ", " modulo "), "%")

        # On peut choisir comme separateur decimal la virgule (convention francaise en particulier)
        if self.changer_separateurs:
            calcul = calcul.replace(self.separateurs_personnels[0], ".").replace(self.separateurs_personnels[1], ",")

        # Transforme les accolades en parentheses (utile par exemple pour les fonctions issues d'OpenOffice.org).
        # De meme, la notation "times" d'OpenOffice.org est convertie.
        if self.OOo:
            calcul = regexp("[ ]left[])}[({]|[ ]right[])}[({]", calcul, "' ' + x[-1]")
            calcul = calcul.replace("{", "(").replace("}", ")")
            calcul = regexp("[^a-zA-Z_]times[^a-zA-Z_0-9]", calcul, "x[0] + '*' + x[-1]")
            calcul = regexp("[^a-zA-Z_]over[^a-zA-Z_0-9]", calcul, "x[0] + '/' + x[-1]")

        calcul = " " + calcul.replace(" ", "") # permet de simplifier certains regexp.
        # Par exemple, pour ne pas remplacer log10() par log10*(), les chiffres doivent etre precedes
        # par un caractere non alphanumerique.
        # L'espace en debut de ligne evite de faire le cas particulier ou log10 est en tout debut de ligne.
        
        if self.calcul_exact:
            # Le code qui suit remplace 2/3 par 2*Frac(1, 1)/3, ce qui donne Frac(2,3).
            calcul = calcul.replace("/", "*frac(1)/")
            # On remplace rac, rc et sqrt par rac, mais sans que Frac devienne FRac (par exemple).
            # il doit y avoir plus elegant !
            calcul = regexp("[^a-zA-Z_]sqrt[(]|[^a-zA-Z_]rc[(]", calcul, "x[0] + 'rac('")
            #calcul = regexp("^sqrt[(]|^rac[(]|^rc[(]", calcul, "'Rac('")
            debug(calcul)
        #else:
            # calcul = calcul.replace("/", "*1./")
            # Si le mode fraction est desactive, on convertit tout en float (sinon, 1/2 reserve des surprises).
        
        # frac est utilise comme alias de Frac
        #calcul = regexp("[^a-zA-Z_]frac[(]", calcul, "x[0] + 'Frac('")
        #calcul = regexp("^frac[(]", calcul, "'Frac('")

        
        # Transforme les listes en matrices (permet le calcul matriciel).
        if self.conversion_crochet:
            calcul = calcul.replace("[", "array([").replace("]", "])")

        # Le code qui suit remplace les expressions style 3x ou 2cos(x) par 3*x et 2*cos(x).
        calcul = regexp("[^a-zA-Z0-9][0-9]+[a-zA-DF-Z]", calcul, "x[:-1] + '*' + x[-1]")
        # On traite separemment le cas ou ca commence par "E", a cause de 2E17 par exemple, qui signifie 2*10**17
        calcul = regexp("[^a-zA-Z0-9.][E][-+]?[0-9]", calcul, "x.replace('E','1E')")
        calcul = regexp("[^a-zA-Z0-9][0-9]+[E][0-9]*[a-zA-Z]", calcul, "x.replace('E','*E', 1)")
        
        # De meme, on rajoute les * entre deux parentheses...
        calcul = calcul.replace(")(",")*(")
        
        calcul = calcul.replace("(", "( ").replace(")", " )")
        # evite les problemes avec "x(x(" par exemple (les regexp se "chevauchent").
 
        #... et devant des parentheses precedees de chiffres.
        # C'est assez delicat, car les noms de fonctions peuvent contenir des chiffres.
        #calcul = regexp("[^a-zA-Z0-9][0-9]+[(]", calcul, "x[:-1] + '*' + x[-1]")
        
        debug(calcul, 1)
        # Si a, b, c... ne sont pas des fonctions, on remplace "a(" par "a*(", etc...
        if self.variables:
            calcul = regexp("[^a-zA-Z0-9][0-9]*[0-9" + self.variables + "][(]", calcul, "x[:-1] + '*' + x[-1]")
        
        # On remplace ")x" par ")*x"
        calcul = regexp("[)][a-zA-Z0-9]", calcul, "x[0] + '*' + x[1]")

        
        # n! devient factoriel(n).
        calcul = regexp("[0-9]+[!]", calcul, "'factoriel(' + x[:-1] + ')'")
        
        return calcul















def formatage_formule(chaine, variables_libres = ""):
    u"""Corrige les abus de notations les plus frquents, dans une formule mathmatique.
    
    Exemples :
    - signes "*" manquants avant une variable : 2x -> 2*x
    - signes "*" manquants entre deux parenthses : (x+5)(x+7)
    - symbole "^" pour les puissances : 2^4 -> 2**4
    - symbole "!" pour factoriel : 4! -> 1*2*3*4
    - symbole "%" pour le pourcentage : 5% -> 5/100
    - "mod" pour les modulos : 5 mod 3 -> 5%3 (ie. 5 modulo 3)
    """
    chaine = chaine.strip().replace("^", "**")
    
    # il est plus intuitif d'utiliser le symbole % pour les pourcentages, et mod pour le modulo.
    chaine = chaine.replace("%", "/100").replace(" mod ", "%").replace(" modulo ", "%")


    chaine = " " + chaine.replace(" ", "") # permet de simplifier certains regexp.
    # Par exemple, pour ne pas remplacer log10() par log10*(), les chiffres doivent tre prcds
    # par un caractre non alphanumrique.
    # L'espace en debut de ligne vite de faire le cas particulier o log10 est en tout dbut de ligne.
    

    # Le code qui suit remplace les expressions style 3x ou 2cos(x) par 3*x et 2*cos(x).
    chaine = regexp("[^a-zA-Z0-9][0-9]+[a-zA-Z]", chaine, "x[:-1] + '*' + x[-1]")
    # On traite separemment le cas ou ca commence par "E", a cause de 2E17 par exemple, qui signifie 2*10**17
    #calcul = regexp("[^a-zA-Z0-9.][E][-+]?[0-9]", calcul, "x.replace('E','1E')")
    #calcul = regexp("[^a-zA-Z0-9][0-9]+[E][0-9]*[a-zA-Z]", calcul, "x.replace('E','*E', 1)")
    
    # De meme, on rajoute les * entre deux parentheses...
    chaine = chaine.replace(")(",")*(")
    
    chaine = chaine.replace("(", "( ").replace(")", " )")
    # evite les problemes avec "x(x(" par exemple (les regexp se "chevauchent").

    #... et devant des parentheses precedees de chiffres.
    # C'est assez delicat, car les noms de fonctions peuvent contenir des chiffres.
    #calcul = regexp("[^a-zA-Z0-9][0-9]+[(]", calcul, "x[:-1] + '*' + x[-1]")
    
    # Si a, b, c... sont des variables libres, on remplace "a(" par "a*(", etc...
    if variables_libres:
        chaine = regexp("[^a-zA-Z0-9][0-9]*[0-9" + variables_libres + "][(]", chaine, "x[:-1] + '*' + x[-1]")
    
    # On remplace ")x" par ")*x"
    return regexp("[)][a-zA-Z0-9]", chaine, "x[0] + '*' + x[1]").strip()

    
    # n! devient factoriel(n).
    calcul = regexp("[0-9]+[!]", calcul, "'factoriel(' + x[:-1] + ')'")
    
    return calcul




def traduire_formule(formule = "", fonctions = (), variables = "[_A-Za-z][_A-Za-z0-9]*", OOo = True, LaTeX = False, changer_separateurs = False, separateurs_personnels = (",", ";")):   
    formule = formule.strip()
    formule = regsub("[ ]+", formule, " ") # un seul caractre d'espacement
    
    if formule.startswith(">"):
        formule = "ans()" + formule
    formule = formule.replace("^", "**").replace(u'',"**2")
    
    for s in ("frac", "hex", "oct", "bin", "deg", "rad"):
        if formule.endswith(">%s" %s):
            formule = "%s(%s)" %(s, formule[:-len(s)-1])
    
    # Transforme les chaines pour pouvoir utiliser simplement factoriser(), ou un alias
    formule = mreplace(formule, ("factor ", "factoriser "), "factorise ")
    formule = mreplace(formule, ("factor(", "factoriser("), "factorise(")
    if formule.startswith("factorise "):
        formule = formule.replace("factorise ", "factorise(") + ")"
    
    # Transforme les chaines pour pouvoir utiliser simplement resoudre(), ou un alias
    formule = mreplace(formule, ("solve ", "resoud "), "resoudre ")
    formule = mreplace(formule, ("solve(", "resoud("), "resoudre(")
    if formule.startswith("resoudre "):
        formule = formule.replace("resoudre ", "resoudre(") + ")"

    print formule
    # les arguments de resoudre() sont formats...
    def formatage_equations(chaine):
        def transformer(chaine):
            chaine = mreplace(chaine, (" et ", " and "), "&")
            chaine = mreplace(chaine, (" ou ", " or "), "|")
            chaine = mreplace(chaine, (" non ", " pas ", " not "), "~")
            chaine = regsub("[-A-Za-z0-9.+/* ()]+[>=<][>=]?[-A-Za-z0-9.+/* ()]+", chaine, lambda s: "(" + s + ")")
            # La variable doit s'appeler 'X'
            # Comme c'est un peu contraignant, s'il y a une seule variable, on la remplace par 'X'
            caracteres = re.findall("[A-Za-z]", chaine)
            if len(set(caracteres)) is 1:
                chaine = chaine.replace(set(caracteres).pop(), "X")
            return chaine
        position = chaine.find("resoudre(")
        if position is not -1:
            # On remplace "=" par "==" si ncessaire:
            chaine = regsub("[^=<>][=][^=]", chaine, lambda s: s.replace("=", "=="))
            liste = split_around_parenthesis(chaine, position)
            if len(liste) is 3:
                return liste[0] + transformer(liste[1]) + formatage_equations(liste[2])
            else:
                return transformer(liste[0])
        else:
            return chaine
    formule = formatage_equations(formule)
        
    # il est plus intuitif d'utiliser le symbole % pour les pourcentages, et mod pour le modulo.
    formule = mreplace(formule.replace("%", "/100"), (" mod ", " modulo "), "%")

    # On peut choisir comme separateur decimal la virgule (convention francaise en particulier)
    if changer_separateurs:
        formule = formule.replace(separateurs_personnels[0], ".").replace(separateurs_personnels[1], ",")

    #Conversion de quelques formules latex ultra-frquentes (comme \frac, \dfrac, \tfrac, \sqrt, suppression de \nombre, etc.).
    if LaTeX:
        formule = formule.replace("\\nombre", "").replace("\\nb", "").replace("\\left", "").replace("\\right", "").replace("\\sqrt", "sqrt").replace("\\%", "%")
        formule = formule.replace("\\sin", "sin").replace("\\cos", "cos").replace("\\tan", "tan").replace("\\times", "*")
        formule = formule.replace("\\infty", "oo").replace("\\oo", "oo")
        formule = formule.replace("\\i", "i").replace("\\e", "e").replace("\\pi", "pi")
        
        # '\dfrac{a}{b}' devient '(a)/(b)' (idem pour \frac et \tfrac)
        for substr in ("\\frac", "\\dfrac", "\\tfrac"):
            i = formule.find(substr)
            while i != -1:
                decoupe1 = split_around_parenthesis(formule, i + len(substr), "{")
                if len(decoupe1) == 3:
                    debut, arg1, reste = decoupe1
                else: # erreur de syntaxe
                    if param.debug:
                        print(u"Il manque l'argument 1 de \\dfrac")
                    break
                decoupe2 = split_around_parenthesis(reste, 0, "{")
                if len(decoupe2) == 3:
                    milieu, arg2, reste = decoupe2
                else: # erreur de syntaxe
                    if param.debug:
                        print(u"Il manque l'argument 2 de \\dfrac")
                    break
                formule = formule[:i] + arg1 + "/" + arg2 + reste # en principe milieu est vide.
                i = formule.find(substr)
        
        #formule = re.sub("\\\\[dt]?frac[ ]?[{]([^}]*)[}][{]([^}]*)[}]", lambda m:"("+m.groups()[0]+")/("+m.groups()[1]+")", formule)
        if not OOo:
            formule = formule.replace("{", "(").replace("}", ")")
        
        
    # Transforme les accolades en parentheses (utile par exemple pour les fonctions issues d'OpenOffice.org).
    # De meme, la notation "times" d'OpenOffice.org est convertie.
    if OOo:
        formule = regsub("[ ]left[])}[({]|[ ]right[])}[({]", formule, lambda s: ' ' + s[-1])
        formule = formule.replace("{", "(").replace("}", ")")
        formule = regsub("[^a-zA-Z_]times[^a-zA-Z_0-9]", formule, lambda s: s[0] + '*' + s[-1])
        formule = regsub("[^a-zA-Z_]over[^a-zA-Z_0-9]", formule, lambda s: s[0] + '/' + s[-1])
        formule = formule.replace(u'\u2013', "-")
        formule = formule.replace('infinity', 'oo')

    ### formule = " " + formule.replace(" ", "") # permet de simplifier certains regexp.
    # Par exemple, pour ne pas remplacer log10() par log10*(), les chiffres doivent etre precedes
    # par un caractere non alphanumerique.
    # L'espace en debut de ligne evite de faire le cas particulier ou log10 est en tout debut de ligne.
    
    


    # Le code qui suit remplace les expressions style 3x ou 2.5cos(x) par 3*x et 2.5*cos(x).
    formule = regsub("(^|[^a-zA-Z0-9.])[0-9.]+[ ]?[a-zA-Z]", formule, lambda s: s[:-1] + '*' + s[-1])
    # On traite separemment le cas o a commence par "E", a cause de 2E-17 par exemple, qui signifie 2*10**-17
    formule = regsub("(^|[^a-zA-Z0-9.])[E][-+]?[0-9]", formule, lambda s: s.replace('E','1E'))
    formule = regsub("(^|[^a-zA-Z0-9.])[0-9.]+[ ]?[E][0-9]*[a-zA-Z]", formule, lambda s: s.replace('E','*E', 1))
    
    # De meme, on rajoute les * entre deux parentheses...
    formule = formule.replace(")(",")*(").replace(") (",")*(")
    
    formule = formule.replace("(", "( ").replace(")", " )")
    # evite les problemes avec "x(x(" par exemple (les regexp se "chevauchent").

    #... et devant des parentheses precedees de chiffres.
    # C'est assez delicat, car les noms de fonctions peuvent contenir des chiffres.
    #formule = regexp("[^a-zA-Z0-9][0-9]+[(]", formule, "x[:-1] + '*' + x[-1]")
    
    # Si a, b, c... ne sont pas des fonctions, on remplace "a(" par "a*(", etc...
    if variables:
        def f(s):
            if s[1:-1] in fonctions:
                return s
            else:
                return s[:-1] + "*("
        # Note: on ajoute un espace avant la formule (qu'on enlve  la fin).
        # C'est juste une astuce pour viter le cas particulier o l'expression  traiter est en dbut de ligne.
        formule = regsub("[^a-zA-Z0-9.]([0-9.]*[0-9]|" + variables + ")[ ]?[(]", " " + formule, f)[1:]
        #  rcrire avec (?=...)
    
    # On remplace ")x" par ")*x"
    formule = regsub("[)][ ]?[a-zA-Z0-9]", formule, lambda s: s[0] + '*' + s[1])

    
    # n! devient factoriel(n).
    formule = regsub("[_0-9A-Za-z]+[ ]?[!]", formule, lambda s: 'factoriel(' + s[:-1] + ')')
    
    debug(formule, "<<formule transformee>>")

    return formule
