# -*- coding: iso-8859-1 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#               Objets mathlib                #
##--------------------------------------#######
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2007  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# version unicode

from fonctions import *


class ObjetMathematique(object):
    u"Classe mre de la plupart des objets mathmatiques."

    def __lt__(self, y):
        return not self >= y

    def __ge__(self, y):
        return self > y or self == y

    def __le__(self, y):
        return not self > y

    def __ne__(self, y):
        return not self.__eq__(y)

    def __nonzero__(self): # la valeur logique de l'objet (la doc officielle est fausse ??)
        return self <> 0    # utilise dans un test "if" par exemple.

    def __truediv__(self, y):
        return self.__div__(y)

    def __rtruediv__(self, y):
        return self.__rdiv__(y)

    def __sub__(self, y):   return self + (- y)

    def __rsub__(self, y):  return y + (- self)

    def __radd__(self,y):   return self.__add__(y)

    def __rmul__(self, y): return self.__mul__(y)

    def __pos__(self):      return self

    def __cmp__(self, y):
        if hasattr(self, "__gt__") and isinstance(y, (ObjetMathematique, float, int, long)):
            if self > y:
                return 1
            elif self < y:
                return -1
            else:
                return 0
        return NotImplemented


class Reel(ObjetMathematique):
    u"""Classe mre pour les classes Frac et Rac.
    Permet de leur appliquer les oprations mathmatiques de pylab."""

    for nom in fonctions_matplotlib:
        exec("""
def %s(self, *args):
    arguments = []
    for arg in args:
        arguments.append(float(arg))
    return %s(float(self), *arguments)""" %(nom, nom))
    del nom




class Frac(Reel):
    u"Une fraction"

    def __init__(self, num = 1, den = 1, simplification = True):
        u"""Dfinit la fraction num/den."""
        if isinstance(num, Frac) or isinstance(num, Frac):
            fraction = num/den
            num = fraction.num; den = fraction.den
        if isinstance(num, float):
            num, d = frac(num)
            den *= d

        if isinstance(den, float):
            den, n = frac(den)
            num *= n

        num, den = int(num), int(den)
        self.numerateur = self.num = num
        self.denominateur = self.den = den
        if simplification:  self.simplifier()


    def simplifier(self):
        a = self.num; b = self.den

        if b < 0:   a = -a; b = -b
        if b == 0: raise ZeroDivisionError, str2(u"le dnominateur ne doit pas tre nul.")

        g = gcd(a, b)
        self.num = a//g; self.den = b//g

        return self


    def __str__(self):
        if self.den == 1:   return str(self.num)
        return str(self.num) + "/" + str(self.den)


    def __repr__(self):
        return u"Frac(%s,%s)" % (repr(self.num), repr(self.den))

    def __add__(self, y):
        if type(y) == float:    return y + float(self)
        if isinstance(y, (int, long)):      return Frac(y*self.den + self.num, self.den)
        if type(y) == types.ArrayType:    return y + array(self)
        if isinstance(y, Frac): return Frac(self.num*y.den + self.den*y.num, self.den*y.den)
        return NotImplemented
        #raise TypeError




    def __neg__(self):      return Frac(-1*self.num, self.den, False)

    def __mul__(self, y):
        if type(y) == float:    return y*float(self)
        if type(y) == types.ArrayType:    return y*array(self)
        if isinstance(y, (int, long)):      return Frac(y*self.num, self.den)
        if isinstance(y, Frac): return Frac(self.num*y.num, self.den*y.den)
        return NotImplemented



    def __abs__(self):      return Frac(abs(self.num), abs(self.den), False)

    def __eq__(self, y):
        if type(y) == float:    return y == float(self)
        if isinstance(y, (int, long)):      return Frac(self.num, self.den) == Frac(y, 1)
        if type(y) == types.ArrayType:    return y == array(self)
        # Frac(4, 2, False) == 2 donne ainsi True.
        if isinstance(y, Frac): return self.num * y.den - self.den * y.num == 0
        if isinstance(y, Reel): return NotImplemented
        return False
        #raise TypeError



    def __float__(self):    return self.num/self.den

    def __int__(self):      return self.num//self.den

    def __ne__(self, y):    return not self.__eq__(y)

    def __gt__(self, y):
        if type(y) == float:    return float(self) > y
        if isinstance(y, (int, long)):      y = Frac(y, 1)
        if type(y) == types.ArrayType:    return y > array(self)
        if isinstance(y, Frac): return (self - y).num > 0
        return NotImplemented
        #raise TypeError


    def __pow__(self, y):
        if type(y) == float:    return pow(float(self), y)
        if int(y) == y: y = int(y)
        if isinstance(y, (int, long)):      return Frac(self.num**y, self.den**y)
        if isinstance(y, Frac): return pow(float(self), float(y))
        return NotImplemented
        #raise TypeError

    def __rpow__(self, y):
        if self.den == 2:
            return y**((self.num - 1)//2)*Rac(y)
        if int(self) == self and type(y) <> float:   return pow(y, int(self))
        if isinstance(y, Frac) or type(y) == float or type(y) == int:  return pow(float(y), float(self))
        if type(y) == types.ArrayType:    return pow(y, array(self))
        return NotImplemented
        #raise TypeError

    def __div__(self, y):
        if type(y) == float:    return float(self)/y
        if type(y) == types.ArrayType:    return array(self)/y
        if isinstance(y, (int, long)):      return Frac(self.num, y*self.den)
        if isinstance(y, Frac): return Frac(y.den*self.num, y.num*self.den)
        return NotImplemented
        #raise TypeError

    def __rdiv__(self, y):
        if type(y) == float:    return y/float(self)
        if type(y) == types.ArrayType:    return y/array(self)
        if isinstance(y, (int, long)):      return Frac(y*self.den, self.num)
        if isinstance(y, Frac): return Frac(self.den*y.num, self.num*y.den)
        return NotImplemented
        #raise TypeError










class Rac(Reel):
    u"Une somme de racines carres"

    def __init__(self, r = 1, n = 1, simplification = True):
        u"""Dfinit la racine n*sqrt(r).
        n, r peuvent tre des entiers, ou des fractions.
        r peut aussi tre une liste de couples [ri, ni].
        Dans ce dernier cas, Rac([[r1, n1], [r2, n2]]) signifie n1*Rac(r1) + n2*Rac(r2)
        Pour l'utilisateur final, la derniere notation, plus explicite, est celle recommande.
        Eventuellement, l'implmentation pourrait en effet changer."""
        self.valeur = iff( type(r) == list, r, [[r, n]] )


        # La somme de racines carrees est stockee dans self.valeur, sous forme [[r1, n1], [r2, n2], ...]
        # Ceci signifie : n1*sqrt(r1) + n2*sqrt(r2) + ...

        self.den_entier()
        if simplification:  self.simplifier()
        self.regrouper()
        self.suppr_zeros()


    def den_entier(self):
        u"Renvoie les racines avec des dnominateurs entiers."
        valeur = []
        for r, n in self.valeur:
            if isinstance(r, Frac):
                n *= Frac(1, r.den); r = r.num*r.den
            valeur.append([r, n])
        self.valeur = valeur
        return self


    def regrouper(self):
        u"Regroupe entre elles les racines : 2V3 + 5V3 = 7V3."
        # on regroupe les termes selon les racines r.
        # on ne doit PAS tenir compte du n, qui peut-etre un nombre complexe (classement impossible !)
        self.valeur.sort(key=lambda x:x[0])
        # x est une liste des couples (ri, ni), y est un couple (r,n).
        def f(x, y):
            if x[-1][0] == y[0]:
                x[-1][1] += y[1]
            else: x.append(y)
            return x
        self.valeur = reduce(f, self.valeur[1:], [self.valeur[0]])
        return self


    def suppr_zeros(self):
        u"""Supprime les zros de la liste de racine.
        Exemple : V5 + 0V3 = V5."""
        self.valeur = [[r, n] for r, n in self.valeur if r<>0 and n<>0] or [[1, 0]]
        return self


    def simplifier(self):
        u"Renvoie les racines sous forme a*Rac(b) avec a le plus grand possible."
        valeur = []
        for r, n in self.valeur:
            r_initial = r
            if r < 0:   raise ValueError, str2(u"racine carre d'un nombre ngatif.")
            if r > 10**7:
                print u"Racine: dpassement de capacit de calcul."
                return self

            # NOTE : Algorithme exponentiel (t <= .5*exp(.5*n)), a revoir si possible.

            m = math.sqrt(r)

            if int(m) == m: # la racine tombe juste
                n *= int(m); r = 1
            else:
                while r%4 == 0:
                    r = r//4; m = m/2
                i = 1
                while i < m:
                    i += 2
                    while r%i**2 == 0:
                        r = r//i**2; m = m/i
                n *= int(math.sqrt(r_initial//r))
            valeur.append([r, n])
        self.valeur = valeur
        return self


    def __str__(self):
        def signe(x):
            return iff(x<0, "-", "+")

        def affiche(racine):
            # representation d'une racine.
            # Exemple: [5, Frac(1,3)] devient sqrt(5)/3
            r, n = racine; d = 1
            if r == 1:   return str(n)
            if isinstance(n, Frac):
                d = n.den; n = n.num
            return signe(n) + iff(abs(n) <> 1, str(abs(n)) + "*") + "sqrt(%s)" %r + iff(d <> 1, "/" + str(d))

        return "".join([affiche(racine) for racine in self.valeur]).lstrip("+")


    def __repr__(self):
        return "Rac(%s)" % repr(self.valeur)



    def __float__(self):    return float(sum([n*sqrt(r) for r,n in self.valeur]))

    def __int__(self):      return int(float(self))

    def __add__(self, y):
        if type(y) == float:   return float(self) + y
        if isinstance(y, (int, long, Frac)):    return self + Rac(1, y)
        if isinstance(y, Rac):  return Rac(y.valeur + self.valeur)
        return NotImplemented



    def __neg__(self):      return Rac([[r, -n] for r,n in self.valeur])

    def __mul__(self, y):
        if type(y) == float:    return y*float(self)
        if isinstance(y, (int, long, Frac)):    return self * Rac(1, y)
#        print [[r*s,m*n] for r,n in self.valeur for s,m in y.valeur]
        if isinstance(y, Rac):  return Rac([[r*s,m*n] for r,n in self.valeur for s,m in y.valeur])
        return NotImplemented


    def __abs__(self):
        if self < 0:
            return -self
        else:
            return self

    def __eq__(self, y):
        if isinstance(y, (int, long)):
            return self.valeur == [[1, y]]
        if isinstance(y, Rac):
            return len(self.valeur) == len(y.valeur) and not sum([i<>j for i,j in zip(self.valeur, y.valeur)])
        return float(self) == y

    def __ne__(self, y):    return not self.__eq__(y)

    def __gt__(self, y):
        return float(self) > y


    def __pow__(self, y):
        if type(y) <> float and int(y) == y: y = int(y)
        if not isinstance(y, (int, long)):    return pow(float(self), y)
        if y == 0: return 1
        if y == 1: return self
        # Si c'est un entier, on applique l'exponentiation indienne.
        return self**(y//2)*self**(y//2)*self**(y%2)

    def __rpow__(self, y):
        return pow(float(y), float(self))

    def __div__(self, y):
        if type(y) == float:
            return float(self)/y
        if isinstance(y, Rac) and len(y.valeur) == 1 and y.valeur[0][0] == 1:
            y = y.valeur[0][1]
        if isinstance(y, (int, long, Frac)):
            return Rac([[r, Frac(n,y)] for r, n in self.valeur])
        if isinstance(y, Rac):
            if len(y.valeur) == 1:
                expression_conjuguee = y
            else:
                expression_conjuguee = Rac(y.valeur[:-1]) - Rac([y.valeur[-1]])
            return (self*expression_conjuguee)/(y*expression_conjuguee)
        return NotImplemented


    def __rdiv__(self, y):
        if type(y) == float:    return y/float(self)
        if isinstance(y, (int, long, Frac)): return Rac(y)/self
        return NotImplemented






class Polynome(ObjetMathematique):
    u"""Un polynome  une variable.

    L'implmentation est assez rudimentaire, et est simplement destine  manier des polynmes de faibles degrs (<200)."""

    def __init__(self, *coeffs):
        if coeffs and isinstance(coeffs[0], (tuple,list, types.ArrayType)):
            coeffs = coeffs[0]
        if len(coeffs) == 0:
            coeffs = [0]
        self.coefficients = self.coeffs = list(coeffs)

        self.simplifier()
        # a poursuivre !

    def simplifier(self): # enleve les coeffs nuls au debut
        for i in range(len(self.coeffs)):
            if self.coeffs[i] <> 0: break
        self.coeffs = self.coeffs[i:]

    #def __call__(self, valeur):


    def degre(self):
        if self.coeffs == [0]:
            return MoinsInfini()
        return len(self.coeffs) - 1

    deg = degre

    def delta(self):
        if self.deg() == 2:
            a, b, c = self.coeffs
            return b**2 - 4*a*c

    def solutions(self):
        print self.coeffs
        if self.deg() == 1:
            a, b = self.coeffs
            return [Frac(-b, a)]

        if self.deg() == 2:
            a, b, c = self.coeffs; delta = self.delta()
            if delta == 0:  return [-b*Frac(1)/(2*a)]
            if delta > 0:   return [(-b - Rac(delta))*Frac(1)/(2*a), (-b + Rac(delta))*Frac(1)/(2*a)]
            return []

    racines = solutions

    def derive(self):
        n = len(self.coeffs) - 1
        return Polynome([(n-i)*self.coeffs[i] for i in range(n)])

    def primitive(self):
        n = len(self.coeffs)
        return Polynome([self.coeffs[n-i]*Frac(1,i) for i in range(n, 0, -1)] + [0])

    def __call__(self, valeur):
        somme = 0 # algorithme de Horner
        for coeff in self.coeffs:
            somme = somme*valeur + coeff
        return somme


    def __iter__(self):
        return iter(self.coeffs)

    def __str__(self):
        d = self.degre()
        if d < 0:
            return "0"
        def formater(i):
            coeff = self.coeffs[i]
            if coeff == 0:
                return ""
            if i == d:
                return str(self.coeffs[i])
            if coeff == 1:
                prefixe = ""
            elif coeff == -1:
                prefixe = "-"
            else:
                prefixe = str(coeff) + "*"
            if i == d-1:
                return prefixe + "X"
            return prefixe + "X^" + str(d-i)
        return ("+".join(formater(i) for i in range(d+1) if formater(i))).replace("+-", "-")

    def __repr__(self):
        return u"Polynome(" + ",".join(repr(coeff) for coeff in self.coeffs) + ")"

    def __add__(self, y):
        if isinstance(y, Polynome):
            d1 = self.degre()
            d2 = y.degre()
            if d1 < 0:
                return y
            elif d2 < 0:
                return self
            m = max(d1, d2)
            P1 = (m - d1)*[0] + self.coeffs # les deux listes de coeffs doivent avoir la mme longueur, on complte avec des "0"
            P2 = (m - d2)*[0] + y.coeffs
            return Polynome(*(i + j for i,j in zip(P1, P2)))
        else:
            liste = self.coeffs[:]
            liste[-1] += y
            return Polynome(*liste)


    def __neg__(self):
        return Polynome(*(-k for k in self.coeffs))


    def __mul__(self, y):
        if isinstance(y, Polynome):
            n = self.degre()
            m = y.degre()
            if n < 0 or m < 0:
                return Polynome(0)
            coeffs = []
            for i in range(n+m+1):
                coeffs.append(0)
                j = 0
                k = i - j # le ieme coefficient est la somme des aj*bk, tels que j+k==i.
                while k >= 0:
                    if j <= n and k <= m:
                        coeffs[i] += self.coeffs[j]*y.coeffs[k]
                    j += 1
                    k = i - j
            return Polynome(*coeffs)
        else:
            return Polynome(*(y*k for k in self.coeffs))


    def __float__(self):
        return Polynome(*(float(coeff) for coeff in self.coeffs))

    def __pow__(self, y):
        if y == 0: return 1
        if y == 1: return self
        # On applique l'exponentiation indienne.
        return self**(y//2)*self**(y//2)*self**(y%2)

    def __div__(self, y):
        if isinstance(y, (int, long)):
            return self.__mul__(Frac(1, y))
        if isinstance(y, (float, Frac, Rac)):
            return self.__mul__(1/y)
        return NotImplemented

    def __gt__(self, y):
        if isinstance(y, Polynome) and y.deg() < 0 or not isinstance(y, Polynome) and y == 0: # bref, si y = 0, mais pos tel quel, on aurait une rcursion infinie !
            from resolution import Systeme
            return Systeme([[(self, 2)]])
        else:
            return self - y > 0

    def __eq__(self, y):
        if isinstance(y, Polynome) and y.deg() < 0 or not isinstance(y, Polynome) and y == 0: # bref, si y = 0, mais pos tel quel, on aurait une rcursion infinie !
            if self.deg() < 0: # ie. le polynome est nul.
                return True
            if self.deg() == 0: # le polynome est une constante non nulle
                return False
            from resolution import Systeme
            return Systeme([[(self, 1)]])
        else:
            return self - y == 0

    def __lt__(self, y):
        return -self > -y

    def __ge__(self, y):
        return (self > y) | (self == y)

    def __le__(self, y):
        return (self < y) | (self == y)

    def __cmp__(self, y):
        if isinstance(y, Polynome):
            if self.deg() < y.deg():
                return -1
            elif self.deg() > y.deg():
                return 1
            else:
                return cmp(self.coeffs, y.coeffs)
        else:
            return 1

    def __ne__(self, y):
        resultat = ~self.__eq__(y)
        if resultat is -1:
            return True
        elif resultat is -2:
            return False
        return resultat

    def __nonzero__(self): # la valeur logique de l'objet (la doc officielle est fausse ??)
        return self.deg() >= 0

    def __floordiv__(self, y):
        if isinstance(y, (int, long, float, Frac, Rac)):
            return self.__div__(y)
        if isinstance(y, Polynome):
            return divise_polynome(self, y)[0]
        return NotImplemented

    def __mod__(self, y):
        if isinstance(y, (int, long, float, Frac, Rac)):
            return 0
        if isinstance(y, Polynome):
            return divise_polynome(self, y)[1]
        return NotImplemented





class PlusInfini(ObjetMathematique):
    def __cmp__(self, y):
        return 1

    def __neg__(self):
        return MoinsInfini()


    def __invert__(self):
        return 0

    def __eq__(self, y):
        return y.__class__ == self.__class__


    def __gt__(self, y):
        return True

    def __add__(self, y):
        if y == -self:
            raise ArithmeticError, str2(u"forme indtermine.")
        return self


    def __mul__(self, y):
        if y > 0:
            return self
        elif y < 0:
            return -self
        else:
            raise ArithmeticError, str2(u"forme indtermine.")

    def __pow__(self, y):
        if int(y) == y and y != 0:
            return self
        return NotImplemented
        
    def __div__(self, y):
        if y > 0:
            return self
        elif y < 0:
            return -self
        else:
            raise ArithmeticError, str2(u"forme indtermine.")

    def __rdiv__(self, y):
        if y <> self and y <> -self:
            return 0
        raise ArithmeticError, str2(u"forme indtermine.")



    def __str__(self):
        return u"+infini"

    def __repr__(self):
        return self.__class__.__name__ + "()"





class MoinsInfini(PlusInfini):
    def __cmp__(self, y):
        return -1

    def __neg__(self):
        return PlusInfini()

    def __gt__(self, y):
        return False

    def __str__(self):
        return u"-infini"

    def __pow__(self, y):
        if int(y) == y and y != 0:
            if int(y)%2:
                return self
            else:
                return -self
        return NotImplemented


oo = infini = Infini = PlusInfini()



if param.version_numerix == "numpy":
    class Mat(matplotlib.numerix.Matrix):
        def __new__(cls, *lignes):
            if lignes and lignes[0] and isinstance(lignes[0][0], (list, tuple, types.ArrayType)):
                lignes = lignes[0]
            lignes = [tuple(ligne) for ligne in lignes]
            return matplotlib.numerix.Matrix.__new__(cls, lignes)


        #def __repr__(self):
        #    return matplotlib.numerix.Matrix.__repr__(self).replace("matrix", "Mat")

    matrix = Matrix = Mat

else: # Numeric (now depreciated)
    class Mat(matplotlib.numerix.Matrix):
        def __init__(self, *lignes):
            if lignes and lignes[0] and isinstance(lignes[0][0], (list, tuple, types.ArrayType)):
                lignes = lignes[0]
            matplotlib.numerix.Matrix.__init__(self, list(lignes))

        def __repr__(self):
            return "Mat(" + str(self.asarray()).replace("\n",",")[1:-1] + ")"



class Product(list):
    u"Classe utilise essentiellement pour des questions de formatage."
    def __init__(self, *args):
        list.__init__(self, args)
        self.sort(cmp = self.cmp)


    def __str__(self):
        self.sort(cmp = self.cmp)
        if len(self) is 1:
            return str(self[0])
        t = "(" + ")(".join(str(elt) for elt in self if elt <> 1) + ")"
        s = ""
        while t <> s:
            s = t
            t = regexp("[(][0-9./*]+[)][(][0-9./*]+[)]", s, "x.replace(')(', '*')")
        s = ""
        while t <> s:
            s = t
            t = regexp("[)][(][0-9./*]+[)]", s, "x.replace(')', '*').replace('*(', ')*')").strip("*")
            t = regexp("[(][0-9./*]+[)][(]", s, "x.replace('(', '*').replace(')*', '*(')").strip("*")
        #s = regexp("[(][0-9./*]+[)]", s, "x.replace('(', '*').replace(')', '*')").strip("*")
        return s

    def cmp(self, x, y):
        # Il y a une erreur dans la doc de Python.
        # x.__cmp__(y) n'est pas quivalent = cmp(x, y)
        # tout dpend du type de y (si y est de type int par ex, cmp(x, y) appelle x.__eq__(y)).
        try:
            return x.__cmp__(y)
        except TypeError, AttributeError:
            try:
                return -y.__cmp__(x)
            except TypeError, AttributeError:
                return cmp(x, y)



