#!/usr/bin/env python
# -*- coding: iso-8859-1 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#                Calculatrice                 #
##--------------------------------------#######
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2007  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# version unicode


from API import *
from GUI import *
from mathlib.fonctions_utilisateur import __fonctions__






class CalculatriceMenuBar(MenuBar):
    def __init__(self, panel):
        MenuBar.__init__(self, panel)
        self.ajouter(u"Fichier", ["quitter"])
        self.ajouter(u"Affichage", ["onglet"])
        liste = []
        for fonction in __fonctions__:
            if fonction is None:
                liste.append(None)
            else:
                liste.append([fonction.__name__.capitalize(), fonction.__doc__.split("\n")[0], None, self.generer_fonction(fonction)])
                
        self.ajouter(u"Fonctions", *liste)
        self.ajouter(u"Outils",
                        [u"Mmoriser le rsultat", u"Copie le resultat du calcul dans le presse-papier, afin de pouvoir l'utiliser ailleurs.", "Ctrl+M", self.panel.vers_presse_papier],
                        [u"Rinitialiser", u"Rinitialiser la calculatrice", "Ctrl+I", self.panel.initialiser],
                        [u"debug"],
                        [u"options"])
        self.ajouter("?")



    def generer_fonction(self, fonction):
        def f(event = None, panel = self.panel, fonction = fonction):
            #panel.entree.SetInsertionPoint(0)
            panel.entree.WriteText(fonction.__name__ + "(")
            panel.entree.SetFocus()
        return f














class Interprete(object):
    def __init__(self, parent):
        self.parent = parent
        # Dictionnaire local (qui contiendra toutes les variables dfinies par l'utilisateur).
        self.locals = {}
        # Dictionnaire global (qui contient les fonctions, variables et constantes prdfinies).
        self.globals = vars(mathlib.liste_fonctions).copy()
        self.globals.update({
                "__builtins__": None,
                "__expression_formelle__": calcul_formel.expression_formelle,
                "__fonction__": calcul_formel.Fonction,
                "ans": self.parent.ans,
                "__vars__": self.vars,
                #"phi":  
                "X": Polynome(1, 0),
                "oo": PlusInfini(),
                "infini": PlusInfini(),
                            })
        # pour viter que les procdures de rcriture des formules ne touchent au mots clefs,
        # on les rfrence comme fonctions (elles seront inaccessibles, mais ce n'est pas grave).
        # ainsi, "c and(a or b)" ne deviendra pas "c and*(a or b)" !
        self.globals.update({}.fromkeys(securite.__keywords_autorises__, lambda:None))
        

    def initialiser(self):
        self.locals.clear()

    def evaluer(self, calcul = "", mode_exact = True):
        
        # calcul = re.sub("[_]+", "_", calcul.strip()) # par mesure de scurit, les "__" sont interdits.
        # Cela permet ventuellement d'interdire l'accs  des fonctions.
        # Warning: inefficace. Cf. "_import _builtins_import _
        
        # Ferme automatiquement les parentheses.
        parentheses = "({[", ")}]"
        for i in range(3):
            correction = parentheses[1][i]*(calcul.count(parentheses[0][i])-calcul.count(parentheses[1][i]))
            calcul = calcul + correction
            if correction: self.parent.warning += u" Attention, il manque des parenthses \"" + parentheses[1][i] + "\"."

        if calcul[0] in "*/^%+" : calcul = "ans()" + calcul

        # Si f est une fonction, il ne faut pas remplacer "f(" par "f*(".
        

        if True in [calcul.startswith(exception + " ") for exception in securite.__keywords_debut_ligne__]:
            self.executer(calcul)
            
        elif re.match("[^=]+=[^=]", calcul):
            var, val = calcul.split("=", 1)
        
            # Il y a ici 2 cas:
            #   - variable
            #   - fonction -> cf. classe Fonction (calcul_formel.py)
            
            var = var.strip()
            i = var.find("(")
            if i == -1: # cas d'une variable. Ex: a=5+3
                if mode_exact:
                    self.executer(var + "=__expression_formelle__(" + repr(self.traduire(val)) + ", True, __vars__())")
                    # Test des rfrences circulaires:
                    _ = self.locals["_"]
                    if isinstance(_, calcul_formel.Expression_formelle) and calcul_formel.Variable_formelle(var) in _.variables():
                        self.locals.pop(var, None)
                        self.locals["_"] = u"Rfrence circulaire (%s dpend de %s)" %(var, var)
                else:
                    self.executer(self.traduire(calcul))
                # exec/eval encodent les chanes cres en utf8.
                if isinstance(self.locals[var], str):
                    self.locals["_"] = self.locals[var] = self.locals[var].decode("utf8").encode(param.encodage)
                    
            else: # cas d'une fonction. Ex: f(x)=x+3
                nom = var[:i]
                variables = var[i+1:-1]
                self.executer(nom + "=__fonction__(" + repr(nom) + "," + repr(variables + ":" + self.traduire(val)) + ", __vars__())")
        else:
            calcul = self.traduire(calcul)
            if mode_exact:
                self.executer("__expression_formelle__(" + repr(calcul) + ", True, __vars__())")
            else:
                self.executer(calcul)
                if isinstance(self.locals["_"], mathlib.calcul_formel.Objet_formel):
                    self.executer(self.parent.formater(self.locals["_"]))
        return self.locals["_"]

    def traduire(self, formule):
        variables = self.globals.copy()
        variables.update(self.locals)
        #callable_types = (  types.BuiltinFunctionType,
                            #types.BuiltinMethodType, 
                            #types.FunctionType, 
                            #types.UfuncType, 
                            #types.MethodType, 
                            #types.TypeType, 
                            #types.ClassType)
        #fonctions = [key for key, val in variables.items() if isinstance(val, callable_types)]
        fonctions = [key for key, val in variables.items() if hasattr(val, "__call__")]
        #print fonctions
        # La fonction traduire_formule de la librairie mathlib.formatage permet d'effectuer un certain nombre de conversions.
        return traduire_formule(formule, fonctions, 
                        OOo = self.parent._param_.formatage_OOo, 
                        LaTeX = self.parent._param_.formatage_LaTeX, 
                        changer_separateurs = self.parent._param_.changer_separateurs, 
                        separateurs_personnels = self.parent._param_.separateurs_personnels)

    def executer(self, instruction):
        # dans certains cas, il ne faut pas affecter le rsultat  la variable "_" (cela provoquerait une erreur de syntaxe)
        if True in [instruction.startswith(exception + " ") for exception in securite.__keywords_debut_ligne__]:
            self.locals["_"] = None
        else:
            instruction = "_=" + instruction
        
        keywords_interdits = [kw + " " for kw in securite.__keywords_interdits__] + [kw + "(" for kw in securite.__keywords_interdits__]
        
        instruction = recursive_mreplace(instruction, keywords_interdits)
        #print self.globals.keys()
        exec(instruction, self.globals, self.locals)
        
    
    def vars(self):
        dictionnaire = self.globals.copy()
        dictionnaire.update(self.locals)
        return dictionnaire


















class Calculatrice(Panel_simple):
    __titre__ = u"Calculatrice" # Donner un titre a chaque module
    
    def __init__(self, parent):
        Panel_simple.__init__(self,parent)


        self.entrees = wx.BoxSizer(wx.HORIZONTAL)
        self.entree = wx.TextCtrl(self, size = (500, -1))
        self.entrees.Add(self.entree, 1, wx.ALL|wx.GROW, 5)
        self.valider = wx.Button(self, wx.ID_OK)
        self.entrees.Add(self.valider, 0, wx.ALL, 5)
                
        self.corps = wx.BoxSizer(wx.HORIZONTAL)
        self.resultats = wx.TextCtrl(self, size = (400,310), style = wx.TE_MULTILINE | wx.TE_READONLY | wx.TE_RICH)
        self.corps.Add(self.resultats, 0, wx.ALL, 5)
        
        
        ### Pave numerique de la calculatrice ###
        # On construit le pav de la calculatrice.
        # Chaque bouton du pav doit provoquer l'insertion de la commande correspondante.
        
        self.pave  = wx.BoxSizer(wx.VERTICAL)
        self.corps.Add(self.pave, 0, wx.ALL, 5)
        boutons = ["2nde", "ans", "ouv", "ferm", "egal", "7", "8", "9", "div", "x", "4", "5", "6", "mul", "y", "1", "2", "3", "minus", "z", "0", "pt", "pow", "plus", "t", "rac", "sin", "cos", "tan", "exp", "i", "pi", "e", "abs", "mod"]
        inserer = ["", "ans()", "(", ")", "=", "7", "8", "9",  "/", "x", "4", "5", "6", "*", "y", "1", "2", "3", "-", "z", "0", ".", "^", "+", "t", "sqrt(", ("sin(", "asin(", "sinus / arcsinus"), ("cos(", "acos(", "cosinus / arccosinus"), ("tan(", "atan(", "tangente / arctangente"), ("exp(", "ln(", "exponentielle / logarithme neperien"), ("i", "cbrt(", "i / racine cubique"), ("pi", "sinh(", "pi / sinus hyperbolique"), ("e", "cosh", "e / cosinus hyperbolique"), ("abs(", "tanh", "valeur absolue / tangente hyperbolique"), (" mod ", "log10(", "modulo / logarithme decimal")]
        
        self.seconde = False # indique si la touche 2nde est activee.
        
        def action(event = None):
            self.seconde = not self.seconde
            if self.seconde: self.message(u"Touche [2nde] active.")
            else: self.message("")
            
        self.actions = [action]
        
        for i in range(len(boutons)):
            # On aligne les boutons de la calculatrice par rangees de 5.
            if i%5 == 0:
                self.rangee = wx.BoxSizer(wx.HORIZONTAL)
                self.pave.Add(self.rangee, 0, wx.ALL,0)
            
            # Ensuite, on construit une liste de fonctions, parallelement  la liste des boutons.
            if i > 0:
                def action(event = None, entree = self.entree, j = i):
                    if type(inserer[j]) == tuple:
                        entree.WriteText(inserer[j][self.seconde])
                    else:
                        entree.WriteText(inserer[j])
                    n = entree.GetInsertionPoint()
                    entree.SetFocus()
                    entree.SetInsertionPoint(n)
                    self.seconde = False
                    self.message("")
                self.actions.append(action)
                
            bmp = png('btn_' + boutons[i])
            bouton = wx.BitmapButton(self, -1, bmp, style=wx.NO_BORDER)
            bouton.SetBackgroundColour(self.GetBackgroundColour())
            espace = 3
            if param.BugWxSizerGtk:           
                espace = 0
            self.rangee.Add(bouton, 0, wx.ALL, espace)
            # A chaque bouton, on associe une fonction de la liste.
            bouton.Bind(wx.EVT_BUTTON, self.actions[i])
            if type(inserer[i]) == tuple: bouton.SetToolTipString(inserer[i][2])

        self.pave.Add(wx.BoxSizer(wx.HORIZONTAL))

        
        ### Liste des options ###
        # En dessous du pav apparait la liste des diffrents modes de fonctionnement de la calculatrice.
        
        self.options = [(u"Calcul formel.", u"calcul_exact"),
                    (u"Ecriture scientifique.", u"ecriture_scientifique"), 
                    (u"Virgule comme sparateur dcimal.", u"changer_separateurs"), 
                    (u"Copie du rsultat dans le presse-papier.", u"copie_automatique"),
                    (u"Formatage OpenOffice.org", u"formatage_OOo"),
                    (u"Formatage LaTeX", u"formatage_LaTeX"),
                        ]
        self.options_box = []
        for i in range(len(self.options)):
            ligne = wx.BoxSizer(wx.HORIZONTAL)
            self.options_box.append(wx.CheckBox(self))
            ligne.Add(self.options_box[i])
            self.options_box[i].SetValue(eval("self._param_." + self.options[i][1]))
            def action(event, chaine = self.options[i][1], entree = self.entree, self = self):
                exec("self._param_." + chaine + "= not self._param_." + chaine)
                entree.SetFocus()
            self.options_box[i].Bind(wx.EVT_CHECKBOX, action)
            ligne.Add(wx.StaticText(self, -1, "  " + self.options[i][0]))
            self.pave.Add(ligne)
            
            
        self.option1 = wx.BoxSizer(wx.HORIZONTAL)
        self.pave.Add(self.option1)
        #wx.CheckBox(self)
        
        self.option2 = wx.BoxSizer(wx.HORIZONTAL)
        self.pave.Add(self.option2)
 
        
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(self.entrees, 0, wx.ALL, 5)
        self.sizer.Add(self.corps, 0, wx.ALL, 5)
        self.SetSizer(self.sizer)
        self.Fit()
        
        # historique des calculs
        self.entree.Bind(wx.EVT_KEY_UP, self.EvtChar)
        self.entree.Bind(wx.EVT_MIDDLE_DOWN, self.EvtMenu)
        #self.Bind(wx.EVT_CHAR, self.EvtChar, self.entree)
        self.valider.Bind(wx.EVT_BUTTON, self.affiche)
        
        self.interprete = Interprete(self)
        
        self.initialiser()
        
        
    # <EN TRAVAUX>
    
    def _sauvegarder(self, fgeo):
        fgeo.contenu["Calculatrice"] = [{}]
        fgeo.contenu["Calculatrice"][0]["Calcul"] = self.historique
        fgeo.contenu["Calculatrice"][0]["Resultats"] = [repr(self.derniers_resultats)]
        fgeo.contenu["Calculatrice"][0]["Affichage"] = [self.resultats.GetValue()]
        fgeo.contenu["Calculatrice"][0]["Dernier_resultat"] = [self.dernier_resultat]
        fgeo.contenu["Calculatrice"][0]["Position"] = [repr(self.position)]
        fgeo.contenu["Calculatrice"][0]["Options"] = [{}]
        for i in range(len(self.options)):
            fgeo.contenu["Calculatrice"][0]["Options"][0][self.options[i][1]] = [str(self.options_box[i].GetValue())]


        
    def _ouvrir(self, fgeo):
        if fgeo.contenu.has_key("Calculatrice"):
            calc = fgeo.contenu["Calculatrice"][0]
            self.initialiser()
            
            self.historique = calc["Calculs"]
            self.derniers_resultats = securite.eval_safe(calc["Resultats"][0])
            self.position = int(calc["Position"][0])
            self.resultats.SetValue(calc["Affichage"][0] + "\n")
            
            liste = calc["Options"][0].items()
            options = [option for aide, option in self.options]
            for key, value in liste:
                value = securite.eval_safe(value[0])
                setattr(self._param_, key, value)
                if key in options:
                    self.options_box[options.index(key)].SetValue(value)
            # il faudrait encore sauvegarder les variables, mais la encore, 2 problemes :
            # - pb de securite pour evaluer les variables
            # - pb pour obtenir le code source d'une fonction.
            # Pour remedier a cela, il faut envisager de :
            # - creer un module d'interpretation securisee.
            # - rajouter a chaque fonction un attribut __code__, ou creer une nouvelle classe.
    
    # </EN TRAVAUX>

    def vers_presse_papier(self, event = None):
        self.clipBoard=wx.TheClipboard
        if self.clipBoard.Open():
            self.clipBoard.AddData(wx.TextDataObject(self.dernier_resultat))
        self.clipBoard.Close()
    
    def initialiser(self, event = None):
        self.dernier_resultat = "" # dernier resultat, sous forme de chaine formatee pour l'affichage
        self.historique = []
        self.position = -1
        self.derniers_resultats = [] # liste des resultats, sous forme de valeurs numeriques
        self.entree.SetFocus()
        self.interprete.initialiser()
        self.resultats.Clear()
    
    def affiche(self, event = None):
        entree = self.entree.GetValue()
        if not entree:
            if self.historique:
                entree = self.historique[-1]
            else:
                return
        try:
            self.warning = ""
            resultat = self.interprete.evaluer(entree, self._param_.calcul_exact)
            longueur = len(self.resultats.GetValue())
            self.resultats.AppendText(u" Calcul n" + str(len(self.derniers_resultats)) + " :   " + unicode2(entree) + u"\n Rsultat :" + " "*(4+len(str(len(self.derniers_resultats)))) + unicode2(self.formater(resultat)) + "\n__________________\n\n")
            self.message(u"Calcul effectu." + self.warning)
            self.historique.append(entree)
            self.derniers_resultats.append(resultat)
            self.position = len(self.historique) - 1
            self.entree.Clear()
            self.resultats.SetInsertionPoint(len(self.resultats.GetValue()))
            self.resultats.SetFocus()
            self.resultats.ScrollLines(1)
            self.entree.SetFocus()
        except Exception, erreur:
            self.message(u"Calcul impossible. " + pylib.erreurs.message(erreur) + (self._param_.calcul_exact and u" Essayez en dcochant l'option 'Calcul formel'." or ""))
            self.entree.SetFocus()
            if param.debug:
                raise


    def EvtMenu(self, event = None):
        def generer_fonction(fonction, self = self):
            def f(event = None, self = self, fonction = fonction):
                self.entree.WriteText(fonction.__name__ + "(")
            return f
        menu = wx.Menu()
        menu.SetTitle(u"Fonctions mathmatiques")
        for fonction in __fonctions__:
            if fonction is None:
                menu.AppendSeparator()
            else:
                i = wx.NewId()
                menu.Append(i, fonction.__name__.capitalize(), fonction.__doc__.split("\n")[0])
                menu.Bind(wx.EVT_MENU, generer_fonction(fonction), id =i)
        self.PopupMenu(menu)
        menu.Destroy()
        self.entree.SetFocus()


    
    def EvtChar(self,event):    # gere la ligne d'entree des calculs.
        code = event.GetKeyCode()
        
        #print code
        if code == wx.WXK_RETURN or code == wx.WXK_NUMPAD_ENTER:
            self.affiche()

        elif code == wx.WXK_UP:   # on remonte dans l'historique des entrees
            if self.position >= 0:
                self.entree.SetValue(self.historique[self.position])
                self.position = self.position - 1

        elif code == wx.WXK_DOWN:   # on redescend dans l'historique vers les entrees recentes
            if self.position < len(self.historique) - 1:
                if self.position == (len(self.historique) - 2):
                    self.entree.Clear()
                else:
                    self.entree.SetValue(self.historique[self.position + 2])
                self.position = self.position + 1
        else:
            #return
            event.Skip()
            
            
        
    #~ def evaluer(self, calcul):
    
                
                
        #~ self.warning = ""
        
                
        #~ # Pour multiplier par 2 le dernier resultat, il suffit de taper "*2", le ans() s'insere avant.
        #~ if calcul[0] in "*/^%+" : calcul = "ans()" + calcul
        
        #~ calcul = calcul.replace("ans(", "self.ans(")
        
        
        #~ # Ferme automatiquement les parentheses.
        #~ parentheses = "({[", ")}]"
        #~ for i in range(3):
            #~ correction = parentheses[1][i]*(calcul.count(parentheses[0][i])-calcul.count(parentheses[1][i]))
            #~ calcul = calcul + correction
            #~ if correction: self.warning += u" Attention, il manque des parenthses \"" + parentheses[1][i] + "\"."


        #~ # Si a, b, c... ne sont pas des fonctions, on remplace "a(" par "a*(", etc...
        #~ #variables = filter(lambda x:type(eval("self." + x)) <> types.FunctionType, "".ascii_lowercase + "".ascii_uppercase)
        #~ variables =""
        #~ for lettre in string.ascii_lowercase + string.ascii_uppercase:
            #~ if type(eval("self." + lettre)) <> types.FunctionType:
                #~ variables += lettre
        

        #~ # La classe Expression de la librairie CALC_objets permet d'effectuer un certain nombre de conversions.
        #~ calcul = Expression(calcul, variables, self._param_.calcul_exact, changer_separateurs = self._param_.changer_separateurs, separateurs_personnels = self._param_.separateurs_personnels).reformuler()
        
        
        #~ # On affecte a la variable a la valeur de self.a, a la variable A celle de self.A, etc... et i vaut 1j.
        #~ for lettre in string.ascii_lowercase + string.ascii_uppercase:
            #~ vars = globals()
            #~ vars.update(locals())
            #~ exec lettre + "=self." + lettre in vars
            
        #~ i = 1j
        #~ e = math.e
        #~ # on evite d'attribuer la valeur 2.718... a E, pour eviter les confusions entre 3E-7 et 3*E-7 (!)
        #~ try:
            #~ del E
        #~ except NameError:
            #~ pass
        #~ pi = PI = math.pi
        #~ X = Polynome(1, 0)

        #~ def rac(x):
            #~ if self._param_.calcul_exact and not isinstance(x, float): return Rac(x)
            #~ return sqrt(x)
        #~ def frac(x):
            #~ if self._param_.calcul_exact: return Frac(x)
            #~ return float(x)


        #~ # Le code qui suit sert a affecter une valeur aux variables a, b, ... z.
        #~ # On remplace a=3, d=sin(2) par self.a=3, self.d=sin(2), et on execute.
        #~ l = re.split("([ ]*?[a-zA-Z][ ]*?=[^=])", calcul)
        #~ if len(l) <> 1:
            #~ if l[1].strip()[0] not in "XEei":
                #~ vars = globals()
                #~ vars.update(locals())
                #~ exec "self." + l[1].strip() + l[2] in vars
                #~ l = (l[1]+l[2]).split("=")
                #~ print eval(l[1])
                #~ print str(eval(l[1])).decode("utf8")
                #~ print str(eval(l[1])).decode(param.encodage)
                #~ print param.encodage
                #~ return u"La variable \"" + l[0].strip() + "\" vaut maintenant " + unicode2(eval(l[1])) + "."
            #~ else:
                #~ return u"""Les constantes "i", "E", "X" et "e" sont protges."""
        
        #~ # On definit des fonctions de la sorte : "f(x)=2x".
        #~ # Attention : a cause des corrections automatiques, une chaine de cette forme arrive ici sous la forme
        #~ # "f*( x)=2*x", donc il faudra supprimer le premier "*".
        #~ l = re.split(u"([a-zA-Z][*]?[(][ ]?[a-zA-Z,]*[ ]?[)]=)", calcul)
        #~ if len(l) <> 1:
            #~ if l[1][0] not in "XEei":
                #~ vars = globals()
                #~ vars.update(locals())
                #~ exec "def " + l[1][:-1].replace("*", "") + ":return " + l[2] in vars # def f(x):return 2*x dans cet exemple
                #~ exec "self." + l[1][0] + "=" + l[1][0] in vars # self.f = f dans cet exemple
                #~ return u"La fonction \"" + l[1][0] + u"\" est dfinie par " + l[1][0] + ": " +   l[1][4:-3] + " -> " + l[2] + "."
            #~ else:
                #~ return u"""Les constantes "i", "E", "X" et "e" sont protges."""
        
        #~ debug(calcul)
        #~ return eval(calcul)


    def ans(self, n = -1):
        n = int(n)
        if self.derniers_resultats: return self.derniers_resultats[n]
        self.warning += u" Ans(): aucun calcul antrieur." 
        return 0
    
    def formater(self, valeur):
        if isinstance(valeur, tuple):
            return ", ".join(self.formater(i) for i in valeur)
        if isinstance(valeur, (list, types.ArrayType)) and not isinstance(valeur, Product):
            return "[" + ", ".join(self.formater(i) for i in valeur) + "]"
            
        resultat = unicode2(valeur)
        # Les nombres complexes sont notes entre parentheses, ce qui n'est pas tres "esthetique"...
        # L'expression reguliere suivante les enleve.
        # resultat = regexp("[^a-zA-Z][(][-]?[0-9]+[.]?[0-9]*[+-][0-9]+[.]?[0-9]*j[)]", " " + resultat, "x[0]+x[2:-1]")[1:]
        # Cela posait probleme dans certains cas (produits de facteurs...), on les enleve seulement dans les cas simples.
        resultat = regsub("^[(][^()]*[)]$", resultat, lambda x:x[1:-1])
        

        # Une bonne part du formatage concerne les nombres complexes : 0j est remplace par 0, par exemple.
        resultat = resultat.replace("+0j", "")
                
        def formater_nb_complexe(chaine):
            prefixe = chaine[0]
            nbr = chaine[1:-1]
            if nbr == "0":
                return prefixe + "0"
            elif nbr == "1":
                return prefixe + "i"
            else:
                return prefixe + nbr + "i"
        # Note: on ajoute un espace avant la formule (qu'on enlve  la fin).
        # C'est juste une astuce pour viter le cas particulier o l'expression  traiter est en dbut de ligne.
        resultat = regsub("[^0-9A-Za-z.][0-9.]*[0-9]j", " " + resultat, formater_nb_complexe)[1:]
        
        
        
        resultat = resultat.replace("-", " - ").replace("+", " + ").replace("e - ", "E-").replace("e + ", "E+")
        
        # Gestion de l'ecriture scientifique
        if self._param_.ecriture_scientifique:
            try:
                if isinstance(valeur, float):
                    resultat = eval(resultat)
                    mantisse = int(math.floor(math.log10(abs(resultat))))
                    resultat = str(resultat/10.**mantisse)+"E"+str(mantisse)
            except: pass
        if self._param_.changer_separateurs:
            resultat = resultat.replace(",", self._param_.separateurs_personnels[1]).replace(".", self._param_.separateurs_personnels[0])
        
        # Cette partie sert au presse-papier.
        self.dernier_resultat = resultat
        if self._param_.copie_automatique: 
            self.vers_presse_papier()
        return resultat
