# -*- coding: iso-8859-1 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#                 Geometre                    #
##--------------------------------------#######
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2007  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA


# version unicode


from API import *
from GUI import *
#import _param_



class GeometreMenuBar(MenuBar):
    def __init__(self, panel):
        MenuBar.__init__(self, panel)
        self.ajouter(u"Fichier", [u"nouveau"], [u"ouvrir"], [u"ouvrir ici"], [u"enregistrer"], [u"enregistrer_sous"], [u"exporter"], [u"exporter&sauver"], None, [u"mise en page"], [u"imprimer"], [u"presse-papier"], None, [u"proprietes"], None, self.panel.doc_ouverts, None, [u"fermer"], [u"quitter"])
        self.ajouter(u"Editer", [u"annuler"], [u"refaire"], [u"modifier"], [u"supprimer"])
        self.ajouter(u"creer")
        self.ajouter("affichage")
        self.ajouter("autres")
        self.ajouter(u"Outils", [u"historique"], [u"securise"], [u"debug"], [u"options"])
        self.ajouter(u"?")






class MultiButton(wx.BitmapButton):
    def __init__(self, parent, raccourci, selectionnable, *liste):
        self.parent = parent
        self.raccourci = raccourci
        self.selectionnable = selectionnable
        self.selected = False
        self.liste = list(liste)
        self.items = []
        for i in xrange(len(self.liste)):
            if self.liste[i] is None:
                self.items.append(None)
            else:
                self.items.append(wx.NewId())
                self.parent.Bind(wx.EVT_MENU, self.OnPopup, id=self.items[-1])
                self.liste[i] = list(self.liste[i])
                # on cre 2 images, nomme segment.png et segment_.png par exemple, la 2e correspondant au bouton slectionn.
                self.liste[i][1] = [png(self.liste[i][1] + s) for s in ("", "_")]
        wx.BitmapButton.__init__(self, parent, style=wx.NO_BORDER)
        self.SetBackgroundColour(self.parent.GetBackgroundColour())
        self.select(False, 0)
        self.Bind(wx.EVT_RIGHT_DOWN, self.OnRightclic)
        self.Bind(wx.EVT_BUTTON, self.OnLeftclic)
        if self.raccourci is not None:
            self.parent.canvas.Bind(wx.EVT_CHAR, self.OnChar)

    def select(self, sel = True, i = None):
        u"Slectionne et affiche le bouton i"
        if i is not None:
            self.selection = i
            if self.raccourci is None:
                touche = ""
            else:
                touche = " (" + self.raccourci.replace("Shift", "Maj") +")"
            aide = self.liste[self.selection][2] + touche
            if len(self.liste) > 1:
                aide += "\n\nClic droit pour plus de choix."
            self.SetToolTipString(aide)
        self.display(sel)
        
    def display(self, selected = None):
        if self.selectionnable:
            if selected is not None:
                self.selected = selected
                if selected is True:
                    if hasattr(self.parent, "selected_button") and self.parent.selected_button is not self:
                        self.parent.selected_button.display(False)
                    self.parent.selected_button = self
        #else:
        #    self.selected = False
        self.SetBitmapLabel(self.liste[self.selection][1][self.selected])
        
    def action(self, event = None):
        self.liste[self.selection][3](event)
        
    def OnLeftclic(self, event = None):
        self.display(True)
        self.action(event)
        
        
    def OnRightclic(self, event = None):
        if len(self.liste) > 1:
            menu = wx.Menu()
            for i in xrange(len(self.liste)):
                if self.liste[i] is None:
                    menu.AppendSeparator()
                    #item = wx.MenuItem(menu, id = wx.ITEM_SEPARATOR, kind = wx.ITEM_SEPARATOR)
                else:
                    item = wx.MenuItem(menu, self.items[i], self.liste[i][0])
                    #item.SetBitmap(self.liste[i][1][0])
                    #item.SetBackgroundColour(self.parent.GetBackgroundColour())
                    #-> ces mthodes sont largement buggues dans WxPython
                    menu.AppendItem(item)
                    del item
            self.PopupMenu(menu)
            menu.Destroy()
        
    def OnPopup(self, event = None):
        self.select(True, self.items.index(event.GetId()))
        self.action(event)

    def OnChar(self, event = None):
        r = self.raccourci.split("+")
        if len(r) is 2:
            m = r[0].lower()
            shift = m == "shift"
            alt = m == "alt"
            ctrl = m == "ctrl"
        else:
            shift = alt = ctrl = False
        if (shift^event.ShiftDown()) or (alt^event.AltDown()) or (ctrl^event.ControlDown()):
            event.Skip()
            return
        keycode = event.GetKeyCode()
        if hasattr(wx, "WXK_" + r[-1]) and getattr(wx, "WXK_" + r[-1]) == keycode:
            self.OnLeftclic(event)
        elif 0 <= keycode < 256 and chr(keycode) == r[-1]:
            self.OnLeftclic(event)
        else:
            event.Skip()
















class Geometre(Panel_API_graphique):

    __titre__ = u"Gomtrie dynamique" # Donner un titre  chaque module
    #_param_ = _param_

    def __init__(self, parent):
        Panel_API_graphique.__init__(self, parent)
    
        self.entrees = wx.BoxSizer(wx.HORIZONTAL)
        
        self.debut_selection = None
        self.debut_zoombox = None
                
        def add(racc, *liste):
            u"Ajoute des boutons multiples, slectionnables."
            button = MultiButton(self, racc, True, *liste)
            i = 5
            if param.BugWxSizerGtk:           
                i = 0
            self.entrees.Add(button, 0, wx.ALL, i)
            return button

        def add2(*args):
            u"Ajoute des boutons simples, non slectionnables."
            button = MultiButton(self, None, False, ("",) + args)
            i = 5
            if param.BugWxSizerGtk:           
                i = 0
            self.entrees.Add(button, 0, wx.ALL, i)
            return button

            
        add2(u"ouvrir", u"Ouvrir un fichier .geo.", self.parent.OpenFile)
        self.btn_sauver = add2(u"sauvegarde", u"Enregistrer le document.", self.parent.SaveFile)
        add2(u"image", u"Exporter comme image.", self.parent.ExportFile)

        if self._param_.afficher_boutons:
            add("F1", (u"Pointeur", u"fleche3", u"Dplacer ou modifier un objet.", self.curseur),
                      (u"Zoomer", u"zoombox2", u"Recentrer la zone d'affichage.", self.zoombox),
                      (u"Slectionner", u"selection", u"Slectionner une partie de la feuille.", self.selectionner)).display(True)
            add("F2", (u"Point", u"point2",u"Crer un point.", self.point))
            add("F3", (u"Milieu", u"milieu2", u"Crer un milieu ou un centre.", self.milieu))
            add("F4", (u"Segment", u"segment2", u"Crer un segment.", self.segment),
                        None,
                        (u"Vecteur", u"vecteur", u"Crer un vecteur.", self.vecteur),
                        (u"Reprsentant d'un vecteur", u"representant", u"Crer un reprsentant d'un vecteur.", self.representant),
                        )
            add("F5", (u"Droite", u"droite2", u"Crer une droite.", self.droite),
                        (u"Demi-droite", u"demidroite", u"Crer une demi-droite.", self.demidroite),
                        None,
                        (u"Parallle", u"parallele", u"Crer une parallle.", self.parallele),
                        (u"Perpendiculaire", u"perpendiculaire", u"Crer une perpendiculaire.", self.perpendiculaire),
                        (u"Tangente", u"tangente", u"Crer une tangente.", self.tangente),
                        None,
                        (u"Mdiatrice", u"mediatrice", u"Crer une mdiatrice.", self.mediatrice),
                        (u"Bissectrice", u"bissectrice", u"Crer une bissectrice.", self.bissectrice),
                        )
            add("F6", (u"Cercle", u"cercle", u"Crer un cercle.", self.cercle),
                        (u"Cercle dfini par son diamtre", u"cerclediametre", u"Crer un cercle dfini par son diamtre.", self.cercle_diametre),
                        (u"Cercle passant par 3 points", u"cercle3points", u"Crer un cercle passant par 3 points.", self.cercle_points),
                        None,
                        (u"Arc de cercle", u"arc", u"Crer un arc de cercle de centre donn.", self.arc),
                        (u"Arc passant par 3 points", u"arc_points", u"Crer un arc de cercle passant par 3 points.", self.arc_points),
                        (u"Arc de cercle orient", u"arc_oriente", u"Crer un arc de cercle orient.", self.arc_oriente),
                        (u"Demi-cercle", u"demicercle", u"Crer un demi-cercle.", self.demicercle),
                        None,
                        (u"Disque", u"disque", u"Crer un disque.", self.disque),
                        )
            add("F7", (u"Polygone", u"polygone", u"Crer un polygone.", self.polygone),
                        (u"Triangle", u"triangle", u"Crer un triangle.", self.triangle),
                        (u"Paralllogramme", u"parallelogramme", u"Crer un paralllogramme.", self.parallelogramme),
                        )
            add("F8", (u"Intersection", u"intersection", u"Crer une intersection.", self.intersection))
            add("F9", (u"Angle", u"angle", u"Crer un angle non orient.", self.angle),
                        (u"Angle orient", u"angle_oriente", u"Crer un angle orient.", self.angle_oriente),
                        )
            add("Shift+F1", (u"Texte", u"texte", u"Crer un texte.", self.texte))
            add("Shift+F2", (u"Masquer", u"masquer", u"Masquer des objets.", self.masque))
            add("Shift+F3", (u"Gommer", u"gomme", u"Supprimer des objets.", self.gomme))
            add("Shift+F4", (u"Copier", u"pinceau", u"Copier le style d'un objet.", self.pinceau))
        
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(self.entrees, 0, wx.ALL, 5)
        self.sizer.Add(self.canvas, 1, wx.LEFT | wx.TOP | wx.GROW)
        self.SetSizer(self.sizer)
        self.Fit()

        #if param.afficher_boutons:
        #    self.canvas.Bind(wx.EVT_CHAR, self.raccourcis)
        # self.Bind(wx.EVT_CHAR, self.raccourcis) ne marche pas ?!? :-(
            #self.selection(3) # l'icone selectionnee DOIT avoir deux images associees : img.png et img_.png par exemple.
        
        
    def rafraichir_titre(self):
        Panel_API_graphique.rafraichir_titre(self)
        if self.actuelle and hasattr(self, "btn_sauver"): # evite les conflits lors de l'initialisation
            self.btn_sauver.selected = not self.actuelle.sauvegarde["modifie"]
            self.btn_sauver.display()
    
    
    def initialiser(self):
        self.cache = [] # objets en memoire
        #self.objet = None # objet recemment cree
        self.actuelle.objet_temporaire(None)
        if self.debut_zoombox is not None or self.debut_selection is not None:
            self.debut_zoombox = None
            self.debut_selection = None
            self.affiche() # on rafraichit l'affichage pour effacer le rectangle de slection ou de zoom


    #def valider_nom_precedent(self):
    #    try:
    #        self.canvas.editeur.init(valider = True)
    #    except NameError:
    #        pass
        

    def executer(self, instruction, nom = None, init = True):
        u"""Excute une instruction dans la console.
        Si nom is not None, le nom de l'objet est edite (points et textes seulement),
        afin de permettre de le renommer dans la foulee.
        Si init = True, le cache des selections est initialise."""
        self.commande.executer_code_interne(instruction)
        if init:
            self.initialiser()
        if nom:
            self.canvas.select = self.actuelle.objets[nom]
            self.canvas.editer()
                
            

        
    
    def interagir(self, *args, **kw):
        u"Marque le bouton n comme slectionn (et initialise le cache des slections)."
        self.initialiser()
        self.canvas.interagir(*args, **kw)

    
    
    def test(self, doublons_interdits = False, **kw):
        u"""On vrifie qu'il y ait encore une feuille, et que le cache ne contienne que des objets
        de la feuille (la feuille a pu changer, ou un objet tre supprim)."""
        if self.actuelle is not None:   # des fois que l'utilisateur ait ferme la feuille entre temps !
            #self.cache = [obj for obj in self.cache if obj in self.actuelle.objets.__dict__.values()]
            self.canvas.editeur.close()
            objets_feuille = self.actuelle.objets()
            for obj in self.cache:
                if obj not in objets_feuille:
                    self.cache.remove(obj)
            # Utiliser l'objet set pour faire l'intersection entre self.cache te self.actuelle.objets.__dict__.values() ??
            if doublons_interdits and self.cache and kw.get("selection", None) is self.cache[-1]:
                # on vite de rentrer 2 fois de suite le mme objet dans le cache (vite de se retrouver avec qqch du style Segment(A,A) par ex.)
                # depuis la version 0.101.3, recliquer sur un objet le supprime du cache.
                self.cache.pop()
                return False
            return True
    
    
    
#----------------------------------------------------------
#      Fonctions interactives de creation d'objets
#----------------------------------------------------------


    def curseur(self, event = None):
        u"Revenir en mode standard (flche simple)."
        self.interagir(None)
    

    
    def zoombox(self, event = False, **kw):
        u"Mode zoom."
        if event is not False:
            self.debut_zoombox = None
            self.interagir(self.zoombox, u"Cliquez pour dlimiter le dbut de la zone  afficher.", self.zoombox_onmotion)
        elif self.actuelle and not self.canvas.fixe:
            if self.debut_zoombox is None:
                self.canvas.message(u"Cliquez pour dlimiter la fin de la zone  afficher.")
                self.debut_zoombox = kw["position"]
            else:
                (x0, y0), (x1, y1) = self.debut_zoombox, self.canvas.fin_zoom
                self.commande.executer_code_interne(u"set_fenetre(" + str((x0, x1, y0, y1))[1:-1] + u")")
                self.canvas.message(u"Cliquez pour dlimiter le dbut de la zone  afficher.")
                self.debut_zoombox = None

    def zoombox_onmotion(self, **kw):
        if self.debut_zoombox is not None:
            self.canvas.debut_zoom = self.debut_zoombox
            self.canvas.gestion_zoombox(kw["pixel"])
            self.canvas.debut_zoom = None
            


    
    def selectionner(self, event = False, **kw):
        u"Slectionner une zone."
        if event is not False:
            self.debut_selection = None
            self.interagir(self.selectionner, u"Cliquez pour dlimiter le dbut de la zone  slectionner.", self.selectionner_onmotion)
        elif self.actuelle:
            if self.debut_selection is None:
                self.canvas.message(u"Cliquez pour dlimiter la fin de la zone  slectionner.")
                self.debut_selection = kw["position"]
            else:
                (x0, y0), (x1, y1) = self.debut_selection, self.canvas.fin_select
                #self.commande.executer_code_interne(u"set_fenetre(" + str((x0, x1, y0, y1))[1:-1] + u")")
                self.canvas.message(u"Cliquez pour dlimiter le dbut de la zone  slectionner.")
                self.debut_selection = None

    def selectionner_onmotion(self, **kw):
        if self.debut_selection is not None:
            self.canvas.debut_select = self.debut_selection
            self.canvas.selection_zone(kw["pixel"])
            self.canvas.debut_select = None





        
    def point(self, event = False, **kw):
        if event is not False:
            self.interagir(self.point, u"Cliquez sur un objet, ou dans l'espace vierge.")
        elif self.actuelle:
            #self.valider_nom_precedent()
            selection = kw["selection"]
            position = kw["position"]

            nom = self.actuelle.nom_alea("M")
            if not selection:
                self.executer(nom + u" = Point(%s, %s, legende=2)" % position, nom, init = False)
            else:
                snom = selection.nom()
                if isinstance(selection, Cercle_generique):
                    self.executer(u"%s = Glisseur_cercle(%s, %s, legende=2)" %(nom, snom, position), nom, init = False)
                elif isinstance(selection, Segment):
                    self.executer(u"%s = Glisseur_segment(%s, %s, legende=2)" %(nom, snom, position), nom, init = False)
                elif isinstance(selection, Droite):
                    self.executer(u"%s = Glisseur_droite(%s, %s, legende=2)" %(nom, snom, position), nom, init = False)
                elif isinstance(selection, Demidroite):
                    self.executer(u"%s = Glisseur_demidroite(%s, %s, legende=2)" %(nom, snom, position), nom, init = False)
                elif isinstance(selection, Arc_generique):
                    self.executer(u"%s = Glisseur_arc_cercle(%s, %s, legende=2)" %(nom, snom, position), nom, init = False)
                else:
                    self.executer(nom + u" = Point(%s, %s, legende=2)" % position, nom, init = False)
            return self.actuelle.objets[nom] # afin de pouvoir l'utiliser comme routine ailleurs
    
 
        
    def milieu(self, event = False, **kw):
        if event is not False:
            self.interagir(self.milieu, u"Choisissez deux points, ou un segment, un cercle, un polygone...")
        elif self.test(True, **kw):
            selection = kw["selection"]
            position = kw["position"]
            if isinstance(selection, Point_generique):
                self.cache.append(selection)
            elif isinstance(selection, Polygone):
                nom = self.actuelle.nom_alea("G")
                self.executer(u"%s = Centre_gravite(%s, legende=2)" %(nom, selection.nom()), nom)
            elif isinstance(selection, Cercle_generique) and not self.cache:
                nom = self.actuelle.nom_alea("O")
                self.executer(u"%s = Centre(%s, legende=2)" %(nom, selection.nom()), nom)
            elif isinstance(selection, Segment) and not self.cache:
                nom = self.actuelle.nom_alea("I")
                self.executer(u"%s = Milieu(%s.point1, %s.point2, legende=2)" %(nom, selection.nom(), selection.nom()), nom)
            else:
                self.cache.append(self.point(**kw))
                
            if len(self.cache) == 1:
                self.actuelle.objet_temporaire(Milieu(self.cache[0], self.actuelle.point_temporaire()))

            if len(self.cache) == 2:
                self.actuelle.objet_temporaire(None)
                nom = self.actuelle.nom_alea("I") # doit etre regenere (il se peut qu'un point ait ete cree entre temps)
                self.executer(u"%s = Milieu(%s, %s, legende=2)" %(nom, self.cache[0].nom(), self.cache[1].nom()), nom)

            if len(self.cache) > 2: # ne se produit que si l'execution a plante...
                self.initialiser()
                
        elif self.cache:
            self.actuelle.objet_temporaire(Milieu(self.cache[0], self.actuelle.point_temporaire()))
        else:
            self.actuelle.objet_temporaire(None)



    #event doit recevoir par defaut False et non None (car event vaut None par dfaut dans MultiButton.action())
        
    def segment(self, event = False, **kw):
        if event is False:
            self.npoints(Segment, "s", **kw)
        else:
            self.interagir(self.segment, u"Choisissez ou crez deux points.")

        
    def vecteur(self, event = False, **kw):
        if event is False:
            self.npoints(Vecteur, "v", **kw)
        else:
            self.interagir(self.vecteur, u"Choisissez ou crez deux points.")
 
        
    def droite(self, event = False, **kw):
        if event is False:
            self.npoints(Droite, "d", **kw)
        else:
            self.interagir(self.droite, u"Choisissez ou crez deux points.")

    def demidroite(self, event = False, **kw):
        if event is False:
            self.npoints(Demidroite, "d", **kw)
        else:
            self.interagir(self.demidroite, u"Choisissez ou crez deux points.")

    def mediatrice(self, event = False, **kw):
        if event is not False:
            self.interagir(self.mediatrice, u"Choisissez un segment ou deux points.")
        elif self.test(**kw):
            selection = kw["selection"]
            position = kw["position"]
            
            if isinstance(selection, Point_generique):
                self.cache.append(selection)
                if len(self.cache) == 1:
                    self.actuelle.objet_temporaire(Mediatrice(selection, self.actuelle.point_temporaire()))
            elif isinstance(selection, Segment):
                nom = self.actuelle.nom_alea("d") # doit etre regenere (il se peut qu'un point ait ete cree entre temps)
                self.executer("%s = Mediatrice(%s.point1, %s.point2)" %(nom, selection.nom(), selection.nom()), nom)
            else:
                self.cache.append(self.point(**kw))
                if len(self.cache) == 1:
                    self.actuelle.objet_temporaire(Mediatrice(self.cache[0], self.actuelle.point_temporaire()))

            if len(self.cache) == 2:
                self.actuelle.objet_temporaire(None)
                nom = self.actuelle.nom_alea("d") # doit etre regenere (il se peut qu'un point ait ete cree entre temps)
                self.executer("%s = Mediatrice(%s, %s)" %(nom, self.cache[0].nom(), self.cache[1].nom()), nom)

            if len(self.cache) > 2: # ne se produit que si l'execution a plante...
                self.initialiser()


            
    def bissectrice(self, event = False, **kw): # A REVOIR
        if event is not False:
            self.interagir(self.bissectrice, u"Choisissez un angle ou trois points.")
        elif self.test(**kw):
            selection = kw["selection"]
            position = kw["position"]
            
            if isinstance(selection, Point_generique):
                self.cache.append(selection)
                if len(self.cache) == 2:
                    self.actuelle.objet_temporaire(Bissectrice(self.cache[0], self.cache[1], self.actuelle.point_temporaire()))
            elif isinstance(selection, Secteur_angulaire):
                nom = self.actuelle.nom_alea("d") # doit etre regenere (il se peut qu'un point ait ete cree entre temps)
                self.executer("%s = Bissectrice(%s.point1, %s.point2, %s.point3)" %(nom, selection.nom(), selection.nom(), selection.nom()), nom)
            else:
                self.cache.append(self.point(**kw))
                if len(self.cache) == 2:
                    self.actuelle.objet_temporaire(Bissectrice(self.cache[0], self.cache[1], self.actuelle.point_temporaire()))

            if len(self.cache) == 3:
                self.actuelle.objet_temporaire(None)
                nom = self.actuelle.nom_alea("d") # doit tre regner (il se peut qu'un point ait t cre entre temps)
                self.executer("%s = Bissectrice(%s, %s, %s)" %(nom, self.cache[0].nom(), self.cache[1].nom(), self.cache[2].nom()), nom)

            if len(self.cache) > 3: # ne se produit que si l'execution a plante...
                self.initialiser()
            
            
    def perpendiculaire(self, event = False, **kw):
        if event is not False:
            self.interagir(self.perpendiculaire, u"Choisissez ou crez un point et une droite.")
        elif self.test(**kw):
            selection = kw["selection"]
            position = kw["position"]
            if len(self.cache) == 0:
                if isinstance(selection, Point_generique):
                    self.cache.append(selection)
                elif isinstance(selection, Ligne):
                    self.cache.append(selection)
                    self.actuelle.objet_temporaire(Perpendiculaire(selection, self.actuelle.point_temporaire()))
                else:
                    self.cache.append(self.point(**kw))
            elif len(self.cache) == 1:
                if isinstance(self.cache[0], Ligne):
                    if isinstance(selection, Point_generique):
                        self.cache.append(selection)
                    else:
                        self.cache.append(self.point(**kw))
                elif isinstance(selection, Ligne):
                    self.cache.append(selection)


            if len(self.cache) == 2:
                self.actuelle.objet_temporaire(None)
                nom = self.actuelle.nom_alea("d") # doit etre regenere (il se peut qu'un point ait ete cree entre temps)
                if isinstance(self.cache[0], Point_generique):
                    self.cache.reverse()
                self.executer("%s = Perpendiculaire(%s, %s)" %(nom, self.cache[0].nom(), self.cache[1].nom()), nom)

            if len(self.cache) > 2: # ne se produit que si l'execution a plante...
                self.initialiser()


    def parallele(self, event = False, **kw):
        if event is not False:
            self.interagir(self.parallele, "Choisissez ou crez un point et une droite.")
        elif self.test(**kw):
            selection = kw["selection"]
            position = kw["position"]
            if len(self.cache) == 0:
                if isinstance(selection, Point_generique):
                    self.cache.append(selection)
                elif isinstance(selection, Ligne):
                    self.cache.append(selection)
                    self.actuelle.objet_temporaire(Parallele(selection, self.actuelle.point_temporaire()))
                else:
                    self.cache.append(self.point(**kw))
            elif len(self.cache) == 1:
                if isinstance(self.cache[0], Ligne):
                    if isinstance(selection, Point_generique):
                        self.cache.append(selection)
                    else:
                        self.cache.append(self.point(**kw))
                elif isinstance(selection, Ligne):
                    self.cache.append(selection)


            if len(self.cache) == 2:
                self.actuelle.objet_temporaire(None)
                nom = self.actuelle.nom_alea("d") # doit etre regenere (il se peut qu'un point ait ete cree entre temps)
                if isinstance(self.cache[0], Point_generique):
                    self.cache.reverse()
                self.executer(u"%s = Parallele(%s, %s)" %(nom, self.cache[0].nom(), self.cache[1].nom()), nom)

            if len(self.cache) > 2: # ne se produit que si l'execution a plante...
                self.initialiser()


    def tangente(self, event = False, **kw):
        if event is not False:
            self.interagir(self.tangente, "Choisissez ou crez un point et un cercle.")
        elif self.test(**kw):
            selection = kw["selection"]
            position = kw["position"]
            if len(self.cache) == 0:
                if isinstance(selection, Point_generique):
                    self.memoire_position = position
                    self.cache.append(selection)
                elif isinstance(selection, Cercle_generique):
                    self.cache.append(selection)
                    self.actuelle.objet_temporaire(Tangente(selection, self.actuelle.point_temporaire()))
                else:
                    self.memoire_position = position
                    self.cache.append(self.point(**kw))
            elif len(self.cache) == 1:
                if isinstance(self.cache[0], Cercle_generique):
                    self.memoire_position = position
                    if isinstance(selection, Point_generique):
                        self.cache.append(selection)
                    else:
                        self.cache.append(self.point(**kw))
                elif isinstance(selection, Cercle_generique):
                    self.cache.append(selection)


            if len(self.cache) == 2:
                self.actuelle.objet_temporaire(None)
                nom = self.actuelle.nom_alea("d") # doit tre regnr (il se peut qu'un point ait t cre entre temps)
                if isinstance(self.cache[0], Point_generique):
                    self.cache.reverse()
                x1, y1 = self.cache[0].centre   # on choisit la tangente qui passe le plus prs du point du cercle slectionn
                x, y = position
                x2, y2 = self.cache[1]
                det = (x1 - x)*(y2 - y) - (x2 - x)*(y1 - y)
                self.executer(u"%s = Tangente(%s, %s, %s)" %(nom, self.cache[0].nom(), self.cache[1].nom(), det>0), nom)

            if len(self.cache) > 2: # ne se produit que si l'excution a plante...
                self.initialiser()


    def representant(self, event = False, **kw):
        if event is not False:
            self.interagir(self.representant, "Choisissez ou crez un point et un vecteur.")
        elif self.test(**kw):
            selection = kw["selection"]
            position = kw["position"]
            if len(self.cache) == 0:
                if isinstance(selection, Point_generique):
                    self.cache.append(selection)
                elif isinstance(selection, Vecteur):
                    self.cache.append(selection)
                    self.actuelle.objet_temporaire(Representant(selection, self.actuelle.point_temporaire(), previsualisation = True))
                else:
                    self.cache.append(self.point(**kw))
            elif len(self.cache) == 1:
                if isinstance(self.cache[0], Vecteur):
                    if isinstance(selection, Point_generique):
                        self.cache.append(selection)
                    else:
                        self.cache.append(self.point(**kw))
                elif isinstance(selection, Vecteur):
                    self.cache.append(selection)


            if len(self.cache) == 2:
                self.actuelle.objet_temporaire(None)
                nom = self.actuelle.nom_alea("v") # doit etre regenere (il se peut qu'un point ait ete cree entre temps)
                if isinstance(self.cache[0], Point_generique):
                    self.cache.reverse()
                self.executer(u"%s = Representant(%s, %s)" %(nom, self.cache[0].nom(), self.cache[1].nom()), nom)

            if len(self.cache) > 2: # ne se produit que si l'execution a plante...
                self.initialiser()



    def cercle(self, event = False, **kw):
        if event is False:
            self.npoints(Cercle_rayon, "c", **kw)
        else:
            self.interagir(self.cercle, u"Choisissez ou crez deux points.")

    def cercle_diametre(self, event = False, **kw):
        if event is False:
            self.npoints(Cercle_diametre, "c", **kw)
        else:
            self.interagir(self.cercle_diametre, u"Choisissez ou crez deux points.")

    def cercle_points(self, event = False, **kw):
        if event is False:
            self.npoints(Cercle_points, "c", 3, **kw)
        else:
            self.interagir(self.cercle_points, u"Choisissez ou crez 3 points.")

    def arc(self, event = False, **kw):
        if event is False:
            self.npoints(Arc_cercle, "c", 3, **kw)
        else:
            self.interagir(self.arc, u"Choisissez ou crez 3 points.")

    def demicercle(self, event = False, **kw):
        if event is False:
            self.npoints(Demicercle, "c", **kw)
        else:
            self.interagir(self.demicercle, u"Choisissez ou crez deux points.")

    def arc_points(self, event = False, **kw):
        if event is False:
            self.npoints(Arc_points, "c", 3, **kw)
        else:
            self.interagir(self.arc_points, u"Choisissez ou crez 3 points.")

    def arc_oriente(self, event = False, **kw):
        if event is False:
            self.npoints(Arc_oriente, "c", 3, **kw)
        else:
            self.interagir(self.arc_oriente, u"Choisissez ou crez 3 points.")

    def disque(self, event = False, **kw):
        if event is not False:
            self.interagir(self.disque, u"Choisissez un cercle.")
        elif self.test(**kw):
            selection = kw["selection"]
            if isinstance(selection, Cercle_generique):
                nom = self.actuelle.nom_alea("d")
                self.executer(u"%s = Disque(%s)" %(nom, selection.nom()))


    def angle(self, event = False, **kw):
        if event is False:
            self.npoints(Angle, "a", 3, **kw)
        else:
            self.interagir(self.angle, u"Choisissez ou crez trois points.")


    def angle_oriente(self, event = False, **kw):
        if event is False:
            self.npoints(Angle_oriente, "a", 3, **kw)
        else:
            self.interagir(self.angle_oriente, u"Choisissez ou crez trois points.")


    def triangle(self, event = False, **kw):
        if event is False:
            self.npoints(Triangle, "p", 3, **kw)
        else:
            self.interagir(self.triangle, u"Choisissez ou crez trois points.")


    def parallelogramme(self, event = False, **kw):
        if event is False:
            self.npoints(Parallelogramme, "p", 3, **kw)
        else:
            self.interagir(self.parallelogramme, u"Choisissez ou crez trois points.")



    def polygone(self, event = False, **kw):
        if event is not False:
            self.interagir(self.polygone, u"Indiquez les sommets, puis cliquez sur le 1er sommet.")
        elif self.test(True, **kw):
            #self.cache = [obj for obj in self.cache if obj.nom() and obj.__feuille__ == self.actuelle]
            selection = kw["selection"]
            position = kw["position"]

            #nom = self.actuelle.nom_alea()
            if isinstance(selection, Point_generique):
                self.cache.append(selection)
            else:
                self.cache.append(self.point(**kw))
                
            if len(self.cache) >= 1:
                self.cache[0].affiche_en_gras(True)
                
            if len(self.cache) > 2 and self.cache[0] is self.cache[-1]:
                self.actuelle.objet_temporaire(None)
                cache = self.cache
                self.cache[0].affiche_en_gras(False)
                self.initialiser()
                nom = self.actuelle.nom_alea("p")
                self.executer(nom + u" = Polygone(" + ",".join([obj.nom() for obj in cache[:-1]]) + ")")
            else:
                tmp = self.actuelle.objet_temporaire() # liste des objets temporaires
                if tmp and isinstance(tmp[-1], Polygone): # on ne recre pas la prvisualisation, on se contente d'ajouter un sommet
                    tmp = tmp[-1]
                    tmp.points = tmp.points[:-1] + (self.cache[-1], self.actuelle.point_temporaire())
                    tmp.refresh()
                else:
                    self.actuelle.objet_temporaire(Polygone(*(self.cache + [self.actuelle.point_temporaire()]), **{"previsualisation": True}))
        elif self.cache:
            tmp = self.actuelle.objet_temporaire()[-1]
            tmp.points = tmp.points[:-2] + (self.actuelle.point_temporaire(),) # on supprime un sommet
            tmp.refresh()
        else:
            self.actuelle.objet_temporaire(None)
            





    def intersection(self, event = False, **kw):
        def inter_dte_cer(dte, cer, position): # sert  detecter l'intersection la plus proche du pointeur
            M = Intersection_droite_cercle(dte, cer)
            
            xy = M.intersections()
            if not xy: # aucune intersection n'existe actuellement
                test = True
            else:
                xy0, xy1 = xy
                if xy0:
                    xy0 = self.canvas.XYcoo2pix(array(xy0), -1)
                    if xy1: # 2 points d'intersection -> on regarde le plus proche du pointeur
                        xy1 = self.canvas.XYcoo2pix(array(xy1), -1)
                        position = self.canvas.XYcoo2pix(array(position), -1)
                        test = norm(array(position) - array(xy0), 2) < norm(array(position) - array(xy1), 2)
                    else:
                        test = True
                else:
                    test = False
            if test: point_proche = M.point_proche # M est bien l'intersection la plus proche
            else:   point_proche = M.point_lointain
            if point_proche is dte.point1:
                i = 1
            else:
                i = 2
            self.executer(u"%s = Intersection_droite_cercle(%s, %s, %s.point%s, legende=2)" %(nom, dte.nom(), cer.nom(), dte.nom(), i), nom)
            
        if event is not False:
            self.interagir(self.intersection, u"Indiquez deux objets, ou le lieu de l'intersection.")
        elif self.test(**kw):
            selection = kw["selection"]
            autres = kw["autres"] # autres objets a proximite
            position = kw["position"]

            nom = self.actuelle.nom_alea("M")
            if isinstance(selection, (Segment, Demidroite, Droite, Cercle_generique, Arc_generique)):
                self.cache.append(selection)
                
            if len(self.cache) == 1:  # permet de construire une intersection en 1 clic.
                objets = [obj for obj in autres if obj is not selection and isinstance(obj,(Segment, Demidroite, Droite, Cercle_generique, Arc_generique))] 
                if len(objets) == 1:  # il n'y a pas d'ambiguite sur le 2eme objet.
                    self.cache += objets
                

            if len(self.cache) == 2:
                obj1, obj2 = self.cache
                #self.initialiser() #inutile
                nom = self.actuelle.nom_alea("M") # doit etre regenere (il se peut qu'un point ait ete cree entre temps)
                if isinstance(obj1, (Droite, Segment, Demidroite)) and isinstance(obj2, (Droite, Segment, Demidroite)):
                    self.executer(u"%s = Intersection_droites(%s, %s, legende=2)" %(nom, obj1.nom(), obj2.nom()), nom)
                elif isinstance(obj1, (Cercle_generique, Arc_generique)) and isinstance(obj2, (Cercle_generique, Arc_generique)):
                    M = Intersection_cercles(obj1, obj2)
                    xy0, xy1 = M.intersections()
                    xy0 = self.canvas.XYcoo2pix(array(xy0), -1)
                    xy1 = self.canvas.XYcoo2pix(array(xy1), -1)
                    position = self.canvas.XYcoo2pix(array(position), -1)
                    test = norm(array(position) - array(xy0), 2) < norm(array(position) - array(xy1), 2)
                    if test: angle = M.angle > 0 # M est bien l'intersection la plus proche
                    else:   angle = M.angle < 0
                    self.executer(u"%s = Intersection_cercles(%s, %s, %s, legende=2)" %(nom, obj1.nom(), obj2.nom(), angle), nom)
                elif isinstance(obj1, (Droite, Segment, Demidroite)) and isinstance(obj2, (Cercle_generique, Arc_generique)):
                    inter_dte_cer(obj1, obj2, position)
                else:
                    inter_dte_cer(obj2, obj1, position)

            if len(self.cache) > 2: # ne se produit que si l'execution a plante...
                self.initialiser()
                

            

    def texte(self, event = False, **kw):
        if event is not False:
            self.interagir(self.texte, u"Cliquez  l'emplacement souhait.")
        elif self.test(**kw):
            position = kw["position"]
            nom = self.actuelle.nom_alea("txt")
            self.executer(nom + u" = Texte('', %s, %s)" % position, nom, init = False)

            

    def masque(self, event = False, **kw):
        if event is False:
            selection = kw["selection"]
            if self.actuelle and selection is not None:
                self.commande.executer_code_interne("%s.cacher()" %selection.nom())
        else:
            self.interagir(self.masque)



    def gomme(self, event = False, **kw):
        if event is False:
            selection = kw["selection"]
            if self.actuelle and selection is not None:
                self.commande.executer_code_interne("%s.supprimer()" %selection.nom())
        else:
            self.interagir(self.gomme)

    def pinceau(self, event = False, **kw):
        if event is not False:
            self.interagir(self.pinceau, u"Slectionnez un objet pour en copier le style.")
        elif self.test(**kw):
            selection = kw["selection"]
            if selection is not None:
                if len(self.cache) == 0:
                    self.cache.append(selection)
                else:
                    self.commande.executer_code_interne("%s.copier_style(%s)" %(selection.nom(), self.cache[0].nom()))
            if len(self.cache):
                self.cache[0].affiche_en_gras(True)





#    def raccourcis(self, event):
#        key = event.GetKeyCode() + 1 - wx.WXK_F1
#        if 1 <= key <= 9:
#            if event.ShiftDown():
#                key += 9
#            self.icones[key + 2][2]()
#        else:
#            event.Skip()
            


    def npoints(self, classe, prefixe, n = 2, **kw):
        u"Cration d'un objet de classe 'classe' ayant 'n' points comme arguments. Le nom de l'objet sera compos de 'prefixe' + 1 numro."
        if self.test(True, **kw):
            self.cache = [obj for obj in self.cache if obj.nom() and obj.__feuille__ == self.actuelle]
            selection = kw["selection"]
            position = kw["position"]

            if isinstance(selection, Point_generique):
                self.cache.append(selection)
                nouveau_point = False
            else:
                self.cache.append(self.point(**kw))
                nouveau_point = True
                
            if len(self.cache) == n-1:
                #if issubclass(classe, Polygone):
                #    self.actuelle.objet_temporaire(classe(*(tuple(self.cache) + (self.actuelle.point_temporaire(),)), **{"previsualisation": True}))
                #else:
                #    self.actuelle.objet_temporaire(classe(*(tuple(self.cache) + (self.actuelle.point_temporaire(),))))
                self.actuelle.objet_temporaire(classe(*(tuple(self.cache) + (self.actuelle.point_temporaire(),)), **{"previsualisation": True}))
                
            elif len(self.cache) == n:
                self.actuelle.objet_temporaire(None)
                nom = self.actuelle.nom_alea(prefixe)
                code = nom + " = " + classe.__name__ + "(" + ",".join(obj.nom() for obj in self.cache) + ")"
                if nouveau_point: # on edite le nom du nouveau point (dernier parametre de self.executer)
                    self.executer(code, self.cache[1].nom())
                else: # si c'est un vieux point, pas besoin d'editer son nom
                    self.executer(code)

            elif len(self.cache) > n: # ne se produit que si l'execution a plante...
                self.initialiser()

        elif self.cache:
            self.actuelle.objet_temporaire(classe(*(tuple(self.cache) + (self.actuelle.point_temporaire(),)), **{"previsualisation": True}))
        else:
            self.actuelle.objet_temporaire(None)
