# -*- coding: iso-8859-1 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#                 Geometre                    #
##--------------------------------------#######
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2007  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA



from API import *
from GUI import *





class GeometreMenuBar(MenuBar):
    def __init__(self, panel):
        MenuBar.__init__(self, panel)
        self.ajouter("Fichier", ["nouveau"], ["ouvrir"], ["ouvrir ici"], ["enregistrer"], ["enregistrer_sous"], ["exporter"], ["exporter&sauver"], None, ["mise en page"], ["imprimer"], None, ["proprietes"], None, self.panel.doc_ouverts, None, ["fermer"], ["quitter"])
        self.ajouter("Editer", ["annuler"], ["refaire"], ["modifier"], ["supprimer"])
        self.ajouter("creer")
        self.ajouter("Affichage", ["onglet"], ["repere"], ["quadrillage"], ["orthonorme"], ["fenetre"], ["zoomer"], ["dezoomer"], ["orthonormaliser"], ["zoom_auto"])
        self.ajouter("Autres actions", ["coder"], ["decoder"])
        self.ajouter("Outils", ["historique"], ["securise"], ["debug"], ["options"])
        self.ajouter("?")






class MultiButton(wx.BitmapButton):
    def __init__(self, parent, raccourci, *liste):
        self.parent = parent
        self.raccourci = raccourci
        self.liste = liste
        self.items = []
        for i in xrange(len(self.liste)):
            self.items.append(wx.NewId())
            self.Bind(wx.EVT_MENU, self.OnPopup, id=self.items[-1])
            self.liste[i] = list(self.liste[i])
            img = wx.Image(emplacement('images', self.liste[i][1] + (self.selected and "_" or "") + '.png'), wx.BITMAP_TYPE_PNG)
            self.liste[i][1] = img.ConvertToBitmap()
        wx.BitmapButton.__init__(self, parent, style=wx.NO_BORDER)
        self.SetBackgroundColour(self.parent.GetBackgroundColour())
        self.select(False, 0)
        self.Bind(wx.EVT_RIGHT_DOWN, self.OnRightclic)

    def select(self, sel = True, i = None):
        "Affiche l'image i"
        self.selected = sel
        if i is not None:
            self.selection = i
        touche = " (" + self.raccourci.replace("Shift", "Maj") +")"
        self.SetToolTipString(self.liste[self.selection][2] + touche)
        self.SetBitmapLabel(self.liste[self.selection][1])
        self.Unbind()
        self.Bind(wx.EVT_BUTTON, self.liste[self.selection][3])
        
        
    def OnRightclic(self, event = None):
        menu = wx.Menu()
        for elt in liste:
            item = wx.MenuItem(menu, self.popupID1, elt[0])
            item.SetBitmap(elt[1])
            menu.AppendItem(item)
        self.PopupMenu(menu)
        menu.Destroy()
        
    def OnPopup(self, event = None):
        self.selection = self.items.index(event.GetId())




class Geometre(Panel_API_graphique):

    __titre__ = "Geometrie dynamique" # Donner un titre a chaque module


    def __init__(self, parent):
        Panel_API_graphique.__init__(self, parent)
    
        self.entrees = wx.BoxSizer(wx.HORIZONTAL)
        
        self.icones = [("ouvrir","Ouvrir un fichier .geo.", self.parent.OpenFile),
                    ("sauvegarde","Enregistrer le document.", self.parent.SaveFile),
                    ("image","Exporter comme image.", self.parent.ExportFile),
                    ("fleche2", "Deplacer ou modifier un objet.", self.curseur),
                    ("point2","Creer un point.", self.creer_point),
                    ("milieu2","Creer un milieu ou un centre.", self.creer_milieu),
                    ("segment2","Creer un segment.", self.creer_segment),
                    ("droite2","Creer une droite.", self.creer_droite),
                    ("cercle2","Creer un cercle.", self.creer_cercle),
                    ("polygone","Creer un polygone.", self.creer_polygone),
                    ("intersection","Creer une intersection.", self.creer_intersection),
                    ("texte","Creer un texte.", self.creer_texte),
                    ("masquer", "Masquer des objets.", self.masquer),
                    ("gomme", "Supprimer des objets.", self.gommer),
                    ]
        
        self.boutons = []
        for i in range(len(self.icones)):
            image = self.icones[i]
            #bmp = wx.Image('images' + os.sep + image[0] + '.png', wx.BITMAP_TYPE_PNG).ConvertToBitmap()
            bouton = wx.BitmapButton(self, -1, self.icone(i), style=wx.NO_BORDER)
            bouton.SetBackgroundColour(self.GetBackgroundColour())
            self.entrees.Add(bouton, 0, wx.ALL,5)
            touche = "" # touche de raccourci
            if i >= 3:
                touche = " ("
                if i >= 12:
                    touche += "Maj+"
                touche += "F%s)" %((i - 3)%9 + 1)
            bouton.SetToolTipString(image[1] + touche)
            if image[2]:
                bouton.Bind(wx.EVT_BUTTON, image[2])
            bouton.Bind(wx.EVT_CHAR, self.raccourcis)
            self.boutons.append(bouton)
            if not param.afficher_boutons:
                break
                
                
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(self.entrees, 0, wx.ALL, 5)
        self.sizer.Add(self.canvas, 1, wx.LEFT | wx.TOP | wx.GROW)
        self.SetSizer(self.sizer)
        self.Fit()

        if param.afficher_boutons:
            self.canvas.Bind(wx.EVT_CHAR, self.raccourcis)
        # self.Bind(wx.EVT_CHAR, self.raccourcis) ne marche pas ?!? :-(
            self.selection(3) # l'icone selectionnee DOIT avoir deux images associees : img.png et img_.png par exemple.
        
        
    def rafraichir_titre(self):
        Panel_API_graphique.rafraichir_titre(self)
        if self.actuelle and hasattr(self, "boutons"): # evite les conflits lors de l'initialisation
            self.boutons[1].SetBitmapLabel(self.icone(1, not self.actuelle.sauvegarde["modifie"]))
    
    
    def initialiser(self):
        self.cache = [] # objets en memoire
        self.objet = None # objet recemment cree
        self.actuelle.objet_temporaire(None)

    #def valider_nom_precedent(self):
    #    try:
    #        self.canvas.editeur.init(valider = True)
    #    except NameError:
    #        pass
        

    def executer(self, instruction, nom = None, init = True):
        """Execute une instruction dans la console.
        Si nom is not None, le nom de l'objet est edite (points et textes seulement),
        afin de permettre de le renommer dans la foulee.
        Si init = True, le cache des selections est initialise."""
        self.canvas.editeur.close()
        self.commande.executer_code_interne(instruction)
        if init:
            self.initialiser()
        if nom:
            self.canvas.select = self.actuelle.objets[nom]
            self.canvas.editer()
                
            

        
    
    def icone(self, n, selection = False):
        "Renvoie l'image du nieme bouton (eventuellement selectionne)."
        nom_complet = 'images' + os.sep + self.icones[n][0] + (selection and "_" or "") + '.png'
        return wx.Image(nom_complet, wx.BITMAP_TYPE_PNG).ConvertToBitmap()
        
    
    def selection(self, n):
        "Marque le bouton n comme selectionne (et initialise le cache des selections)."
        self.initialiser()
        for i in range(3, len(self.icones)):
            if i <> n:
                self.boutons[i].SetBitmapLabel(self.icone(i))
            else:
                self.boutons[i].SetBitmapLabel(self.icone(i, True)) # bouton selectionne               

    
    
    def test(self):
        """On verifie qu'il y ait encore une feuille, et que le cache ne contienne que des objets
        de la feuille (la feuille a pu changer, ou un objet etre supprime).
        On en profite pour valider les noms d'objets en cours d'dition."""
        if self.actuelle:   # des fois que l'utilisateur ait ferme la feuille entre temps !
            self.cache = [obj for obj in self.cache if obj.nom() and obj.__feuille__ == self.actuelle]
            #self.valider_nom_precedent()
            return True
    
    
    
#----------------------------------------------------------
#      Fonctions interactives de creation d'objets
#----------------------------------------------------------


    def curseur(self, event = None):
        "Revenir en mode standard (fleche simple)."
        self.selection(3)
        self.canvas.interagir(None)
    
    
    def creer_point(self, event = None):
        self.selection(4)
        self.canvas.interagir(self.point, "Cliquez sur un objet, ou dans l'espace vierge.")
        
        
    def point(self, **kw):
        if self.actuelle:
            #self.valider_nom_precedent()
            selection = kw["selection"]
            position = kw["position"]

            nom = self.actuelle.nom_alea("M")
            if not selection:
                self.executer(nom + " = Point(%s, %s, legende=2)" % position, nom, init = False)
            else:
                snom = selection.nom()
                if isinstance(selection, Cercle_generique):
                    self.executer("%s = Glisseur_cercle(%s, %s, legende=2)" %(nom, snom, position), nom, init = False)
                elif isinstance(selection, Segment):
                    self.executer("%s = Glisseur_segment(%s, %s, legende=2)" %(nom, snom, position), nom, init = False)
                elif isinstance(selection, Droite):
                    self.executer("%s = Glisseur_droite(%s, %s, legende=2)" %(nom, snom, position), nom, init = False)
                elif isinstance(selection, Demidroite):
                    self.executer("%s = Glisseur_demidroite(%s, %s, legende=2)" %(nom, snom, position), nom, init = False)
                elif isinstance(selection, Arc_generique):
                    self.executer("%s = Glisseur_arc_cercle(%s, %s, legende=2)" %(nom, snom, position), nom, init = False)
                else:
                    self.executer(nom + " = Point(%s, %s, legende=2)" % position, nom, init = False)
            return self.actuelle.objets[nom] # afin de pouvoir l'utiliser comme routine ailleurs
    
            
    def creer_milieu(self, event = None):
        self.selection(5)
        self.canvas.interagir(self.milieu, "Choisissez deux points, ou un segment, un cercle, un polygone...")
 
        
    def milieu(self, **kw):
        if self.test():
            selection = kw["selection"]
            position = kw["position"]
            if isinstance(selection, Point_generique):
                self.cache.append(selection)
            elif isinstance(selection, Polygone):
                nom = self.actuelle.nom_alea("G")
                self.executer("%s = Centre_gravite(%s, legende=2)" %(nom, selection.nom()), nom)
            elif isinstance(selection, Cercle_generique) and not self.cache:
                nom = self.actuelle.nom_alea("O")
                self.executer("%s = Centre(%s, legende=2)" %(nom, selection.nom()), nom)
            elif isinstance(selection, Segment) and not self.cache:
                nom = self.actuelle.nom_alea("I")
                self.executer("%s = Milieu(%s.point1, %s.point2, legende=2)" %(nom, selection.nom(), selection.nom()), nom)
            else:
                self.cache.append(self.point(**kw))
                
            if len(self.cache) == 1:
                self.actuelle.objet_temporaire(Milieu(self.cache[0], self.actuelle.point_temporaire()))

            if len(self.cache) == 2:
                self.actuelle.objet_temporaire(None)
                nom = self.actuelle.nom_alea("I") # doit etre regenere (il se peut qu'un point ait ete cree entre temps)
                self.executer("%s = Milieu(%s, %s, legende=2)" %(nom, self.cache[0].nom(), self.cache[1].nom()), nom)

            if len(self.cache) > 2: # ne se produit que si l'execution a plante...
                self.initialiser()


    def creer_segment(self, event = None):
        self.selection(6)
        self.canvas.interagir(self.segment, "Choisissez ou crez deux points.")
 
        
    def segment(self, **kw):
        if self.test():
            selection = kw["selection"]
            position = kw["position"]
            #nom = self.actuelle.nom_alea()
            if isinstance(selection, Point_generique):
                self.cache.append(selection)
                nouveau_point = False
            else:
                self.cache.append(self.point(**kw))
                nouveau_point = True
                
            if len(self.cache) == 1:
                self.actuelle.objet_temporaire(Segment(self.cache[0], self.actuelle.point_temporaire()))

            if len(self.cache) == 2:
                self.actuelle.objet_temporaire(None)
                nom = self.actuelle.nom_alea("s")
                if nouveau_point: # on edite le nom du nouveau point (dernier parametre de self.executer)
                    self.executer("%s = Segment(%s, %s)" %(nom, self.cache[0].nom(), self.cache[1].nom()), self.cache[1].nom())
                else: # si c'est un vieux point, pas besoin d'editer son nom
                    self.executer("%s = Segment(%s, %s)" %(nom, self.cache[0].nom(), self.cache[1].nom()))

            if len(self.cache) > 2: # ne se produit que si l'execution a plante...
                self.initialiser()


    def creer_droite(self, event = None):
        self.selection(7)
        self.canvas.interagir(self.droite, "Choisissez ou crez deux points.")
 
        
    def droite(self, **kw):
        if self.test():
            selection = kw["selection"]
            position = kw["position"]

            #nom = self.actuelle.nom_alea()
            if isinstance(selection, Point_generique):
                self.cache.append(selection)
                nouveau_point = False
            else:
                self.cache.append(self.point(**kw))
                nouveau_point = True
                
            if len(self.cache) == 1:
                self.actuelle.objet_temporaire(Droite(self.cache[0], self.actuelle.point_temporaire()))

            if len(self.cache) == 2:
                self.actuelle.objet_temporaire(None)
                nom = self.actuelle.nom_alea("d")
                if nouveau_point: # on edite le nom du nouveau point
                    self.executer("%s = Droite(%s, %s)" %(nom, self.cache[0].nom(), self.cache[1].nom()), self.cache[1].nom())
                else:
                    self.executer("%s = Droite(%s, %s)" %(nom, self.cache[0].nom(), self.cache[1].nom()))

            if len(self.cache) > 2: # ne se produit que si l'execution a plante...
                self.initialiser()


    def creer_cercle(self, event = None):
        self.selection(8)
        self.canvas.interagir(self.cercle, "Indiquez le centre, et un point du cercle.")
 
        
    def cercle(self, **kw):
        if self.test():
            self.cache = [obj for obj in self.cache if obj.nom() and obj.__feuille__ == self.actuelle]
            selection = kw["selection"]
            position = kw["position"]

            #nom = self.actuelle.nom_alea()
            if isinstance(selection, Point_generique):
                self.cache.append(selection)
                nouveau_point = False
            else:
                self.cache.append(self.point(**kw))
                nouveau_point = True
                
            if len(self.cache) == 1:
                self.actuelle.objet_temporaire(Cerlce_rayon(self.cache[0], self.actuelle.point_temporaire()))

            if len(self.cache) == 2:
                self.actuelle.objet_temporaire(None)
                nom = self.actuelle.nom_alea("c")
                if nouveau_point: # on edite le nom du nouveau point (dernier parametre de self.executer)
                    self.executer("%s = Cercle_rayon(%s, %s)" %(nom, self.cache[0].nom(), self.cache[1].nom()), self.cache[1].nom())
                else: # si c'est un vieux point, pas besoin d'editer son nom
                    self.executer("%s = Cercle_rayon(%s, %s)" %(nom, self.cache[0].nom(), self.cache[1].nom()))

            if len(self.cache) > 2: # ne se produit que si l'execution a plante...
                self.initialiser()


    def creer_polygone(self, event = None):
        self.selection(9)
        self.canvas.interagir(self.polygone, "Indiquez les sommets, puis cliquez sur le 1er sommet.")


    def polygone(self, **kw):
        if self.test():
            #self.cache = [obj for obj in self.cache if obj.nom() and obj.__feuille__ == self.actuelle]
            selection = kw["selection"]
            position = kw["position"]

            #nom = self.actuelle.nom_alea()
            if isinstance(selection, Point_generique):
                self.cache.append(selection)
            else:
                self.cache.append(self.point(**kw))
                
            if len(self.cache) >= 1:
                self.cache[0].affiche_en_gras(True)
                
            if len(self.cache) > 2 and self.cache[0] is self.cache[-1]:
                self.actuelle.objet_temporaire(None)
                cache = self.cache
                self.cache[0].affiche_en_gras(False)
                self.initialiser()
                nom = self.actuelle.nom_alea("p")
                self.executer(nom + " = Polygone(" + ",".join([obj.nom() for obj in cache[:-1]]) + ")")
            else:
                self.actuelle.objet_temporaire(FastPolygone(*(self.cache + [self.actuelle.point_temporaire()])))



    def creer_intersection(self, event = None):
        self.selection(10)
        self.canvas.interagir(self.intersection, "Indiquez deux objets, ou le lieu de l'intersection.")


    def intersection(self, **kw):
        def inter_dte_cer(dte, cer, position): # sert a detecter l'intersection la plus proche du pointeur
            M = Intersection_droite_cercle(dte, cer)
            
            xy = M.coordonnees(bipoint = True)
            if not xy: # aucune intersection n'existe actuellement
                test = True
            else:
                xy0, xy1 = xy
                if xy0:
                    xy0 = self.canvas.XYcoo2pix(array(xy0), -1)
                    if xy1: # 2 points d'intersection -> on regarde le plus proche du pointeur
                        xy1 = self.canvas.XYcoo2pix(array(xy1), -1)
                        position = self.canvas.XYcoo2pix(array(position), -1)
                        test = norm(array(position) - array(xy0), 2) < norm(array(position) - array(xy1), 2)
                    else:
                        test = True
                else:
                    test = False
            if test: point_proche = M.point_proche # M est bien l'intersection la plus proche
            else:   point_proche = M.point_lointain
            self.executer("%s = Intersection_droite_cercle(%s, %s, %s, legende=2)" %(nom, dte.nom(), cer.nom(), point_proche.nom()), nom)
            
        if self.test():
            selection = kw["selection"]
            autres = kw["autres"] # autres objets a proximite
            position = kw["position"]

            nom = self.actuelle.nom_alea("M")
            if isinstance(selection, (Segment, Demidroite, Droite, Cercle_generique, Arc_generique)):
                self.cache.append(selection)
                
            if len(self.cache) == 1:  # permet de construire une intersection en 1 clic.
                objets = [obj for obj in autres if obj is not selection and isinstance(obj,(Segment, Demidroite, Droite, Cercle_generique, Arc_generique))] 
                if len(objets) == 1:  # il n'y a pas d'ambiguite sur le 2eme objet.
                    self.cache += objets
                

            if len(self.cache) == 2:
                obj1, obj2 = self.cache
                #self.initialiser() #inutile
                nom = self.actuelle.nom_alea("M") # doit etre regenere (il se peut qu'un point ait ete cree entre temps)
                if isinstance(obj1, (Droite, Segment, Demidroite)) and isinstance(obj2, (Droite, Segment, Demidroite)):
                    self.executer("%s = Intersection_droites(%s, %s, legende=2)" %(nom, obj1.nom(), obj2.nom()), nom)
                elif isinstance(obj1, (Cercle_generique, Arc_generique)) and isinstance(obj2, (Cercle_generique, Arc_generique)):
                    M = Intersection_cercles(obj1, obj2)
                    xy0, xy1 = M.coordonnees(bipoint = True)
                    xy0 = self.canvas.XYcoo2pix(array(xy0), -1)
                    xy1 = self.canvas.XYcoo2pix(array(xy1), -1)
                    position = self.canvas.XYcoo2pix(array(position), -1)
                    test = norm(array(position) - array(xy0), 2) < norm(array(position) - array(xy1), 2)
                    if test: angle = M.angle > 0 # M est bien l'intersection la plus proche
                    else:   angle = M.angle < 0
                    self.executer("%s = Intersection_cercles(%s, %s, %s, legende=2)" %(nom, obj1.nom(), obj2.nom(), angle), nom)
                elif isinstance(obj1, (Droite, Segment, Demidroite)) and isinstance(obj2, (Cercle_generique, Arc_generique)):
                    inter_dte_cer(obj1, obj2, position)
                else:
                    inter_dte_cer(obj2, obj1, position)

            if len(self.cache) > 2: # ne se produit que si l'execution a plante...
                self.initialiser()
                

    def creer_texte(self, event = None):
        self.selection(11)
        self.canvas.interagir(self.texte, "Cliquez  l'emplacement souhait.")
            

    def texte(self, **kw):
        if self.test():
            position = kw["position"]
            nom = self.actuelle.nom_alea("txt")
            self.executer(nom + " = Texte('', %s, %s)" % position, nom, init = False)



    def masquer(self, event = None):
        self.selection(12)
        self.canvas.interagir(self.masque)
            

    def masque(self, **kw):
        selection = kw["selection"]

        if self.actuelle and selection is not None:
            self.commande.executer_code_interne("%s.cacher()" %selection.nom())



    def gommer(self, event = None):
        self.selection(13)
        self.canvas.interagir(self.gomme)


    def gomme(self, **kw):
        selection = kw["selection"]

        if self.actuelle and selection is not None:
            self.commande.executer_code_interne("%s.supprimer()" %selection.nom())

            



    def raccourcis(self, event):
        key = event.GetKeyCode() + 1 - wx.WXK_F1
        if 1 <= key <= 9:
            if event.ShiftDown():
                key += 9
            self.icones[key + 2][2]()
        else:
            event.Skip()
            


