# -*- coding: iso-8859-1 -*-
from __future__ import division # 1/2 == .5 (par defaut, 1/2 == 0)

##--------------------------------------#######
#                Probabilits                 #
##--------------------------------------#######
#    WxGeometrie
#    Dynamic geometry, graph plotter, and more for french mathematic teachers.
#    Copyright (C) 2005-2007  Nicolas Pourcelot
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# version unicode


from API import *
from GUI import *


class ProbaMenuBar(MenuBar):
    def __init__(self, panel):
        MenuBar.__init__(self, panel)

        self.ajouter(u"Fichier", [u"nouveau"], [u"ouvrir"], [u"ouvrir ici"], [u"enregistrer"], [u"enregistrer_sous"], [u"exporter"], [u"exporter&sauver"], None, [u"mise en page"], [u"imprimer"], [u"presse-papier"], None, [u"proprietes"], None, self.panel.doc_ouverts, None, ["fermer"], ["quitter"])
        self.ajouter(u"Editer", ["annuler"], ["refaire"], ["modifier"], ["supprimer"])
        self.ajouter(u"creer")
        self.ajouter(u"Affichage", ["onglet"], None, ["repere"], ["quadrillage"], ["orthonorme"], None, ["fenetre"], ["zoomer"], ["dezoomer"], ["orthonormaliser"], [u"zoom_auto"])
        self.ajouter(u"Autres actions", [u"detecter"])
        self.ajouter(u"Outils", [u"Style des sommets", u"Modifier le style des sommets de l'arbre.", None, self.panel.proprietes_sommets], [u"Style des artes", u"Modifier le style des artes de l'arbre.", None, self.panel.proprietes_aretes], [u"Style de la lgende", u"Modifier le style des titres de chaque niveau.", None, self.panel.proprietes_titres], None, [u"historique"], [u"securise"], [u"debug"], [u"options"])
        self.ajouter(u"?")




class Probabilites(Panel_API_graphique):
    
    __titre__ = u"Arbre de probabilits" # Donner un titre a chaque module
    
    def __init__(self, parent):
        
    
        self.couleurs = u"bgrmkcy"


        Panel_API_graphique.__init__(self, parent)

        self.entrees = wx.BoxSizer(wx.VERTICAL)
        
        
        self.entrees.Add(wx.StaticText(self, -1, u" Instructions :"), 0, wx.ALL,5)
        
        self.instructions = wx.TextCtrl(self, size = (200, 300), style = wx.TE_MULTILINE)
        self.instructions.SetValue("""||Tirage 1|Tirage 2|Tirage 3
omega
>A:0,7
>>B:0,2
>>C:0,8
>&A:0,3
>>E:0,1
>>>F
>>>G
>>>H
>>&E:0,9""")
        self.entrees.Add(self.instructions, 0, wx.ALL,5)
        self.appliquer = wx.Button(self, label = "Appliquer")
        self.appliquer.Bind(wx.EVT_BUTTON, self.Appliquer)
        self.entrees.Add(self.appliquer, 0, wx.ALL,5)
                
        #self.dessiner = wx.Button(self, wx.ID_REFRESH)
        #self.entrees.Add(self.dessiner, 0, wx.ALL, 5)
        self.sizer = wx.BoxSizer(wx.HORIZONTAL)
        i = 1
        if param.BugWxSizerGtk:
            i = 0 # bug de WxGtk
            self.canvas.SetSize(wx.Size(600, 400))
        self.sizer.Add(self.canvas, i, wx.LEFT | wx.TOP | wx.GROW)
        self.sizer.Add(self.entrees, 0, wx.ALL, 5)
        self.SetSizer(self.sizer)
        self.Fit()


    
    def _sauvegarder(self, fgeo):
        Panel_API_graphique._sauvegarder(self, fgeo)
        fgeo.contenu[u"Instructions"] = [self.instructions.GetValue()]
        
    
    def _ouvrir(self, fgeo):
        Panel_API_graphique._ouvrir(self, fgeo)
        if fgeo.contenu.has_key(u"Instructions"):
            self.instructions.SetValue(fgeo.contenu[u"Instructions"][0])
        #self.affiche()

    def Appliquer(self, event):
        self.nouveau()
        instructions = [instruction for instruction in self.instructions.GetValue().split("\n") if instruction]
        nbr_colonnes = 0
        if instructions[0].startswith("|"):
            legende = instructions[0]
            instructions = instructions[1:]
            self.canvas.set_fenetre(-.1, 1.1, -.1, 1.2)
        else:
            legende = None
            self.canvas.set_fenetre(-.1, 1.1, -.1, 1.1)
            
        for instruction in instructions:
            if instruction.startswith(">"):
                nbr_colonnes = max(nbr_colonnes, len(instruction) - len(instruction.lstrip(">")))
        if instructions[0].startswith(">"):
            instructions = [""] + instructions
        nbr_colonnes += 1
        print "nbr colonnes: ", nbr_colonnes
        nbr_lignes = []
        for i in xrange(nbr_colonnes):
            nbr_lignes.append(len([instruction for instruction in instructions if instruction.startswith(i*">") and not instruction.startswith((i+1)*">")]))
        print "nbr lignes: ", nbr_lignes
        
        #intersection, union : \cap \cup
        
        # Interprtation des instructions sous forme de listes de listes
        # >A
        # >>B
        # >>C
        # >D
        # >>E
        # >>>F
        # >>>G
        # >>H
        # >E
        # devient :
        # [{'liste': [{'liste': [{'liste': [], 'texte': u'B'}, {'liste': [], 'texte': u'C'}], 'texte': u'A'}, {'liste': [{'liste': [{'liste': [], 'texte': u'F'}, {'liste': [], 'texte': u'G'}], 'texte': u'E'}, {'liste': [], 'texte': u'H'}], 'texte': u'D'}, {'liste': [], 'texte': u'E'}], 'texte': ''}]

        #niveau = 0
        #if instructions[0].startswith(">"):
        #    arbre = [{"texte": "", "liste": []}]
        arbre = []
        ligne_precedente = [-1 for i in xrange(nbr_colonnes)] # numro de ligne atteint pour chaque colonne
        #ligne_precedente[0] = 0
        #adresse_precedent = [None for i in xrange(nbr_colonnes)] # adresse de l'arbre  utiliser comme rfrence
        for instruction in instructions:
            colonne = len(instruction) - len(instruction.lstrip(">"))
            print "c:", colonne
            branche = arbre
            for i in xrange(colonne): # on se dplace de branche en branche ;)
                branche = branche[ligne_precedente[i]]["liste"]
            branche.append({"texte": instruction.lstrip(">"), "liste": []})
            ligne_precedente[colonne] += 1
            for i in xrange(colonne + 1, nbr_colonnes):
                ligne_precedente[i] = -1
        print arbre
        
        # on parcourt l'arbre pour compter le nombre de ramifications

        def compter_ramifications(branche):
            if len(branche["liste"]) > 0:
                return builtin.sum(compter_ramifications(tige) for tige in branche["liste"])
            else:
                return 1
        
        ramifications = builtin.sum(compter_ramifications(branche) for branche in arbre)
        print ramifications
        
        def creer_point(x, y, texte):
            if texte:
                if texte.startswith("&"):
                    texte = r"\bar{" + texte[1:] + "}"
                texte = "$" + texte + "$"
                texte = texte.replace(" inter ", r"\  \cap \ ").replace(" union ", r"\  \cup \ ").replace("omega", r"\Omega")
            M = Point(x, y, legende = wxgeo.TEXTE, label = texte, style = "o", couleur = "w", taille = 0)
            M.etiquette.style(_rayon_ = 0, niveau = 15, alignement_vertical = "center", alignement_horizontal = "center", fond = "w")
            return M


        ramification = [0] # astuce pour avoir un "objet modifiable" (a mutable object)
        def parcourir_branche(branche, n, ramification = ramification):
            txt = branche["texte"].split(":", 1)
            if len(txt) == 2:
                txt_pt, txt_segm = txt
            else:
                txt_pt = txt[0]
                txt_segm = ""
            if len(branche["liste"]) > 0:
                l = []
                for tige in branche["liste"]:
                    l.append(parcourir_branche(tige, n + 1))
                
                M = creer_point(n/(nbr_colonnes - 1), .5*(l[0][0].ordonnee()+l[-1][0].ordonnee()), txt_pt)
                setattr(self.actuelle.objets, self.actuelle.nom_alea("Probabilites_point"), M)
                for point, txt_s in l:
                    s = Segment(point, M, legende = wxgeo.TEXTE, label = txt_s)
                    s.etiquette.style(_rayon_ = 0, niveau = 15, alignement_vertical = "center", alignement_horizontal = "center", fond = "w")
                    setattr(self.actuelle.objets, self.actuelle.nom_alea("Probabilites_segment"), s)
                return M, txt_segm
            else:
                M = creer_point(n/(nbr_colonnes - 1), 1 - ramification[0]/(ramifications - 1), txt_pt)
                ramification[0] += 1
                setattr(self.actuelle.objets, self.actuelle.nom_alea("Probabilites_point"), M)
                return M, txt_segm
            
        for branche in arbre:
            parcourir_branche(branche, 0)
            
        if legende is not None:
            decalage = -0.5
            for i in legende:
                if i <> "|":
                    break
                decalage += .5
            legende = legende.strip("|").split("|")
            for n in xrange(len(legende)):
                t = Texte(legende[n], (n + decalage)/(nbr_colonnes - 1), 1.1)
                setattr(self.actuelle.objets, self.actuelle.nom_alea("Probabilites_legende"), t)
            
        self.canvas.regenerer_liste = True
        self.canvas.detecter()
#        dernier_point = [None for i in xrange(nbr_colonnes)]
#        print self.actuelle
#        for instruction in instructions:
#            colonne = len(instruction) - len(instruction.lstrip(">"))
#            print colonne
#            ligne = lignes[colonne]
#            print ligne
#            lignes[colonne] += 1
#            M = Point(colonne/nbr_colonnes, -ligne/nbr_lignes[colonne])
#            setattr(self.actuelle.objets, self.actuelle.nom_alea("M"), M)
#            dernier_point[colonne] = M
#            if colonne > 0:
#                s = Segment(dernier_point[colonne - 1], M)
#                setattr(self.actuelle.objets, self.actuelle.nom_alea("s"), s)
        
            
    def proprietes_sommets(self, event = None):
        objets = self.actuelle.objets(type = Point)
        from GUI.proprietes_objets import Proprietes
        win = Proprietes(self, objets)
        win.Show(True)
    
    
    
    def proprietes_aretes(self, event = None):
        objets = self.actuelle.objets(type = Segment)
        from GUI.proprietes_objets import Proprietes
        win = Proprietes(self, objets)
        win.Show(True)
        

    def proprietes_titres(self, event = None):
        objets = self.actuelle.objets(type = Texte)
        from GUI.proprietes_objets import Proprietes
        win = Proprietes(self, objets)
        win.Show(True)
        

    def _affiche(self):
        pass


